/*
**  linux/amiga/amicon.c
**
**  Copyright (C) 1994 David Carter  [carter@compsci.bristol.ac.uk]
**      Modified 20-Apr-94 by William Rucklidge (wjr@cs.cornell.edu)
**          - AGA modes
**      Modified 03-May-94 by Geert Uytterhoeven (Geert.Uytterhoeven@cs.kuleuven.ac.be)
**          - additional AGA modes         
**          - 24-bit color palette
**      Modified 19-June-94 by Geert Uytterhoeven
**          - cursor flashing for better cursor visibility
**	Modified 24-Jan-95 by Geert Uytterhoeven
**	    - frame buffer update
**      Modified 18-August-94 by Jes Sorensen (jds@kom.auc.dk)
**          - ECS support. The following screen-modes are included:
**              multiscan, multiscan-lace, super72, super72-lace, dblntsc,
**              dblpal & euro72.
**            I suggest that we remove the old AGA screenmodes, as they are
**              non-standard, and some of them doesn't work under ECS.
**      Minor cleanup on 15-November-94, first release of the ECS-screenmodes.
**            (Jes)
**
**  Interface based on the original 'amicon.c' code:
**      Copyright (C) 1993 Hamish Macdonald and Greg Harp
**  although the contents have been radically altered.
**
** This file is subject to the terms and conditions of the GNU General
** Public License.  See the file README.legal in the main directory of this
** archive for more details.
*/

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/tty.h>
#include <linux/console.h>
#include <linux/string.h>

#include <linux/amigahw.h>
#include <linux/config.h>       /* New: DPC */
#include <linux/bootinfo.h>
#include <linux/interrupt.h>    /* New: DPC */
#include <linux/amigaints.h>

#include <asm/system.h>

#include "amifb.h"

struct fb_info *fb_info;
struct display *disp;


static int amicon_switch (struct condata *conp);


static void amicon_setup(int con, int setcol, int cls)
{
	struct display *display = &disp[con];
	struct condata *conp = &vc_cons[con];

	conp->vc_cols = display->scr_width / display->fontwidth;
	conp->vc_rows = display->scr_height / display->fontheight;
	conp->vc_can_do_color = 0;
	if (cls)
		/* clear screen (a dirty hack...) */
		con_clear(con);

   conp->vc_flags |= CON_INITED;
}


static int amicon_changevar(int con)
{
	amicon_setup(con, 1, 1);
	return(0);
}


static long amicon_init(struct condata *conp, long kmem_start, char**display_desc)
{
	int unit = conp - vc_cons;
	static int master_init = 0;
   
	if (!master_init) {
		master_init = 1;
		fb_info = amifb_init(&kmem_start);
		disp = fb_info->disp;
		fb_info->changevar = &amicon_changevar;
		*display_desc = fb_info->modename;
	} else {
		/* Copy all from the first console, maybe allow different
		   display modes later. */
		disp[unit] = disp[0];
	}
	amicon_setup(unit, 1, 0);
	return(kmem_start);
}


static int amicon_deinit (struct condata *conp)
{
	struct display *p = &disp[conp - vc_cons];

	/* You can't deinit a non inited screen nor the console 0. */
	if (!(conp->vc_flags & CON_INITED) || p == disp)
		return -1;

	amicon_switch (&vc_cons[0]);	/* Always switch to the console 0 */
	memset (p, 0, sizeof (struct display));
	conp->vc_flags &= ~CON_INITED;

	return 0;
}

/* ====================================================================== */

/* Those of a delicate disposition might like to skip the next couple of
 * pages.
 *
 * These functions are drop in replacements for memmove and
 * memset(_, 0, _). However their five instances add at least a kilobyte
 * to the object file. You have been warned.
 *
 * Not a great fan of assembler for the sake of it, but I think
 * that these routines are at least 10 times faster than their C
 * equivalents for large blits, and thats important to the lowest level of
 * a graphics driver. Question is whether some scheme with the blitter
 * would be faster. I suspect not for simple text system - not much
 * asynchronisity.
 *
 * Code is very simple, just gruesome expansion. Basic strategy is to
 * increase data moved/cleared at each step to 16 bytes to reduce
 * instruction per data move overhead. movem might be faster still
 * For more than 15 bytes, we try to align the write direction on a
 * longword boundary to get maximum speed. This is even more gruesome.
 * Unaligned read/write used requires 68020+ - think this is a problem?
 *
 * Sorry!
*/


#if defined (__GNUC__) && defined (__mc68000__)

static __inline__ void *mymemclear_small(void * s, size_t count)
{
__asm__("movel  %1,d0 ; beq 4f ; movel %0,a0 ; movel #0,d1\n\t"
        "lsrl   #1,d0 ; jcc 1f ; moveb d1,a0@+\n\t"
     "1: lsrl   #1,d0 ; jcc 2f ; movew d1,a0@+\n\t"
     "2: lsrl   #1,d0 ; jcc 3f ; movel d1,a0@+\n\t"
     "3: lsrl   #1,d0 ; jcc 4f ; movel d1,a0@+ ; movel d1,a0@+\n\t"
     "4: subql  #1,d0 ; jcs 6f\n\t"
     "5: movel d1,a0@+; movel d1,a0@+ ; movel d1,a0@+ ; movel d1,a0@+\n\t"
        "dbra d0,5b\n\t"
     "6:"
      :
           : "g" (s), "g" (count)
           : "a0", "d0", "d1");

   return(0);              
}

static __inline__ void *mymemclear(void * s, size_t count)
{
    if (count < 16) 
__asm__("movel  %1,d0 ; beq 4f ; movel %0,a0   ; movel #0,d1\n\t"
        "lsrl   #1,d0 ; jcc 1f ; moveb d1,a0@+\n\t"
      "1:lsrl   #1,d0 ; jcc 2f ; movew d1,a0@+\n\t"
      "2:lsrl   #1,d0 ; jcc 3f ; movel d1,a0@+\n\t"
      "3:lsrl   #1,d0 ; jcc 4f ; movel d1,a0@+ ; movel d1,a0@+\n\t"
      "4:"
   :
        : "g" (s), "g" (count)
   : "a0", "d0", "d1");
    else
__asm__("movel  %1,d0 ; movel  %0,a0 ; movel a0,d1\n\t"
        "lsrl   #1,d1 ; jcc 1f ; clrb a0@+ ; subqw #1,d0\n\t"
        "lsrl   #1,d1 ; jcs 2f\n\t"  /* a0 increased=>bit 2 switched*/
        "clrw   a0@+  ; subqw  #2,d0 ; jra 2f\n\t"
     "1: lsrl   #1,d1 ; jcc 2f\n\t"
        "clrw   a0@+  ; subqw  #2,d0\n\t"
     "2: movew d0,sp@-; lsrl #2,d0 ; jeq 6f ; moveql #0,d1\n\t"
        "lsrl   #1,d0 ; jcc 3f ; movel  d1,a0@+\n\t"        
     "3: lsrl   #1,d0 ; jcc 4f ; movel d1,a0@+ ; movel d1,a0@+\n\t"
     "4: subql  #1,d0 ; jcs 6f\n\t"
     "5: movel d1,a0@+; movel d1,a0@+ ; movel d1,a0@+ ; movel d1,a0@+\n\t"
         "dbra d0,5b; subl #65536,d0; jcc 5b\n\t"
     "6: movew sp@+,d0; btst #1,d0 ; jeq 7f ; movew d1,a0@+\n\t"
     "7:              ; btst #0,d0 ; jeq 8f ; moveb d1,a0@+\n\t"
     "8:"
   :
        : "g" (s), "g" (count)
   : "a0", "d0", "d1");
   return(0);              
}

static __inline__ void *mymemmove(void * d, void * s, size_t count)
{
    if (d < s) {
        if (count < 16) {
__asm__("movel  %2,d0 ; beq 4f ; movel %0,a0   ; movel %1, a1\n\t"
        "lsrl   #1,d0 ; jcc 1f ; moveb a1@+,a0@+\n\t"
      "1:lsrl   #1,d0 ; jcc 2f ; movew a1@+,a0@+\n\t"
      "2:lsrl   #1,d0 ; jcc 3f ; movel a1@+,a0@+\n\t"
      "3:lsrl   #1,d0 ; jcc 4f ; movel a1@+,a0@+ ; movel a1@+,a0@+\n\t"
      "4:"
   :
        : "g" (d), "g" (s), "g" (count)
   : "a0", "a1", "d0");
       } else {
__asm__("movel  %2,d0 ; movel  %0,a0 ; movel %1,a1; movel a0,d1\n\t"
        "lsrl   #1,d1 ; jcc 1f ; moveb a1@+,a0@+ ; subqw #1,d0\n\t"
        "lsrl   #1,d1 ; jcs 2f\n\t"  /* a0 increased=>bit 2 switched*/
        "movew  a1@+,a0@+  ; subqw  #2,d0 ; jra 2f\n\t"
     "1: lsrl   #1,d1 ; jcc 2f\n\t"
        "movew  a1@+,a0@+  ; subqw  #2,d0\n\t"
     "2: movew d0,sp@-; lsrl #2,d0 ; jeq 6f\n\t"
        "lsrl   #1,d0 ; jcc 3f ; movel a1@+,a0@+\n\t"        
     "3: lsrl   #1,d0 ; jcc 4f ; movel a1@+,a0@+ ; movel a1@+,a0@+\n\t"
     "4: subql  #1,d0 ; jcs 6f\n\t"
          "5: movel a1@+,a0@+;movel a1@+,a0@+\n\t"
             "movel a1@+,a0@+;movel a1@+,a0@+\n\t"
             "dbra d0,5b ; subl #65536,d0; jcc 5b\n\t"
     "6: movew sp@+,d0; btst #1,d0 ; jeq 7f ; movew a1@+,a0@+\n\t"
     "7:              ; btst #0,d0 ; jeq 8f ; moveb a1@+,a0@+\n\t"
     "8:"
   :
        : "g" (d), "g" (s), "g" (count)
   : "a0", "a1", "d0", "d1");
      }
   } else {
    if (count < 16) {
__asm__("movel  %2,d0 ; beq 4f ; movel %0,a0   ; movel %1, a1\n\t"
        "addal  d0,a0 ; addal d0,a1\n\t"
        "lsrl   #1,d0 ; jcc 1f ; moveb a1@-,a0@-\n\t"
      "1:lsrl   #1,d0 ; jcc 2f ; movew a1@-,a0@-\n\t"
      "2:lsrl   #1,d0 ; jcc 3f ; movel a1@-,a0@-\n\t"
      "3:lsrl   #1,d0 ; jcc 4f ; movel a1@-,a0@- ; movel a1@-,a0@-\n\t"
      "4:"
   :
        : "g" (d), "g" (s), "g" (count)
   : "a0", "a1", "d0");
      } else {
__asm__("movel  %2,d0 ; movel  %0,a0 ; movel %1,a1\n\t"
        "addal  d0,a0 ; addal d0,a1; movel a0,d1\n\t"
        "lsrl   #1,d1 ; jcc 1f ; moveb a1@-,a0@- ; subqw #1,d0\n\t"
        "lsrl   #1,d1 ; jcs 2f\n\t"  /* a0 increased=>bit 2 switched*/
        "movew  a1@-,a0@-  ; subqw  #2,d0 ; jra 2f\n\t"
     "1: lsrl   #1,d1 ; jcc 2f\n\t"
        "movew  a1@-,a0@-  ; subqw  #2,d0\n\t"
     "2: movew d0,sp@-; lsrl #2,d0 ; jeq 6f\n\t"
        "lsrl   #1,d0 ; jcc 3f ; movel a1@-,a0@-\n\t"        
     "3: lsrl   #1,d0 ; jcc 4f ; movel a1@-,a0@- ; movel a1@-,a0@-\n\t"
     "4: subql  #1,d0 ; jcs 6f\n\t"
          "5: movel a1@-,a0@-;movel a1@-,a0@-\n\t"
             "movel a1@-,a0@-;movel a1@-,a0@-\n\t"
             "dbra d0,5b ; subl #65536,d0; jcc 5b\n\t"
     "6: movew sp@+,d0; btst #1,d0 ; jeq 7f ; movew a1@-,a0@-\n\t"
     "7:              ; btst #0,d0 ; jeq 8f ; moveb a1@-,a0@-\n\t"
     "8:"
   :
        : "g" (d), "g" (s), "g" (count)
   : "a0", "a1", "d0", "d1");
      }
   }
   return(0);              
}
#else
#define mymemmove(d, s, count)       memmove((d),(s),(count))
#define mymemclear(d, count)         memset((d), 0 ,(count))
#define mymemclear_small(d, count)   memset((d), 0 ,(count))
#endif

/* ====================================================================== */

/* amicon_XXX routines - interface used by the world
 *
 * This system is now divided into two levels because of complications
 * caused by hardware scrolling. Top level functions:
 *
 *    amicon_bmove(), amicon_clear(), amicon_putc()
 *
 * handles y values in range [0, scr_height-1] that correspond to real
 * screen positions. y_wrap shift means that first line of bitmap may be
 * anywhere on this display. These functions convert lineoffsets to
 * bitmap offsets and deal with the wrap-around case by splitting blits.
 *
 *    amicon_bmove_physical_8()   -- These functions fast implementations
 *    amicon_clear_physical_8()   -- of original amicon_XXX fns.
 *    amicon_putc_physical_8()    -- (fontwidth  != 8) may be added later
 *
 * WARNING:
 *
 * At the moment amicon_putc() cannot blit across vertical wrap boundary
 * Implies should only really hardware scroll in rows. Only reason for
 * restriction is simplicity & efficiency at the moment.
 *
*/

static __inline__ int real_y_8(register struct display *p, int y)
{
   int y_bottom_size = p->y_wrap >> 3;
   int y_top_size    = (p->scr_height - p->y_wrap) >> 3;

   return( (y < y_top_size) ? (y + y_bottom_size) : (y-y_top_size) );
}

static void amicon_bmove_physical_8 (register struct display *p,
                                    int sy, int sx, int dy, int dx,
                                    int height, int width)
{

    if (sx == 0 && sy == 0 && width == p->bytes_per_row) {
        register ushort i;
        for (i = 0 ; i < p->scr_depth ; i++)
            mymemmove(p->bitplane[i] + dy * width * p->fontheight,
                      p->bitplane[i] + sy * width * p->fontheight,
                      width * height * p->fontheight);
    } else {
        register ushort rows, i;
        register u_char *src;
        register u_char *dst;
        register ushort bytes      = p->bytes_per_row;
        register ushort linesize   = bytes * p->fontheight;
        
        if (dy <= sy) {
            for (i = 0 ; i < p->scr_depth ; i++) {
                src  = p->bitplane[i] + sy * linesize + sx;
                dst  = p->bitplane[i] + dy * linesize + dx;
                for (rows = height*p->fontheight ; rows-- ;) {
                    mymemmove(dst, src, width);
                    src += bytes;
                    dst += bytes;
                }
            }
        } else {
            for (i = 0 ; i < p->scr_depth ; i++) {
                src  = p->bitplane[i] + (sy+height) * linesize - bytes + sx;
                dst  = p->bitplane[i] + (dy+height) * linesize - bytes + dx;

                for (rows = height*p->fontheight ; rows-- ;) {
                    mymemmove(dst, src, width);
                    src -= bytes;
                    dst -= bytes;
                }
            }
        }
    }
}

static int amicon_bmove (struct condata *conp,
          int sy, int sx, int dy, int dx,
          int height, int width)
{
    int unit=0;
/*    int unit = conp - vc_cons;*/
    struct display *p = &disp[unit];


    if (p->fontwidth == 8) {
        /* Split blits that cross physical y_wrap case.
         * Pathological case involves 4 blits, better to use recursive
         * code rather than unrolled case
        */

        ushort y_break = (p->scr_height - p->y_wrap) >> 3;

        if (sy < y_break && (sy+height-1) >= y_break) {
            ushort b = y_break - sy;
            if (dy < sy) {  /* Avoid trashing self */
                amicon_bmove(conp, sy,   sx, dy,   dx, b,        width);
                amicon_bmove(conp, sy+b, sx, dy+b, dx, height-b, width);
            } else {
                amicon_bmove(conp, sy+b, sx, dy+b, dx, height-b, width);
                amicon_bmove(conp, sy,   sx, dy,   dx, b,        width);
            }
            return(0);
        }
       
        if (dy < y_break && (dy+height-1) >= y_break) {
            ushort b = y_break - dy;
            if (dy < sy) {  /* Avoid trashing self */
                amicon_bmove(conp, sy,   sx, dy,   dx, b,        width);
                amicon_bmove(conp, sy+b, sx, dy+b, dx, height-b, width);
            } else {
                amicon_bmove(conp, sy+b, sx, dy+b, dx, height-b, width);
                amicon_bmove(conp, sy,   sx, dy,   dx, b,        width);
            }
            return(0);
        }
        amicon_bmove_physical_8(p, real_y_8(p,sy),   sx,
                                   real_y_8(p,dy),   dx,
                                   height,           width);
    } else {
        /* No support (fontwidth != 8) yet */
    }

    return(0);
}

static void amicon_clear_physical_8 (register struct display *p,
                                     int sy, int sx, int height, int width)
{
    ulong offset;
    register u_char *start;
    register ushort i, rows;
    register ushort bytes = p->bytes_per_row;
    register ulong size;

    if (sx == 0 && width == bytes) {
        offset = sy     * bytes * p->fontheight;
        size   = height * bytes * p->fontheight;
        
        for (i = 0 ; i < p->scr_depth ; i++)
            mymemclear(p->bitplane[i]+offset, size);
    } else {
        offset = (sy * bytes * p->fontheight) + sx;
        
        for (i = 0 ; i < p->scr_depth ; i++) {
            start = p->bitplane[i]+offset;
                
            for (rows = height*p->fontheight; rows-- ; start += bytes)
                mymemclear_small(start, width);
        }
    }
}

static int amicon_clear (struct condata *conp,
                         int sy, int sx, int height, int width)
{
    int unit = 0;
    /*int unit = conp - vc_cons;*/
    register struct display *p = &disp[unit];

    if (p->fontwidth == 8) {
        /* Split blits that cross physical y_wrap boundary */

        ushort y_break = (p->scr_height - p->y_wrap) >> 3;

        if (sy < y_break && (sy+height-1) >= y_break) {
            ushort b = y_break - sy;
            amicon_clear_physical_8(p, real_y_8(p,sy),   sx,
                                       b,                width);
            amicon_clear_physical_8(p, real_y_8(p,sy+b), sx,
                                       height-b,         width);

        } else
            amicon_clear_physical_8(p, real_y_8(p,sy),   sx,
                                       height,           width);
    } else {
        /* No support (fontwidth != 8) yet */
    }

    return(0);
}

static void amicon_putc_physical_8 (struct display *p,
                                    int c, int y, int x, int attr)
{
    register u_char   *dest;
    register u_char   *cdat;
    register ushort   rows;
    register ushort   i;
    u_short           bytes;
    ulong             offset;
        
    if ((c >= 32) && (c < 127))
   /* lower half of charset -- normal ASCII */
   c -=  32;
    else if ((c >= 160) && (c < 255))
   /* upper half of charset -- special characters */
   c -= 64;
    else
   /* not defined in charset -- non-printable character */
   return;

    bytes  = p->bytes_per_row;
    offset = y * p->fontheight * bytes + x;
    dest  = p->bitplane[0] + offset;
    cdat  = p->fontdata + (c * p->fontheight);

    /* Arbitary decision to write chars as colour 1 for the moment,
     * regardless of number of bitplanes. Later will write/clear
     * based on colour bitmask.
     */

    for (rows = p->fontheight ; rows-- ; dest += bytes)
      if (attr & 8)
	*dest = ~*cdat++;
      else
	*dest = *cdat++;

    for (i = 1 ; i < p->scr_depth ; i++) {
        dest = p->bitplane[i] + offset;
        for (rows = p->fontheight ; rows-- ; dest += bytes)
            *dest = (u_char)0;
    }
}

static int amicon_putc (struct condata *conp, int c, int y, int x)
{
    int unit = 0;
    /*int unit = conp - vc_cons;*/
    register struct display *p = &disp[unit];

    if (p->fontwidth == 8) {
        amicon_putc_physical_8(p, c, real_y_8(p, y) , x, conp->vc_attr);
    } else {
        /* Currently no support for (fontwidth != 8) */
    }

    return(0);
}

static void amicon_putcs_physical_8 (struct display *p, const char *s,
                                     int count, int y, int x, int attr)
{
    register u_char   *dest, *cdat;
    register ushort   rows, bytes;
    register u_char   c;
    u_char   *dest0;
    ulong    offset;
    ushort   i;

    bytes  = p->bytes_per_row;
    offset = y * p->fontheight * bytes + x;

    for (i = 1 ; i < p->scr_depth ; i++) {
        dest = p->bitplane[i] + offset;
        for (rows = p->fontheight ; rows-- ; dest += bytes)
            mymemclear_small(dest, count);
    }

    dest0  = p->bitplane[0] + offset;

#if defined (__GNUC__) && defined (__mc68000__)
    if (p->fontheight == 8) {
        /* Ultra special support for 8x8 fonts, as its nice and easy
         * and the console driver seems to spend most of its time in
         * here. Would take another 4 instructions to support,
         * (fontheight != 8) but I have an inbuilt suspicion of mulu's.
         * A hangover from my 68000 days, I guess.
         *
         * Both code and register use mirrors the C code directly below:
         *
         * a0 -- dest & dest0  [uses stack] 
         * a1 -- cdat     ; a2 -- s      ; a3 -- fontdata
         * d0 -- scratch  ; d1 -- count  ; d2 -- bytes
         * 
         */
	if (!(attr & 0x8))
	  __asm__("movel  %1,d1   ; beq 5f      ; subqw #1,d1\n\t"
		  "movel  %0,a0   ; movel %2,a2\n\t"
		  "movel %3,a3;movel %4,d2\n\t"
		  "1: moveb  a2@+,d0\n\t"
		  "cmpb   #32,d0  ; jlt 2f ; cmpb #127,d0 ; jge 2f\n\t"
		  "subb   #32,d0  ; jra 3f\n\t"
		  "2: cmpb   #160,d0 ; jlt 1b ; cmpb #255,d0 ; jeq 1b\n\t"
		  "subb   #64,d0\n\t"
		  "3: lea    a3@(d0:w:8),a1 ; movel a0,sp@-\n\t"
		  " moveb  a1@+,a0@; lea a0@(d2:w),a0\n\t" /* 0 */
		  " moveb  a1@+,a0@; lea a0@(d2:w),a0\n\t" /* 1 */
		  " moveb  a1@+,a0@; lea a0@(d2:w),a0\n\t" /* 2 */
		  " moveb  a1@+,a0@; lea a0@(d2:w),a0\n\t" /* 3 */
		  " moveb  a1@+,a0@; lea a0@(d2:w),a0\n\t" /* 4 */
		  " moveb  a1@+,a0@; lea a0@(d2:w),a0\n\t" /* 5 */
		  " moveb  a1@+,a0@; lea a0@(d2:w),a0\n\t" /* 6 */
		  " moveb  a1@+,a0@; lea a0@(d2:w),a0\n\t" /* 7 */
		  "movel  sp@+,a0 ; lea a0@(1),a0   ; dbra d1,1b\n\t"
		  "5:"
		  :
		  : "g" (dest0), "g" (count), "g" (s),
		  "g" (p->fontdata), "g" ((ulong) bytes)
		  : "a0", "a1", "a2", "a3", "d0", "d1", "d2");
	else
	  __asm__("movel  %1,d1   ; beq 5f      ; subqw #1,d1\n\t"
		  "movel  %0,a0   ; movel %2,a2\n\t"
		  "movel %3,a3;movel %4,d2\n\t"
		  "1: moveb  a2@+,d0\n\t"
		  "cmpb   #32,d0  ; jlt 2f ; cmpb #127,d0 ; jge 2f\n\t"
		  "subb   #32,d0  ; jra 3f\n\t"
		  "2: cmpb   #160,d0 ; jlt 1b ; cmpb #255,d0 ; jeq 1b\n\t"
		  "subb   #64,d0\n\t"
		  "3: lea    a3@(d0:w:8),a1 ; movel a0,sp@-\n\t"
		  " moveb  a1@+,d3; notb d3; moveb d3,a0@; lea a0@(d2:w),a0\n\t" /* 0 */
		  " moveb  a1@+,d3; notb d3; moveb d3,a0@; lea a0@(d2:w),a0\n\t" /* 1 */
		  " moveb  a1@+,d3; notb d3; moveb d3,a0@; lea a0@(d2:w),a0\n\t" /* 2 */
		  " moveb  a1@+,d3; notb d3; moveb d3,a0@; lea a0@(d2:w),a0\n\t" /* 3 */
		  " moveb  a1@+,d3; notb d3; moveb d3,a0@; lea a0@(d2:w),a0\n\t" /* 4 */
		  " moveb  a1@+,d3; notb d3; moveb d3,a0@; lea a0@(d2:w),a0\n\t" /* 5 */
		  " moveb  a1@+,d3; notb d3; moveb d3,a0@; lea a0@(d2:w),a0\n\t" /* 6 */
		  " moveb  a1@+,d3; notb d3; moveb d3,a0@; lea a0@(d2:w),a0\n\t" /* 7 */
		  "movel  sp@+,a0 ; lea a0@(1),a0   ; dbra d1,1b\n\t"
		  "5:"
		  :
		  : "g" (dest0), "g" (count), "g" (s),
		  "g" (p->fontdata), "g" ((ulong) bytes)
		  : "a0", "a1", "a2", "a3", "d0", "d1", "d2", "d3");
	return;
    }
#endif

    printk (__FILE__ ": Line %d shouldn't be executed\n", __LINE__);

    while (count--) {
        c    = *(s++); dest = dest0++;
            
        /* Arbitary decision to write chars as colour 1 for the moment,
         * regardless of number of bitplanes. Later will write/clear
         * based on colour bitmask.
         */

        if ((c >= 32) && (c < 127))
            /* lower half of charset -- normal ASCII */
            c -=  32;
        else if ((c >= 160) && (c < 255))
            /* upper half of charset -- special characters */
            c -= 64;
        else
            continue;

        cdat  = p->fontdata + (c * p->fontheight);
        for (rows = p->fontheight ; rows-- ; dest += bytes)
	  if (!(attr & 8))
	    *dest = *cdat++;
	  else
	    *dest = ~*cdat++;
    }
}

static int amicon_putcs (struct condata *conp, const char *s,
                         int count, int y, int x)
{
    int unit = 0;
    /*int unit = conp - vc_cons;*/
    register struct display *p = &disp[unit];

    if (p->fontwidth == 8) {
        amicon_putcs_physical_8(p, s, count, real_y_8(p, y) , x, conp->vc_attr);
    } else {
        /* Currently no support for (fontwidth != 8) */
    }

    return(0);
}

static int amicon_cursor (struct condata *conp,
           int mode)
{
   int unit=0;
   /*int unit = conp - vc_cons;*/
   struct display *p = &disp[unit];

   if (mode == CM_ERASE)
		p->cursor_visible = 0;
   else {
		p->cursor_visible = 1;
		p->cursor_latch = 0;
		p->cursor_x     = (ushort) conp->vc_x;
		p->cursor_y     = (ushort) conp->vc_y;
		p->cursor_latch = 1;
   }

    return 0;
}

static int amicon_scroll (struct condata *conp,
                          int t, int b, int dir, int count)
{
    int unit=0;
    /*int unit                   = conp - vc_cons;*/
    register struct display *p = &disp[unit];

    amicon_cursor (conp, CM_ERASE);

    switch (dir) {
    case SM_UP:
        if (t == 0 && b == conp->vc_rows) {
            /* Can hardware scroll */

            p->scroll_latch = 0;
            if (count > conp->vc_rows)         /* Maximum realistic size */
                count = conp->vc_rows;
            
            p->y_wrap += p->fontheight * count;
            if (p->y_wrap > p->scr_height)     /* Deal with wrap */
                p->y_wrap -= p->scr_height;

            p->scroll_latch = 1;
        } else
            amicon_bmove(conp,t + count, 0, t, 0,
                         (b-t-count), p->bytes_per_row);

   amicon_clear(conp,b-count,0,count,p->bytes_per_row);
        break;

    case SM_DOWN:
        if (t == 0 && b == conp->vc_rows) {
            /* Can hardware scroll */

            p->scroll_latch = 0;
            if (count > conp->vc_rows)         /* Maximum realistic size */
                count = conp->vc_rows;

            if (((short)p->y_wrap - p->fontheight * count) < 0)
                p->y_wrap += p->scr_height;    /* y_wrap is U_short */

            p->y_wrap -= p->fontheight * count;
            p->scroll_latch = 1;
        } else
            amicon_bmove(conp,t,0,t+count,0,
                         (b-t-count), p->bytes_per_row);

        /* Fixed bmove() should end Arno's frustration with copying?
         * Confusius say:
         *    Man who copies in wrong direction, end up with trashed data
        */

   amicon_clear(conp, t, 0, count, p->bytes_per_row);
        break;

    case SM_LEFT:
    case SM_RIGHT:
        /* XXX Nothing here yet XXX */

        break;
    }

    return 0;
}

static int amicon_switch (struct condata *conp)
{
	if (!(conp->vc_flags & CON_INITED))
		return -1;

	if (fb_info && fb_info->switch_con)
		(*fb_info->switch_con) (conp-vc_cons);
	return 0;
}

static int amicon_blank (int blank)
{
	(*fb_info->blank)(blank);
	return 0;
}


/*
 * The console "switch" structure for the Amiga native console
 */

struct consw ami_con =
{
    amicon_init, amicon_deinit, amicon_clear, amicon_putc, amicon_putcs,
    amicon_cursor, amicon_scroll, amicon_bmove, amicon_switch, amicon_blank
};
