/*
 * atari/atafb.c -- Low level implementation of Atari frame buffer device
 * amiga/amifb.c -- Low level implementation of Amiga frame buffer device
 *
 *  Copyright (C) 1994 Martin Schaller & Roman Hodek & Geert Uytterhoeven
 *  
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 *
 * History:
 *   - 03 Jan 95: Original version my Martin Schaller: The TT driver and
 *                all the device independant stuff
 *   - 09 Jan 95: Roman: I've added the hardware abstraction (hw_switch)
 *                and wrote the Falcon, ST(E), and External drivers
 *                based on the original TT driver.
 *   - 26 Jan 95: Geert: Amiga version
 *   - 19 Feb 95: Hamish: added Jes Sorensen's ECS patches to the Amiga
 *		  frame buffer device.  This provdes ECS support and the
 *		  following screen-modes: multiscan, multiscan-lace,
 * 		  super72, super72-lace, dblntsc, dblpal & euro72.
 *		  He suggests that we remove the old AGA screenmodes,
 *		  as they are non-standard, and some of them doesn't work
 *		  under ECS.
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/amigahw.h>
#include <linux/bootinfo.h>
#include <linux/tty.h>
#include <linux/malloc.h>
#include <linux/delay.h>
#include <linux/config.h>
#include <linux/interrupt.h>
#include <linux/amigaints.h>
#include <asm/segment.h>
#include <asm/system.h>
#include "amifb.h"


/*
 * A bunch of register definitions. Doing some things in terms of bitfields
 * might be nice eventually.
 */

/* BPLCON0 defines */
#define	BPC0_HIRES		0x8000
#define	BPC0_BPU3		0x0010	/* AGA */
#define	BPC0_BPU2		0x4000	/* Bit plane used count */
#define	BPC0_BPU1		0x2000
#define	BPC0_BPU0		0x1000
#define	BPC0_HAM		0x0800	/* HAM mode */
#define	BPC0_DPF		0x0400	/* Double playfield */
#define	BPC0_COLOR		0x0200	/* Enable colorburst */
#define	BPC0_GAUD		0x0100	/* Genlock audio enable */
#define	BPC0_UHRES		0x0080	/* Ultrahi res enable */
#define	BPC0_SHRES		0x0040	/* Super hi res mode */
#define	BPC0_BYPASS		0x0020	/* Bypass LUT - AGA */
#define	BPC0_LPEN		0x0008	/* Light pen enable */
#define	BPC0_LACE		0x0004	/* Interlace */
#define	BPC0_ERSY		0x0002	/* External resync */
#define	BPC0_ECSENA		0x0001	/* ECS emulation disable */

/* BPLCON1 controlls horizontal scrolling */

#define	BPC2_KILLEHB		0x0200	/* Kill EHB mode - AGA */
#define	BPC2_RDRAM		0x0100	/* Color table accesses read, not write - AGA */
#define	BPC2_SOGEN		0x0080	/* SOG output pin high - AGA */
#define	BPC2_PF2PRI		0x0040	/* PF2 priority over PF1 */
#define	BPC2_PF2P2		0x0020	/* PF2 priority wrt sprites */
#define	BPC2_PF2P1		0x0010
#define	BPC2_PF2P0		0x0008
#define	BPC2_PF1P2		0x0004	/* ditto PF1 */
#define	BPC2_PF1P1		0x0002
#define	BPC2_PF1P0		0x0001

/* BPLCON3 defines - AGA */
#define	BPC3_BANK2		0x8000	/* Bits to select color register bank */
#define	BPC3_BANK1		0x4000
#define	BPC3_BANK0		0x2000
#define	BPC3_PF2OF2		0x1000	/* Bits for color table offset when PF2 */
#define	BPC3_PF2OF1		0x0800
#define	BPC3_PF2OF0		0x0400
#define	BPC3_LOCT		0x0200	/* Color register writes go to low bits */
#define	BPC3_SPRES1		0x0080	/* Sprite resolution bits */
#define	BPC3_SPRES0		0x0040
#define	BPC3_BRDRBLNK		0x0020	/* Border blanked? */
#define	BPC3_BRDRTRAN		0x0010	/* Border transparent? */
#define	BPC3_BRDRSPRT		0x0002	/* Sprites in border? */
#define	BPC3_EXTBLKEN		0x0001	/* BLANK programmable */

/* BEAMCON0 */
#define	BMC0_HARDDIS	0x4000	/* Disable hardware limits */
#define	BMC0_LPENDIS	0x2000	/* Disable light pen latch */
#define	BMC0_VARVBEN	0x1000	/* Enable variable vertical blank */
#define	BMC0_LOLDIS	0x0800	/* Disable long/short line toggle */
#define	BMC0_CSCBEN	0x0400	/* Composite sync/blank */
#define	BMC0_VARVSYEN	0x0200	/* Enable variable vertical sync */
#define	BMC0_VARHSYEN	0x0100	/* Enable variable horizontal sync? */
#define	BMC0_VARBEAMEN	0x0080	/* Enable variable beam counters */
#define	BMC0_PAL	0x0020	/* Set decodes for PAL? */
#define	BMC0_VARCSYEN	0x0010	/* Enable variable composite sync? */
#define	BMC0_BLANKEN	0x0008	/* Blank enable? */
#define	BMC0_CSYTRUE	0x0004	/* CSY polarity */
#define	BMC0_VSYTRUE	0x0002	/* VSY polarity */
#define	BMC0_HSYTRUE	0x0001	/* HSY polarity */

/* FMODE - AGA */
#define	FMODE_SSCAN2	0x8000	/* Sprite scan-doubling */
#define	FMODE_BSCAN2	0x4000	/* Use PF2 modulus every other line */
#define	FMODE_SPAGEM	0x0008	/* Sprite page mode */
#define	FMODE_SPR32	0x0004	/* Sprite 32 bit fetch */
#define	FMODE_BPAGEM	0x0002	/* Bitplane page mode */
#define	FMODE_BPL32	0x0001	/* Bitplane 32 bit fetch */


#define CUSTOM_OFS(fld) ((long)&((struct CUSTOM*)0)->fld)


/*
** Externals
*/

extern unsigned char
    fontdata[];

extern int
    fontheight,
    fontwidth;

extern char
    fontname[];

static struct display disp[NR_CONSOLES];
static struct fb_info fb_info;

static int node;			/* node of the /dev/fb?current file */

static struct amiga_fb_par {
	u_long smem_start;
	u_long smem_len;
	struct geometry *geometry;
} current_par;


static ushort cursor_data[] =
{
    0x2c81,0x2d00,
    0xf000,0x0000,
    0x0000,0x0000
};


/*
 *  Color definitions
 */

#define FG_COLOR		(0x000000)	/* black */
#define BG_COLOR		(0xaaaaaa)	/* lt. grey */
#define CRSR_COLOR		(0xff0000)	/* bright red */

#define FG_COLOR_INV		BG_COLOR
#define BG_COLOR_INV		FG_COLOR
#define CRSR_COLOR_INV		(0x4477aa)	/* a blue-ish color */

/*
 *  Split 24-bit RGB colors in 12-bit MSB (for OCS/ECS/AGA) and LSB (for AGA)
 */

#define COLOR_MSB(rgb)	(((rgb>>12)&0xf00)|((rgb>>8)&0x0f0)|((rgb>>4)&0x00f))
#define COLOR_LSB(rgb)	(((rgb>>8) &0xf00)|((rgb>>4)&0x0f0)|((rgb)   &0x00f))

/* Cursor definitions */

#define CRSR_FLASH		1	/* Cursor flashing on(1)/off(0) */
#define CRSR_RATE		24	/* Number of frames/flash toggle */
#define CRSR_BLOCK		1	/* Block(1) or line(0) cursor */


/* controlling screen blanking (read in VBL handler) */
static int do_blank;
static int do_unblank;
static unsigned short save_bplcon3;

/*
 * ecs_color_zero is used to keep custom.color[0] for the special ECS color-
 * table, as custom.color[0] is cleared at vblank interrupts.
 * -Jes (jds@kom.auc.dk)
 */

static ushort ecs_color_zero;

static struct {
	int right_count;
	int done;
} vblank;


static __inline__ void init_vblank(void)
{
	vblank.right_count = 0;
	vblank.done = 0;
}

/* Geometry structure contains all useful information about given mode.
 *
 * Strictly speaking `scr_max_height' and `scr_max_width' is redundant
 * information with DIWSTRT value provided. Might be useful if modes
 * can be hotwired by user in future. It fits for the moment.
 *
 * At the moment, the code only distinguishes between OCS and AGA. ECS
 * lies somewhere in between - someone more familiar with it could make
 * appropriate modifications so that some advanced display modes are
 * available, without confusing the poor chipset. OCS modes use only the
 * bplcon0, diwstrt, diwstop, ddfstrt, ddfstop registers (a few others could
 * be used as well). -wjr
 *
 * The code now supports ECS aswell, except for FMODE all control registers
 * are the same under ECS. A special color-table has to be generated though.
 * -Jes
 */
struct geometry {
    char *modename;	/* Name this thing */
    char isOCS;		/* Is it OCS or ECS/AGA */
    ushort bplcon0;	/* Values for bit plane control register 0 */
    ushort scr_width;
    ushort scr_height;
    ushort scr_depth;
    ushort scr_max_width;     /* Historical, might be useful still */
    ushort scr_max_height;
    ushort diwstrt_h;	/* Where the display window starts */
    ushort diwstrt_v;
    ushort alignment;	/* Pixels per scanline must be a multiple of this */
    /* OCS doesn't need anything past here */
    ushort bplcon2;
    ushort bplcon3;
    /* The rest of these control variable sync */
    ushort htotal;	/* Total hclocks */
    ushort hsstrt;	/* HSYNC start and stop */
    ushort hsstop;
    ushort hbstrt;	/* HBLANK start and stop */
    ushort hbstop;
    ushort vtotal;	/* Total vlines */
    ushort vsstrt;	/* VSYNC, VBLANK ditto */
    ushort vsstop;
    ushort vbstrt;
    ushort vbstop;
    ushort hcenter;	/* Center of line, for interlaced modes */
    ushort beamcon0;	/* Beam control */
    ushort fmode;	/* Memory fetch mode */
};

#define MAX_COP_LIST_ENTS 64
#define COP_MEM_REQ       (MAX_COP_LIST_ENTS*4*2)
#define SPR_MEM_REQ       (24)


static struct geometry modes[] = {
	/* NTSC modes: !!! can't guarantee anything about overscan modes !!! */
	{
		"ntsc-lace", 1,
		BPC0_HIRES | BPC0_LACE,
		640, 400, 1,
		704, 480,
		0x71, 0x18,		/* diwstrt h,v */
		16			/* WORD aligned */
	}, {
		"ntsc", 1,
		BPC0_HIRES,
		640, 200, 1,
		704, 240,
		0x71, 0x18,
		16			/* WORD aligned */
	}, {
		"ntsc-lace-over", 1,
		BPC0_HIRES | BPC0_LACE,
		704, 480, 1,
		704, 480,
		0x71, 0x18,
		16			/* WORD aligned */
	}, {
		"ntsc-over", 1,
		BPC0_HIRES,
		704, 240, 1,
		704, 240,
		0x71, 0x18,
		16			/* WORD aligned */
	},
	/* PAL modes. Warning:
	 * 
	 * PAL overscan causes problems on my machine because maximum diwstop_h
	 * value seems to be ~0x1c2, rather than 0x1e0+ inferred by RKM 1.1
	 * and 0x1d5 inferred by original `amicon.c' source. Is this a hardware
	 * limitation of OCS/pal or 1084?. Or am I doing something stupid here?
	 *
	 * Included a couple of overscan modes that DO work on my machine,
	 * although not particularly useful.
	 */
	{
		"pal-lace", 1,
		BPC0_HIRES | BPC0_LACE,
		640, 512, 1,
		704, 592,
		0x71, 0x18,
		16			/* WORD aligned */
	}, {
		"pal", 1,
		BPC0_HIRES,
		640, 256, 1,
		704, 296,
		0x71, 0x18,
		16			/* WORD aligned */
	}, {
		"pal-lace-over", 1,
		BPC0_HIRES | BPC0_LACE,
		704, 592, 1,
		704, 582,
		0x5b, 0x18,
		16			/* WORD aligned */
	}, {
		"pal-over", 1,
		BPC0_HIRES,
		704, 296, 1,
		704, 296,
		0x5b, 0x18,
		16			/* WORD aligned */

	},
	/* ECS modes, these are real ECS modes */
	{
		"multiscan", 0,
		BPC0_SHRES | BPC0_ECSENA,	/* bplcon0 */
		640, 480, 1,
		640, 480,
		0x0041, 0x002c,			/* diwstrt h,v */
		64,	        		/* 64-bit aligned */
		BPC2_KILLEHB,			/* bplcon2 */
		BPC3_PF2OF1 | BPC3_PF2OF0 | BPC3_SPRES1 | BPC3_SPRES0 |
			BPC3_BRDRBLNK | BPC3_EXTBLKEN,	/* bplcon3 */

		0x0072,				/* htotal */
		0x000a,				/* hsstrt */
		0x0013,				/* hsstop */
		0x0002,				/* hbstrt */
		0x001c,				/* hbstop */
		0x020c,				/* vtotal */
		0x0008,				/* vsstrt */
		0x0011,				/* vsstop */
		0x0000,				/* vbstrt */
		0x001c,				/* vbstop */
		0x0043,				/* hcenter */
		BMC0_VARVBEN | BMC0_LOLDIS | BMC0_VARVSYEN |
			BMC0_VARHSYEN | BMC0_VARBEAMEN | BMC0_BLANKEN,
						/* beamcon0 */
		FMODE_BPAGEM | FMODE_BPL32	/* fmode */
	},
	{
		"multiscan-lace", 0,
		BPC0_SHRES | BPC0_LACE | BPC0_ECSENA,	/* bplcon0 */
		640, 960, 1,
		640, 960,
		0x0041, 0x002c,			/* diwstrt h,v */
		64,	        		/* 64-bit aligned */
		BPC2_KILLEHB,			/* bplcon2 */
		BPC3_PF2OF1 | BPC3_PF2OF0 | BPC3_SPRES1 | BPC3_SPRES0 |
			BPC3_BRDRBLNK | BPC3_EXTBLKEN,	/* bplcon3 */

		0x0072,				/* htotal */
		0x000a,				/* hsstrt */
		0x0013,				/* hsstop */
		0x0002,				/* hbstrt */
		0x001c,				/* hbstop */
		0x020c,				/* vtotal */
		0x0008,				/* vsstrt */
		0x0011,				/* vsstop */
		0x0000,				/* vbstrt */
		0x001c,				/* vbstop */
		0x0043,				/* hcenter */
		BMC0_VARVBEN | BMC0_LOLDIS | BMC0_VARVSYEN |
			BMC0_VARHSYEN | BMC0_VARBEAMEN | BMC0_BLANKEN,
						/* beamcon0 */
		FMODE_BPAGEM | FMODE_BPL32	/* fmode */
	},
	/* Super 72 - 800x300 72Hz noninterlaced mode. */
	{
		"super72", 0,
		BPC0_SHRES | BPC0_ECSENA,	/* bplcon0 */
		800, 304, 1,			/* need rows%8 == 0 */
		800, 304,			/* (cols too) */
		0x0051, 0x0021,			/* diwstrt h,v */
		64,				/* 64-bit aligned */
		BPC2_KILLEHB,			/* bplcon2 */
		BPC3_PF2OF1 | BPC3_PF2OF0 | BPC3_SPRES1 | BPC3_SPRES0 |
			BPC3_BRDRBLNK | BPC3_EXTBLKEN,	/* bplcon3 */
		0x0091,				/* htotal */
		0x000a,				/* hsstrt */
		0x0013,				/* hsstop */
		0x0001,				/* hbstrt */
		0x001e,				/* hbstop */
		0x0156,				/* vtotal */
		0x0009,				/* vsstrt */
		0x0012,				/* vsstop */
		0x0000,				/* vbstrt */
		0x001c,				/* vbstop */
		0x0052,				/* hcenter */
		BMC0_VARVBEN | BMC0_LOLDIS | BMC0_VARVSYEN |
			BMC0_VARHSYEN | BMC0_VARBEAMEN | BMC0_BLANKEN,
						/* beamcon0 */
		FMODE_BPAGEM | FMODE_BPL32	/* fmode */
	},
	/* Super 72 lace - 800x600 72Hz interlaced mode. */
	{
		"super72-lace", 0,
		BPC0_SHRES | BPC0_LACE | BPC0_ECSENA,	/* bplcon0 */
		800, 600, 1,			/* need rows%8 == 0 */
		800, 600,			/* (cols too) */
		0x0051, 0x0021,			/* diwstrt h,v */
		64,     			/* 64-bit aligned */
		BPC2_KILLEHB,			/* bplcon2 */
		BPC3_PF2OF1 | BPC3_PF2OF0 | BPC3_SPRES1 | BPC3_SPRES0 |
			BPC3_BRDRBLNK | BPC3_EXTBLKEN,
						/* bplcon3 */
		0x0091,				/* htotal */
		0x000a,				/* hsstrt */
		0x0013,				/* hsstop */
		0x0001,				/* hbstrt */
		0x001e,				/* hbstop */
		0x0150,				/* vtotal */
		0x0009,				/* vsstrt */
		0x0012,				/* vsstop */
		0x0000,				/* vbstrt */
		0x001c,				/* vbstop */
		0x0052,				/* hcenter */
		BMC0_VARVBEN | BMC0_LOLDIS | BMC0_VARVSYEN |
			BMC0_VARHSYEN | BMC0_VARBEAMEN | BMC0_BLANKEN,
						/* beamcon0 */
		FMODE_BPAGEM | FMODE_BPL32	/* fmode */
	},
	/* DblNtsc - 640x400 59Hz noninterlaced mode. */
	{
		"dblntsc", 0,
		BPC0_SHRES | BPC0_ECSENA,	/* bplcon0 */
		640, 400, 1,			/* need rows%8 == 0 */
		640, 400,			/* (cols too) */
		0x0049, 0x0021,			/* diwstrt h,v */
		64,				/* 64-bit aligned */
		BPC2_KILLEHB,			/* bplcon2 */
		BPC3_PF2OF1 | BPC3_PF2OF0 | BPC3_SPRES1 | BPC3_SPRES0 |
			BPC3_BRDRBLNK | BPC3_EXTBLKEN,
						/* bplcon3 */
		0x0079,				/* htotal */
		0x0007,				/* hsstrt */
		0x0013,				/* hsstop */
		0x0001,				/* hbstrt */
		0x001e,				/* hbstop */
		0x01ec,				/* vtotal */
		0x0008,				/* vsstrt */
		0x0010,				/* vsstop */
		0x0000,				/* vbstrt */
		0x0019,				/* vbstop */
		0x0046,				/* hcenter */
		BMC0_VARVBEN | BMC0_LOLDIS | BMC0_VARVSYEN |
			BMC0_VARHSYEN | BMC0_VARBEAMEN | BMC0_BLANKEN,
						/* beamcon0 */
		FMODE_BPAGEM | FMODE_BPL32	/* fmode */
	},
	/* DblPal - 640x512 52Hz noninterlaced mode. */
	{
		"dblpal", 0,
		BPC0_SHRES | BPC0_ECSENA,	/* bplcon0 */
		640, 512, 1,			/* need rows%8 == 0 */
		640, 512,			/* (cols too) */
		0x0049, 0x0021,			/* diwstrt h,v */
		64,				/* 64-bit aligned */
		BPC2_KILLEHB,			/* bplcon2 */
		BPC3_PF2OF1 | BPC3_PF2OF0 | BPC3_SPRES1 | BPC3_SPRES0 |
			BPC3_BRDRBLNK | BPC3_EXTBLKEN,
						/* bplcon3 */
		0x0079,				/* htotal */
		0x0007,				/* hsstrt */
		0x0013,				/* hsstop */
		0x0001,				/* hbstrt */
		0x001e,				/* hbstop */
		0x0234,				/* vtotal */
		0x0008,				/* vsstrt */
		0x0010,				/* vsstop */
		0x0000,				/* vbstrt */
		0x0019,				/* vbstop */
		0x0046,				/* hcenter */
		BMC0_VARVBEN | BMC0_LOLDIS | BMC0_VARVSYEN |
			BMC0_VARHSYEN | BMC0_VARBEAMEN | BMC0_BLANKEN,
						/* beamcon0 */
		FMODE_BPAGEM | FMODE_BPL32	/* fmode */
	},
	/* Euro72 - productivity - 640x400 71Hz noninterlaced mode. */
	{
		"euro72", 0,
		BPC0_SHRES | BPC0_ECSENA,	/* bplcon0 */
		640, 400, 1,			/* need rows%8 == 0 */
		640, 400,			/* (cols too) */
		0x0041, 0x0021,			/* diwstrt h,v */
		64,				/* 64-bit aligned */
		BPC2_KILLEHB,			/* bplcon2 */
		BPC3_PF2OF1 | BPC3_PF2OF0 | BPC3_SPRES1 | BPC3_SPRES0 |
			BPC3_BRDRBLNK | BPC3_EXTBLKEN,
						/* bplcon3 */
		0x0071,				/* htotal */
		0x0009,				/* hsstrt */
		0x0013,				/* hsstop */
		0x0001,				/* hbstrt */
		0x001e,				/* hbstop */
		0x01be,				/* vtotal */
		0x0008,				/* vsstrt */
		0x0016,				/* vsstop */
		0x0000,				/* vbstrt */
		0x001f,				/* vbstop */
		0x0041,				/* hcenter */
		BMC0_VARVBEN | BMC0_LOLDIS | BMC0_VARVSYEN |
			BMC0_VARHSYEN | BMC0_VARBEAMEN | BMC0_BLANKEN,
						/* beamcon0 */
		FMODE_BPAGEM | FMODE_BPL32	/* fmode */
	},
	/* AGA modes */
	{
	/*
	 * A 640x480, 60Hz noninterlaced AGA mode. It would be nice to be
	 * able to have some of these values computed dynamically, but that
	 * requires more knowledge of AGA than I have. At the moment,
	 * the values make it centered on my 1960 monitor. -wjr
	 *
	 * For random reasons to do with the way arguments are parsed,
	 * these names can't start with a digit.
	 *
	 * Don't count on being able to reduce scr_width and scr_height
	 * and ending up with a smaller but well-formed screen - this
	 * doesn't seem to work well at the moment.
	 */
		"aga640x480", 0,
		BPC0_SHRES | BPC0_ECSENA,			/* bplcon0 */
		640, 480, 1,
		640, 480,
		0x0041, 0x002b,						/* diwstrt h,v */
		64,										/* 64-bit aligned */
		BPC2_KILLEHB,							/* bplcon2 */
		BPC3_PF2OF1 | BPC3_PF2OF0 | BPC3_SPRES1 | BPC3_SPRES0 |
		BPC3_BRDRBLNK | BPC3_EXTBLKEN,	/* bplcon3 */
		0x0071,									/* htotal */
		0x000c,									/* hsstrt */
		0x001c,									/* hsstop */
		0x0008,									/* hbstrt */
		0x001e,									/* hbstop */
		0x020c,									/* vtotal */
		0x0001,									/* vsstrt */
		0x0003,									/* vsstop */
		0x0000,									/* vbstrt */
		0x000f,									/* vbstop */
		0x0046,									/* hcenter */
		BMC0_HARDDIS | BMC0_VARVBEN | BMC0_LOLDIS | BMC0_VARVSYEN |
			BMC0_VARHSYEN | BMC0_VARBEAMEN | BMC0_BLANKEN,	/* beamcon0 */
		FMODE_BPAGEM | FMODE_BPL32			/* fmode */
	}, {
		/* An 800x600 72Hz interlaced mode. */
		"aga800x600", 0,
		BPC0_SHRES | BPC0_LACE | BPC0_ECSENA,	/* bplcon0 */
		896, 624, 1,							/* need rows%8 == 0 */
		896, 624,								/* (cols too) */
		0x0041, 0x001e,						/* diwstrt h,v */
		64,						/* 64-bit aligned */
		BPC2_KILLEHB,							/* bplcon2 */
		BPC3_PF2OF1 | BPC3_PF2OF0 | BPC3_SPRES1 | BPC3_SPRES0 |
		BPC3_BRDRBLNK | BPC3_EXTBLKEN,	/* bplcon3 */
		0x0091,									/* htotal */
		0x000e,									/* hsstrt */
		0x001d,									/* hsstop */
		0x000a,									/* hbstrt */
		0x001e,									/* hbstop */
		0x0156,									/* vtotal */
		0x0001,									/* vsstrt */
		0x0003,									/* vsstop */
		0x0000,									/* vbstrt */
		0x000f,									/* vbstop */
		0x0050,									/* hcenter */
		BMC0_HARDDIS | BMC0_VARVBEN | BMC0_LOLDIS | BMC0_VARVSYEN |
		BMC0_VARHSYEN | BMC0_VARBEAMEN | BMC0_BLANKEN,	/* beamcon0 */
		FMODE_BPAGEM | FMODE_BPL32			/* fmode */
	},
	/*
	 * Additional AGA modes by Geert Uytterhoeven
	 */
	{
		/*
		 * A 720x400, 70 Hz noninterlaced AGA mode (29.27 kHz)
		 */
		"aga720x400", 0,
		BPC0_SHRES | BPC0_ECSENA,			/* bplcon0 */
		720, 400, 1,
		720, 400,
		0x0041, 0x0013,						/* diwstrt h,v */
		64,										/* 64-bit aligned */
		BPC2_KILLEHB,							/* bplcon2 */
		BPC3_PF2OF1 | BPC3_PF2OF0 | BPC3_SPRES1 | BPC3_SPRES0 |
		BPC3_BRDRBLNK | BPC3_EXTBLKEN,	/* bplcon3 */
		0x0079,									/* htotal */
		0x000e,									/* hsstrt */
		0x0018,									/* hsstop */
		0x0001,									/* hbstrt */
		0x0021,									/* hbstop */
		0x01a2,									/* vtotal */
		0x0003,									/* vsstrt */
		0x0005,									/* vsstop */
		0x0000,									/* vbstrt */
		0x0012,									/* vbstop */
		0x0046,									/* hcenter */
		BMC0_VARVBEN | BMC0_LOLDIS | BMC0_VARVSYEN | BMC0_VARHSYEN |
		BMC0_VARBEAMEN | BMC0_PAL | BMC0_VARCSYEN | BMC0_CSYTRUE |
		BMC0_VSYTRUE,							/* beamcon0 */
		FMODE_BPAGEM | FMODE_BPL32			/* fmode */
	}, {
		/*
		 * A 640x400, 76 Hz noninterlaced AGA mode (31.89 kHz)
		 */
		"aga640x400", 0,
		BPC0_SHRES | BPC0_ECSENA,			/* bplcon0 */
		640, 400, 1,
		640, 400,
		0x0041, 0x0015,						/* diwstrt h,v */
		64,										/* 64-bit aligned */
		BPC2_KILLEHB,							/* bplcon2 */
		BPC3_PF2OF1 | BPC3_PF2OF0 | BPC3_SPRES1 | BPC3_SPRES0 |
		BPC3_BRDRBLNK | BPC3_EXTBLKEN,	/* bplcon3 */
		0x006f,									/* htotal */
		0x000d,									/* hsstrt */
		0x0018,									/* hsstop */
		0x0001,									/* hbstrt */
		0x0021,									/* hbstop */
		0x01a4,									/* vtotal */
		0x0003,									/* vsstrt */
		0x0005,									/* vsstop */
		0x0000,									/* vbstrt */
		0x0014,									/* vbstop */
		0x0046,									/* hcenter */
		BMC0_VARVBEN | BMC0_LOLDIS | BMC0_VARVSYEN | BMC0_VARHSYEN |
		BMC0_VARBEAMEN | BMC0_PAL | BMC0_VARCSYEN | BMC0_CSYTRUE |
		BMC0_VSYTRUE,							/* beamcon0 */
		FMODE_BPAGEM | FMODE_BPL32			/* fmode */
	}, {
		/*
		 * A 640x480, 64 Hz noninterlaced AGA mode (31.89 kHz)
		 */
		"aga640x480a", 0,
		BPC0_SHRES | BPC0_ECSENA,			/* bplcon0 */
		640, 480, 1,
		640, 480,
		0x0041, 0x0015,						/* diwstrt h,v */
		64,										/* 64-bit aligned */
		BPC2_KILLEHB,							/* bplcon2 */
		BPC3_PF2OF1 | BPC3_PF2OF0 | BPC3_SPRES1 | BPC3_SPRES0 |
		BPC3_BRDRBLNK | BPC3_EXTBLKEN,	/* bplcon3 */
		0x006f,									/* htotal */
		0x000e,									/* hsstrt */
		0x0018,									/* hsstop */
		0x0001,									/* hbstrt */
		0x0021,									/* hbstop */
		0x01f4,									/* vtotal */
		0x0003,									/* vsstrt */
		0x0005,									/* vsstop */
		0x0000,									/* vbstrt */
		0x0014,									/* vbstop */
		0x0046,									/* hcenter */
		BMC0_VARVBEN | BMC0_LOLDIS | BMC0_VARVSYEN | BMC0_VARHSYEN |
		BMC0_VARBEAMEN | BMC0_PAL | BMC0_VARCSYEN | BMC0_CSYTRUE |
		BMC0_VSYTRUE,							/* beamcon0 */
		FMODE_BPAGEM | FMODE_BPL32			/* fmode */
	}
};

#define	NMODES	(sizeof(modes) / sizeof(struct geometry))

struct fb_var_screeninfo amiga_fb_predefined[] = {
	{ /* autodetect */
		0, 0, 0, 0, 0, 0, 0, 0,   		/* xres-grayscale */
		{0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, 	/* red green blue tran*/
		0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0 }
};

int num_amiga_fb_predefined= sizeof(amiga_fb_predefined)/sizeof(struct fb_var_screeninfo);



/* Some default modes */
#define OCS_PAL_LOWEND_DEFMODE	5	/* PAL non-laced for 500/2000 */
#define OCS_PAL_3000_DEFMODE		4	/* PAL laced for 3000 */
#define OCS_NTSC_LOWEND_DEFMODE	1	/* NTSC non-laced for 500/2000 */
#define OCS_NTSC_3000_DEFMODE		0	/* NTSC laced for 3000 */
#define AGA_DEFMODE					8	/* 640x480 non-laced for AGA */

static int amicon_inverse = 0;
static int amicon_mode = -1;

void video_setup (char *options, int *ints)
{
	char *this_opt;
	int i;

	if (!options || !*options)
		return;

	for (this_opt = strtok(options,","); this_opt; this_opt = strtok(NULL,","))
		if (!strcmp (this_opt, "inverse"))
			amicon_inverse = 1;
		else
			for (i = 0; i < NMODES; i++)
				if (!strcmp(this_opt, modes[i].modename)) {
					amicon_mode = i;
					break;
				}
}

/* Notes about copper scrolling:
 *
 * 1. The VBLANK routine dynamically rewrites a LIVE copper list that is
 *    currently being executed. Don't mess with it unless you know the
 *    complications. Fairly sure that double buffered lists doesn't
 *    make our life any easier.
 *
 * 2. The vblank code starts executing at logical line 0. Display must be
 *    set up and ready to run by line DIWSTRT_V, typically 0x2c, minimum
 *    value is 0x18 for maximum overscan.
 *
 *    Tests on my A500/030 for dynamically generating a 37 element copper
 *    list during the VBLANK period under AmigaDos required between
 *    0x10 and 0x14 scanlines. This should be pathological case, and
 *    should do better under Linux/68k. It is however IMPERATIVE that I am
 *    first in the VBLANK isr chain. Try to keep 'buildclist' as fast as
 *    possible. Don't think that it justifies assembler thou'
 *
 * 3. PAL 640x256 display (no overscan) has copper-wait y positions in range
 *    0x02c -> 0x12c. NTSC overscan uses values > 256 too. However counter
 *    is 8 bit, will wrap. RKM 1.1 suggests use of a WAIT(0x00,0xff),
 *    WAIT(x,y-0x100) pair to handle this case. This is WRONG - must use
 *    WAIT(0xe2,0xff) to ensure that wrap occures by next copper
 *    instruction. Argghh!
 *
 * 4. RKM 1.1 suggests Copper-wait x positions are in range [0,0xe2].
 *    Horizontal blanking occurs in range 0x0f -> 0x35. Black screen
 *    shown in range 0x04 -> 0x47.
 *
 *    Experiments suggest that using WAIT(0x00,y), we can replace up to
 *    7 bitplane pointers before display fetch start. Using a 
 *    WAIT(0xe0,y-1) instead, we can replace 8 pointers that should be
 *    all that we need for a full AGA display. Should work because of
 *    fetch latency with bitmapped display.
 *
 *    I think that this works. Someone please tell me if something breaks.
 *
 * Is diwstop_h the right value to use for "close to the end of line"?
 * It seems to work for me, at least for the modes I've defined. -wjr
 *
 * I changed the Wait(diwstop_h, 0xff) for 256-line chunk skipping to
 * Wait(diwstop_h-2, 0xff) to make it work with the additional
 * `get-all-you-can-get-out-of-it' AGA modes. Maybe we should derive the
 * wait position from the HTOTAL value? - G.U.
 *
 * The Wait(diwstop_h-2, 0xff) didn't work in Super72 under ECS, instead
 * I changed it to Wait(htotal-4, 0xff). Dunno whether it works under AGA,
 * and don't ask my why it works. I'm trying to get some facts on this issue
 * from Commodore.
 * -Jes
 */

static __inline__ ushort *build_clist_hdr(register struct display *p,
														ushort *cop,
														ushort *othercop)	/* Interlace: List for next frame */
{
	int i;
	ushort diwstrt_v = p->diwstrt_v;
	ushort diwstop_h = p->diwstop_h;

	if (othercop) {
		*cop++ = CUSTOM_OFS(cop1lc);
		*cop++ = (long)othercop >> 16;
		*cop++ = CUSTOM_OFS(cop1lc) + 2;
		*cop++ = (long)othercop;
	}

	/* Point Sprite 0 at cursor sprite: */
	*cop++ = CUSTOM_OFS(sprpt[0]);
	*cop++ = (ushort)((long)p->cursor >> 16);
	*cop++ = CUSTOM_OFS(sprpt[0]) + 2;
	*cop++ = (ushort)((long)p->cursor & 0x0000ffff);

	/* Point Sprites 1-7 at dummy sprite: */
	for (i=1; i<8; i++) {
		*cop++ = CUSTOM_OFS(sprpt[i]);
		*cop++ = (ushort)((long)p->dummy >> 16);
		*cop++ = CUSTOM_OFS(sprpt[i]) + 2;
		*cop++ = (ushort)((long)p->dummy & 0x0000ffff);
	}

	/* Halt copper until we have rebuilt the display list */

	*cop++ = ((diwstrt_v - 2) << 8) | (diwstop_h >> 1) | 0x1;
	*cop++ = 0xfffe;

	return(cop);
}

static __inline__ ushort *build_clist_dyn(register struct display *p,
														ushort *cop,
														int shf)				/* Interlace: Short frame */
{
	ushort diwstrt_v = p->diwstrt_v;
	ushort diwstop_h = p->diwstop_h;
	ushort y_wrap = p->y_wrap;
	ulong offset = y_wrap * p->bytes_per_row;
	long scrmem;
	int i;

	/* Set up initial bitplane ptrs */

	for (i = 0 ; i < p->scr_depth ; i++) {
		scrmem    = ((long)p->bitplane[i]) + offset;

		if (shf)
			scrmem += p->bytes_per_row;

		*cop++ = CUSTOM_OFS(bplpt[i]);
		*cop++ = (long)scrmem >> 16;
		*cop++ = CUSTOM_OFS(bplpt[i]) + 2;
		*cop++ = (long)scrmem;
	}

	/* If wrapped frame needed - wait for line then switch bitplXs */

	if (y_wrap) {
		ushort line;
        
		if (p->bplcon0 & BPC0_LACE)
			line = diwstrt_v + (p->scr_height - y_wrap)/2;
		else
			line = diwstrt_v + p->scr_height - y_wrap;

		/* Handle skipping over 256-line chunks */
		while (line > 256) {
			/* Hardware limitation - 8 bit counter    */
			/* Wait(diwstop_h-2, 0xff) */
			if (p->bplcon0 & BPC0_SHRES)
				/*
				 * htotal-4 is used in SHRES-mode, as diwstop_h-2 doesn't work under ECS.
				 * Does this work under AGA?
				 * -Jes
				 */
				*cop++ = 0xff00 | ((p->htotal-4) | 1);
			else 
				*cop++ = 0xff00 | ((diwstop_h-2) >> 1) | 0x1;

			*cop++ = 0xfffe;
			/* Wait(0, 0) - make sure we're in the new segment */
			*cop++ = 0x0001;
			*cop++ = 0xfffe;
			line -= 256;

			/*
			 * Under ECS we have to keep color[0], as it is part of a special color-table.
			 */

			if (boot_info.bi_amiga.chipset == CS_ECS && p->bplcon0 & BPC0_ECSENA) {
				*cop++ = 0x0180;
				*cop++ = ecs_color_zero;
			}
		}
	
		/* Wait(diwstop_h, line - 1) */
		*cop++ = ((line - 1)   << 8) | (diwstop_h >> 1) | 0x1;
		*cop++ = 0xfffe;

		for (i = 0 ; i < p->scr_depth ; i++) {
			scrmem = (long)p->bitplane[i];
			if (shf)
				scrmem += p->bytes_per_row;

				*cop++ = CUSTOM_OFS(bplpt[i]); 
				*cop++ = (long)scrmem >> 16;
				*cop++ = CUSTOM_OFS(bplpt[i]) + 2;
				*cop++ = (long)scrmem;
		}
	}
    
	/* End of Copper list */
	*cop++ = 0xffff;
	*cop++ = 0xfffe;

	return(cop);
}


static __inline__ void build_cursor(register struct display *p)
{
	int vs, hs, ve;
	ushort diwstrt_v = p->diwstrt_v;
	ushort diwstrt_h = p->diwstrt_h;

	if (p->bplcon0 & BPC0_LACE) {
		vs = diwstrt_v + (p->cursor_y * p->fontheight)/2; 
		ve = vs + p->fontheight/2;
	} else {
		vs = diwstrt_v + (p->cursor_y * p->fontheight); 
		ve = vs + p->fontheight;
	}

	if (p->bplcon0 & BPC0_ECSENA)
		/*
		 * It's an AGA mode. We'll assume that the sprite was set
		 * into 35ns resolution by the appropriate SPRES bits in bplcon3.
		 */
		hs = diwstrt_h  * 4 + (p->cursor_x * p->fontwidth) - 4;
	else
		hs = diwstrt_h + (p->cursor_x * p->fontwidth) / 2 - 1;

	if (p->bplcon0 & BPC0_ECSENA) {
		/* There are some high-order bits on the sprite position */
		*((ulong *) p->cursor) =
		((((vs & 0xff) << 24) | ((vs & 0x100) >> 6) |
		((vs & 0x200) >> 3)) |
		(((hs & 0x7f8) << 13) | ((hs & 0x4) >> 2) |
		((hs & 0x3) << 3)) |
		(((ve & 0xff) << 8) | ((ve & 0x100) >> 7) |
		((ve & 0x200) >> 4)));
	} else {
		*((ulong *) p->cursor) =
		((vs << 24) | ((vs & 0x00000100) >> 6) |
		((hs & 0x000001fe) << 15) | (hs & 0x00000001) | 
		((ve & 0x000000ff) << 8) | ((ve & 0x00000100) >> 7));
	}
}

static void build_ecs_colors(ushort color1, ushort color2, ushort color3, 
                      ushort color4, ushort *table)
{
/*
 * This function calculates the special ECS color-tables needed when running
 * new screen-modes available under ECS. See the hardware reference manual
 * 3rd edition for details.
 * -Jes
 */
ushort  t;

        t = (color1 & 0x0ccc);
        table[0] = t;
        table[4] = t;
        table[8] = t;
        table[12] = t;
        t = t >> 2;
        table[0] = (table[0] | t);
        table[1] = t;
        table[2] = t;
        table[3] = t;

        t = (color2 & 0x0ccc);
        table[1] = (table[1] | t);
        table[5] = t;
        table[9] = t;
        table[13] = t;
        t = t >> 2;
        table[4] = (table[4] | t);
        table[5] = (table[5] | t);
        table[6] = t;
        table[7] = t;

        t = (color3 & 0x0ccc);
        table[2] = (table[2] | t);
        table[6] = (table[6] | t);
        table[10] = t;
        table[14] = t;
        t = t >> 2;
        table[8] = (table[8] | t);
        table[9] = (table[9] | t);
        table[10] = (table[10] | t);
        table[11] = t;

        t = (color4 & 0x0ccc);
        table[3] = (table[3] | t);
        table[7] = (table[7] | t);
        table[11] = (table[11] | t);
        table[15] = t;
        t = t >> 2;
        table[12] = (table[12] | t);
        table[13] = (table[13] | t);
        table[14] = (table[14] | t);
        table[15] = (table[15] | t);

}

/* display_init():
 *
 *    Fills out (struct display *) given a geometry structure
 */

static void display_init(struct display *p,
                         struct geometry *geom, ushort inverse)
{
	ushort ecs_table[16];
	int    i;
	char   *chipptr;
	ushort diwstrt_v, diwstop_v;
	ushort diwstrt_h, diwstop_h;
	ushort diw_min_h, diw_min_v;
	ushort bplmod, diwstrt, diwstop, diwhigh, ddfstrt, ddfstop;
	ushort cursorheight, cursormask = 0;
	u_long size;

	/* Decide colour scheme */

	if (inverse) {
		p->fgcol   = FG_COLOR_INV;
		p->bgcol   = BG_COLOR_INV;
		p->crsrcol = CRSR_COLOR_INV;
	} else {
		p->fgcol   = FG_COLOR;
		p->bgcol   = BG_COLOR;
		p->crsrcol = CRSR_COLOR;
	}

	/* Define screen geometry */
   
	p->scr_max_height = geom->scr_max_height;
	p->scr_max_width  = geom->scr_max_width; 
	p->scr_height     = geom->scr_height;
	p->scr_width      = geom->scr_width;
	p->scr_depth      = geom->scr_depth;
	p->fontdata       = fontdata;
	p->fontheight     = fontheight;
	p->fontwidth      = fontwidth;
	p->bplcon0        = geom->bplcon0 | BPC0_COLOR;
	p->htotal         = geom->htotal;

	/* htotal was added, as I use it to calc the pal-line. -Jes */

	if (p->scr_depth < 8)
		p->bplcon0 |= (p->scr_depth << 12);
	else {
		/* must be exactly 8 */
		p->bplcon0 |= BPC0_BPU3;
	}

	diw_min_v         = geom->diwstrt_v;
	diw_min_h         = geom->diwstrt_h;

	/* We can derive everything else from this, at least for OCS */
	/*
	 * For AGA: we don't use the finer position control available for
	 * diw* yet (could be set by 35ns increments).
	 */

	/* Calculate line and plane size while respecting the alignment restrictions */
	p->bytes_per_row  = ((p->scr_width+geom->alignment-1)&~(geom->alignment-1)) >> 3;
	p->plane_size     = p->bytes_per_row * p->scr_height;


	/*
	 *		Quick hack for frame buffer mmap():
	 *
	 *		plane_size must be a multiple of the page size
	 */

	p->plane_size = PAGE_ALIGN(p->plane_size);


	p->y_wrap   = 0;                  p->scroll_latch = 1;
	p->cursor_x = 0; p->cursor_y = 0; p->cursor_latch = 1;

	if (p->bplcon0 & BPC0_LACE) {
		bplmod = p->bytes_per_row;
		diwstrt_v = diw_min_v + (p->scr_max_height - p->scr_height)/4;
		diwstop_v = (diwstrt_v + p->scr_height/2);
	} else {
		bplmod = 0;
		diwstrt_v = diw_min_v + (p->scr_max_height - p->scr_height)/2;
		diwstop_v = (diwstrt_v + p->scr_height);
	}

	if (p->bplcon0 & BPC0_HIRES) {
		diwstrt_h =  diw_min_h + (p->scr_max_width - p->scr_width)/4;
		diwstop_h = (diwstrt_h + p->scr_width/2);
		/* ??? Where did 0x1d5 come from in original code ??? */
	} else if (p->bplcon0 & BPC0_SHRES) {
		diwstrt_h =  diw_min_h + (p->scr_max_width - p->scr_width)/8;
		diwstop_h = (diwstrt_h + p->scr_width/4);
	} else {
		diwstrt_h =  diw_min_h + (p->scr_max_width - p->scr_width)/2;
		diwstop_h = (diwstrt_h + p->scr_width);
	}

	if (p->bplcon0 & BPC0_HIRES) {
		ddfstrt = (diwstrt_h >> 1) - 4;
		ddfstop = ddfstrt + (4 * (p->bytes_per_row>>1)) - 8;
	} else if (p->bplcon0 & BPC0_SHRES && boot_info.bi_amiga.chipset == CS_AGA) {
		/* There may be some interaction with FMODE here... -8 is magic. */

		/*
		 * This should be fixed, so it supports all different
		 * FMODE's.  FMODE varies the speed with 1,2 & 4 the
		 * standard ECS speed.  Someone else has to do it, as
		 * I don't have an AGA machine with MMU available
		 * here.
		 *
		 * This particular speed looks like FMODE = 3 to me.
		 * ddfstop should be changed so it depends on FMODE under AGA.
		 * -Jes
		 */
		ddfstrt = (diwstrt_h >> 1) - 8;
		ddfstop = ddfstrt + (2 * (p->bytes_per_row>>1)) - 8;
	} else if (p->bplcon0 & BPC0_SHRES && boot_info.bi_amiga.chipset == CS_ECS){
		/* 
		 * Normal speed for ECS, should be the same for FMODE = 0
		 * -Jes
		 */
		ddfstrt = (diwstrt_h >> 1) - 2;
		ddfstop = ddfstrt + (2 * (p->bytes_per_row>>1)) - 8;
	} else {
		ddfstrt = (diwstrt_h >> 1) - 8;
		ddfstop = ddfstrt + (8 * (p->bytes_per_row>>1)) - 8;
	}

	if (p->bplcon0 & BPC0_LACE)
		cursorheight = p->fontheight/2;
	else
		cursorheight = p->fontheight;

	/*
	 *		Quick hack for frame buffer mmap():
	 *
	 *		chipptr must be at a page boundary
	 */

	size = p->scr_depth*p->plane_size+COP_MEM_REQ+SPR_MEM_REQ+4*(cursorheight-1);
	size += PAGE_SIZE-1;
	chipptr = amiga_chip_alloc(size);
	chipptr = (char *)PAGE_ALIGN((u_long)chipptr);

   
	/* locate the bitplanes */
	/* These MUST be 64 bit aligned for full AGA compatibility!! */

	current_par.smem_start = (u_long)chipptr;
	current_par.smem_len = p->plane_size*p->scr_depth;
	current_par.geometry = geom;

	for (i = 0 ; i < p->scr_depth ; i++, chipptr += p->plane_size) {
		p->bitplane[i] = (u_char *) chipptr;
		memset ((void *)chipptr, 0, p->plane_size);  /* and clear */
	}

	/* locate the copper lists */
	p->coplist1hdr = (ushort *) chipptr;  chipptr += MAX_COP_LIST_ENTS * 4;
	p->coplist2hdr = (ushort *) chipptr;  chipptr += MAX_COP_LIST_ENTS * 4;

	/* locate the sprite data */
	p->cursor      = (ushort *) chipptr;  chipptr += 8+4*cursorheight;
	p->dummy       = (ushort *) chipptr;  chipptr += 12;

	/* create the sprite data for the cursor image */
	memset((void *)p->cursor, 0, 8+4*cursorheight);
	/*
	 * Only AGA supplies hires sprites.
	 */
	if (p->bplcon0 & BPC0_ECSENA && boot_info.bi_amiga.chipset == CS_AGA)
		/* AGA cursor is SHIRES, ECS sprites differ */
		for (i = 0; (i < p->fontwidth) && (i < 16); i++)
			cursormask |= 1<<(15-i);
	else
		/* For OCS & ECS sprites are pure LORES 8-< */
		for (i = 0; (i < p->fontwidth/2) && (i < 8); i++)
			cursormask |= 1<<(15-i);

	p->cursor[0] = cursor_data[0];
	p->cursor[1] = cursor_data[1];

#if (CRSR_BLOCK == 1)
	for (i = 0; i < cursorheight; i++)
#else
	for (i = cursorheight-2; i < cursorheight; i++)
#endif
		p->cursor[2+2*i] = cursormask;

	/* set dummy sprite data to a blank sprite */
	memset((void *)p->dummy, 0, 12);
  
	/* set cursor flashing */
	p->cursor_flash = CRSR_FLASH;

	/* Make the cursor visible */
	p->cursor_visible = 1;
 
	/* Initialise the chipregs */
	p->diwstrt_v = diwstrt_v;
	p->diwstrt_h = diwstrt_h;
	p->diwstop_v = diwstop_v;
	p->diwstop_h = diwstop_h;
	diwstrt = ((diwstrt_v << 8) | diwstrt_h);
	diwstop = ((diwstop_v & 0xff) << 8) | (diwstop_h & 0xff);

	custom.bplcon0   = p->bplcon0;	/* set the display mode */
	custom.bplcon1   = 0;		/* Needed for horizontal scrolling */
	custom.bplcon2   = 0;
	custom.bpl1mod   = bplmod;
	custom.bpl2mod   = bplmod;
	custom.diwstrt   = diwstrt;
	custom.diwstop   = diwstop;
	custom.ddfstrt   = ddfstrt;
	custom.ddfstop   = ddfstop;

	custom.color[0]  = COLOR_MSB(p->bgcol);
	custom.color[1]  = COLOR_MSB(p->fgcol);
	custom.color[17] = COLOR_MSB(p->crsrcol); /* Sprite 0 color */

	if (boot_info.bi_amiga.chipset == CS_AGA) {
		/* Fill in the LSB of the 24 bit color palette */
		/* Must happen after MSB */
		custom.bplcon3   = geom->bplcon3 | BPC3_LOCT;
		custom.color[0]  = COLOR_LSB(p->bgcol);
		custom.color[1]  = COLOR_LSB(p->fgcol);
		custom.color[17] = COLOR_LSB(p->crsrcol);
		custom.bplcon3   = geom->bplcon3;
	}

	if (boot_info.bi_amiga.chipset == CS_ECS && p->bplcon0 & BPC0_ECSENA) {
		/*
		 * Calculation of the special ECS color-tables for
		 * planes and sprites is done in the function
		 * build_ecs_table
		 */

		/*
		 * Calcs a special ECS colortable for the bitplane,
		 * and copies it to the custom registers
		 */
		build_ecs_colors(COLOR_MSB(p->bgcol), COLOR_MSB(p->fgcol),
				 0, 0, ecs_table); 

		for (i = 0; i < 8; i++){
			custom.color[i]   = ecs_table[i*2];
			custom.color[i+8] = ecs_table[i*2+1];
		}

		ecs_color_zero = ecs_table[0];

		/*
		 * Calcs a special ECS colortable for the cursor
		 * sprite, and copies it to the appropriate custom
		 * registers
		 */
		build_ecs_colors(0, COLOR_MSB(p->crsrcol), 0, 0, ecs_table);

		for (i = 0; i < 16; i++){
			custom.color[i+16] = ecs_table[i];
		}
	}

	if (!(geom->isOCS)) {
		/* Need to set up a bunch more regs */
		/* Assumes that diwstrt is in the (0,0) sector, but stop might not be */
		diwhigh = (diwstop_v & 0x700) | ((diwstop_h & 0x100) << 5);

		custom.bplcon2   = geom->bplcon2;
		custom.bplcon3   = geom->bplcon3;
		/* save bplcon3 for blanking */
		save_bplcon3 = geom->bplcon3;

		custom.diwhigh   = diwhigh;	/* must happen AFTER diwstrt, stop */

		custom.htotal	= geom->htotal;
		custom.hsstrt	= geom->hsstrt;
		custom.hsstop	= geom->hsstop;
		custom.hbstrt	= geom->hbstrt;
		custom.hbstop	= geom->hbstop;
		custom.vtotal	= geom->vtotal;
		custom.vsstrt	= geom->vsstrt;
		custom.vsstop	= geom->vsstop;
		custom.vbstrt	= geom->vbstrt;
		custom.vbstop	= geom->vbstop;
		custom.hcenter	= geom->hcenter;
		custom.beamcon0	= geom->beamcon0;
		if (boot_info.bi_amiga.chipset == CS_AGA) {
			custom.fmode	= geom->fmode;
		}
		/* 
		 * fmode does NOT! exist under ECS, weird things might happen
		 */

		/* We could load 8-bit colors here, if we wanted */
	}


	/* Build initial copper lists. sprites must be set up, and p->diwstrt. */

	if (p->bplcon0 & BPC0_LACE) {
		p->coplist1dyn = build_clist_hdr(p,p->coplist1hdr, p->coplist2hdr),
		build_clist_dyn(p, p->coplist1dyn, 0);

		p->coplist2dyn = build_clist_hdr(p,p->coplist2hdr, p->coplist1hdr),
		build_clist_dyn(p, p->coplist2dyn, 1);
	} else {
		p->coplist1dyn = build_clist_hdr(p,p->coplist1hdr, NULL),
		build_clist_dyn(p, p->coplist1dyn, 0);
	}


	/* Get ready to run first copper list */
	custom.cop1lc = p->coplist1hdr;
	custom.copjmp1 = 0;

	/* turn on DMA for bitplane and sprites */
	custom.dmacon = DMAF_SETCLR | DMAF_RASTER | DMAF_COPPER | DMAF_SPRITE;

	if (p->bplcon0 & BPC0_LACE) {
		/* Make sure we get the fields in the right order */

		/* wait for LOF frame bit to go low */
		while (custom.vposr & 0x8000)
			;

		/* wait for LOF frame bit to go high */
		while (!(custom.vposr & 0x8000))
			;

		/* start again at the beginning of copper list 1 */
		custom.cop1lc = p->coplist1hdr;
		custom.copjmp1 = 0;
	}
}


static void amicon_interrupt(struct intframe *fp, void *data)
{
	int unit = 0;
	register struct display *p = &disp[unit];

	static ushort cursorcount = 0;
	static ushort cursorstate = 0;

	/* I *think* that you should only change display lists on long frame.
	 * At least it goes awfully perculiar on my A500 without the following
	 * test. Not really in a position to test this hypothesis, so sorry
	 * for the slow scrolling, all you flicker-fixed souls
	 */

	if (!(p->bplcon0 & BPC0_LACE) || (custom.vposr & 0x8000)) {
		if (p->scroll_latch || p->cursor_latch)
			build_cursor(p);

		if (p->scroll_latch)
			if (p->bplcon0 & BPC0_LACE) {
				build_clist_dyn(p, disp[unit].coplist1dyn, 0);
				build_clist_dyn(p, disp[unit].coplist2dyn, 1);
			} else
				build_clist_dyn(p, disp[unit].coplist1dyn, 0);
			p->scroll_latch = 0;
			p->cursor_latch = 0;
	}

	if (!(custom.potgor & (1<<10)))
		vblank.right_count++;

	if (p->cursor_visible) {
		if (p->cursor_flash) {
			if (cursorcount)
				cursorcount--;
			else {
				cursorcount = CRSR_RATE;
				if ((cursorstate = !cursorstate))
					custom.dmacon = DMAF_SETCLR | DMAF_SPRITE;
				else
					custom.dmacon = DMAF_SPRITE;
			}
		}
	} else
		custom.dmacon = DMAF_SPRITE;

	if (do_blank) {
		custom.dmacon = DMAF_RASTER | DMAF_SPRITE;
		custom.color[0] = 0;
		if (boot_info.bi_amiga.chipset == CS_AGA) {
			/* Fill in the LSB of the 24 bit color palette */
			/* Must happen after MSB */
			custom.bplcon3 = save_bplcon3 | BPC3_LOCT;
			custom.color[0]= 0;
			custom.bplcon3 = save_bplcon3;
		}
		do_blank = 0;
	}

	if (do_unblank) {
		if (p->cursor_visible)
			custom.dmacon = DMAF_SETCLR | DMAF_RASTER | DMAF_SPRITE;
		else
			custom.dmacon = DMAF_SETCLR | DMAF_RASTER;
		custom.color[0] = COLOR_MSB(p->bgcol);
		if (boot_info.bi_amiga.chipset == CS_AGA) {
			/* Fill in the LSB of the 24 bit color palette */
			/* Must happen after MSB */
			custom.bplcon3 = save_bplcon3 | BPC3_LOCT;
			custom.color[0] = COLOR_LSB(p->bgcol);
			custom.bplcon3 = save_bplcon3;
		}
		/* color[0] is set to ecs_color_zero under ECS */
		if (boot_info.bi_amiga.chipset == CS_ECS && p->bplcon0 & BPC0_ECSENA) {
			custom.color[0]  = ecs_color_zero;
		}
		do_unblank = 0;
	}

	vblank.done = 1;
}

#ifdef DEBUG
static void __debug__print_fb_fix_screeninfo(struct fb_fix_screeninfo *fix)
{
	printk("  id = `%15s'  smem_start = 0x%08lx  smem_len = 0x%08lx\n",
			 fix->id, fix->smem_start, fix->smem_len);
	printk("  type = %5d  aux = %5d  visual = %5d\n",
			 fix->type, fix->type_aux, fix->visual);
}
#endif



static int amiga_fb_get_fix(struct fb_fix_screeninfo *fix, int con)
{
#ifdef DEBUG
	printk("amiga_fb_get_fix() {\n");
#endif

	strcpy(fix->id, current_par.geometry->modename);
	fix->smem_start = current_par.smem_start;
	fix->smem_len = current_par.smem_len;

	/*
	 *		Only monochrome bitmap at the moment
	 */

	fix->type = FB_TYPE_PACKED_PIXELS;

	fix->type_aux = 0;
	if (amicon_inverse)
		fix->visual = FB_VISUAL_MONO10;
	else
		fix->visual = FB_VISUAL_MONO01;


#ifdef DEBUG
	__debug__print_fb_fix_screeninfo(fix);
	printk("}\n");
#endif

	return(0);
}


#ifdef DEBUG
static void __debug__print_fb_var_screeninfo(struct fb_var_screeninfo *var)
{
	printk("  xres     =  %4d  yres     = %4d  xres_v = %4d  yres_v = %4d\n",
			  var->xres, var->yres, var->xres_virtual, var->yres_virtual);
	printk("  xoffset  =  %4d  yoffset  = %4d  bpp    = %4d  graysc = %4d\n",
			  var->xoffset, var->yoffset, var->bits_per_pixel, var->grayscale);
	printk("  nonstd   =  %4d  activate = %4d  height = %4d  width  = %4d  accel = %4d\n",
			  var->nonstd, var->activate, var->height, var->width, var->accel);
	printk("  pixclock = %5ld  left     = %4ld  right  = %4ld  upper  = %4ld  lower = %4ld\n",
			  var->pixclock, var->left_margin, var->right_margin, var->upper_margin, var->lower_margin);
	printk("  hsync    =  %4ld  vsync    = %4ld  sync   = %4d  vmode  = %4d\n",
			  var->hsync_len, var->vsync_len, var->sync, var->vmode);
}
#endif


static int amiga_fb_get_var(struct fb_var_screeninfo *var, int con)
{
#ifdef DEBUG
	printk("amiga_fb_get_var() {\n");
#endif

	var->xres = current_par.geometry->scr_width;
	var->yres = current_par.geometry->scr_height;
	var->xres_virtual = var->xres;
	var->yres_virtual = var->yres;
	var->xoffset = 0;
	var->yoffset = 0;

	var->bits_per_pixel = current_par.geometry->scr_depth;
	var->grayscale = 0;

	if (boot_info.bi_amiga.chipset == CS_AGA) {
		var->red.offset = 0;
		var->red.length = 8;
		var->red.msb_right = 0;
		var->green = var->red;
		var->blue = var->red;
	} else {
		var->red.offset = 0;
		var->red.length = 4;
		var->red.msb_right = 0;
		var->green = var->red;
		var->blue = var->red;
	}

	var->nonstd = 0;
	var->activate = 0;

	var->width = -1;
	var->height = -1;

	var->accel = FB_ACCEL_NONE;

	var->pixclock = 35242;
	var->left_margin = (current_par.geometry->hbstop-current_par.geometry->hsstrt)*8;
	var->right_margin = (current_par.geometry->hsstrt-current_par.geometry->hbstrt)*8;
	var->upper_margin = (current_par.geometry->vbstop-current_par.geometry->vsstrt)*8;
	var->lower_margin = (current_par.geometry->vsstrt-current_par.geometry->vbstrt)*8;
	var->hsync_len = (current_par.geometry->hsstop-current_par.geometry->hsstrt)*8;
	var->vsync_len = (current_par.geometry->vsstop-current_par.geometry->vsstrt)*8;
	var->sync = 0;
	if (current_par.geometry->bplcon0 & BPC0_LACE)
		var->vmode = FB_VMODE_INTERLACED;
	else if ((boot_info.bi_amiga.chipset == CS_AGA) && (current_par.geometry->fmode & FMODE_BSCAN2))
		var->vmode = FB_VMODE_DOUBLE;
	else
		var->vmode = FB_VMODE_NONINTERLACED;


#ifdef DEBUG
	__debug__print_fb_var_screeninfo(var);
	printk("}\n");
#endif


	return(0);
}


static int amiga_fb_set_var(struct fb_var_screeninfo *var, int con)
{
#ifdef DEBUG
	printk("amiga_fb_set_var() {\n");
	__debug__print_fb_var_screeninfo(var);
	printk("}\n");
#endif

	/*
	 *		Not yet implemented
	 */
	return 0;
	return(-EINVAL);
}


static short red_normal[] = {
	((BG_COLOR & 0xff0000)>>8) | ((BG_COLOR & 0xff0000)>>16),
	((FG_COLOR & 0xff0000)>>8) | ((FG_COLOR & 0xff0000)>>16)
};
static short green_normal[] = {
	((BG_COLOR & 0x00ff00)) | ((BG_COLOR & 0x00ff00)>>8),
	((FG_COLOR & 0x00ff00)) | ((FG_COLOR & 0x00ff00)>>8)
};
static short blue_normal[] = {
	((BG_COLOR & 0x0000ff)<<8) | ((BG_COLOR & 0x0000ff)),
	((FG_COLOR & 0x0000ff)<<8) | ((FG_COLOR & 0x0000ff))
};

static short red_inverse[] = {
	((BG_COLOR_INV & 0xff0000)>>8) | ((BG_COLOR_INV & 0xff0000)>>16),
	((FG_COLOR_INV & 0xff0000)>>8) | ((FG_COLOR_INV & 0xff0000)>>16)
};
static short green_inverse[] = {
	((BG_COLOR_INV & 0x00ff00)) | ((BG_COLOR_INV & 0x00ff00)>>8),
	((FG_COLOR_INV & 0x00ff00)) | ((FG_COLOR_INV & 0x00ff00)>>8)
};
static short blue_inverse[] = {
	((BG_COLOR_INV & 0x0000ff)<<8) | ((BG_COLOR_INV & 0x0000ff)),
	((FG_COLOR_INV & 0x0000ff)<<8) | ((FG_COLOR & 0x0000ff))
};

struct fb_cmap default_cmap_normal = { 0, 2, red_normal, green_normal, blue_normal, NULL };
struct fb_cmap default_cmap_inverse = { 0, 2, red_inverse, green_inverse, blue_inverse, NULL };

static int amiga_fb_get_cmap(struct fb_cmap *cmap, int kspc, int con)
{
	int i, start;
	unsigned short *red, *green, *blue, *transp;
	unsigned int hred, hgreen, hblue, htransp;
	struct fb_cmap *def_cmap;

#ifdef DEBUG
	printk("amiga_fb_get_cmap() {\n");
#endif


	red = cmap->red;
	green = cmap->green;
	blue = cmap->blue;
	transp = cmap->transp;
	start = cmap->start;
	if (start < 0) {


#ifdef DEBUG
		printk("  Error = EINVAL\n");
		printk("}\n");
#endif


		return(-EINVAL);
	}

	if (amicon_inverse)
		def_cmap = &default_cmap_inverse;
	else
		def_cmap = &default_cmap_normal;

	for (i = 0; i < cmap->len; i++) {
		if (i < def_cmap->len) {
			hred = def_cmap->red[i];
			hgreen = def_cmap->green[i];
			hblue = def_cmap->blue[i];
			if (def_cmap->transp)
				htransp = def_cmap->transp[i];
			else
				htransp = 0;
		} else
			hred = hgreen = hblue = htransp = 0;
		if (kspc) {
			*red = hred;
			*green = hgreen;
			*blue = hblue;
			if (transp)
				*transp = htransp;
		} else {
			put_user_word(hred, red);
			put_user_word(hgreen, green);
			put_user_word(hblue, blue);
			if (transp)
				put_user_word(htransp, transp);
		}
		red++;
		green++;
		blue++;
		if (transp)
			transp++;
	}


#ifdef DEBUG
	printk("}\n");
#endif


	return(0);
}


static int amiga_fb_set_cmap(struct fb_cmap *cmap, int kspc, int con)
{
#ifdef DEBUG
	printk("amiga_fb_set_cmap()\n");
#endif


	/*
	 *		Not yet implemented
	 */
	return(-EINVAL);
}


static int amiga_fb_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
								  unsigned long arg, int con)
{
#ifdef DEBUG
	printk("amiga_fb_ioctl()\n");
#endif


	return(-EINVAL);
}

static struct fb_ops amiga_fb_ops = {
	amiga_fb_get_fix, amiga_fb_get_var, amiga_fb_set_var, 
	amiga_fb_get_cmap, amiga_fb_set_cmap, amiga_fb_ioctl	
};













static int amifb_switch (int con)
{
	struct display *p = &disp[con];
	struct display *old = &disp[fg_console];

	p->bitplane[0] = old->bitplane[0];	/* syncing the base address */
	p->bitplane[1] = old->bitplane[1];	/* syncing the base address */
	p->bitplane[2] = old->bitplane[2];	/* syncing the base address */
	p->bitplane[3] = old->bitplane[3];	/* syncing the base address */
	p->bitplane[4] = old->bitplane[4];	/* syncing the base address */
	p->bitplane[5] = old->bitplane[5];	/* syncing the base address */
	p->bitplane[6] = old->bitplane[6];	/* syncing the base address */
	p->bitplane[7] = old->bitplane[7];	/* syncing the base address */

	/*
	 *		Reset hardware scrolling to make sure the display is continous
	 *		(X doesn't look very well on split displays :-)
	 */

	disp->y_wrap = 0;
	disp->scroll_latch = 1;

	return(0);
}


static int amifb_updatevar(int con)
{
	struct display *p = &disp[con];
	p->cursor_latch = 1;
	p->scroll_latch = 1;
	return(0);
}


static void amifb_blank(int blank)
{
	if (blank)
		do_blank = 1;
	else
		do_unblank = 1;
}


struct fb_info *amifb_init(long *mem_start)
{
	int mode = amicon_mode;
	ulong model;
	int inverse_video = amicon_inverse;
	int err;

	err=register_framebuffer("Amiga Builtin", &node, &amiga_fb_ops,  num_amiga_fb_predefined,
									 amiga_fb_predefined);

	model = boot_info.bi_un.bi_ami.model;
	if (mode == -1)
		if (boot_info.bi_amiga.chipset == CS_AGA)
			mode = AGA_DEFMODE;
		else if (model == AMI_3000)
			mode = (boot_info.bi_un.bi_ami.vblank == 50) ? OCS_PAL_3000_DEFMODE : OCS_NTSC_3000_DEFMODE;
		else
			mode = (boot_info.bi_un.bi_ami.vblank == 50) ? OCS_PAL_LOWEND_DEFMODE : OCS_NTSC_LOWEND_DEFMODE;

	init_vblank();
	display_init(disp, &modes[mode], inverse_video);
	if (!mach_add_isr(IRQ_AMIGA_VERTB, (isrfunc)amicon_interrupt, 0, NULL))
		panic("Couldn't add vblank interrupt");

#ifdef DEBUG
	printk("amifb_init: fired up mode %s\n", modes[mode].modename);
	printk("disp->diwstrt = (%x, %x), diwstop = (%x, %x)\n", disp->diwstrt_h, disp->diwstrt_v,
			 disp->diwstop_h, disp->diwstop_v);
#endif

	amiga_fb_get_var(&disp->var, 0);
	if (amicon_inverse)
		disp->cmap = default_cmap_inverse;
	else
		disp->cmap = default_cmap_normal;

	strcpy(fb_info.modename, "Amiga Builtin ");
	fb_info.disp = disp;
	fb_info.switch_con = &amifb_switch;
	fb_info.updatevar = &amifb_updatevar;
	fb_info.blank = &amifb_blank;	
	strcat(fb_info.modename, modes[mode].modename);
	return(&fb_info);
}
