/*
 * linux/amiga/amisound.c
 *
 * amiga sound driver for 680x0 Linux
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 */

#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/amigahw.h>
#include <linux/bootinfo.h>

#include <asm/system.h>

static u_short *snd_data = NULL;
static const signed char sine_data[] = {
	0,  39,  75,  103,  121,  127,  121,  103,  75,  39,
	0, -39, -75, -103, -121, -127, -121, -103, -75, -39
};
#define DATA_SIZE	(sizeof(sine_data)/sizeof(sine_data[0]))

#define NTSC_CLOCK	(3579545/DATA_SIZE)
#define NTSC_MIN_PERIOD	(124)

#define PAL_CLOCK	(3546895/DATA_SIZE)
#define PAL_MIN_PERIOD	(123)

#define MAX_PERIOD	(65535)

#define WIERD_PC_VAL	(1193180)

static u_short min_period;
static u_long clock_constant;

static void init_sound(void)
{
	snd_data = amiga_chip_alloc(sizeof(sine_data));
	if (!snd_data) {
		printk (KERN_CRIT "amiga init_sound: failed to allocate chipmem\n");
		return;
	}
	memcpy (snd_data, sine_data, sizeof(sine_data));

	/* setup min period and divisor */
	min_period = boot_info.bi_amiga.eclock == NTSC_ECLOCK ?
		NTSC_MIN_PERIOD : PAL_MIN_PERIOD;
	clock_constant = boot_info.bi_amiga.eclock == NTSC_ECLOCK ?
		NTSC_CLOCK : PAL_CLOCK;
}

static void nosound( unsigned long ignored );
static struct timer_list sound_timer = { NULL, NULL, 0, 0, nosound };

void amiga_mksound( unsigned int count, unsigned int ticks )
{
	static int inited = 0;
	unsigned long flags;

	if (!inited) {
		init_sound();
		inited = 1;
	}

	if (!snd_data)
		return;

	save_flags(flags);
	cli();
	del_timer( &sound_timer );

	if (count) {
		u_long period = (clock_constant * count) / WIERD_PC_VAL;

		if (period < min_period)
			period = min_period;
		if (period > MAX_PERIOD)
			period = MAX_PERIOD;

		/* setup pointer to data, period, length and volume */
		custom.aud[0].audlc = snd_data;
		custom.aud[0].audlen = sizeof(sine_data)/2;
		custom.aud[0].audper = (u_short)period;
		custom.aud[0].audvol = 64; /* maxvol */
	
		if (ticks) {
			sound_timer.expires = ticks;
			add_timer( &sound_timer );
		}

		/* turn on DMA for audio channel 0 */
		custom.dmacon = DMAF_SETCLR | DMAF_AUD0;

		restore_flags(flags);
		return;
	} else {
		nosound( 0 );
		restore_flags(flags);
		return;
	}
}


static void nosound( unsigned long ignored )
{
	/* turn off DMA for audio channel 0 */
	custom.dmacon = DMAF_AUD0;
}	
