/*
**  linux/atari/atacon.c
**
**  Copyright (C) 1993 Bjoern Brauel and Roman Hodek
**
**
**  11/06/94 Guenther Kelleter:
**  Added auto detection of screen resolution at boot time.
**
** This file is subject to the terms and conditions of the GNU General Public
** License.  See the file README.legal in the main directory of this archive
** for more details.
*/

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/tty.h>
#include <linux/console.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/bootinfo.h>
#include <linux/delay.h>

#include <linux/malloc.h>
#include <linux/mm.h>

#include <linux/atarihw.h>
#include <linux/atariints.h>

#include <asm/system.h>

#include "atafb.h"

/* The following symbols select what modes are supported. They should
 * be settable by the user ("make config") later.
 * True color (= 16 plane) mode isn't implemented yet!
 */
#define	CONFIG_ATARI_2PLANE
#define	CONFIG_ATARI_4PLANE
#define	CONFIG_ATARI_8PLANE
#undef	CONFIG_ATARI_16PLANE

/* If you want to see underlined characters in color modes, define the
 * following flag. But note that this effect cannot be restored when
 * switching VC's!
 */
#undef COLOR_UNDERLINING

/* To do:
 *
 *  - Do proper Falcon video mode setting, analogous to the TT (but
 *    much more and better modes possible!)
 *  - Implement 16 plane mode.
 *
 */


/*
** Macros
*/


/* Define if the font contains all 256 characters (undefined: only
 * 32..127 and 160..255, this is the font in font.c of the Amiga
 * distribution)
 */

#define	WHOLE_FONT


/* Import console_blanked from console.c */

extern int console_blanked;

struct fb_info *fb_info;
struct display *disp;

/* This macro should return true, if in this mode the
 * blanking/unblanking is done by blackening and redrawing the screen
 */
#define REDRAW_BLANK_MODE(cons)	(! disp[(cons)].wrclut)

/* FONT_CNV does the character conversion necessary if the font
 * doesn't contain all characters. It returns 0 if the characters is
 * not in the font (WHOLE_FONT undefined).
 * 
 * If we have a full font, FONT_CNV has to do the jobs of translating
 * the character to be printed according to the console's current
 * translation table. This is just a quick hack to enable the
 * different translations (normal, graphic, ...) that VT102 allows. It
 * should really be done in console.c, where it is commented out
 * currently, and I didn't want to interfere with the people
 * maintaining that code.
 *                                                             Roman
 */

/* ++andreas: Translation now done in console.c */

#ifndef WHOLE_FONT

#define	FONT_CNV(c) (((c) -= 32) >= 0 \
		     && ((c) <= 127 - 32 || ((c) -= 32) >= 160 - 64))

#else

#define	FONT_CNV(c) (1)

#endif


/*
** Local Variables
*/

struct display_switch {
	void (*bmove)( register struct display *p, int sy, int sx, int dy,
				   int dx, int height, int width);
	void (*clear)( struct condata *conp, register struct display *p,
				   int sy, int sx, int height, int width);
	void (*putc)( struct condata *conp, struct display *p, int c,
				  int y, int x);
	void (*putcs)( struct condata *conp, struct display *p,
				   const char *s, int count, int y, int x);
	void (*rev_char)( struct display *display, int x, int y );
};	

static int cursor_drawn = 0;

/* Time to elapse before the cursor is really drawn, in VBL units
 * (60 Hz: 17 ms, 72 Hz: 14 ms)
 */
#define	CURSOR_DRAW_DELAY	2

/* # VBL ints between cursor state changes */
#define	DEFAULT_CURSOR_BLINK_RATE	42

static int		vbl_cursor_cnt = 0;
static int		cursor_on = 0;
static int		cursor_blink_rate = DEFAULT_CURSOR_BLINK_RATE;

static __inline__ int
CURSOR_UNDRAWN (void)
{
  int cursor_was_drawn;
  vbl_cursor_cnt = 0;
  cursor_was_drawn = cursor_drawn;
  cursor_drawn = 0;
  return cursor_was_drawn;
}


/***************************** Prototypes *****************************/

static long atacon_init( struct condata *conp, long mem_start, char **display_desc);
static int atacon_deinit( struct condata *conp);
static __inline__ void *mymemclear_small( void * s, size_t count);
static __inline__ void *mymemclear( void * s, size_t count);
static __inline__ void *mymemmove( void * d, void * s, size_t count);
static __inline__ void memclear_4p_col( void *d, size_t h, u_long val,
                                        int bpr );
static __inline__ void memset_even_4p( void *d, size_t count, u_long val1,
                                       u_long val2 );
static __inline__ void memmove_4p_col (void *d, void *s, int h, int bpr);
static __inline__ unsigned long expand4l( unsigned char c );
static __inline__ void expand4dl( unsigned char c, unsigned long *ret1,
                                  unsigned long *ret2 );
static __inline__ unsigned long dup4l( unsigned char c );
static __inline__ void memclear_8p_col( void *d, size_t h, u_long val1,
                                        u_long val2, int bpr);
static __inline__ void memset_even_8p( void *d, size_t count, u_long val1,
                                       u_long val2, u_long val3, u_long
                                       val4 );
static __inline__ void memmove_8p_col (void *d, void *s, int h, int bpr);
static __inline__ void expand8dl( unsigned char c, u_long *ret1, u_long
                                  *ret2 );
static __inline__ void memclear_2p_col( void *d, size_t h, u_short val,
				       int bpr);
static __inline__ void memset_even_2p( void *d, size_t count, u_long val );
static __inline__ void memmove_2p_col (void *d, void *s, int h, int bpr);
static __inline__ unsigned short expand2w( unsigned char c );
static __inline__ unsigned short dup2w( unsigned char c );
static int atacon_bmove( struct condata *conp, int sy, int sx, int dy, int
                         dx, int height, int width);
static int atacon_clear( struct condata *conp, int sy, int sx, int height,
                         int width);
static int atacon_putc( struct condata *conp, int c, int y, int x);
static int atacon_putcs( struct condata *conp, const char *s, int count,
                         int y, int x);
static int atacon_scroll( struct condata *conp, int t, int b, int dir, int
                          count);
static int atacon_switch( struct condata *conp);
static void atacon_bmove_1_plane( register struct display *p, int sy, int
                                  sx, int dy, int dx, int height, int
                                  width);
static void atacon_clear_1_plane( struct condata *conp, register struct
                                  display *p, int sy, int sx, int height,
                                  int width );
static void atacon_putc_1_plane( struct condata *conp, struct display *p,
                                 int c, int y, int x);
static void atacon_putcs_1_plane( struct condata *conp, struct display *p,
                                  const char *s, int count, int y, int x);
static void atacon_rev_char_1_plane( struct display *display, int x, int y );
#ifdef CONFIG_ATARI_2PLANE
static void atacon_bmove_2_plane( register struct display *p, int sy, int
                                  sx, int dy, int dx, int height, int
                                  width);
static void atacon_clear_2_plane( struct condata *conp, register struct
                                  display *p, int sy, int sx, int height,
                                  int width );
static void atacon_putc_2_plane( struct condata *conp, struct display *p,
                                 int c, int y, int x);
static void atacon_putcs_2_plane( struct condata *conp, struct display *p,
                                  const char *s, int count, int y, int x);
static void atacon_rev_char_2_plane( struct display *display, int x, int y );
#endif
#ifdef CONFIG_ATARI_4PLANE
static void atacon_bmove_4_plane( register struct display *p, int sy, int
                                  sx, int dy, int dx, int height, int
                                  width);
static void atacon_clear_4_plane( struct condata *conp, register struct
                                  display *p, int sy, int sx, int height,
                                  int width );
static void atacon_putc_4_plane( struct condata *conp, struct display *p,
                                 int c, int y, int x);
static void atacon_putcs_4_plane( struct condata *conp, struct display *p,
                                  const char *s, int count, int y, int x);
static void atacon_rev_char_4_plane( struct display *display, int x, int y );
#endif
#ifdef CONFIG_ATARI_8PLANE
static void atacon_bmove_8_plane( register struct display *p, int sy, int
                                  sx, int dy, int dx, int height, int
                                  width);
static void atacon_clear_8_plane( struct condata *conp, register struct
                                  display *p, int sy, int sx, int height,
                                  int width );
static void atacon_putc_8_plane( struct condata *conp, struct display *p,
                                 int c, int y, int x);
static void atacon_putcs_8_plane( struct condata *conp, struct display *p,
                                  const char *s, int count, int y, int x);
static void atacon_rev_char_8_plane( struct display *display, int x, int y );
#endif
static void atacon_vbl_handler( struct intframe *fp, void *dummy );
static int atacon_cursor( struct condata *conp, int mode);

/************************* End of Prototypes **************************/


struct display_switch
	dispsw_1_plane = {
		atacon_bmove_1_plane, atacon_clear_1_plane, atacon_putc_1_plane,
		atacon_putcs_1_plane, atacon_rev_char_1_plane
	}
#ifdef CONFIG_ATARI_2PLANE
	, dispsw_2_plane = {
		atacon_bmove_2_plane, atacon_clear_2_plane, atacon_putc_2_plane,
		atacon_putcs_2_plane, atacon_rev_char_2_plane
	}
#endif
#ifdef CONFIG_ATARI_4PLANE
    , dispsw_4_plane = {
		atacon_bmove_4_plane, atacon_clear_4_plane, atacon_putc_4_plane,
		atacon_putcs_4_plane, atacon_rev_char_4_plane
	}
#endif
#ifdef CONFIG_ATARI_8PLANE
	, dispsw_8_plane = {
		atacon_bmove_8_plane, atacon_clear_8_plane, atacon_putc_8_plane,
		atacon_putcs_8_plane, atacon_rev_char_8_plane
	}
#endif
#ifdef CONFIG_ATARI_16PLANE
	, dispsw_16_plane = {
		atacon_bmove_16_plane, atacon_clear_16_plane, atacon_putc_16_plane,
		atacon_putcs_16_plane, atacon_rev_char_16_plane
	}
#endif
;
	


static void
atacon_setup(int con, int setcol, int cls)
{
	struct display *display=&disp[con];
	struct condata *conp=&vc_cons[con];

	int oldrows,oldcols;
	int fgcol=7;
	int bgcol=0;

	if (display->var.bits_per_pixel == 1) {
		display->dispsw = &dispsw_1_plane;
		fgcol = 1;
	}
#ifdef CONFIG_ATARI_2PLANE
	else if (display->var.bits_per_pixel == 2) {
		display->dispsw = &dispsw_2_plane;
		fgcol = 3;
	}
#endif
#ifdef CONFIG_ATARI_4PLANE
	else if (display->var.bits_per_pixel == 4) 
		display->dispsw = &dispsw_4_plane;
#endif
#ifdef CONFIG_ATARI_8PLANE
	else if (display->var.bits_per_pixel == 8)
		display->dispsw = &dispsw_8_plane;
#endif
#ifdef CONFIG_ATARI_16PLANE
	else if (display->var.bits_per_pixel == 16) {
		display->dispsw = &dispsw_16_plane;
		fgcol = 0xbbbbbb;
	}
#endif
	else {
		printk( "atacon_init: %d planes not supported, using 4 planes instead\n",
				display->var.bits_per_pixel );
		display->dispsw = &dispsw_4_plane;
	}

	if (setcol) {
		display->fgcol=fgcol;
		display->bgcol=bgcol;
	}
	oldcols=conp->vc_cols;
	oldrows=conp->vc_rows;
	conp->vc_cols = display->var.xres / display->fontwidth;
	conp->vc_rows = display->var.yres / display->fontheight;
	conp->vc_can_do_color = (display->var.bits_per_pixel != 1);
	if (cls) {
		/* clear screen (a dirty hack...) */
		con_clear(con);
	}
	conp->vc_flags |= CON_INITED;
}
	
static int atacon_changevar(int con)
{
	atacon_setup(con, 1, 1);
	return 0;
}

static long atacon_init(struct condata *conp, long mem_start, char **display_desc)
{
	int unit = conp - vc_cons;	
	static int master_init = 0;
	
	if (!master_init) {
		master_init = 1;
		fb_info = atafb_init(&mem_start);
		disp=fb_info->disp;
		*display_desc = fb_info->modename;
		fb_info->changevar=&atacon_changevar;

		add_isr( IRQ4, atacon_vbl_handler, 0, NULL );
	}
	else
		/* Copy all from the first console, maybe allow different
		   display modes later. */
		disp[unit] = disp[0];
	
	atacon_setup(unit, 1, 0);
	
	return mem_start;
}

static int atacon_deinit (struct condata *conp)
{
  struct display *p = &disp[conp - vc_cons];

  /* You can't deinit a non inited screen nor the console 0. */
  if (!(conp->vc_flags & CON_INITED) || p == disp)
    return -1;

  atacon_switch (&vc_cons[0]);	/* Always switch to the console 0 */
  memset (p, 0, sizeof (struct display));
  conp->vc_flags &= ~CON_INITED;
  return 0;
}


/* ================================================================= */
/*                      Utility Assembler Functions                  */
/* ================================================================= */


/* ====================================================================== */

/* Those of a delicate disposition might like to skip the next couple of
 * pages.
 *
 * These functions are drop in replacements for memmove and
 * memset(_, 0, _). However their five instances add at least a kilobyte
 * to the object file. You have been warned.
 *
 * Not a great fan of assembler for the sake of it, but I think
 * that these routines are at least 10 times faster than their C
 * equivalents for large blits, and thats important to the lowest level of
 * a graphics driver. Question is whether some scheme with the blitter
 * would be faster. I suspect not for simple text system - not much
 * asynchronisity.
 *
 * Code is very simple, just gruesome expansion. Basic strategy is to
 * increase data moved/cleared at each step to 16 bytes to reduce
 * instruction per data move overhead. movem might be faster still
 * For more than 15 bytes, we try to align the write direction on a
 * longword boundary to get maximum speed. This is even more gruesome.
 * Unaligned read/write used requires 68020+ - think this is a problem?
 *
 * Sorry!
 */

/* ++roman: I've optimized Robert's original versions in some minor
 * aspects, e.g. moveq instead of movel, let gcc choose the registers,
 * use movem in some places...
 * For other modes than 1 plane, lots of more such assembler functions
 * were needed (e.g. the ones using movep or expanding color values).
 */

/* ++andreas: more optimizations:
   subl #65536,d0 replaced by clrw d0; subql #1,d0 for dbcc
   addal is faster than addaw
   movep is rather expensive compared to ordinary move's
   some functions rewritten in C for clearity, no speed loss */

static __inline__ void *mymemclear_small(void * s, size_t count)
{
	if (!count) return(0);
	
	__asm__ __volatile__
	   ("lsrl   #1,%1 ; jcc 1f ; moveb %2,%0@-\n\t"
     "1: lsrl   #1,%1 ; jcc 1f ; movew %2,%0@-\n\t"
     "1: lsrl   #1,%1 ; jcc 1f ; movel %2,%0@-\n\t"
     "1: lsrl   #1,%1 ; jcc 1f ; movel %2,%0@- ; movel %2,%0@-\n\t"
     "1: subql  #1,%1 ; jcs 3f\n\t"
     "2: moveml %2/%3/%4/%5,%0@-\n\t"
        "dbra %1,2b\n\t"
	 "3:"
		: "=a" (s), "=d" (count)
	    :  "d" (0), "d" (0), "d" (0), "d" (0),
	       "0" ((char *) s + count), "1" (count)
	   );

   return(0);              
}

static __inline__ void *mymemclear(void * s, size_t count)
{
	if (!count)
		return(0);
	
    else if (count < 16) {
		__asm__ __volatile__
		("lsrl   #1,%1 ; jcc 1f ; clrb %0@+\n\t"
	   "1:lsrl   #1,%1 ; jcc 1f ; clrw %0@+\n\t"
       "1:lsrl   #1,%1 ; jcc 1f ; clrl %0@+\n\t"
       "1:lsrl   #1,%1 ; jcc 1f ; clrl %0@+ ; clrl %0@+\n\t"
	   "1:"
		 : "=a" (s), "=d" (count)
		 : "0" (s), "1" (count)
		);
	}
	
    else {
		long	tmp;
		__asm__ __volatile__
		("movel %1,%2\n\t"
         "lsrl   #1,%2 ; jcc 1f ; clrb %0@+ ; subqw #1,%1\n\t"
         "lsrl   #1,%2 ; jcs 2f\n\t"  /* %0 increased=>bit 2 switched*/
         "clrw   %0@+  ; subqw  #2,%1 ; jra 2f\n\t"
	  "1: lsrl   #1,%2 ; jcc 2f\n\t"
         "clrw   %0@+  ; subqw  #2,%1\n\t"
      "2: movew %1,%2; lsrl #2,%1 ; jeq 6f\n\t"
         "lsrl   #1,%1 ; jcc 3f ; clrl %0@+\n\t"        
      "3: lsrl   #1,%1 ; jcc 4f ; clrl %0@+ ; clrl %0@+\n\t"
      "4: subql  #1,%1 ; jcs 6f\n\t"
      "5: clrl %0@+; clrl %0@+ ; clrl %0@+ ; clrl %0@+\n\t"
         "dbra %1,5b   ; clrw %1; subql #1,%1; jcc 5b\n\t"
      "6: movew %2,%1; btst #1,%1 ; jeq 7f ; clrw %0@+\n\t"
      "7:            ; btst #0,%1 ; jeq 8f ; clrb %0@+\n\t"
      "8:"
		 : "=a" (s), "=d" (count), "=d" (tmp)
		 : "0" (s), "1" (count)
		);
	}
	
	return(0);              
}

static __inline__ void *mymemset(void * s, size_t count)

{
  if (!count) return(0);
  
  __asm__ __volatile__
    ("lsrl   #1,%1 ; jcc 1f ; moveb %2,%0@-\n\t"
     "1: lsrl   #1,%1 ; jcc 1f ; movew %2,%0@-\n\t"
     "1: lsrl   #1,%1 ; jcc 1f ; movel %2,%0@-\n\t"
     "1: lsrl   #1,%1 ; jcc 1f ; movel %2,%0@- ; movel %2,%0@-\n\t"
     "1: subql  #1,%1 ; jcs 3f\n\t"
     "2: moveml %2/%3/%4/%5,%0@-\n\t"
     "dbra %1,2b\n\t"
     "3:"
     : "=a" (s), "=d" (count)
     :  "d" (-1), "d" (-1), "d" (-1), "d" (-1),
     "0" ((char *) s + count), "1" (count)
     );
  
  return(0);              
}

static __inline__ void *mymemmove(void * d, void * s, size_t count)
{
    if (d < s) {
        if (count < 16) {
			__asm__ __volatile__
       ("lsrl   #1,%2 ; jcc 1f ; moveb %1@+,%0@+\n\t"
      "1:lsrl   #1,%2 ; jcc 1f ; movew %1@+,%0@+\n\t"
      "1:lsrl   #1,%2 ; jcc 1f ; movel %1@+,%0@+\n\t"
      "1:lsrl   #1,%2 ; jcc 1f ; movel %1@+,%0@+ ; movel %1@+,%0@+\n\t"
      "1:"
		: "=a" (d), "=a" (s), "=d" (count)
        : "0" (d), "1" (s), "2" (count)
	   );
       } else {
		   long tmp;
		   __asm__ __volatile__
	   ("movel  %0,%3\n\t"
        "lsrl   #1,%3 ; jcc 1f ; moveb %1@+,%0@+ ; subqw #1,%2\n\t"
        "lsrl   #1,%3 ; jcs 2f\n\t"  /* %0 increased=>bit 2 switched*/
        "movew  %1@+,%0@+  ; subqw  #2,%2 ; jra 2f\n\t"
     "1: lsrl   #1,%3 ; jcc 2f\n\t"
        "movew  %1@+,%0@+  ; subqw  #2,%2\n\t"
     "2: movew  %2,sp@-; lsrl #2,%2 ; jeq 6f\n\t"
        "lsrl   #1,%2 ; jcc 3f ; movel %1@+,%0@+\n\t"        
     "3: lsrl   #1,%2 ; jcc 4f ; movel %1@+,%0@+ ; movel %1@+,%0@+\n\t"
     "4: subql  #1,%2 ; jcs 6f\n\t"
     "5: movel  %1@+,%0@+;movel %1@+,%0@+\n\t"
        "movel  %1@+,%0@+;movel %1@+,%0@+\n\t"
        "dbra   %2,5b ; clrw %2; subql #1,%2; jcc 5b\n\t"
     "6: movew sp@+,%2; btst #1,%2 ; jeq 7f ; movew %1@+,%0@+\n\t"
     "7:              ; btst #0,%2 ; jeq 8f ; moveb %1@+,%0@+\n\t"
     "8:"
		: "=a" (d), "=a" (s), "=d" (count), "=d" (tmp)
        : "0" (d), "1" (s), "2" (count)
	   );
      }
   } else {
    if (count < 16) {
		__asm__ __volatile__
       ("lsrl   #1,%2 ; jcc 1f ; moveb %1@-,%0@-\n\t"
      "1:lsrl   #1,%2 ; jcc 1f ; movew %1@-,%0@-\n\t"
      "1:lsrl   #1,%2 ; jcc 1f ; movel %1@-,%0@-\n\t"
      "1:lsrl   #1,%2 ; jcc 1f ; movel %1@-,%0@- ; movel %1@-,%0@-\n\t"
      "1:"
		: "=a" (d), "=a" (s), "=d" (count)
        : "0" ((char *) d + count), "1" ((char *) s + count), "2" (count)
	   );
      } else {
		  long tmp;
		  __asm__ __volatile__
       ("movel %0,%3\n\t"
        "lsrl   #1,%3 ; jcc 1f ; moveb %1@-,%0@- ; subqw #1,%2\n\t"
        "lsrl   #1,%3 ; jcs 2f\n\t"  /* %0 increased=>bit 2 switched*/
        "movew  %1@-,%0@-  ; subqw  #2,%2 ; jra 2f\n\t"
     "1: lsrl   #1,%3 ; jcc 2f\n\t"
        "movew  %1@-,%0@-  ; subqw  #2,%2\n\t"
     "2: movew %2,sp@-; lsrl #2,%2 ; jeq 6f\n\t"
        "lsrl   #1,%2 ; jcc 3f ; movel %1@-,%0@-\n\t"        
     "3: lsrl   #1,%2 ; jcc 4f ; movel %1@-,%0@- ; movel %1@-,%0@-\n\t"
     "4: subql  #1,%2 ; jcs 6f\n\t"
          "5: movel %1@-,%0@-;movel %1@-,%0@-\n\t"
             "movel %1@-,%0@-;movel %1@-,%0@-\n\t"
             "dbra %2,5b ; clrw %2; subql #1,%2; jcc 5b\n\t"
     "6: movew sp@+,%2; btst #1,%2 ; jeq 7f ; movew %1@-,%0@-\n\t"
     "7:              ; btst #0,%2 ; jeq 8f ; moveb %1@-,%0@-\n\t"
     "8:"
		: "=a" (d), "=a" (s), "=d" (count), "=d" (tmp)
        : "0" ((char *) d + count), "1" ((char *) s + count), "2" (count)
	   );
      }
   }
   return(0);              
}

/* ++andreas: Simple and fast version of memmove, assumes size is
   divisible by 16, suitable for moving the whole screen bitplane */
static __inline__ void
fast_memmove (char *dst, char *src, size_t size)
{
  if (!size)
    return;
  if (dst < src)
    __asm__ __volatile__
      ("1:"
       "  moveml %0@+,d0/d1/a0/a1\n"
       "  moveml d0/d1/a0/a1,%1@\n"
       "  addql #8,%1; addql #8,%1\n"
       "  dbra %2,1b\n"
       "  clrw %2; subql #1,%2\n"
       "  jcc 1b"
       : "=a" (src), "=a" (dst), "=d" (size)
       : "0" (src), "1" (dst), "2" (size / 16 - 1)
       : "d0", "d1", "a0", "a1", "memory");
  else
    __asm__ __volatile__
      ("1:"
       "  subql #8,%0; subql #8,%0\n"
       "  moveml %0@,d0/d1/a0/a1\n"
       "  moveml d0/d1/a0/a1,%1@-\n"
       "  dbra %2,1b\n"
       "  clrw %2; subql #1,%2\n"
       "  jcc 1b"
       : "=a" (src), "=a" (dst), "=d" (size)
       : "0" (src + size), "1" (dst + size), "2" (size / 16 - 1)
       : "d0", "d1", "a0", "a1", "memory");
}

/* Sets the bytes in the visible column at d, height h, to the value
 * val for a 4 plane screen. The the bis of the color in 'color' are
 * moved (8 times) to the respective bytes. This means:
 *
 * for( h times; d += bpr )
 *   *d     = (color & 1) ? 0xff : 0;
 *   *(d+2) = (color & 2) ? 0xff : 0;
 *   *(d+4) = (color & 4) ? 0xff : 0;
 *   *(d+6) = (color & 8) ? 0xff : 0;
 */

static __inline__ void memclear_4p_col( void *d, size_t h,
				       u_long val, int bpr)

{
	__asm__ __volatile__
		("1: movepl %4,%0@(0)\n\t"
		  "addal  %5,%0\n\t"
		  "dbra	  %1,1b"
		  : "=a" (d), "=d" (h)
		  : "0" (d), "1" (h - 1), "d" (val), "r" (bpr)
		);
}

/* Sets a 4 plane region from 'd', length 'count' bytes, to the color
 * in val1/val2. 'd' has to be an even address and count must be divisible
 * by 8, because only whole words and all planes are accessed. I.e.:
 *
 * for( count/8 times )
 *   *d     = *(d+1) = (color & 1) ? 0xff : 0;
 *   *(d+2) = *(d+3) = (color & 2) ? 0xff : 0;
 *   *(d+4) = *(d+5) = (color & 4) ? 0xff : 0;
 *   *(d+6) = *(d+7) = (color & 8) ? 0xff : 0;
 */

static __inline__ void memset_even_4p( void *d, size_t count,
									   u_long val1, u_long val2 )

{	
  u_long *dd = d;

  count /= 8;
  while (count--)
    {
      *dd++ = val1;
      *dd++ = val2;
    }
}

/* Copies a 4 plane column from 's', height 'h', to 'd'. */

static __inline__ void memmove_4p_col (void *d, void *s, int h, int bpr)
{
  u_char *dd = d, *ss = s;

  while (h--)
    {
      dd[0] = ss[0];
      dd[2] = ss[2];
      dd[4] = ss[4];
      dd[6] = ss[6];
      dd += bpr;
      ss += bpr;
    }
}


/* This expands a 4 bit color into a long for movepl (4 plane) operations. */

static __inline__ unsigned long expand4l( unsigned char c )

{	unsigned long	rv;
	
	__asm__ __volatile__
		( "lsrb	 #1,%2\n\t"
		  "scs	 %0\n\t"
		  "lsll	 #8,%0\n\t"
		  "lsrb	 #1,%2\n\t"
		  "scs	 %0\n\t"
		  "lsll	 #8,%0\n\t"
		  "lsrb	 #1,%2\n\t"
		  "scs	 %0\n\t"
		  "lsll	 #8,%0\n\t"
		  "lsrb	 #1,%2\n\t"
		  "scs	 %0\n\t"
		  : "=&d" (rv), "=d" (c)
		  : "1" (c)
		);
	return( rv );
}

/* This expands a 4 bit color into two longs for two movel operations
 * (4 planes).
 */

static __inline__ void expand4dl( unsigned char c, unsigned long *ret1,
							   unsigned long *ret2 )

{	unsigned long	rv1, rv2;
	
	__asm__ __volatile__
		( "lsrb	 #1,%3\n\t"
		  "scs	 %0\n\t"
		  "extw	 %0\n\t"
		  "swap	 %0\n\t"
		  "lsrb	 #1,%3\n\t"
		  "scs	 %0\n\t"
		  "extw	 %0\n\t"
		  "lsrb	 #1,%3\n\t"
		  "scs	 %1\n\t"
		  "extw	 %1\n\t"
		  "swap	 %1\n\t"
		  "lsrb	 #1,%3\n\t"
		  "scs   %1\n\t"
		  "extw	 %1"
		  : "=&d" (rv1), "=&d" (rv2), "=d" (c)
		  : "2" (c)
		);
	*ret1 = rv1;
	*ret2 = rv2;
}


/* This duplicates a byte 4 times into a long. */

static __inline__ unsigned long dup4l( unsigned char c )

{	ushort	tmp;
	ulong	rv;

	__asm__ __volatile__
		( "moveb  %2,%0\n\t"
		  "lslw   #8,%0\n\t"
		  "moveb  %2,%0\n\t"
		  "movew  %0,%1\n\t"
		  "swap   %0\n\t"
		  "movew  %1,%0"
		  : "=&d" (rv), "=d" (tmp)
		  : "d" (c)
		);

	return( rv );
}


/* Sets the bytes in the visible column at d, height h, to the value
 * val1,val2 for a 8 plane screen. The the bis of the color in 'color' are
 * moved (8 times) to the respective bytes. This means:
 *
 * for( h times; d += bpr )
 *   *d      = (color & 1) ? 0xff : 0;
 *   *(d+2)  = (color & 2) ? 0xff : 0;
 *   *(d+4)  = (color & 4) ? 0xff : 0;
 *   *(d+6)  = (color & 8) ? 0xff : 0;
 *   *(d+8)  = (color & 16) ? 0xff : 0;
 *   *(d+10) = (color & 32) ? 0xff : 0;
 *   *(d+12) = (color & 64) ? 0xff : 0;
 *   *(d+14) = (color & 128) ? 0xff : 0;
 */

static __inline__ void memclear_8p_col( void *d, size_t h, u_long val1,
				       u_long val2, int bpr)

{
	__asm__ __volatile__
		("1: movepl %4,%0@(0)\n\t"
	      "movepl %5,%0@(8)\n\t"
		  "addal  %6,%0\n\t"
		  "dbra	  %1,1b"
		  : "=a" (d), "=d" (h)
		  : "0" (d), "1" (h - 1), "d" (val1), "d" (val2), "r" (bpr)
		);
}

/* Sets a 8 plane region from 'd', length 'count' bytes, to the color
 * val1..val4. 'd' has to be an even address and count must be divisible
 * by 16, because only whole words and all planes are accessed. I.e.:
 *
 * for( count/16 times )
 *   *d      = *(d+1)  = (color & 1) ? 0xff : 0;
 *   *(d+2)  = *(d+3)  = (color & 2) ? 0xff : 0;
 *   *(d+4)  = *(d+5)  = (color & 4) ? 0xff : 0;
 *   *(d+6)  = *(d+7)  = (color & 8) ? 0xff : 0;
 *   *(d+8)  = *(d+9)  = (color & 16) ? 0xff : 0;
 *   *(d+10) = *(d+11) = (color & 32) ? 0xff : 0;
 *   *(d+12) = *(d+13) = (color & 64) ? 0xff : 0;
 *   *(d+14) = *(d+15) = (color & 128) ? 0xff : 0;
 */

static __inline__ void memset_even_8p( void *d, size_t count,
									   u_long val1, u_long val2,
									   u_long val3, u_long val4 )

{	
  u_long *dd = d;

  count /= 16;
  while (count--)
    {
      *dd++ = val1;
      *dd++ = val2;
      *dd++ = val3;
      *dd++ = val4;
    }
}

/* Copies a 8 plane column from 's', height 'h', to 's'. */

static __inline__ void memmove_8p_col (void *d, void *s, int h, int bpr )
{
  u_char *dd = d, *ss = s;

  while (h--)
    {
      dd[0] = ss[0];
      dd[2] = ss[2];
      dd[4] = ss[4];
      dd[6] = ss[6];
      dd[8] = ss[8];
      dd[10] = ss[10];
      dd[12] = ss[12];
      dd[14] = ss[14];
      dd += bpr;
      ss += bpr;
    }
}


/* This expands a 8 bit color into two longs for two movepl (8 plane)
 * operations.
 */

static __inline__ void expand8dl( unsigned char c, u_long *ret1,
								  u_long *ret2 )

{	unsigned long	rv1, rv2;
	
	__asm__ __volatile__
		( "lsrb	 #1,%3\n\t"
		  "scs	 %0\n\t"
		  "lsll	 #8,%0\n\t"
		  "lsrb	 #1,%3\n\t"
		  "scs	 %0\n\t"
		  "lsll	 #8,%0\n\t"
		  "lsrb	 #1,%3\n\t"
		  "scs	 %0\n\t"
		  "lsll	 #8,%0\n\t"
		  "lsrb	 #1,%3\n\t"
		  "scs	 %0\n\t"
		  "lsrb	 #1,%3\n\t"
		  "scs	 %1\n\t"
		  "lsll	 #8,%1\n\t"
		  "lsrb	 #1,%3\n\t"
		  "scs	 %1\n\t"
		  "lsll	 #8,%1\n\t"
		  "lsrb	 #1,%3\n\t"
		  "scs	 %1\n\t"
		  "lsll	 #8,%1\n\t"
		  "lsrb	 #1,%3\n\t"
		  "scs	 %1"
		  : "=&d" (rv1), "=&d" (rv2),"=d" (c)
		  : "2" (c)
		);

	*ret1 = rv1;
	*ret2 = rv2;
}

/* This expands a 8 bit color into four longs for four movel operations
 * (8 planes).
 */

/* ++andreas: use macro to avoid taking address of return values */
#define expand8ql(c, rv1, rv2, rv3, rv4) \
do {	u_char tmp = c;						\
	__asm__ __volatile__					\
		( "lsrb	 #1,%5\n\t"				\
		  "scs	 %0\n\t"				\
		  "extw	 %0\n\t"				\
		  "swap	 %0\n\t"				\
		  "lsrb	 #1,%5\n\t"				\
		  "scs	 %0\n\t"				\
		  "extw	 %0\n\t"				\
		  "lsrb	 #1,%5\n\t"				\
		  "scs	 %1\n\t"				\
		  "extw	 %1\n\t"				\
		  "swap	 %1\n\t"				\
		  "lsrb	 #1,%5\n\t"				\
		  "scs   %1\n\t"				\
		  "extw	 %1\n\t"				\
		  "lsrb	 #1,%5\n\t"				\
		  "scs	 %2\n\t"				\
		  "extw	 %2\n\t"				\
		  "swap	 %2\n\t"				\
		  "lsrb	 #1,%5\n\t"				\
		  "scs	 %2\n\t"				\
		  "extw	 %2\n\t"				\
		  "lsrb	 #1,%5\n\t"				\
		  "scs	 %3\n\t"				\
		  "extw	 %3\n\t"				\
		  "swap	 %3\n\t"				\
		  "lsrb	 #1,%5\n\t"				\
		  "scs   %3\n\t"				\
		  "extw	 %3"					\
		  : "=&d" (rv1), "=&d" (rv2), "=&d" (rv3),	\
		    "=&d" (rv4), "=d" (tmp)			\
		  : "4" (tmp)					\
		);						\
} while (0)



/* Sets the bytes in the visible column at d, height h, to the value
 * val for a 2 plane screen. The the bis of the color in 'color' are
 * moved (8 times) to the respective bytes. This means:
 *
 * for( h times; d += bpr )
 *   *d     = (color & 1) ? 0xff : 0;
 *   *(d+2) = (color & 2) ? 0xff : 0;
 */

static __inline__ void memclear_2p_col( void *d, size_t h,
				       u_short val, int bpr)

{
	__asm__ __volatile__
		("1: movepw %4,%0@(0)\n\t"
		  "addal  %5,%0\n\t"
		  "dbra	  %1,1b"
		  : "=a" (d), "=d" (h)
		  : "0" (d), "1" (h - 1), "d" (val), "r" (bpr)
		);
}

/* Sets a 2 plane region from 'd', length 'count' bytes, to the color
 * in val1. 'd' has to be an even address and count must be divisible
 * by 8, because only whole words and all planes are accessed. I.e.:
 *
 * for( count/4 times )
 *   *d     = *(d+1) = (color & 1) ? 0xff : 0;
 *   *(d+2) = *(d+3) = (color & 2) ? 0xff : 0;
 */

static __inline__ void memset_even_2p( void *d, size_t count, u_long val )

{	
  u_long *dd = d;

  count /= 4;
  while (count--)
    *dd++ = val;
}

/* Copies a 2 plane column from 's', height 'h', to 's'. */

static __inline__ void memmove_2p_col (void *d, void *s, int h, int bpr)
{
  u_char *dd = d, *ss = s;

  while (h--)
    {
      dd[0] = ss[0];
      dd[2] = ss[2];
      dd += bpr;
      ss += bpr;
    }
}


/* This expands a 2 bit color into a short for movepw (2 plane) operations. */

static __inline__ unsigned short expand2w( unsigned char c )

{	unsigned short	rv;
	
	__asm__ __volatile__
		( "lsrb	 #1,%2\n\t"
		  "scs	 %0\n\t"
		  "lsll	 #8,%0\n\t"
		  "lsrb	 #1,%2\n\t"
		  "scs	 %0\n\t"
		  : "=&d" (rv), "=d" (c)
		  : "1" (c)
		);
	return( rv );
}

/* This expands a 2 bit color into one long for a movel operation
 * (2 planes).
 */

static __inline__ u_long expand2l( unsigned char c)

{	unsigned long	rv;
	
	__asm__ __volatile__
		( "lsrb	 #1,%2\n\t"
		  "scs	 %0\n\t"
		  "extw	 %0\n\t"
		  "swap	 %0\n\t"
		  "lsrb	 #1,%2\n\t"
		  "scs	 %0\n\t"
		  "extw	 %0\n\t"
		  : "=&d" (rv), "=d" (c)
		  : "1" (c)
		);

	return rv;
}


/* This duplicates a byte 2 times into a short. */

static __inline__ unsigned short dup2w( unsigned char c )

{	ushort	rv;

	__asm__ __volatile__
		( "moveb  %1,%0\n\t"
		  "lslw   #8,%0\n\t"
		  "moveb  %1,%0\n\t"
		  : "=&d" (rv)
		  : "d" (c)
		);

	return( rv );
}



/* ====================================================================== */

/* atacon_XXX routines - interface used by the world
 *
 * This system is now divided into two levels because of complications
 * caused by hardware scrolling. Top level functions:
 *
 *    atacon_bmove(), atacon_clear(), atacon_putc()
 *
 * handles y values in range [0, scr_height-1] that correspond to real
 * screen positions.
 *
 *    atacon_bmove_physical_8()   -- These functions fast implementations
 *    atacon_clear_physical_8()   -- of original atacon_XXX fns.
 *    atacon_putc_physical_8()    -- (fontwidth  != 8) may be added later
 *
 */


static int atacon_bmove (struct condata *conp,
          int sy, int sx, int dy, int dx,
          int height, int width)
{
    int unit = conp - vc_cons;
    struct display *p = &disp[unit];

	if (REDRAW_BLANK_MODE(unit) && console_blanked)
		return( 0 );
	
	if ((sy <= p->cursor_y && p->cursor_y < sy+height &&
		 sx <= p->cursor_x && p->cursor_x < sx+width)   ||
		(dy <= p->cursor_y && p->cursor_y < dy+height &&
		 dx <= p->cursor_x && p->cursor_x < dx+width))
		atacon_cursor( conp, CM_ERASE );
	
	p->dispsw->bmove( p, sy, sx, dy, dx, height, width );
    return(0);
}

static int atacon_clear (struct condata *conp,
                         int sy, int sx, int height, int width)
{
    int unit = conp - vc_cons;
    register struct display *p = &disp[unit];

	if (REDRAW_BLANK_MODE(unit) && console_blanked)
		return( 0 );
	
	if (sy <= p->cursor_y && p->cursor_y < sy+height &&
		sx <= p->cursor_x && p->cursor_x < sx+width)
		CURSOR_UNDRAWN();

	p->dispsw->clear( conp, p, sy, sx, height, width );

    return(0);
}

static int atacon_putc (struct condata *conp, int c, int y, int x)
{
    int unit = conp - vc_cons;
    register struct display *p = &disp[unit];

	if (REDRAW_BLANK_MODE(unit) && console_blanked)
		return( 0 );
	
	if (p->cursor_x == x && p->cursor_y == y)
		CURSOR_UNDRAWN();
	
	p->dispsw->putc( conp, p, c, y, x );

    return(0);
}

static int atacon_putcs (struct condata *conp, const char *s,
                         int count, int y, int x)
{
    int unit = conp - vc_cons;
    register struct display *p = &disp[unit];

	if (REDRAW_BLANK_MODE(unit) && console_blanked)
		return( 0 );
	
	if (p->cursor_y == y &&
		x <= p->cursor_x && p->cursor_x < x+count)
		CURSOR_UNDRAWN();
	
	p->dispsw->putcs( conp, p, s, count, y, x );

    return(0);
}


static int atacon_scroll (struct condata *conp,
                          int t, int b, int dir, int count)
{
	if (REDRAW_BLANK_MODE(conp-vc_cons) && console_blanked)
		return( 0 );
	
    switch (dir) {

    case SM_UP:
      /* Use hardware scrolling if possible */
      if (t == 0 && b == conp->vc_rows)
	{
	  int unit = conp - vc_cons;
	  int yoff;
	  struct display *p = &disp[unit];
	  int countlines=p->fontheight * count;

	  atacon_cursor (conp, CM_ERASE);

	  yoff=p->var.yoffset + countlines;
	  if (yoff + p->var.yres > p->var.yres_virtual)
	    {
	      yoff=0;
	      /* We are at the end of the bitplane */
	      fast_memmove (p->screen_base, 
			     p->bitplane+countlines*p->bytes_per_row,
			    (p->var.yres-countlines) * p->bytes_per_row);

	    }
	  p->var.yoffset=yoff;
	  fb_info->updatevar(unit);
	}
      else
		atacon_bmove( conp, t + count, 0, t, 0,
					  (b-t-count), conp->vc_cols );
		atacon_clear( conp, b-count, 0, count, conp->vc_cols );
        break;

    case SM_DOWN:
      /* Use hardware scrolling if possible */
      if (t == 0 && b == conp->vc_rows)
	{
	  int unit = conp - vc_cons;
	  int yoff;
	  struct display *p = &disp[unit];
	  int countlines=p->fontheight * count;

	  atacon_cursor (conp, CM_ERASE);

          yoff=p->var.yoffset - countlines;
	  if (yoff < 0)  
	    {
	      yoff=p->var.yres_virtual - p->var.yres;
	      /* We are at the start of the screen. */
	      fast_memmove (p->screen_base+(yoff+countlines)*p->bytes_per_row,
			    p->bitplane,
			    (p->var.yres-countlines) * p->bytes_per_row);
	    }
	  p->var.yoffset=yoff;
	  fb_info->updatevar(unit);
	}
      else
		atacon_bmove( conp, t, 0, t+count, 0,
					  (b-t-count), conp->vc_cols );

        /* Fixed bmove() should end Arno's frustration with copying?
         * Confusius say:
         *    Man who copies in wrong direction, end up with trashed data
		 */

		atacon_clear( conp, t, 0, count, conp->vc_cols );
        break;

	  case SM_LEFT:
		atacon_bmove( conp, 0, t + count, 0, t,
					  conp->vc_rows, (b-t-count) );
		atacon_clear( conp, 0, b-count, conp->vc_rows, count );
        break;

	  case SM_RIGHT:
		atacon_bmove( conp, 0, t, 0, t+count,
					  conp->vc_rows, (b-t-count) );
		atacon_clear( conp, 0, t, conp->vc_rows, count );
        break;
    }

    return 0;
}

static int atacon_switch (struct condata *conp)
{

  if (!(conp->vc_flags & CON_INITED))
    return -1;

  if (fb_info && fb_info->switch_con)
    (*fb_info->switch_con) (conp-vc_cons);    
  return 0;
}

static int atacon_blank (int blank)
{
  atacon_cursor( &vc_cons[fg_console], blank ? CM_ERASE : CM_DRAW );
  
  if (REDRAW_BLANK_MODE(fg_console)) {
    /* Special treatment for TT High mode: Changing the palette
     * registers isn't possible, so the screen is filled with black
     * for blanking and completely restored for unblanking.
     */
    if (blank) {
        struct display *p = &disp[fg_console];
			mymemset( p->bitplane, p->bytes_per_row * p->var.yres );
        return( 0 );
    }
    else {
        /* Tell console.c that it has to completely restore the
         * screen itself
         */
        return( 1 );
    }
  }

	(*fb_info->blank) (blank);    
  return 0;
}



/* ================================================================= */
/*                           1 Plane Functions                       */
/* ================================================================= */


static void atacon_bmove_1_plane( register struct display *p,
								  int sy, int sx, int dy, int dx,
								  int height, int width)
{

    if (sx == 0 && dx == 0 && width == p->bytes_per_row) {
		mymemmove(p->bitplane + dy * width * p->fontheight,
				  p->bitplane + sy * width * p->fontheight,
				  width * height * p->fontheight );
    } else {
        register ushort rows;
        register u_char *src;
        register u_char *dst;
        register int bytes = p->bytes_per_row;
        register int linesize = bytes * p->fontheight;
        
        if (dy < sy || (dy == sy && dx < sx)) {
			src  = p->bitplane + sy * linesize + sx;
			dst  = p->bitplane + dy * linesize + dx;
			for (rows = height*p->fontheight ; rows-- ;) {
				mymemmove(dst, src, width);
				src += bytes;
				dst += bytes;
			}
        } else {
			src  = p->bitplane + (sy+height) * linesize + sx - bytes;
			dst  = p->bitplane + (dy+height) * linesize + dx - bytes;

			for (rows = height*p->fontheight ; rows-- ;) {
				mymemmove(dst, src, width);
				src -= bytes;
				dst -= bytes;
			}
        }
    }
}


static void atacon_clear_1_plane( struct condata *conp, register struct
								  display *p, int sy, int sx, int height,
								  int width )
{
    ulong offset;
    register u_char *start;
    register int rows;
    register int bytes = p->bytes_per_row;
    register ulong size;

    if (sx == 0 && width == bytes) {
        offset = sy     * bytes * p->fontheight;
        size   = height * bytes * p->fontheight;
		mymemclear( p->bitplane+offset, size );
    } else {
        offset = (sy * bytes * p->fontheight) + sx;
		start = p->bitplane+offset;
		
		for (rows = height*p->fontheight; rows-- ; start += bytes)
			mymemclear_small(start, width);
    }
}


/* The text effects (bold, reverse, light, underlined, blink) are represented
 * in the attribute byte as follows: (like a monochrome PC graphics adapter):
 * 
 *  - (attr & 0x80) == 0x00 =>  not blinking
 *                  == 0x80 =>  blinking (not implemented, ignored)
 *  - (attr & 0x70) != 0    =>  reverse
 *                  == 0    =>  not reverse
 *  - (attr & 0x03) == 0x01 =>  underlined
 *                  != 0x01 =>  not underlined
 *  - (attr & 0x08) == 0    =>  normal intensity
 *    (attr & 0x08) != 0 && (attr & 0x77) == 0 => light (half intensity)
 *                                        != 0 => bright (bold, double intensity)
 *
 * Note that several combinations of effects are not possible with
 * that representation. Light mode isn't possible if underlined and/or
 * reverse mode are active. Characters are printed bright in that
 * cases instead of light :-( Maybe the representation should be
 * changed (separate bits for each effect)...
 *
 */

static void atacon_putc_1_plane( struct condata *conp, struct display *p,
								 int c, int y, int x)
{
    register u_char   *dest;
    register u_char   *cdat;
    register int rows;
    int bytes;
    ulong             offset;

    c &= 0xff;
    if (!FONT_CNV(c)) return;

    bytes  = p->bytes_per_row;
    offset = y * p->fontheight * bytes + x;
    dest  = p->bitplane + offset;
    cdat  = p->fontdata + (c * p->fontheight);

    if ((conp->vc_attr & 0x0b) != 0x01) {
      register u_char	andmask, xormask, c;
      register int	bold;
		
      xormask = (conp->vc_attr & 0x08) ? 0xff : 0;           /* reverse */
      andmask = ((conp->vc_attr & 0x03) == 0) ? 0x55 : 0xff; /* light */
      bold    = ((conp->vc_attr & 0x03) == 2);               /* bold/bright */
		
      for (rows = p->fontheight ; rows-- ; dest += bytes) {
	c = *cdat++;
	if (rows == 0 && (conp->vc_attr & 0x04)) c = 0xff;
	c = (c & andmask) ^ xormask;
	if (bold) c |= c >> 1;
	*dest = c;
      }
    }
    else {
      for (rows = p->fontheight ; rows-- ; dest += bytes)
	*dest = *cdat++;

      if (conp->vc_attr & 0x04) /* underlined */
	dest[-bytes] ^= ~0;
    }
}


static void atacon_putcs_1_plane( struct condata *conp, struct
								  display *p, const char *s,
								  int count, int y, int x)
{
    register u_char   *dest, *cdat;
    register int rows, bytes;
    register u_char   c;
    u_char   *dest0;
    ulong    offset;
    int      effects = (conp->vc_attr & 0x0f) != 0x01;
	
    if (!count) return;
	
    bytes  = p->bytes_per_row;
    offset = y * p->fontheight * bytes + x;
    dest0  = p->bitplane + offset;

    if (!effects && p->fontheight == 8) {
        /* Ultra special support for 8x8 fonts, as its nice and easy
         * and the console driver seems to spend most of its time in
         * here. Would take another 4 instructions to support,
         * (fontheight != 8) but I have an inbuilt suspicion of mulu's.
         * A hangover from my 68000 days, I guess.
         */
        /* ++andreas: rewritten in C, gcc is good enough */
        u_char *fntp;
	while (count--)
	  {
	    c = *s++;
	    dest = dest0++;
	    if (!FONT_CNV (c))
	      continue;

	    fntp = p->fontdata + 8 * c;
	    *dest = *fntp++; dest += bytes; /* 0 */
	    *dest = *fntp++; dest += bytes; /* 1 */
	    *dest = *fntp++; dest += bytes; /* 2 */
	    *dest = *fntp++; dest += bytes; /* 3 */
	    *dest = *fntp++; dest += bytes; /* 4 */
	    *dest = *fntp++; dest += bytes; /* 5 */
	    *dest = *fntp++; dest += bytes; /* 6 */
	    *dest = *fntp++; dest += bytes; /* 7 */
	  }
        return;
    }
    if (!effects && p->fontheight == 16) {
	/* ++andreas: the same for 8x16 */
	u_char *fntp;
	while (count--)
	  {
	    c = *s++;
	    dest = dest0++;
	    if (!FONT_CNV (c))
	      continue;

	    fntp = p->fontdata + 16 * c;
	    *dest = *fntp++; dest += bytes; /* 0 */
	    *dest = *fntp++; dest += bytes; /* 1 */
	    *dest = *fntp++; dest += bytes; /* 2 */
	    *dest = *fntp++; dest += bytes; /* 3 */
	    *dest = *fntp++; dest += bytes; /* 4 */
	    *dest = *fntp++; dest += bytes; /* 5 */
	    *dest = *fntp++; dest += bytes; /* 6 */
	    *dest = *fntp++; dest += bytes; /* 7 */
	    *dest = *fntp++; dest += bytes; /* 8 */
	    *dest = *fntp++; dest += bytes; /* 9 */
	    *dest = *fntp++; dest += bytes; /* 10 */
	    *dest = *fntp++; dest += bytes; /* 11 */
	    *dest = *fntp++; dest += bytes; /* 12 */
	    *dest = *fntp++; dest += bytes; /* 13 */
	    *dest = *fntp++; dest += bytes; /* 14 */
	    *dest = *fntp++; dest += bytes; /* 15 */
	  }
        return;
    }

	if (effects) {
		register u_char	andmask, xormask, ch;
		register int	bold;
		
		xormask = (conp->vc_attr & 0x08) ? 0xff : 0;
		andmask = ((conp->vc_attr & 0x03) == 0) ? 0x55 : 0xff;
		bold = ((conp->vc_attr & 0x03) == 2);

		while (count--) {
			c = *(s++);
			dest = dest0++;
			if (!FONT_CNV(c)) continue;

			cdat  = p->fontdata + (c * p->fontheight);

			for (rows = p->fontheight ; rows-- ; dest += bytes) {
				ch = *cdat++;
				if (rows == 0 && (conp->vc_attr & 0x04)) ch = 0xff;
				ch = (ch & andmask) ^ xormask;
				if (bold) ch |= ch >> 1;
				*dest = ch;
			}
		}
	}
	else {

		while (count--) {
			c = *(s++);
			dest = dest0++;
			if (!FONT_CNV(c)) continue;

			cdat  = p->fontdata + (c * p->fontheight);

			for (rows = p->fontheight ; rows-- ; dest += bytes)
				*dest = *cdat++;
		}
	}
}


static void atacon_rev_char_1_plane( struct display *display, int x, int y )

{	u_char	*p;
	int		offs;
	int j;
	int bytes;
	
	offs = y * display->fontheight * display->bytes_per_row + x;

	p = display->bitplane + offs;
	j = display->fontheight;
	bytes = display->bytes_per_row;
	while (j--)
	  {
	    *p = ~*p;
	    p += bytes;
	  }
}



#ifdef CONFIG_ATARI_2PLANE

/* ================================================================= */
/*                           2 Plane Functions                       */
/* ================================================================= */


/* Increment/decrement 2 plane addresses */

#define	INC_2P(p)	do { if (!((long)(++(p)) & 1)) (p) += 2; } while(0)
#define	DEC_2P(p)	do { if ((long)(--(p)) & 1) (p) -= 2; } while(0)

/* Convert a standard 4 bit color to our 2 bit color assignment:
 * If at least two RGB channels are active, the low bit is turned on;
 * The intensity bit (b3) is shifted into b1.
 */

#define	COLOR_2P(c)	(((c & 7) >= 3 && (c & 7) != 4) | (c & 8) >> 2)


static void atacon_bmove_2_plane( register struct display *p,
								  int sy, int sx, int dy, int dx,
								  int height, int width)
{
	/* bmove() has to distinguish two major cases: If both, source and
	 * destination, start at even addresses or both are at odd
	 * addresses, just the first odd and last even column (if present)
	 * require special treatment (memmove_col()). The rest between
	 * then can be copied by normal operations, because all adjancent
	 * bytes are affected and are to be stored in the same order.
	 *   The pathological case is when the move should go from an odd
	 * address to an even or vice versa. Since the bytes in the plane
	 * words must be assembled in new order, it seems wisest to make
	 * all movements by memmove_col().
	 */

    if (sx == 0 && dx == 0 && width == p->bytes_per_row/2) {
		/* Special (but often used) case: Moving whole lines can be
		 * done with memmove()
		 */
		mymemmove( p->bitplane + dy * p->bytes_per_row * p->fontheight,
				   p->bitplane + sy * p->bytes_per_row * p->fontheight,
				   p->bytes_per_row * height * p->fontheight );
    } else {
        register int rows, cols;
        register u_char *src;
        register u_char *dst;
        register int bytes = p->bytes_per_row;
        register int linesize = bytes * p->fontheight;
		       unsigned colsize  = height * p->fontheight;
		       unsigned upwards  = (dy < sy) || (dy == sy && dx < sx);

		if ((sx & 1) == (dx & 1)) {
			/* odd->odd or even->even */

			if (upwards) {

				src = p->bitplane + sy * linesize + (sx>>1)*4 + (sx & 1);
				dst = p->bitplane + dy * linesize + (dx>>1)*4 + (dx & 1);

				if (sx & 1) {
					memmove_2p_col( dst, src, colsize, bytes );
					src += 3;
					dst += 3;
					--width;
				}

				if (width > 1) {
					for( rows = colsize; rows > 0; --rows ) {
						mymemmove( dst, src, (width>>1)*4 );
						src += bytes;
						dst += bytes;
					}
				}
				
				if (width & 1) {
					src -= colsize * bytes;
					dst -= colsize * bytes;
					memmove_2p_col( dst + (width>>1)*4, src + (width>>1)*4,
									colsize, bytes );
				}
			}
			else {

				if (!((sx+width-1) & 1)) {
					src = p->bitplane + sy * linesize + ((sx+width-1)>>1)*4;
					dst = p->bitplane + dy * linesize + ((dx+width-1)>>1)*4;
					memmove_2p_col( dst, src, colsize, bytes );
					--width;
				}

				src = p->bitplane + sy * linesize + (sx>>1)*4 + (sx & 1);
				dst = p->bitplane + dy * linesize + (dx>>1)*4 + (dx & 1);

				if (width > 1) {
					src += colsize * bytes + (sx & 1)*3;
					dst += colsize * bytes + (sx & 1)*3;
					for( rows = colsize; rows > 0; --rows ) {
						src -= bytes;
						dst -= bytes;
						mymemmove( dst, src, (width>>1)*4 );
					}
				}

				if (width & 1) {
					memmove_2p_col( dst-3, src-3, colsize, bytes );
				}

			}
		}
		else {
			/* odd->even or even->odd */

			if (upwards) {
				src = p->bitplane + sy * linesize + (sx>>1)*4 + (sx & 1);
				dst = p->bitplane + dy * linesize + (dx>>1)*4 + (dx & 1);
				for( cols = width; cols > 0; --cols ) {
					memmove_2p_col( dst, src, colsize, bytes );
					INC_2P( src );
					INC_2P( dst );
				}
			}
			else {
				sx += width-1;
				dx += width-1;
				src = p->bitplane + sy * linesize + (sx>>1)*4 + (sx & 1);
				dst = p->bitplane + dy * linesize + (dx>>1)*4 + (dx & 1);
				for( cols = width; cols > 0; --cols ) {
					memmove_2p_col( dst, src, colsize, bytes );
					DEC_2P( src );
					DEC_2P( dst );
				}
			}
		}
			
		
    }
}


static void atacon_clear_2_plane( struct condata *conp, register struct
								  display *p, int sy, int sx, int height,
								  int width )
{
    ulong offset;
    register u_char *start;
    register int rows;
    register int bytes = p->bytes_per_row;
    int lines = height * p->fontheight;
    register ulong  size;
	u_long          cval;
	u_short			pcval;
	
    cval = expand2l (COLOR_2P (conp->vc_video_erase_char >> 12));
	
    if (sx == 0 && width == bytes/2) {

        offset = sy * bytes * p->fontheight;
        size   = lines * bytes;
		memset_even_2p( p->bitplane+offset, size, cval );

    } else {

        offset = (sy * bytes * p->fontheight) + (sx>>1)*4 + (sx & 1);
		start = p->bitplane + offset;
		pcval = expand2w( COLOR_2P(conp->vc_video_erase_char >> 12) );
		
		/* Clears are split if the region starts at an odd column or
		 * end at an even column. These extra columns are spread
		 * across the interleaved planes. All in between can be
		 * cleared by normal mymemclear_small(), because both bytes of
		 * the single plane words are affected.
		 */

		if (sx & 1) {
			memclear_2p_col( start, lines, pcval, bytes );
			start += 3;
			width--;
		}

		if (width & 1) {
			memclear_2p_col( start + (width>>1)*4, lines, pcval, bytes );
			width--;
		}

		if (width) {
			for( rows = lines; rows-- ; start += bytes )
				memset_even_2p( start, width*2, cval );
		}
    }
}


static void atacon_putc_2_plane( struct condata *conp, struct display *p,
								 int c, int y, int x)

{	register u_char   *dest;
    register u_char   *cdat;
    register int rows;
    register int bytes = p->bytes_per_row;
	ulong			  eorx, fgx, bgx, fdx;

	c &= 0xff;
	if (!FONT_CNV(c)) return;

    dest  = p->bitplane + y * p->fontheight * bytes + (x>>1)*4 + (x & 1);
    cdat  = p->fontdata + (c * p->fontheight);

	fgx   = expand2w( COLOR_2P(conp->vc_attr) );
	bgx   = expand2w( COLOR_2P(conp->vc_attr >> 4) );
	eorx  = fgx ^ bgx;
		
    for( rows = p->fontheight ; rows-- ; dest += bytes ) {
		fdx = dup2w( *cdat++ );
		__asm__ __volatile__ ( "movepw %1,%0@(0)" : /* no outputs */
							   : "a" (dest), "d" ((fdx & eorx) ^ bgx) );
	}


#ifdef COLOR_UNDERLINING
	if (conp->vc_underline) {
		__asm__ __volatile__ ( "movepw %1,%0@(0)" : /* no outputs */
							   : "a" (dest-bytes), "d" (fgx) );
	}
#endif
}


static void atacon_putcs_2_plane( struct condata *conp, struct
								  display *p, const char *s,
								  int count, int y, int x)

{	register u_char   *dest, *dest0;
    register u_char   *cdat, c;
    register int rows;
    register int bytes;
	ulong			  eorx, fgx, bgx, fdx;

    bytes = p->bytes_per_row;
    dest0 = p->bitplane + y * p->fontheight * bytes + (x>>1)*4 + (x & 1);
	fgx   = expand2w( COLOR_2P(conp->vc_attr) );
	bgx   = expand2w( COLOR_2P(conp->vc_attr >> 4) );
	eorx  = fgx ^ bgx;

	while (count--) {

		c = *s++;
		if (!FONT_CNV(c)) continue;
		cdat  = p->fontdata + (c * p->fontheight);
	
		for( rows = p->fontheight, dest = dest0; rows-- ; dest += bytes ) {
			fdx = dup2w( *cdat++ );
			__asm__ __volatile__ ( "movepw %1,%0@(0)" : /* no outputs */
								   : "a" (dest), "d" ((fdx & eorx) ^ bgx) );
		}

#ifdef COLOR_UNDERLINING
		if (conp->vc_underline) {
			__asm__ __volatile__ ( "movepw %1,%0@(0)" : /* no outputs */
								   : "a" (dest-bytes), "d" (fgx) );
		}
#endif

		INC_2P(dest0);
	}
}


static void atacon_rev_char_2_plane( struct display *display, int x, int y )

{	u_char	*p;
	int j;
	int bytes;
	
	p = display->bitplane +
		y * display->fontheight * display->bytes_per_row +
		(x>>1)*4 + (x & 1);
	j = display->fontheight;
	bytes = display->bytes_per_row;
	while (j--)
	  {
		/* This should really obey the individual character's
		 * background and foreground colors instead of simply
		 * inverting.
		 */
	    p[0] = ~p[0];
	    p[2] = ~p[2];
	    p += bytes;
	  }
}


#endif /* CONFIG_ATARI_2PLANE */



#ifdef CONFIG_ATARI_4PLANE

/* ================================================================= */
/*                           4 Plane Functions                       */
/* ================================================================= */


/* Increment/decrement 4 plane addresses */

#define	INC_4P(p)	do { if (!((long)(++(p)) & 1)) (p) += 6; } while(0)
#define	DEC_4P(p)	do { if ((long)(--(p)) & 1) (p) -= 6; } while(0)


static void atacon_bmove_4_plane( register struct display *p,
								  int sy, int sx, int dy, int dx,
								  int height, int width)
{
	/* bmove() has to distinguish two major cases: If both, source and
	 * destination, start at even addresses or both are at odd
	 * addresses, just the first odd and last even column (if present)
	 * require special treatment (memmove_col()). The rest between
	 * then can be copied by normal operations, because all adjancent
	 * bytes are affected and are to be stored in the same order.
	 *   The pathological case is when the move should go from an odd
	 * address to an even or vice versa. Since the bytes in the plane
	 * words must be assembled in new order, it seems wisest to make
	 * all movements by memmove_col().
	 */

    if (sx == 0 && dx == 0 && width == p->bytes_per_row/4) {
		/* Special (but often used) case: Moving whole lines can be
		 * done with memmove()
		 */
		mymemmove( p->bitplane + dy * p->bytes_per_row * p->fontheight,
				   p->bitplane + sy * p->bytes_per_row * p->fontheight,
				   p->bytes_per_row * height * p->fontheight );
    } else {
        register int rows, cols;
        register u_char *src;
        register u_char *dst;
        register int bytes = p->bytes_per_row;
        register int linesize = bytes * p->fontheight;
		       unsigned colsize  = height * p->fontheight;
		       unsigned upwards  = (dy < sy) || (dy == sy && dx < sx);

		if ((sx & 1) == (dx & 1)) {
			/* odd->odd or even->even */

			if (upwards) {

				src = p->bitplane + sy * linesize + (sx>>1)*8 + (sx & 1);
				dst = p->bitplane + dy * linesize + (dx>>1)*8 + (dx & 1);

				if (sx & 1) {
					memmove_4p_col( dst, src, colsize, bytes );
					src += 7;
					dst += 7;
					--width;
				}

				if (width > 1) {
					for( rows = colsize; rows > 0; --rows ) {
						mymemmove( dst, src, (width>>1)*8 );
						src += bytes;
						dst += bytes;
					}
				}
				
				if (width & 1) {
					src -= colsize * bytes;
					dst -= colsize * bytes;
					memmove_4p_col( dst + (width>>1)*8, src + (width>>1)*8,
									colsize, bytes );
				}
			}
			else {

				if (!((sx+width-1) & 1)) {
					src = p->bitplane + sy * linesize + ((sx+width-1)>>1)*8;
					dst = p->bitplane + dy * linesize + ((dx+width-1)>>1)*8;
					memmove_4p_col( dst, src, colsize, bytes );
					--width;
				}

				src = p->bitplane + sy * linesize + (sx>>1)*8 + (sx & 1);
				dst = p->bitplane + dy * linesize + (dx>>1)*8 + (dx & 1);

				if (width > 1) {
					src += colsize * bytes + (sx & 1)*7;
					dst += colsize * bytes + (sx & 1)*7;
					for( rows = colsize; rows > 0; --rows ) {
						src -= bytes;
						dst -= bytes;
						mymemmove( dst, src, (width>>1)*8 );
					}
				}

				if (width & 1) {
					memmove_4p_col( dst-7, src-7, colsize, bytes );
				}

			}
		}
		else {
			/* odd->even or even->odd */

			if (upwards) {
				src = p->bitplane + sy * linesize + (sx>>1)*8 + (sx & 1);
				dst = p->bitplane + dy * linesize + (dx>>1)*8 + (dx & 1);
				for( cols = width; cols > 0; --cols ) {
					memmove_4p_col( dst, src, colsize, bytes );
					INC_4P( src );
					INC_4P( dst );
				}
			}
			else {
				sx += width-1;
				dx += width-1;
				src = p->bitplane + sy * linesize + (sx>>1)*8 + (sx & 1);
				dst = p->bitplane + dy * linesize + (dx>>1)*8 + (dx & 1);
				for( cols = width; cols > 0; --cols ) {
					memmove_4p_col( dst, src, colsize, bytes );
					DEC_4P( src );
					DEC_4P( dst );
				}
			}
		}
			
		
    }
}


static void atacon_clear_4_plane( struct condata *conp, register struct
								  display *p, int sy, int sx, int height,
								  int width )
{
    ulong offset;
    register u_char *start;
    register int rows;
    register int bytes = p->bytes_per_row;
    int lines = height * p->fontheight;
    register ulong  size;
	u_long          cval1, cval2, pcval;

	expand4dl( conp->vc_video_erase_char >> 12, &cval1, &cval2 );
	
    if (sx == 0 && width == bytes/4) {

        offset = sy * bytes * p->fontheight;
        size   = lines * bytes;
		memset_even_4p( p->bitplane+offset, size, cval1, cval2 );

    } else {

        offset = (sy * bytes * p->fontheight) + (sx>>1)*8 + (sx & 1);
		start = p->bitplane + offset;
		pcval = expand4l( conp->vc_video_erase_char >> 12 );
		
		/* Clears are split if the region starts at an odd column or
		 * end at an even column. These extra columns are spread
		 * across the interleaved planes. All in between can be
		 * cleared by normal mymemclear_small(), because both bytes of
		 * the single plane words are affected.
		 */

		if (sx & 1) {
			memclear_4p_col( start, lines, pcval, bytes );
			start += 7;
			width--;
		}

		if (width & 1) {
			memclear_4p_col( start + (width>>1)*8, lines, pcval, bytes );
			width--;
		}

		if (width) {
			for( rows = lines; rows-- ; start += bytes )
				memset_even_4p( start, width*4, cval1, cval2 );
		}
    }
}


static void atacon_putc_4_plane( struct condata *conp, struct display *p,
								 int c, int y, int x)
{	register u_char   *dest;
    register u_char   *cdat;
    register int rows;
    register int bytes = p->bytes_per_row;
	ulong			  eorx, fgx, bgx, fdx;

	c &= 0xff;
	if (!FONT_CNV(c)) return;

    dest  = p->bitplane + y * p->fontheight * bytes + (x>>1)*8 + (x & 1);
    cdat  = p->fontdata + (c * p->fontheight);

	fgx   = expand4l( conp->vc_attr );
	bgx   = expand4l( conp->vc_attr >> 4 );
	eorx  = fgx ^ bgx;
		
    for( rows = p->fontheight ; rows-- ; dest += bytes ) {
		fdx = dup4l( *cdat++ );
		__asm__ __volatile__ ( "movepl %1,%0@(0)" : /* no outputs */
							   : "a" (dest), "d" ((fdx & eorx) ^ bgx) );
	}

#ifdef COLOR_UNDERLINING
	if (conp->vc_underline) {
		__asm__ __volatile__ ( "movepl %1,%0@(0)" : /* no outputs */
							   : "a" (dest-bytes), "d" (fgx) );
	}
#endif
}


static void atacon_putcs_4_plane( struct condata *conp, struct
								  display *p, const char *s,
								  int count, int y, int x)
{	register u_char   *dest, *dest0;
    register u_char   *cdat, c;
    register int rows;
    register int bytes;
	ulong			  eorx, fgx, bgx, fdx;

    bytes = p->bytes_per_row;
    dest0 = p->bitplane + y * p->fontheight * bytes + (x>>1)*8 + (x & 1);
	fgx   = expand4l( conp->vc_attr );
	bgx   = expand4l( conp->vc_attr >> 4 );
	eorx  = fgx ^ bgx;

	while (count--) {

		/* I think, unrolling the loops like in the 1 plane case isn't
		 * practicable here, because the body is much longer for 4
		 * planes (mostly the dup4l()). I guess, unrolling this would
		 * need more than 256 bytes and so exceed the instruction
		 * cache :-(
		 */
		
		c = *s++;
		if (!FONT_CNV(c)) continue;
		cdat  = p->fontdata + (c * p->fontheight);
	
		for( rows = p->fontheight, dest = dest0; rows-- ; dest += bytes ) {
			fdx = dup4l( *cdat++ );
			__asm__ __volatile__ ( "movepl %1,%0@(0)" : /* no outputs */
								   : "a" (dest), "d" ((fdx & eorx) ^ bgx) );
		}

#ifdef COLOR_UNDERLINING
		if (conp->vc_underline) {
			__asm__ __volatile__ ( "movepl %1,%0@(0)" : /* no outputs */
								   : "a" (dest-bytes), "d" (fgx) );
		}
#endif

		INC_4P(dest0);
	}
}


static void atacon_rev_char_4_plane( struct display *display, int x, int y )

{	u_char	*p;
	int j;
	int bytes;
	
	p = display->bitplane +
		y * display->fontheight * display->bytes_per_row +
		(x>>1)*8 + (x & 1);
	j = display->fontheight;
	bytes = display->bytes_per_row;

	while (j--)
	  {
		/* This should really obey the individual character's
		 * background and foreground colors instead of simply
		 * inverting.
		 */
	    p[0] = ~p[0];
	    p[2] = ~p[2];
	    p[4] = ~p[4];
	    p[6] = ~p[6];
	    p += bytes;
	  }
}


#endif /* CONFIG_ATARI_4PLANE */



#ifdef CONFIG_ATARI_8PLANE

/* ================================================================= */
/*                           8 Plane Functions                       */
/* ================================================================= */


/* In 8 plane mode, 256 colors would be possible, but only the first
 * 16 are used by the console code (the upper 4 bits are
 * background/unused). For that, the following functions mask off the
 * higher 4 bits of each color.
 */

/* Increment/decrement 8 plane addresses */

#define	INC_8P(p)	do { if (!((long)(++(p)) & 1)) (p) += 14; } while(0)
#define	DEC_8P(p)	do { if ((long)(--(p)) & 1) (p) -= 14; } while(0)


static void atacon_bmove_8_plane( register struct display *p,
								  int sy, int sx, int dy, int dx,
								  int height, int width)
{
	/* bmove() has to distinguish two major cases: If both, source and
	 * destination, start at even addresses or both are at odd
	 * addresses, just the first odd and last even column (if present)
	 * require special treatment (memmove_col()). The rest between
	 * then can be copied by normal operations, because all adjancent
	 * bytes are affected and are to be stored in the same order.
	 *   The pathological case is when the move should go from an odd
	 * address to an even or vice versa. Since the bytes in the plane
	 * words must be assembled in new order, it seems wisest to make
	 * all movements by memmove_col().
	 */

    if (sx == 0 && dx == 0 && width == p->bytes_per_row/8) {
		/* Special (but often used) case: Moving whole lines can be
		 * done with memmove()
		 */
      fast_memmove (p->bitplane + dy * p->bytes_per_row * p->fontheight,
		    p->bitplane + sy * p->bytes_per_row * p->fontheight,
		    p->bytes_per_row * height * p->fontheight);
    } else {
        register int rows, cols;
        register u_char *src;
        register u_char *dst;
        register int bytes = p->bytes_per_row;
        register int linesize = bytes * p->fontheight;
		       unsigned colsize  = height * p->fontheight;
		       unsigned upwards  = (dy < sy) || (dy == sy && dx < sx);

		if ((sx & 1) == (dx & 1)) {
			/* odd->odd or even->even */

			if (upwards) {

				src = p->bitplane + sy * linesize + (sx>>1)*16 + (sx & 1);
				dst = p->bitplane + dy * linesize + (dx>>1)*16 + (dx & 1);

				if (sx & 1) {
					memmove_8p_col( dst, src, colsize, bytes );
					src += 15;
					dst += 15;
					--width;
				}

				if (width > 1) {
					for( rows = colsize; rows > 0; --rows ) {
						fast_memmove (dst, src, (width >> 1) * 16);
						src += bytes;
						dst += bytes;
					}
				}
				
				if (width & 1) {
					src -= colsize * bytes;
					dst -= colsize * bytes;
					memmove_8p_col( dst + (width>>1)*16, src + (width>>1)*16,
									colsize, bytes );
				}
			}
			else {

				if (!((sx+width-1) & 1)) {
					src = p->bitplane + sy * linesize + ((sx+width-1)>>1)*16;
					dst = p->bitplane + dy * linesize + ((dx+width-1)>>1)*16;
					memmove_8p_col( dst, src, colsize, bytes );
					--width;
				}

				src = p->bitplane + sy * linesize + (sx>>1)*16 + (sx & 1);
				dst = p->bitplane + dy * linesize + (dx>>1)*16 + (dx & 1);

				if (width > 1) {
					src += colsize * bytes + (sx & 1)*15;
					dst += colsize * bytes + (sx & 1)*15;
					for( rows = colsize; rows > 0; --rows ) {
						src -= bytes;
						dst -= bytes;
						fast_memmove (dst, src, (width>>1)*16);
					}
				}

				if (width & 1) {
					memmove_8p_col( dst-15, src-15, colsize, bytes );
				}

			}
		}
		else {
			/* odd->even or even->odd */

			if (upwards) {
				src = p->bitplane + sy * linesize + (sx>>1)*16 + (sx & 1);
				dst = p->bitplane + dy * linesize + (dx>>1)*16 + (dx & 1);
				for( cols = width; cols > 0; --cols ) {
					memmove_8p_col( dst, src, colsize, bytes );
					INC_8P( src );
					INC_8P( dst );
				}
			}
			else {
				sx += width-1;
				dx += width-1;
				src = p->bitplane + sy * linesize + (sx>>1)*16 + (sx & 1);
				dst = p->bitplane + dy * linesize + (dx>>1)*16 + (dx & 1);
				for( cols = width; cols > 0; --cols ) {
					memmove_8p_col( dst, src, colsize, bytes );
					DEC_8P( src );
					DEC_8P( dst );
				}
			}
		}
			
		
    }
}


static void atacon_clear_8_plane( struct condata *conp, register struct
								  display *p, int sy, int sx, int height,
								  int width )
{
    ulong offset;
    register u_char *start;
    register int rows;
    register int bytes = p->bytes_per_row;
    int lines = height * p->fontheight;
    register ulong  size;
	u_long          cval1, cval2, cval3, cval4, pcval1, pcval2;

	expand8ql( (conp->vc_video_erase_char >> 12) & 0xf,
		  cval1, cval2, cval3, cval4 );
	
    if (sx == 0 && width == bytes/8) {

        offset = sy * bytes * p->fontheight;
        size   = lines * bytes;
		memset_even_8p( p->bitplane+offset, size, cval1, cval2, cval3, cval4 );

    } else {

        offset = (sy * bytes * p->fontheight) + (sx>>1)*16 + (sx & 1);
		start = p->bitplane + offset;
		expand8dl( (conp->vc_video_erase_char >> 12) & 0xf, &pcval1, &pcval2 );
		
		/* Clears are split if the region starts at an odd column or
		 * end at an even column. These extra columns are spread
		 * across the interleaved planes. All in between can be
		 * cleared by normal mymemclear_small(), because both bytes of
		 * the single plane words are affected.
		 */

		if (sx & 1) {
			memclear_8p_col( start, lines, pcval1, pcval2, bytes );
			start += 7;
			width--;
		}

		if (width & 1) {
			memclear_8p_col( start + (width>>1)*16, lines, pcval1,
							 pcval2, bytes );
			width--;
		}

		if (width) {
			for( rows = lines; rows-- ; start += bytes )
				memset_even_8p( start, width*8, cval1, cval2, cval3, cval4 );
		}
    }
}


static void atacon_putc_8_plane( struct condata *conp, struct display *p,
								 int c, int y, int x)
{	register u_char   *dest;
    register u_char   *cdat;
    register int rows;
    register int bytes = p->bytes_per_row;
	ulong			  eorx1, eorx2, fgx1, fgx2, bgx1, bgx2, fdx;

	c &= 0xff;
	if (!FONT_CNV(c)) return;

    dest  = p->bitplane + y * p->fontheight * bytes + (x>>1)*16 + (x & 1);
    cdat  = p->fontdata + (c * p->fontheight);

	expand8dl( conp->vc_attr & 0x0f, &fgx1, &fgx2 );
	expand8dl( (conp->vc_attr >> 4) & 0x0f, &bgx1, &bgx2 );
	eorx1  = fgx1 ^ bgx1; eorx2  = fgx2 ^ bgx2;
		
    for( rows = p->fontheight ; rows-- ; dest += bytes ) {
		fdx = dup4l( *cdat++ );
		__asm__ __volatile__
			( "movepl %1,%0@(0)\n\t"
			  "movepl %2,%0@(8)"
			  : /* no outputs */
			  : "a" (dest), "d" ((fdx & eorx1) ^ bgx1),
			    "d" ((fdx & eorx2) ^ bgx2)
			);
	}

#ifdef COLOR_UNDERLINING
	if (conp->vc_underline) {
		__asm__ __volatile__
			( "movepl %1,%0@(0)\n\t"
			  "movepl %2,%0@(8)"
			  : /* no outputs */
			  : "a" (dest-bytes), "d" (fgx1), "d" (fgx2)
			);
	}
#endif
}


static void atacon_putcs_8_plane( struct condata *conp, struct
								  display *p, const char *s,
								  int count, int y, int x)

{	register u_char   *dest, *dest0;
    register u_char   *cdat, c;
    register int rows;
    register int bytes;
	ulong			  eorx1, eorx2, fgx1, fgx2, bgx1, bgx2, fdx;

    bytes = p->bytes_per_row;
    dest0 = p->bitplane + y * p->fontheight * bytes + (x>>1)*16 + (x & 1);

	expand8dl( conp->vc_attr & 0x0f, &fgx1, &fgx2 );
	expand8dl( (conp->vc_attr >> 4) & 0x0f, &bgx1, &bgx2 );
	eorx1  = fgx1 ^ bgx1; eorx2  = fgx2 ^ bgx2;

	while (count--) {

		/* I think, unrolling the loops like in the 1 plane case isn't
		 * practicable here, because the body is much longer for 4
		 * planes (mostly the dup4l()). I guess, unrolling this would
		 * need more than 256 bytes and so exceed the instruction
		 * cache :-(
		 */
		
		c = *s++;
		if (!FONT_CNV(c)) continue;
		cdat  = p->fontdata + (c * p->fontheight);
	
		for( rows = p->fontheight, dest = dest0; rows-- ; dest += bytes ) {
			fdx = dup4l( *cdat++ );
			__asm__ __volatile__
				( "movepl %1,%0@(0)\n\t"
				  "movepl %2,%0@(8)"
				  : /* no outputs */
				  : "a" (dest), "d" ((fdx & eorx1) ^ bgx1),
				  "d" ((fdx & eorx2) ^ bgx2)
				);
		}

#ifdef COLOR_UNDERLINING
		if (conp->vc_underline) {
			__asm__ __volatile__
				( "movepl %1,%0@(0)\n\t"
				  "movepl %2,%0@(8)"
				  : /* no outputs */
				  : "a" (dest-bytes), "d" (fgx1), "d" (fgx2)
				);
		}
#endif

		INC_8P(dest0);
	}
}


static void atacon_rev_char_8_plane( struct display *display, int x, int y )

{	u_char	*p;
	int j;
	int bytes;
	
	p = display->bitplane +
		y * display->fontheight * display->bytes_per_row +
		(x>>1)*16 + (x & 1);
	j = display->fontheight;
	bytes = display->bytes_per_row;

	while (j--)
	  {
		/* This should really obey the individual character's
		 * background and foreground colors instead of simply
		 * inverting. For 8 plane mode, only the lower 4 bits of the
		 * color are inverted, because only that color registers have
		 * been set up.
		 */
	    p[0] = ~p[0];
	    p[2] = ~p[2];
	    p[4] = ~p[4];
	    p[6] = ~p[6];
	    p += bytes;
	  }
}


#endif /* CONFIG_ATARI_8PLANE */


/* ================================================================= */
/*                            Cursor Functions                       */
/* ================================================================= */


static void atacon_vbl_handler( struct intframe *fp, void *dummy )

{	struct display *display;

	if (!cursor_on) return;
	
	if (vbl_cursor_cnt && --vbl_cursor_cnt == 0) {
		/* Here no check is possible for console changing. The console
		 * switching code should set vbl_cursor_cnt to an appropriate value.
		 */
		display = &disp[fg_console];
		display->dispsw->rev_char( display, display->cursor_x,
								   display->cursor_y );
		cursor_drawn ^= 1;
		vbl_cursor_cnt = cursor_blink_rate;
	}
}


static int atacon_cursor (struct condata *conp, int mode)

{	struct display *display = &disp[conp - vc_cons];

	if (CURSOR_UNDRAWN ())
	  display->dispsw->rev_char (display, display->cursor_x,
				     display->cursor_y);
	display->cursor_x = conp->vc_x;
	display->cursor_y = conp->vc_y;

	switch( mode ) {

	  case CM_ERASE:
		cursor_on = 0;
		break;

	  case CM_MOVE:
	  case CM_DRAW:
		vbl_cursor_cnt = CURSOR_DRAW_DELAY;
		cursor_on = 1;
		break;
	}

	return( 0 );
}

/*
 * The console "switch" structure for the Atari native console
 */


struct consw ata_con =
{
    atacon_init, atacon_deinit, atacon_clear, atacon_putc, atacon_putcs,
    atacon_cursor, atacon_scroll, atacon_bmove, atacon_switch, atacon_blank
};

