/*
 * atari/atafb.c -- Low level implementation of Atari frame buffer device
 *
 *  Copyright (C) 1994 Martin Schaller & Roman Hodek
 *  
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 *
 * History:
 *   - 03 Jan 95: Original version my Martin Schaller: The TT driver and
 *                all the device independant stuff
 *   - 09 Jan 95: Roman: I've added the hardware abstraction (hw_switch)
 *                and wrote the Falcon, ST(E), and External drivers
 *                based on the original TT driver.
 *
 * To do:
 *   - For the Falcon it is not yet possible to set video modes, only
 *     the one detected at boot time can be used. Seems a bit more
 *     complicated to program the VIDEL...
 *
 *   - The external driver should be extended somehow to allow
 *     colormap operations.
 *     
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/atarihw.h>
#include <linux/bootinfo.h>
#include <linux/tty.h>
#include <linux/malloc.h>
#include <linux/delay.h>
#include <asm/segment.h>
#include "atafb.h"

static int default_par=0;	/* default resolution (0=none) */

static int node;		/* node of the /dev/fb?current file */

static unsigned long default_mem_req=0;

static int default_font=-1;

static int hwscroll=-1;

static struct atari_fb_par {
	unsigned long screen_base;
	union {
		struct {
			int mode;
			int sync;
		} tt, st;
		struct {
			int mono;
			/* Sorry, nothing here yet, because mode switching for the
			 * Falcon isn't implemented yet.
			 * Here should be fields for storing a video mode, either
			 * as direct parameters for the hardware or as a mode code.
			 */
		} falcon;
		/* Nothing needed for external mode */
	} hw;
} current_par;

static struct fb_info fb_info;

static unsigned long screen_base;
static int screen_len;
static int current_par_valid=0; 

static int currcon=0;

static int mono_moni=0;

static struct display disp[NR_CONSOLES];

/* external video handling */

static unsigned			external_xres;
static unsigned			external_yres;
static unsigned			external_depth;
static int				external_pmode;
static unsigned long	external_addr = 0;
static unsigned long	external_len;



extern int fontheight_8x8;
extern int fontwidth_8x8;
extern unsigned char fontdata_8x8[];

extern int fontheight_8x16;
extern int fontwidth_8x16;
extern unsigned char fontdata_8x16[];


/* ++roman: This structure abstracts from the underlying hardware (ST(e),
 * TT, or Falcon.
 *
 * int (*detect)( void )
 *   This function should detect the current video mode settings and
 *   store them in atari_fb_predefined[0] for later reference by the
 *   user. Return the index+1 of an equivalent predefined mode or 0
 *   if there is no such.
 * 
 * int (*encode_fix)( struct fb_fix_screeninfo *fix,
 *                    struct atari_fb_par *par )
 *   This function should fill in the 'fix' structure based on the
 *   values in the 'par' structure.
 *   
 * int (*decode_var)( struct fb_var_screeninfo *var,
 *                    struct atari_fb_par *par )
 *   Get the video params out of 'var'. If a value doesn't fit, round
 *   it up, if it's too big, return EINVAL.
 *   Round up in the following order: bits_per_pixel, xres, yres, 
 *   xres_virtual, yres_virtual, xoffset, yoffset, grayscale, bitfields, 
 *   horizontal timing, vertical timing.
 *
 * int (*encode_var)( struct fb_var_screeninfo *var,
 *                    struct atari_fb_par *par );
 *   Fill the 'var' structure based on the values in 'par' and maybe
 *   other values read out of the hardware.
 *   
 * void (*get_par)( struct atari_fb_par *par )
 *   Fill the hardware's 'par' structure.
 *   
 * void (*set_par)( struct atari_fb_par *par )
 *   Set the hardware according to 'par'.
 *   
 * int (*setcolreg)( unsigned regno, unsigned red,
 *                   unsigned green, unsigned blue,
 *                   unsigned transp )
 *   Set a single color register. The values supplied are already
 *   rounded down to the hardware's capabilities (according to the
 *   entries in the var structure). Return != 0 for invalid regno.
 *
 * int (*getcolreg)( unsigned regno, unsigned *red,
 *                   unsigned *green, unsigned *blue,
 *                   unsigned *transp )
 *   Read a single color register and split it into
 *   colors/transparent. Return != 0 for invalid regno.
 *
 * void (*set_screen_base)( unsigned long s_base )
 *   Set the base address of the displayed frame buffer. Only called
 *   if yres_virtual > yres or xres_virtual > xres.
 *
 */

static struct fb_hwswitch {
	int  (*detect)( void );
	int  (*encode_fix)( struct fb_fix_screeninfo *fix,
						struct atari_fb_par *par );
	int  (*decode_var)( struct fb_var_screeninfo *var,
						struct atari_fb_par *par );
	int  (*encode_var)( struct fb_var_screeninfo *var,
						struct atari_fb_par *par );
	void (*get_par)( struct atari_fb_par *par );
	void (*set_par)( struct atari_fb_par *par );
	int  (*getcolreg)( unsigned regno, unsigned *red,
					   unsigned *green, unsigned *blue,
					   unsigned *transp );
	int  (*setcolreg)( unsigned regno, unsigned red,
					   unsigned green, unsigned blue,
					   unsigned transp );
	void (*set_screen_base)( unsigned long s_base );
} *fbhw;

static char *autodetect_names[] = {"autodetect", NULL};
static char *stlow_names[] = {"stlow", NULL};
static char *stmid_names[] = {"stmid", "default5", NULL};
static char *sthigh_names[] = {"sthigh", "default4", NULL};
static char *ttlow_names[] = {"ttlow", NULL};
static char *ttmid_names[]= {"ttmid", "default1", NULL};
static char *tthigh_names[]= {"tthigh", "default2", NULL};
static char *vga2_names[] = {"vga2", NULL};
static char *vga4_names[] = {"vga4", NULL};
static char *vga16_names[] = {"vga16", "default3", NULL};
static char *vga256_names[] = {"vga256", NULL};
static char *falh2_names[] = {"falh2", NULL};
static char *falh16_names[] = {"falh16", NULL};

char **fb_var_names[] = {
	/* Writing the name arrays directly in this array (via "(char *[]){...}")
	 * crashes gcc 2.5.8 (sigsegv) if the inner array
	 * contains more than two items. I've also seen that all elements
	 * were identical to the last (my cross-gcc) :-(*/
	autodetect_names,
	stlow_names,
	stmid_names,
	sthigh_names,
	ttlow_names,
	ttmid_names,
	tthigh_names,
	vga2_names,
	vga4_names,
	vga16_names,
	vga256_names,
	falh2_names,
	falh16_names,
	NULL
	/* ,NULL */ /* this causes a sigsegv on my gcc-2.5.8 */
};

struct fb_var_screeninfo atari_fb_predefined[] = {
 	{ /* autodetect */
	  0, 0, 0, 0, 0, 0, 0, 0,   		/* xres-grayscale */
	  {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, {0, 0, 0}, 	/* red green blue tran*/
	  0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0 },
 	{ /* st low */
	  320, 200, 320, 200, 0, 0, 4, 0,   		/* xres-grayscale */
	  {0, 4, 0}, {0, 4, 0}, {0, 4, 0}, {0, 0, 0}, 	/* red green blue tran*/
	  0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ /* st mid */
	  640, 200, 640, 200, 0, 0, 2, 0,		/* xres-grayscale */
	  {0, 4, 0}, {0, 4, 0}, {0, 4, 0}, {0, 0, 0}, 	/* red green blue tran*/
	  0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ /* st high */
	  640, 400, 640, 400, 0, 0, 1, 0,		/* xres-grayscale */
	  {0, 4, 0}, {0, 4, 0}, {0, 4, 0}, {0, 0, 0}, 	/* red green blue tran*/
	  0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ /* tt low */
	  320, 480, 320, 480, 0, 0, 8, 0,		/* xres-grayscale */
	  {0, 4, 0}, {0, 4, 0}, {0, 4, 0}, {0, 0, 0}, 	/* red green blue tran*/
	  0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ /* tt mid */
	  640, 480, 640, 480, 0, 0, 4, 0,		/* xres-grayscale */
	  {0, 4, 0}, {0, 4, 0}, {0, 4, 0}, {0, 0, 0}, 	/* red green blue tran*/
	  0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ /* tt high */
	  1280, 960, 1280, 960, 0, 0, 1, 0,		/* xres-grayscale */
	  {0, 4, 0}, {0, 4, 0}, {0, 4, 0}, {0, 0, 0}, 	/* red green blue tran*/
	  0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ /* vga2 */
	  640, 480, 640, 480, 0, 0, 1, 0,   		/* xres-grayscale */
	  {0, 6, 0}, {0, 6, 0}, {0, 6, 0}, {0, 0, 0}, 	/* red green blue tran*/
	  0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ /* vga4 */
	  640, 480, 640, 480, 0, 0, 2, 0,		/* xres-grayscale */
	  {0, 4, 0}, {0, 4, 0}, {0, 4, 0}, {0, 0, 0}, 	/* red green blue tran*/
	  0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ /* vga16 */
	  640, 480, 640, 480, 0, 0, 4, 0,		/* xres-grayscale */
	  {0, 6, 0}, {0, 6, 0}, {0, 6, 0}, {0, 0, 0}, 	/* red green blue tran*/
	  0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ /* vga256 */
	  640, 480, 640, 480, 0, 0, 8, 0,		/* xres-grayscale */
	  {0, 6, 0}, {0, 6, 0}, {0, 6, 0}, {0, 0, 0}, 	/* red green blue tran*/
	  0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ /* falh2 */
	  896, 608, 896, 608, 0, 0, 1, 0,		/* xres-grayscale */
	  {0, 6, 0}, {0, 6, 0}, {0, 6, 0}, {0, 0, 0}, 	/* red green blue tran*/
	  0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ /* falh16 */
	  896, 608, 896, 608, 0, 0, 4, 0,		/* xres-grayscale */
	  {0, 6, 0}, {0, 6, 0}, {0, 6, 0}, {0, 0, 0}, 	/* red green blue tran*/
	  0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0 }
	};	

int num_atari_fb_predefined=
	sizeof(atari_fb_predefined)/sizeof(struct fb_var_screeninfo);


static int
get_video_mode(char *vname)
{
    char ***name_list;
    char **name;
    int i;
    name_list=fb_var_names;
    for (i = 0 ; i < num_atari_fb_predefined ; i++) {
	name=*(name_list++);
	if (! name || ! *name)
	    break;
	while (*name) {
	    if (! strcmp(vname, *name))
		return (i+1);
	    name++;
	}
    }
    return (0);
}



/* ------------------- TT specific functions ---------------------- */


static int tt_encode_fix( struct fb_fix_screeninfo *fix,
						  struct atari_fb_par *par )

{
	int mode;

	strcpy(fix->id,"Atari Builtin");
	fix->smem_start=screen_base;
	fix->smem_len=screen_len & PAGE_MASK;
	fix->type=FB_TYPE_INTERLEAVED_PLANES;
	fix->type_aux=2;
	fix->visual=FB_VISUAL_PSEUDOCOLOR;
	mode = par->hw.tt.mode & TT_SHIFTER_MODEMASK;
	if (mode == TT_SHIFTER_TTHIGH || mode == TT_SHIFTER_STHIGH) {
		fix->type=FB_TYPE_PACKED_PIXELS;
		fix->type_aux=0;
		if (mode == TT_SHIFTER_TTHIGH)
			fix->visual=FB_VISUAL_MONO01;
	}
	return 0;
}


static int tt_decode_var( struct fb_var_screeninfo *var,
						  struct atari_fb_par *par )
{
	int xres=var->xres;
	int yres=var->yres;
	int bpp=var->bits_per_pixel;
	int linelen;

	if (mono_moni) {
		if (bpp > 1 || xres > 1280 || yres >960)
			return -EINVAL;
		par->hw.tt.mode=TT_SHIFTER_TTHIGH;
		xres=1280;
		yres=960;
		bpp=1;
	} else {
		if (bpp > 8 || xres > 640 || yres > 480)
			return -EINVAL;
		if (bpp > 4) {
			if (xres > 320 || yres > 480)
				return -EINVAL;
			par->hw.tt.mode=TT_SHIFTER_TTLOW;
			xres=320;
			yres=480;
			bpp=8;
		}
		else if (bpp > 2) {
			if (xres > 640 || yres > 480)
				return -EINVAL;
			if (xres > 320 || yres > 200) {
				par->hw.tt.mode=TT_SHIFTER_TTMID;
				xres=640;
				yres=480;
				bpp=4;
			}
			else {
				par->hw.tt.mode=TT_SHIFTER_STLOW;
				xres=320;
				yres=200;
				bpp=4;
			}
		}
		else if (bpp > 1) {
			if (xres > 640 || yres > 200)
				return -EINVAL;
			par->hw.tt.mode=TT_SHIFTER_STMID;
			xres=640;
			yres=200;
			bpp=2;
		}
		else if (var->xres > 640 || var->yres > 400) {
			return -EINVAL;
		}
		else {
			par->hw.tt.mode=TT_SHIFTER_STHIGH;
			xres=640;
			yres=400;
			bpp=1;
		}
	}
	if (var->sync & FB_SYNC_EXT)
		par->hw.tt.sync=0;
	else
		par->hw.tt.sync=1;
	linelen=xres*bpp/8;
	if ((var->yoffset + yres)*linelen > screen_len && screen_len)
		return -EINVAL;
	par->screen_base=screen_base+ var->yoffset*linelen;
	return 0;
}

static int tt_encode_var( struct fb_var_screeninfo *var,
						  struct atari_fb_par *par )
{
	int linelen;
	var->red.offset=0;
	var->red.length=4;
	var->red.msb_right=0;
	var->grayscale=0;

	var->pixclock=31041;
	var->left_margin=120;		/* these may be incorrect 	*/
	var->right_margin=100;
	var->upper_margin=8;
	var->lower_margin=16;
	var->hsync_len=140;
	var->vsync_len=30;

	var->height=-1;
	var->width=-1;

	if (par->hw.tt.sync & 1)
		var->sync=0;
	else
		var->sync=FB_SYNC_EXT;

	switch (par->hw.tt.mode & TT_SHIFTER_MODEMASK) {
	case TT_SHIFTER_STLOW:
		var->xres=320;
		var->yres=200;
		var->bits_per_pixel=4;
		break;
	case TT_SHIFTER_STMID:
		var->xres=640;
		var->yres=200;
		var->bits_per_pixel=2;
		break;
	case TT_SHIFTER_STHIGH:
		var->xres=640;
		var->yres=400;
		var->bits_per_pixel=1;
		break;
	case TT_SHIFTER_TTLOW:
		var->xres=320;
		var->yres=480;
		var->bits_per_pixel=8;
		break;
	case TT_SHIFTER_TTMID:
		var->xres=640;
		var->yres=480;
		var->bits_per_pixel=4;
		break;
	case TT_SHIFTER_TTHIGH:
		var->red.length=0;
		var->xres=1280;
		var->yres=960;
		var->bits_per_pixel=1;
		break;
	}		
	var->blue=var->green=var->red;
	var->transp.offset=0;
	var->transp.length=0;
	var->transp.msb_right=0;
	var->xres_virtual=var->xres;
	linelen=var->xres * var->bits_per_pixel / 8;
	if (screen_len)
		var->yres_virtual=screen_len/linelen;
	else {
		if (hwscroll < 0)
			var->yres_virtual = 2 * var->yres;
		else if (default_font > 0)
			var->yres_virtual=var->yres+hwscroll * default_font;
		else
			var->yres_virtual=var->yres+hwscroll * 16;
	}
	var->xoffset=0;
	if (screen_base)
		var->yoffset=(par->screen_base - screen_base)/linelen;
	else
		var->yoffset=0;
	var->nonstd=0;
	var->activate=0;
	return 0;
}


static void tt_get_par( struct atari_fb_par *par )
{
	unsigned long addr;
	par->hw.tt.mode=shifter_tt.tt_shiftmode;
	par->hw.tt.sync=shifter.syncmode;
	addr = ((shifter.bas_hi & 0xff) << 16) |
	       ((shifter.bas_md & 0xff) << 8)  |
	       ((shifter.bas_lo & 0xff));
	par->screen_base = PTOV(addr);
}

static void tt_set_par( struct atari_fb_par *par )
{
	shifter_tt.tt_shiftmode=par->hw.tt.mode;
	shifter.syncmode=par->hw.tt.sync;
	/* only set screen_base if really necessary */
	if (current_par.screen_base != par->screen_base)
		fbhw->set_screen_base(par->screen_base);
}


static int tt_getcolreg( unsigned regno, unsigned *red,
						 unsigned *green, unsigned *blue,
						 unsigned *transp )
{
	if ((shifter_tt.tt_shiftmode & TT_SHIFTER_MODEMASK) == TT_SHIFTER_STHIGH)
		regno += 254;
	if (regno > 255)
		return 1;
	*blue = tt_palette[regno];
	*green = (*blue >> 4) & 0xf;
	*red = (*blue >> 8) & 0xf;
	*blue &= 0xf;
	*transp = 0;
	return 0;
}


static int tt_setcolreg( unsigned regno, unsigned red,
						 unsigned green, unsigned blue,
						 unsigned transp )
{
	if ((shifter_tt.tt_shiftmode & TT_SHIFTER_MODEMASK) == TT_SHIFTER_STHIGH)
		regno += 254;
	if (regno > 255)
		return 1;
	tt_palette[regno] = (red << 8) | (green << 4) | blue;
	if ((shifter_tt.tt_shiftmode & TT_SHIFTER_MODEMASK) ==
		TT_SHIFTER_STHIGH && regno == 254)
		tt_palette[0] = 0;
	return 0;
}

						  
static int tt_detect( void )

{	struct atari_fb_par par;

	/* Determine the connected monitor: The DMA sound must be
	 * disabled before reading the MFP GPIP, because the Sound
	 * Done Signal and the Monochrome Detect are XORed together!
 	 *
 	 * Even on a TT, we should look if there is a DMA sound. It was
 	 * announced that the Eagle is TT compatible, but only the PCM is
 	 * missing...
	 */
	if (boot_info.bi_atari.hw_present & ATARIHW_PCM_8BIT) { 
		tt_dmasnd.ctrl = DMASND_CTRL_OFF;
		udelay( 20 );	/* wait a while for things to settle down */
	}
	mono_moni = (mfp.par_dt_reg & 0x80) == 0;

	tt_get_par( &par );
	tt_encode_var( &atari_fb_predefined[0], &par );

	switch( par.hw.tt.mode & TT_SHIFTER_MODEMASK ) {
	  case TT_SHIFTER_STLOW:
		return( get_video_mode("stlow") );
		break;
	  case TT_SHIFTER_STMID:
		return( get_video_mode("stmid") );
		break;
	  case TT_SHIFTER_STHIGH:
		return( get_video_mode("sthigh") );
		break;
	  case TT_SHIFTER_TTLOW:
		return( get_video_mode("ttlow") );
		break;
	  case TT_SHIFTER_TTMID:
		return( get_video_mode("ttmid") );
		break;
	  case TT_SHIFTER_TTHIGH:
		return( get_video_mode("tthigh") );
		break;
	}
	return( 0 );
}


/* ------------------- Falcon specific functions ---------------------- */


static int falcon_encode_fix( struct fb_fix_screeninfo *fix,
							  struct atari_fb_par *par )

{
	strcpy(fix->id,"Atari Builtin");
	fix->smem_start=screen_base;
	fix->smem_len=screen_len & PAGE_MASK;
	fix->type=FB_TYPE_INTERLEAVED_PLANES;
	fix->type_aux=2;
	fix->visual=FB_VISUAL_PSEUDOCOLOR;
	if (par->hw.falcon.mono) {
		fix->type=FB_TYPE_PACKED_PIXELS;
		fix->type_aux=0;
	}
	return 0;
}


static int falcon_decode_var( struct fb_var_screeninfo *var,
							  struct atari_fb_par *par )
{
	struct fb_var_screeninfo *myvar = &atari_fb_predefined[0];
	int linelen;
	
	if (var->bits_per_pixel > myvar->bits_per_pixel ||
		var->xres > myvar->xres ||
		var->yres > myvar->yres)
		return -EINVAL;
	linelen=myvar->xres*myvar->bits_per_pixel/8;
	if ((var->yoffset + myvar->yres)*linelen > screen_len && screen_len)
		return -EINVAL;
	par->screen_base=screen_base+ var->yoffset*linelen;
	par->hw.falcon.mono = var->bits_per_pixel==1;
	return 0;
}

static int falcon_encode_var( struct fb_var_screeninfo *var,
							  struct atari_fb_par *par )
{
	struct fb_var_screeninfo *myvar = &atari_fb_predefined[0];
	int linelen;

	var->red.offset=0;
	var->red.length= (myvar->bits_per_pixel == 2) ? 4 : 6;
	var->red.msb_right=0;
	var->grayscale=0;

	var->pixclock=31041;
	var->left_margin=120;		/* these are surely incorrect 	*/
	var->right_margin=100;
	var->upper_margin=8;
	var->lower_margin=16;
	var->hsync_len=140;
	var->vsync_len=30;

	var->height=-1;
	var->width=-1;

	var->sync=0;

	var->xres = myvar->xres;
	var->yres = myvar->yres;
	var->bits_per_pixel = myvar->bits_per_pixel;
	
	var->blue=var->green=var->red;
	var->transp.offset=0;
	var->transp.length=0;
	var->transp.msb_right=0;
	var->xres_virtual=var->xres;
	linelen=var->xres * var->bits_per_pixel / 8;
	if (screen_len)
		var->yres_virtual=screen_len/linelen;
	else {
			if (hwscroll < 0)
				var->yres_virtual = 2 * var->yres;
			else if (default_font > 0)
				var->yres_virtual=var->yres+hwscroll * default_font;
			else
				var->yres_virtual=var->yres+hwscroll * 16;
		}
	var->xoffset=0;
	if (screen_base)
		var->yoffset=(par->screen_base - screen_base)/linelen;
	else
		var->yoffset=0;
	var->nonstd=0;
	var->activate=0;
	return 0;
}


static void falcon_get_par( struct atari_fb_par *par )
{
	unsigned long addr;
	addr = ((shifter.bas_hi & 0xff) << 16) |
	       ((shifter.bas_md & 0xff) << 8)  |
	       ((shifter.bas_lo & 0xff));
	par->screen_base = PTOV(addr);
}

static void falcon_set_par( struct atari_fb_par *par )
{
	/* only set screen_base if really necessary */
	if (current_par.screen_base != par->screen_base)
		fbhw->set_screen_base(par->screen_base);
}


static int falcon_getcolreg( unsigned regno, unsigned *red,
							 unsigned *green, unsigned *blue,
							 unsigned *transp )
{	unsigned long col;
	
	if (regno > 255)
		return 1;
	col = f030_col[regno];
	*red = (col >> 26) & 0x3f;
	*green = (col >> 18) & 0x3f;
	*blue = (col >> 2) & 0x3f;
	*transp = 0;
	return 0;
}


static int falcon_setcolreg( unsigned regno, unsigned red,
							 unsigned green, unsigned blue,
							 unsigned transp )
{
	if (regno > 255)
		return 1;
	f030_col[regno] = (red << 26) | (green << 18) | (blue << 2);
	if (regno < 16)
		shifter_tt.color_reg[regno] =
			(((red & 0xe) >> 1) | ((red & 1) << 3) << 8) |
			(((green & 0xe) >> 1) | ((green & 1) << 3) << 4) |
			((blue & 0xe) >> 1) | ((blue & 1) << 3);
	return 0;
}

						  
static int falcon_detect( void )

{
	int xres;
	int yres;
	int depth;
	int interlace;
	int doubleline;
	int spshift, shift;
	int compatibility = 0;
	struct fb_var_screeninfo *myvar = &atari_fb_predefined[0];
	struct atari_fb_par dummy_par;
	
	interlace = (f030_creg[1] & 0x2) >> 1;
	doubleline = f030_creg[1] & 0x1;

	/* to get depth of display, we must examine spshift and shift.
	 * spshift has priority over shift, if any of bits no. 10, 8 or 4
	 * is set. Priority in spshift is: 10 ">" 8 ">" 4, that means
	 * if bit 10 set then bit 8 and bit 4 don't care.
	 * If all these bits are 0, then display depth is coded in shift
	 * (as for ST and STE)
	 */
	spshift = f030_sreg[3];
	shift = (f030_sreg[0] >> 8) & 0x3;
	if (spshift & 0x400)		/* 2 color mode */
		depth = 1;
	else if (spshift & 0x100)	/* truecolor mode */
		depth = 16;
	else if (spshift & 0x010)	/* 8 bitplane mode */
		depth = 8;
	else if (shift == 0)
		depth = 4;
	else if (shift == 1)
	{
		depth = 2;
		/* 4-color mode uses ST palette, even on VGA! */
		compatibility = 1;
	}
	else /* if (shift == 2) */
		depth = 1;
#if 0
	else
		/* shift==3 is reserved,illegal. This should not happen */
		return 0;
#endif
	
	/* Y resolution is calculated from vertical control registers:
	 * Displaying pixels starts at line f030_yreg[4] and ends at line
	 * f030_yreg[3]. If interlace mode off  the VC-registers count
	 * half lines, if interlace on they count whole lines.
	 * If doubleline mode on, every line is displayed 2 times.
	 */
	yres = f030_yreg[4] - f030_yreg[3];
	if ( !interlace )
		yres >>= 1;
	if ( doubleline )
		yres >>= 1;

	/* X resolution is calculated from screen width in words, and
	 * display depth.
	 * I can't use horizontal control registers as in calculation of yres,
	 * since I dont know the unit of these counters.
	 * Who knows?
	 */
	xres = shifter_f030.scn_width * 16 / depth;

/*	printk("Detected FALCON screen mode: X=%d Y=%d D=%d "
	       "IL=%d DL=%d \n",
	       xres, yres, depth, interlace, doubleline);*/

	myvar->xres = xres;
	myvar->yres = yres;
	myvar->bits_per_pixel = depth;
	falcon_encode_var( myvar, &dummy_par );

	return( 1 );
}



/* ------------------- ST(E) specific functions ---------------------- */


static int stste_encode_fix( struct fb_fix_screeninfo *fix,
							 struct atari_fb_par *par )

{
	int mode;

	strcpy(fix->id,"Atari Builtin");
	fix->smem_start=screen_base;
	fix->smem_len=screen_len & PAGE_MASK;
	fix->type=FB_TYPE_INTERLEAVED_PLANES;
	fix->type_aux=2;
	fix->visual=FB_VISUAL_PSEUDOCOLOR;
	mode = par->hw.st.mode & 3;
	if (mode == ST_HIGH) {
		fix->type=FB_TYPE_PACKED_PIXELS;
		fix->type_aux=0;
		fix->visual=FB_VISUAL_MONO01;
	}
	return 0;
}


static int stste_decode_var( struct fb_var_screeninfo *var,
						  struct atari_fb_par *par )
{
	int xres=var->xres;
	int yres=var->yres;
	int bpp=var->bits_per_pixel;
	int linelen;

	if (mono_moni) {
		if (bpp > 1 || xres > 640 || yres > 400)
			return -EINVAL;
		par->hw.st.mode=ST_HIGH;
		xres=640;
		yres=400;
		bpp=1;
	} else {
		if (bpp > 4 || xres > 640 || yres > 400)
			return -EINVAL;
		if (bpp > 2) {
			if (xres > 320 || yres > 200)
				return -EINVAL;
			par->hw.st.mode=ST_LOW;
			xres=320;
			yres=200;
			bpp=4;
		}
		else if (bpp > 1) {
			if (xres > 640 || yres > 200)
				return -EINVAL;
			par->hw.st.mode=ST_MID;
			xres=640;
			yres=200;
			bpp=2;
		}
		else
			return -EINVAL;
	}
	if (var->sync & FB_SYNC_EXT)
		par->hw.st.sync=(par->hw.st.sync & ~1) | 1;
	else
		par->hw.st.sync=(par->hw.st.sync & ~1);
	linelen=xres*bpp/8;
	if ((var->yoffset + yres)*linelen > screen_len && screen_len)
		return -EINVAL;
	par->screen_base=screen_base+ var->yoffset*linelen;
	return 0;
}

static int stste_encode_var( struct fb_var_screeninfo *var,
						  struct atari_fb_par *par )
{
	int linelen;
	var->red.offset=0;
	var->red.length= (boot_info.bi_atari.hw_present&ATARIHW_EXTD_SHIFTER) ? 4 : 3;
	var->red.msb_right=0;
	var->grayscale=0;

	var->pixclock=31041;
	var->left_margin=120;		/* these are incorrect */
	var->right_margin=100;
	var->upper_margin=8;
	var->lower_margin=16;
	var->hsync_len=140;
	var->vsync_len=30;

	var->height=-1;
	var->width=-1;

	if (!(par->hw.st.sync & 1))
		var->sync=0;
	else
		var->sync=FB_SYNC_EXT;

	switch (par->hw.st.mode & 3) {
	case ST_LOW:
		var->xres=320;
		var->yres=200;
		var->bits_per_pixel=4;
		break;
	case ST_MID:
		var->xres=640;
		var->yres=200;
		var->bits_per_pixel=2;
		break;
	case ST_HIGH:
		var->xres=640;
		var->yres=400;
		var->bits_per_pixel=1;
		break;
	}		
	var->blue=var->green=var->red;
	var->transp.offset=0;
	var->transp.length=0;
	var->transp.msb_right=0;
	var->xres_virtual=var->xres;
	linelen=var->xres * var->bits_per_pixel / 8;
	if (screen_len)
		var->yres_virtual=screen_len/linelen;
	else {
 		if (boot_info.bi_atari.hw_present & ATARIHW_EXTD_SHIFTER) {
			if (hwscroll < 0)
				var->yres_virtual = 2 * var->yres;
			else if (default_font > 0)
				var->yres_virtual=var->yres+hwscroll * default_font;
			else
				var->yres_virtual=var->yres+hwscroll * 16;
		}
		else
			/* no hardware scrolling on standard ST shifters (no base low) */
			var->yres_virtual = var->yres;
	}
	var->xoffset=0;
	if (screen_base)
		var->yoffset=(par->screen_base - screen_base)/linelen;
	else
		var->yoffset=0;
	var->nonstd=0;
	var->activate=0;
	return 0;
}


static void stste_get_par( struct atari_fb_par *par )
{
	unsigned long addr;
	par->hw.st.mode=shifter_tt.st_shiftmode;
	par->hw.st.sync=shifter.syncmode;
	addr = ((shifter.bas_hi & 0xff) << 16) |
	       ((shifter.bas_md & 0xff) << 8);
	if (boot_info.bi_atari.hw_present & ATARIHW_EXTD_SHIFTER)
		addr |= (shifter.bas_lo & 0xff);
	par->screen_base = PTOV(addr);
}

static void stste_set_par( struct atari_fb_par *par )
{
	shifter_tt.st_shiftmode=par->hw.st.mode;
	shifter.syncmode=par->hw.st.sync;
	/* only set screen_base if really necessary */
	if (current_par.screen_base != par->screen_base)
		fbhw->set_screen_base(par->screen_base);
}


static int stste_getcolreg( unsigned regno, unsigned *red,
							unsigned *green, unsigned *blue,
							unsigned *transp )
{	unsigned col;
	
	if (regno > 15)
		return 1;
	col = shifter_tt.color_reg[regno];
	if (boot_info.bi_atari.hw_present & ATARIHW_EXTD_SHIFTER) {
		*red = ((col >> 7) & 0xe) | ((col >> 11) & 1);
		*green = ((col >> 3) & 0xe) | ((col >> 7) & 1);
		*blue = ((col << 1) & 0xe) | ((col >> 3) & 1);
	}
	else {
		*red = (col >> 8) & 0x7;
		*green = (col >> 4) & 0x7;
		*blue = col & 0x7;
	}
	*transp = 0;
	return 0;
}


static int stste_setcolreg( unsigned regno, unsigned red,
						 unsigned green, unsigned blue,
						 unsigned transp )
{
	if (regno > 15)
		return 1;
	if (boot_info.bi_atari.hw_present & ATARIHW_EXTD_SHIFTER)
		shifter_tt.color_reg[regno] =
			(((red & 0xe) >> 1) | ((red & 1) << 3) << 8) |
			(((green & 0xe) >> 1) | ((green & 1) << 3) << 4) |
			((blue & 0xe) >> 1) | ((blue & 1) << 3);
	else
		shifter_tt.color_reg[regno] =
			((red & 0x7) << 8) |
			((green & 0x7) << 4) |
			(blue & 0x7);
	return 0;
}

						  
static int stste_detect( void )

{	struct atari_fb_par par;

	/* Determine the connected monitor: The DMA sound must be
	 * disabled before reading the MFP GPIP, because the Sound
	 * Done Signal and the Monochrome Detect are XORed together!
	 */
	if (boot_info.bi_atari.hw_present & ATARIHW_PCM_8BIT) { 
		tt_dmasnd.ctrl = DMASND_CTRL_OFF;
		udelay( 20 );	/* wait a while for things to settle down */
	}
	mono_moni = (mfp.par_dt_reg & 0x80) == 0;

	stste_get_par( &par );
	stste_encode_var( &atari_fb_predefined[0], &par );

	switch( par.hw.st.mode & 3 ) {
	  case ST_LOW:
		return( get_video_mode("stlow") );
		break;
	  case ST_MID:
		return( get_video_mode("stmid") );
		break;
	  case ST_HIGH:
		return( get_video_mode("sthigh") );
		break;
	}
	return( 0 );
}

static void stste_set_screen_base(unsigned long s_base)
{
	unsigned long addr;
	addr= VTOP(s_base);
	/* Setup Screen Memory */
	shifter.bas_hi=(unsigned char) ((addr & 0xff0000) >> 16);
  	shifter.bas_md=(unsigned char) ((addr & 0x00ff00) >> 8);
	if (boot_info.bi_atari.hw_present & ATARIHW_EXTD_SHIFTER)
		shifter.bas_lo=(unsigned char)  (addr & 0x0000ff);
}


/* ------------------- External Video ---------------------- */


static int ext_encode_fix( struct fb_fix_screeninfo *fix,
						   struct atari_fb_par *par )

{
	strcpy(fix->id,"Unknown Extern");
	fix->smem_start=external_addr;
	fix->smem_len=(external_len + PAGE_SIZE -1) & PAGE_MASK;
	if (external_depth == 1) {
		fix->type = FB_TYPE_PACKED_PIXELS;
		/* The letters 'n' and 'i' in the "atavideo=external:" stand
		 * for "normal" and "inverted", rsp., in the monochrome case */
		fix->visual =
			(external_pmode == FB_TYPE_INTERLEAVED_PLANES) ?
				FB_VISUAL_MONO10 :
					FB_VISUAL_MONO01;
	}
	else {
		switch( external_pmode ) {
			/* All visuals are STATIC, because we don't know how to change
			 * colors :-(
			 */
		    case -1:              /* truecolor */
			fix->type=FB_TYPE_PACKED_PIXELS;
			fix->visual=FB_VISUAL_TRUECOLOR;
			break;
		    case FB_TYPE_PACKED_PIXELS:
			fix->type=FB_TYPE_PACKED_PIXELS;
			fix->visual=FB_VISUAL_STATIC_PSEUDOCOLOR;
			break;
		    case FB_TYPE_PLANES:
			fix->type=FB_TYPE_PLANES;
			fix->visual=FB_VISUAL_STATIC_PSEUDOCOLOR;
			break;
		    case FB_TYPE_INTERLEAVED_PLANES:
			fix->type=FB_TYPE_INTERLEAVED_PLANES;
			fix->type_aux=2;
			fix->visual=FB_VISUAL_STATIC_PSEUDOCOLOR;
			break;
		}
	}
	return 0;
}


static int ext_decode_var( struct fb_var_screeninfo *var,
						   struct atari_fb_par *par )
{
	struct fb_var_screeninfo *myvar = &atari_fb_predefined[0];
	
	if (var->bits_per_pixel > myvar->bits_per_pixel ||
		var->xres > myvar->xres ||
		var->yres > myvar->yres ||
		var->xoffset > 0 ||
		var->yoffset > 0)
		return -EINVAL;
	return 0;
}


static int ext_encode_var( struct fb_var_screeninfo *var,
						   struct atari_fb_par *par )
{
	var->red.offset=0;
	var->red.length=(external_pmode == -1) ? external_depth/3 : 0;
	var->red.msb_right=0;
	var->grayscale=0;

	var->pixclock=31041;
	var->left_margin=120;		/* these are surely incorrect 	*/
	var->right_margin=100;
	var->upper_margin=8;
	var->lower_margin=16;
	var->hsync_len=140;
	var->vsync_len=30;

	var->height=-1;
	var->width=-1;

	var->sync=0;

	var->xres = external_xres;
	var->yres = external_yres;
	var->bits_per_pixel = external_depth;
	
	var->blue=var->green=var->red;
	var->transp.offset=0;
	var->transp.length=0;
	var->transp.msb_right=0;
	var->xres_virtual=var->xres;
	var->yres_virtual=var->yres;
	var->xoffset=0;
	var->yoffset=0;
	var->nonstd=0;
	var->activate=0;
	return 0;
}


static void ext_get_par( struct atari_fb_par *par )
{
	par->screen_base = external_addr;
}

static void ext_set_par( struct atari_fb_par *par )
{
}


static int ext_getcolreg( unsigned regno, unsigned *red,
						  unsigned *green, unsigned *blue,
						  unsigned *transp )
{
	return 1;
}
	
static int ext_setcolreg( unsigned regno, unsigned red,
						  unsigned green, unsigned blue,
						  unsigned transp )
{
	return 1;
}
	

static int ext_detect( void )

{
	struct fb_var_screeninfo *myvar = &atari_fb_predefined[0];
	struct atari_fb_par dummy_par;

	myvar->xres = external_xres;
	myvar->yres = external_yres;
	myvar->bits_per_pixel = external_depth;
	ext_encode_var( myvar, &dummy_par );
	return( 1 );
}




/* ------ This is the same for most hardware types -------- */

static void set_screen_base(unsigned long s_base)
{
	unsigned long addr;
	addr= VTOP(s_base);
	/* Setup Screen Memory */
	shifter.bas_hi=(unsigned char) ((addr & 0xff0000) >> 16);
  	shifter.bas_md=(unsigned char) ((addr & 0x00ff00) >> 8);
  	shifter.bas_lo=(unsigned char)  (addr & 0x0000ff);
}




/* ------------ Interfaces to hardware functions ------------ */


struct fb_hwswitch tt_switch = {
	tt_detect, tt_encode_fix, tt_decode_var, tt_encode_var,
	tt_get_par, tt_set_par, tt_getcolreg, tt_setcolreg,
	set_screen_base
};

struct fb_hwswitch falcon_switch = {
	falcon_detect, falcon_encode_fix, falcon_decode_var, falcon_encode_var,
	falcon_get_par, falcon_set_par, falcon_getcolreg,
	falcon_setcolreg, set_screen_base
};

struct fb_hwswitch st_switch = {
	stste_detect, stste_encode_fix, stste_decode_var, stste_encode_var,
	stste_get_par, stste_set_par, stste_getcolreg, stste_setcolreg,
	stste_set_screen_base
};

struct fb_hwswitch ext_switch = {
	ext_detect, ext_encode_fix, ext_decode_var, ext_encode_var,
	ext_get_par, ext_set_par, ext_getcolreg, ext_setcolreg, NULL
};



static void atari_fb_get_par( struct atari_fb_par *par )
{
	if (current_par_valid) {
		*par=current_par;
	}
	else
		fbhw->get_par( par );
}


static void atari_fb_set_par( struct atari_fb_par *par )
{
	fbhw->set_par( par );
	current_par=*par;
	current_par_valid=1;
}



/* =========================================================== */
/* ============== Hardware Independant Functions ============= */
/* =========================================================== */


/* used for hardware scrolling */

static int
fb_update_var(int con)
{
	current_par.screen_base=screen_base + 
			   disp[con].var.yoffset * disp[con].bytes_per_row;

	disp[con].bitplane=(unsigned char *) current_par.screen_base;
	if (fbhw->set_screen_base)
		fbhw->set_screen_base(current_par.screen_base);
	return 0;
}

static int
do_fb_set_var(struct fb_var_screeninfo *var, int isactive)
{
	int err,activate;
	struct atari_fb_par par;
	if ((err=fbhw->decode_var(var, &par)))
		return err;
	activate=var->activate;
	if (((var->activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_NOW) && isactive)
		atari_fb_set_par(&par);
	fbhw->encode_var(var, &par);
	var->activate=activate;
	return 0;
}

/* Functions for handling colormap */

/* there seems to be a bug in gcc 2.5.8 which inhibits using an other solution */
/* I always get a sigsegv */

static short red16[]=
	{ 0x0000,0x0000,0x0000,0x0000,0xc000,0xc000,0xc000,0xc000,
	  0x8000,0x0000,0x0000,0x0000,0xffff,0xffff,0xffff,0xffff};
static short green16[]=
	{ 0x0000,0x0000,0xc000,0xc000,0x0000,0x0000,0xc000,0xc000,
	  0x8000,0x0000,0xffff,0xffff,0x0000,0x0000,0xffff,0xffff};
static short blue16[]=
	{ 0x0000,0xc000,0x0000,0xc000,0x0000,0xc000,0x0000,0xc000,
	  0x8000,0xffff,0x0000,0xffff,0x0000,0xffff,0x0000,0xffff};

static short red4[]=
	{ 0x0000,0xc000,0x8000,0xffff};
static short green4[]=
	{ 0x0000,0xc000,0x8000,0xffff};
static short blue4[]=
	{ 0x0000,0xc000,0x8000,0xffff};

static short red2[]=
	{ 0x0000,0xffff};
static short green2[]=
	{ 0x0000,0xffff};
static short blue2[]=
	{ 0x0000,0xffff};

struct fb_cmap default_16_colors = { 0, 16, red16, green16, blue16, NULL };
struct fb_cmap default_4_colors = { 0, 4, red4, green4, blue4, NULL };
struct fb_cmap default_2_colors = { 0, 2, red2, green2, blue2, NULL };

static struct fb_cmap *
get_default_colormap(int bpp)
{
	if (bpp == 1)
		return &default_2_colors;
	if (bpp == 2)
		return &default_4_colors;
	return &default_16_colors;
}

#define CNVT_TOHW(val,width)	(((val) << (width)) + 0x7fff - (val)) >> 16
#define CNVT_FROMHW(val,width) ((width)?((((val) << 16) - (val)) / ((1<<(width))-1)):0)


static int
do_fb_get_cmap(struct fb_cmap *cmap, struct fb_var_screeninfo *var, int kspc)
{
	int i,start;
	unsigned short *red,*green,*blue,*transp;
	unsigned int hred,hgreen,hblue,htransp;

	red=cmap->red;
	green=cmap->green;
	blue=cmap->blue;
	transp=cmap->transp;
	start=cmap->start;
	if (start < 0)
		return EINVAL;
	for (i=0 ; i < cmap->len ; i++)
	{
		if (fbhw->getcolreg( start++, &hred, &hgreen, &hblue, &htransp ))
			return 0;
		hred=CNVT_FROMHW(hred,var->red.length);
		hgreen=CNVT_FROMHW(hgreen,var->green.length);
		hblue=CNVT_FROMHW(hblue,var->blue.length);
		htransp=CNVT_FROMHW(htransp,var->transp.length);
		if (kspc) {
			*red=hred;
			*green=hgreen;
			*blue=hblue;
			if (transp) *transp=htransp;
		}
		else {
			put_user_word(hred, red);
			put_user_word(hgreen, green);
			put_user_word(hblue, blue);
			if (transp) put_user_word( htransp, transp );
		}
		red++;
		green++;
		blue++;
		if (transp) transp++;
	}
	return 0;
}

static int
do_fb_set_cmap(struct fb_cmap *cmap, struct fb_var_screeninfo *var, int kspc)
{
	int i,start;
	unsigned short *red,*green,*blue,*transp;
	unsigned int hred,hgreen,hblue,htransp;

	red=cmap->red;
	green=cmap->green;
	blue=cmap->blue;
	transp=cmap->transp;
	start=cmap->start;

	if (start < 0)
		return -EINVAL;
	for (i=0 ; i < cmap->len ; i++)
	{
		if (kspc) {
			hred=*red;
			hgreen=*green;
			hblue=*blue;
			htransp=(transp) ? *transp : 0;
		}
		else {
			hred=get_user_word(red);
			hgreen=get_user_word(green);
			hblue=get_user_word(blue);
			htransp=(transp)?get_user_word(transp):0;
		}
		hred=CNVT_TOHW(hred,var->red.length);
		hgreen=CNVT_TOHW(hgreen,var->green.length);
		hblue=CNVT_TOHW(hblue,var->blue.length);
		htransp=CNVT_TOHW(htransp,var->transp.length);
		red++;
		green++;
		blue++;
		if (transp) transp++;
		if (fbhw->setcolreg( start++, hred, hgreen, hblue, htransp ))
			return 0;
	}
	return 0;
}

static void
do_install_cmap(int con)
{
	if (con != currcon)
		return;
	if (disp[con].cmap.len)
		do_fb_set_cmap(&disp[con].cmap, &(disp[con].var), 1);
	else
		do_fb_set_cmap(get_default_colormap(
				disp[con].var.bits_per_pixel), &(disp[con].var), 1);		
}

static void
memcpy_fs(int fsfromto, void *to, void *from, int len)
{
	switch (fsfromto) {
	case 0:
		memcpy(to,from,len);
		return;
	case 1:
		memcpy_fromfs(to,from,len);
		return;
	case 2:
		memcpy_tofs(to,from,len);
		return;
	}
}

static void
copy_cmap(struct fb_cmap *from, struct fb_cmap *to, int fsfromto)
{
	int size;
	int tooff=0, fromoff=0;

	if (to->start > from->start)
		fromoff=to->start-from->start;
	else
		tooff=from->start-to->start;			
	size=to->len-tooff;
	if (size > from->len-fromoff)
		size=from->len-fromoff;
	if (size < 0)
		return;
	size*=sizeof(unsigned short);
	memcpy_fs(fsfromto, to->red+tooff, from->red+fromoff, size);
	memcpy_fs(fsfromto, to->green+tooff, from->green+fromoff, size);
	memcpy_fs(fsfromto, to->blue+tooff, from->blue+fromoff, size);
	if (from->transp && to->transp)
		memcpy_fs(fsfromto, to->transp+tooff, from->transp+fromoff, size);
}
 
static int
alloc_cmap(struct fb_cmap *cmap,int len,int transp)
{
	int size=len*sizeof(unsigned short);
	if (cmap->len != len) {
		if (cmap->red)
			kfree(cmap->red);
		if (cmap->green)
			kfree(cmap->green);
		if (cmap->blue)
			kfree(cmap->blue);
		if (cmap->transp)
			kfree(cmap->transp);
		cmap->red=cmap->green=cmap->blue=cmap->transp=NULL;
		cmap->len=0;
		if (! len)
			return 0;
		if (! (cmap->red=kmalloc(size, GFP_ATOMIC)))
			return -1;
		if (! (cmap->green=kmalloc(size, GFP_ATOMIC)))
			return -1;
		if (! (cmap->blue=kmalloc(size, GFP_ATOMIC)))
			return -1;
		if (transp) {
			if (! (cmap->transp=kmalloc(size, GFP_ATOMIC)))
				return -1;
		}
		else
			cmap->transp=NULL;
	}
	cmap->start=0;
	cmap->len=len;
	copy_cmap(get_default_colormap(len), cmap, 0);
	return 0;
}	

static int
atari_fb_get_fix(struct fb_fix_screeninfo *fix, int con)
{
	struct atari_fb_par par;
	if (con == -1)
		atari_fb_get_par(&par);
	else
		fbhw->decode_var(&disp[con].var,&par);
	return (fbhw->encode_fix(fix, &par));
}
	
static int
atari_fb_get_var(struct fb_var_screeninfo *var, int con)
{
	struct atari_fb_par par;
	if (con == -1) {
		atari_fb_get_par(&par);
		fbhw->encode_var(var, &par);
	}
	else
		*var=disp[con].var;
	return 0;
}

static void
atari_fb_set_disp(int con)
{
	int font=default_font;
	struct fb_fix_screeninfo fix;

	atari_fb_get_fix(&fix, con);
	if (con == -1)
		con=0;
	disp[con].bytes_per_row=disp[con].var.xres*
				disp[con].var.bits_per_pixel/8;
	disp[con].screen_base=(unsigned char *) fix.smem_start;
	disp[con].bitplane=disp[con].screen_base+
			disp[con].var.yoffset*disp[con].bytes_per_row;
	disp[con].wrclut=0; /* a clut can be written */

	if (fix.visual == FB_VISUAL_PSEUDOCOLOR || 
	    fix.visual == FB_VISUAL_DIRECTCOLOR) {
		disp[con].wrclut=1;
	}

	if (font != 8 && font != 16) {
		if (disp[con].var.yres < 400)
			font=8;
		else
			font=16;
	}
	if (font == 8) { 
		disp[con].fontdata=fontdata_8x8;
		disp[con].fontheight=fontheight_8x8;
		disp[con].fontwidth=fontwidth_8x8;
	}
	else {
		disp[con].fontdata=fontdata_8x16;
		disp[con].fontheight=fontheight_8x16;
		disp[con].fontwidth=fontwidth_8x16;
	}
}

static int
atari_fb_set_var(struct fb_var_screeninfo *var, int con)
{
	int err,oldxres,oldyres,oldbpp;
	if ((err=do_fb_set_var(var, con==currcon)))
		return err;
	if ((var->activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_NOW) {
		oldxres=disp[con].var.xres;
		oldyres=disp[con].var.yres;
		oldbpp=disp[con].var.bits_per_pixel;
		disp[con].var=*var;
		disp[con].bitplane=disp[con].screen_base+
			disp[con].var.yoffset*disp[con].bytes_per_row;
		if (oldxres != var->xres || oldyres != var->yres 
		    || oldbpp != var->bits_per_pixel) {
			atari_fb_set_disp(con);
			(*fb_info.changevar)(con);
			alloc_cmap(&disp[con].cmap, 0, 0);
			do_install_cmap(con);
		}
	}
	var->activate=0;
	return 0;
}



static int
atari_fb_get_cmap(struct fb_cmap *cmap, int kspc, int con)
{
	if (con == currcon) /* current console ? */
		return do_fb_get_cmap(cmap, &(disp[con].var), kspc);
	else
		if (disp[con].cmap.len) /* non default colormap ? */
			copy_cmap(&disp[con].cmap, cmap, kspc ? 0 : 2);
		else
			copy_cmap(get_default_colormap(
			    disp[con].var.bits_per_pixel), cmap, kspc ? 0 : 2);
	return 0;
}

static int
atari_fb_set_cmap(struct fb_cmap *cmap, int kspc, int con)
{
	int err;
	if (! disp[con].cmap.len) { /* no colormap allocated ? */
		if ((err = alloc_cmap(&disp[con].cmap, 
					1 << disp[con].var.bits_per_pixel, 0)))
		return err;
	}
	if (con == currcon) /* current console ? */
		return do_fb_set_cmap(cmap, &(disp[con].var), kspc);
	else
		copy_cmap(cmap, &disp[con].cmap, kspc ? 0 : 1);
	return 0;
}

static int
atari_fb_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	       unsigned long arg, int con)
{
	return -EINVAL;
}

static struct fb_ops atari_fb_ops = {
	atari_fb_get_fix, atari_fb_get_var, atari_fb_set_var, 
	atari_fb_get_cmap, atari_fb_set_cmap, atari_fb_ioctl	
};

static void
check_default_par( int detected_mode )
{
	char default_name[10];
	int i;
	struct fb_var_screeninfo var;
	unsigned long min_mem;

	/* First try the user supplied mode */
	if (default_par) {
		var=atari_fb_predefined[default_par-1];
		var.activate = FB_ACTIVATE_TEST;
		if (do_fb_set_var(&var,1))
			default_par=0;		/* failed */
	}
	/* Next is the autodetected one */
	if (! default_par) {
		var=atari_fb_predefined[detected_mode-1]; /* autodetect */
		var.activate = FB_ACTIVATE_TEST;
		if (!do_fb_set_var(&var,1))
			default_par=detected_mode;
	}
	/* If that also failed, try some default modes... */
	if (! default_par) {
		/* try default1, default2... */
		for (i=1 ; i < 10 ; i++) {
			sprintf(default_name,"default%d",i);
			default_par=get_video_mode(default_name);
			if (! default_par)
				panic("can't set default video mode\n");
			var=atari_fb_predefined[default_par-1];
			var.activate = FB_ACTIVATE_TEST;
			if (! do_fb_set_var(&var,1))
				break;	/* ok */
		}
	}
	min_mem=var.xres_virtual * var.yres_virtual * var.bits_per_pixel/8;
	if (default_mem_req < min_mem)
		default_mem_req=min_mem;
}

static int
atafb_switch(int con)
{
	/* Do we have to save the colormap ? */
	if (disp[currcon].cmap.len)
		do_fb_get_cmap(&disp[currcon].cmap, &(disp[currcon].var), 1);
	do_fb_set_var(&disp[con].var,1);
	currcon=con;
	/* Install new colormap */
	do_install_cmap(con);
	return 0;
}

static void
atafb_blank(int blank)
{
	unsigned short black[16];
	struct fb_cmap cmap;
	if (blank) {
		memset(black, 0, 16*sizeof(unsigned short));
		cmap.red=black;
		cmap.green=black;
		cmap.blue=black;
		cmap.transp=NULL;
		cmap.start=0;
		cmap.len=16;
		do_fb_set_cmap(&cmap, &(disp[currcon].var), 1);
	}
	else
		do_install_cmap(currcon);
}

struct fb_info *
atafb_init(long *mem_start)
{
	int err;
	int pad;
	int detected_mode;
	unsigned long mem_req;
	struct fb_var_screeninfo *var;
	
	err=register_framebuffer("Atari Builtin", &node, &atari_fb_ops, 
			num_atari_fb_predefined, atari_fb_predefined);
	if (err < 0)
	  panic ("Cannot register frame buffer\n");
	if (external_addr)
		fbhw = &ext_switch;
	else if (boot_info.bi_atari.hw_present & ATARIHW_TT_SHIFTER)
		fbhw = &tt_switch;
	else if (boot_info.bi_atari.hw_present & ATARIHW_VIDEL_SHIFTER)
		fbhw = &falcon_switch;
	else if (boot_info.bi_atari.hw_present &
			 (ATARIHW_STND_SHIFTER|ATARIHW_EXTD_SHIFTER))
		fbhw = &st_switch;
	else {
		fbhw = &st_switch;
		printk( "Cannot determine video hardware; defaulting to ST(e)\n" );
	}
	detected_mode = fbhw->detect();
	check_default_par( detected_mode );
	if (!external_addr) {
		mem_req = default_mem_req;
		mem_req = ((mem_req + PAGE_SIZE - 1) & PAGE_MASK) + PAGE_SIZE;
		screen_base = (unsigned long) atari_stram_alloc( mem_req, mem_start);
		pad = ((screen_base + PAGE_SIZE-1) & PAGE_MASK) - screen_base;
		screen_base+=pad;
		screen_len=mem_req-pad;
	}
	else {
		/* Map the video memory (physical address given) to somewhere
		 * in the kernel address space.
		 */
		*mem_start = (*mem_start+PAGE_SIZE-1) & ~(PAGE_SIZE-1);
		external_addr = kernel_map( external_addr, external_len,
									KERNELMAP_NO_COPYBACK, mem_start );
		screen_base = external_addr;
		screen_len  = external_len;
		memset ((char *) screen_base, 0, screen_len);
	}

	strcpy(fb_info.modename, "Atari Builtin ");
	fb_info.disp=disp;
	fb_info.switch_con=&atafb_switch;
	fb_info.updatevar=&fb_update_var;
	fb_info.blank=&atafb_blank;
	var=atari_fb_predefined+default_par-1;
	do_fb_set_var(var,1);
	strcat(fb_info.modename,fb_var_names[default_par-1][0]);

	atari_fb_get_var(&disp[0].var, -1);
	atari_fb_set_disp(-1);
	do_install_cmap(0);
	return &fb_info;
}


void atari_video_setup( char *options, int *ints )
{	
    char *this_opt;
    int temp;
	char ext_str[80] = ""; 
	
    if (!options || !*options)
        return;
     
    for(this_opt=strtok(options,","); this_opt; this_opt=strtok(NULL,",")) {
	if (!*this_opt) continue;
	if ((temp=get_video_mode(this_opt)))
		default_par=temp;
	else if (! strcmp(this_opt, "f8x8"))
		default_font=8;
	else if (! strcmp(this_opt, "f8x16"))
		default_font=16;
	else if (! strncmp(this_opt, "hwscroll_",9)) {
		hwscroll=simple_strtoul(this_opt+9, NULL, 10);
		if (hwscroll < 0)
			hwscroll = 0;
		if (hwscroll > 200)
			hwscroll = 200;
	}
	else if (!strncmp( this_opt,"external:",9)) {
		strcpy( ext_str, this_opt+9 );
	}
	}

	if (*ext_str) {
		int		xres, yres, depth, planes;
		unsigned long addr, len;
		char *p;

		/* Format is: <xres>;<yres>;<depth>;<plane organ.>;
		 *            <screen mem addr>[;<screen mem length>]
		 */
		if (!(p = strtok( ext_str, ";" )) ||!*p) goto ext_invalid;
		xres = simple_strtoul(p, NULL, 10);
		if (xres <= 0) goto ext_invalid;

		if (!(p = strtok( NULL, ";" )) ||!*p) goto ext_invalid;
		yres = simple_strtoul(p, NULL, 10);
		if (yres <= 0) goto ext_invalid;

		if (!(p = strtok( NULL, ";" )) ||!*p) goto ext_invalid;
		depth = simple_strtoul(p, NULL, 10);
		if (depth != 1 && depth != 2 && depth != 4 && depth != 8 &&
			depth != 16 && depth != 24) goto ext_invalid;

		if (!(p = strtok( NULL, ";" )) ||!*p) goto ext_invalid;
		if (*p == 'i')
			planes = FB_TYPE_INTERLEAVED_PLANES;
		else if (*p == 'p')
			planes = FB_TYPE_PACKED_PIXELS;
		else if (*p == 'n')
			planes = FB_TYPE_PLANES;
		else if (*p == 't')
			planes = -1; /* true color */
		else
			goto ext_invalid;

		if (!(p = strtok( NULL, ";" )) ||!*p) goto ext_invalid;
		addr = simple_strtoul(p, NULL, 0);

		if (!(p = strtok( NULL, ";" )) ||!*p)
			len = xres*yres*depth/8;
		else
			len = simple_strtoul(p, NULL, 0);

		external_xres  = xres;
		external_yres  = yres;
		external_depth = depth;
		external_pmode = planes;
		external_addr  = addr;
		external_len   = len;

	  ext_invalid:
		;
	}
}


