/*
 * ataints.c -- atari Linux interrupt handling code
 *
 * 5/2/94 Roman Hodek:
 *  Added support for TT interrupts; setup for TT SCU (may someone has
 *  twiddled there and we won't get the right interrupts :-()
 *
 *  Major change: The device-independant code in m68k/ints.c didn't know
 *  about non-autovec ints yet. It hardcoded the number of possible ints to
 *  7 (IRQ1...IRQ7). But the Atari has lots of non-autovec ints! I made the
 *  number of possible ints a constant defined in interrupt.h, which is
 *  47 for the Atari. So we can call add_isr() for all Atari interrupts just
 *  the normal way. Additionally, all vectors >= 48 are initialized to call
 *  trap() instead of inthandler(). This must be changed here, too.
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 *
 */

#include <linux/types.h>
#include <linux/config.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/traps.h>
#include <linux/kernel_stat.h>

#include <asm/system.h>

#include <linux/atarihw.h>
#include <linux/atariints.h>
#include <linux/atari_stdma.h>
#include <linux/bootinfo.h>

/* isrfunc variables for atari interrupt sources */
static isr_node_t *ata_lists[NUM_ATARI_SOURCES];

/*
 * void atari_init_INTS (void)
 *
 * Parameters:	None
 *
 * Returns:	Nothing
 *
 * This function should be called during kernel startup to initialize
 * the atari IRQ handling routines.
 */

/* GK:
 * HBL IRQ handler for Falcon. Nobody needs it :-)
 * ++andreas: raise ipl to disable further interrupts.
 */
asmlinkage void falcon_hblhandler(void);
asm(".text
_falcon_hblhandler:
	moveml	d0/a0,sp@-
	btst	#4,sp@(14)	/* throwaway frame? */
	beq	1f
	movec	msp,a0		/* yes, use master stack */
	bra	2f
1:	lea	sp@(8),a0	/* else, use interrupt stack */
2:	movew	a0@,d0
	andw	#0xf8ff,d0
	orw	#0x0200,d0	/* set saved ipl to 2 */
	movew	d0,a0@
	moveml	sp@+,d0/a0
	rte");

extern void atari_microwire_cmd( int cmd );

void atari_init_INTS (void)

{
	/* Initialize the MFP(s) */
	
	mfp.vec_adr  = 0x40;	/* Automatic EOI-Mode */
	mfp.int_en_a =		    /* turn off MFP-Ints */
	mfp.int_en_b = 0x00;
	mfp.int_mk_a =			/* no Masking */
	mfp.int_mk_b = 0xff;

	if (boot_info.bi_atari.hw_present & ATARIHW_TT_MFP) {
		tt_mfp.vec_adr  = 0x50;		/* Automatic EOI-Mode */
		tt_mfp.int_en_a =			/* turn off MFP-Ints */
		tt_mfp.int_en_b = 0x00;
		tt_mfp.int_mk_a =			/* no Masking */
		tt_mfp.int_mk_b = 0xff;
	}

	if (boot_info.bi_atari.hw_present & ATARIHW_SCU) {
		/* init the SCU if present */
		tt_scu.sys_mask = 0x10;		/* enable VBL (for the cursor) and
									 * disable HSYNC interrupts (how
									 * needs them?)  MFP and SCC are
									 * enabled in VME mask
									 */
		tt_scu.vme_mask = 0x60;		/* enable MFP and SCC ints */
	}
	else {
		/* If no SCU, the HSYNC interrupt needs to be disabled this
		 * way. (Else _inthandler in kernel/sys_call.S gets overruns)
		 */
		vectors[VEC_INT2] = falcon_hblhandler;
	}

	if (boot_info.bi_atari.hw_present & ATARIHW_PCM_8BIT) {
		/* Initialize the LM1992 Sound Controller to enable the PSG sound.
		 * This is misplaced here, it should be in a atasound_init(),
		 * that doesn't exist yet.
		 */
		atari_microwire_cmd( MW_LM1992_PSG_HIGH );
	}
	
	stdma_init();

	/* Initialize the PSG: all sounds off, both ports output */
	sound_ym.rd_data_reg_sel = 7;
	sound_ym.wd_data = 0xff;
}


/*
 * atari_add_isr : add an interrupt service routine for a particular
 *		   machine specific interrupt source.
 *		   If the addition was successful, it returns 1, otherwise
 *		   it returns 0.  It will fail if another routine is already
 *		   bound into the specified source.
 *   Note that the "pri" argument is currently unused.
 */

int atari_add_isr (unsigned long source, isrfunc isr, int pri, void *data)
{
  unsigned long atari_source = source & ~IRQ_MACHSPEC;
  isr_node_t *p;

  if (atari_source >= NUM_ATARI_SOURCES
      || (atari_source >= TTMFP_SOURCE_BASE
	  && !(boot_info.bi_atari.hw_present & ATARIHW_TT_MFP))
      || (atari_source >= SCC_SOURCE_BASE
	  /* the SCC uses only half of the vectors */
	  && atari_source & 1))
    {
      printk ("atari_add_isr: Unknown uservector: %ld\n", atari_source);
      return 0;
    }

  p = new_isr_node();
  p->isr = isr;
  p->pri = pri;
  p->data = data;
  p->next = NULL;

  insert_isr (&ata_lists[atari_source], p);

  return 1;
}

void atari_process_int (int level, struct intframe *fp)
{
  ushort sr;
  int mask;

#if 0
  /* Lower the ipl to let interrupts with higher priority come through */
  asm volatile ("movew sr,%0" : "=dm" (sr));
  sr -= 0x100;
  asm volatile ("movew %0,sr" : : "dm" (sr));
#endif

  if (level >= NUM_ATARI_SOURCES)
    panic ("Unknown user interrupt level %d", level);

  if (level >= SCC_SOURCE_BASE)
    /* SCC interrupts are at level 5 */
    kstat.interrupts[5]++;
  else
    /* MFP interrupts are at level 6 */
    kstat.interrupts[6]++;

  call_isr_list (ata_lists[level], fp);

#if 0
  /* Raise ipl again to avoid stacked interrupts.  */
  asm volatile ("movew sr,%0" : "=dm" (sr));
  sr += 0x100;
  asm volatile ("movew %0,sr" : : "dm" (sr));

  /* acknowledge the interrupt */
  mask = ~(1 << (level & 7));
  switch (level >> 3)
    {
    case 0:
      mfp.int_sv_b = mask;
      break;
    case 1:
      mfp.int_sv_a = mask;
      break;
    case 2:
      tt_mfp.int_sv_b = mask;
      break;
    case 3:
      tt_mfp.int_sv_a = mask;
      break;
    case 4:
    case 5:
      scc.cha_a_ctrl = 070;	/* Reset highest IUS */
      break;
    }
#endif
}
