/*
 * linux/atari/atakeyb.c
 *
 * Atari Keyboard driver for 680x0 Linux
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 */

/*
 * Atari support by Robert de Vries
 * enhanced by Bjoern Brauel and Roman Hodek
 */

#include <linux/config.h>
#include <linux/sched.h>
#include <linux/keyboard.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/kd.h>

#include <linux/atariints.h>
#include <linux/atarihw.h>
#include <linux/atarikb.h>
#include <linux/atari_mouse.h>
#include <linux/atari_joystick.h>

extern void process_scancode (int);
unsigned char mach_keyboard_type;
static void atakeyb_rep( unsigned long ignore );

#define ATAKEY_CAPS	(58)
#define BREAK_MASK	(0x80)

/*
 * ++roman: The following changes were applied manually:
 *
 *  - The Alt (= Meta) key works in combination with Shift and
 *    Control, e.g. Alt+Shift+a sends Meta-A (0xc1), Alt+Control+A sends
 *    Meta-Ctrl-A (0x81) ...
 *
 *  - The parentheses on the keypad send '(' and ')' with all
 *    modifiers (as would do e.g. keypad '+'), but they cannot be used as
 *    application keys (i.e. sending Esc O c).
 *
 *  - HELP and UNDO are mapped to be F21 and F24, resp, that send the
 *    codes "\E[M" and "\E[P". (This is better than the old mapping to
 *    F11 and F12, because these codes are on Shift+F1/2 anyway.) This
 *    way, applications that allow their own keyboard mappings
 *    (e.g. tcsh, X Windows) can be configured to use them in the way
 *    the label suggests (providing help or undoing).
 *
 *  - Console switching is done with Alt+Fx (consoles 1..10) and
 *    Shift+Alt+Fx (consoles 11..20).
 *
 *  - The misc. special function implemented in the kernel are mapped
 *    to the following key combinations:
 *
 *      ClrHome          -> Home/Find
 *      Shift + ClrHome  -> End/Select
 *      Shift + Up       -> Page Up
 *      Shift + Down     -> Page Down
 *      Alt + Help       -> show system status
 *      Shift + Help     -> show memory info
 *      Ctrl + Help      -> show registers
 *      Ctrl + Alt + Del -> Reboot
 *      Alt + Undo       -> switch to last console
 *      Shift + Undo     -> send interrupt
 *      Alt + Insert     -> stop/start output (same as ^S/^Q)
 *      Alt + Up         -> Scroll back console (if implemented)
 *      Alt + Down       -> Scroll forward console (if implemented)
 *      Alt + CapsLock   -> NumLock
 *
 */

#define K_F21	K_MACRO
#define K_F24	K_PAUSE

unsigned short atakey_map[NR_KEYMAPS][NR_KEYS] = {
    { /* no modifiers */
/* 00-03 */	K_HOLE,	'['&0x1f,	'1',	'2',
/* 04-07 */	'3',	'4',	'5',	'6',
/* 08-0b */	'7',	'8',	'9',	'0',
/* 0c-0f */	'-',	'=',	'H'&0x1f,	'I'&0x1f,
/* 10-13 */	K(KT_LETTER,'q'),	K(KT_LETTER,'w'),	K(KT_LETTER,'e'),	K(KT_LETTER,'r'),
/* 14-17 */	K(KT_LETTER,'t'),	K(KT_LETTER,'y'),	K(KT_LETTER,'u'),	K(KT_LETTER,'i'),
/* 18-1b */	K(KT_LETTER,'o'),	K(KT_LETTER,'p'),	'[',	']',
/* 1c-1f */	K_ENTER,	K_CTRL,	K(KT_LETTER,'a'),	K(KT_LETTER,'s'),
/* 20-23 */	K(KT_LETTER,'d'),	K(KT_LETTER,'f'),	K(KT_LETTER,'g'),	K(KT_LETTER,'h'),
/* 24-27 */	K(KT_LETTER,'j'),	K(KT_LETTER,'k'),	K(KT_LETTER,'l'),	';',
/* 28-2b */	'\'',	'`',	K_SHIFT,	'\\',
/* 2c-2f */	K(KT_LETTER,'z'),	K(KT_LETTER,'x'),	K(KT_LETTER,'c'),	K(KT_LETTER,'v'),
/* 30-33 */	K(KT_LETTER,'b'),	K(KT_LETTER,'n'),	K(KT_LETTER,'m'),	',',
/* 34-37 */	'.',	'/',	K_SHIFT,	K_HOLE,
/* 38-3b */	K_ALT,	' ',	K_CAPS,	K_F1,
/* 3c-3f */	K_F2,	K_F3,	K_F4,	K_F5,
/* 40-43 */	K_F6,	K_F7,	K_F8,	K_F9,
/* 44-47 */	K_F10,	K_HOLE,	K_HOLE,	K_FIND,
/* 48-4b */	K_UP,	K_HOLE,	K_PMINUS,	K_LEFT,
/* 4c-4f */	K_HOLE,	K_RIGHT,	K_PPLUS,	K_HOLE,
/* 50-53 */	K_DOWN,	K_HOLE,	K_INSERT,	'?'|0x40,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 5c-5f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 60-63 */	K_HOLE,	K_F24,	K_F21,	'(',
/* 64-67 */	')',	K_PSLASH,	K_PSTAR,	K_P7,
/* 68-6b */	K_P8,	K_P9,	K_P4,	K_P5,
/* 6c-6f */	K_P6,	K_P1,	K_P2,	K_P3,
/* 70-73 */	K_P0,	K_PDOT,	K_PENTER,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* shift */
/* 00-03 */	K_HOLE,	'['&0x1f,	'!',	'@',
/* 04-07 */	'#',	'$',	'%',	'^',
/* 08-0b */	'&',	'*',	'(',	')',
/* 0c-0f */	'_',	'+',	'H'&0x1f,	'I'&0x1f,
/* 10-13 */	K(KT_LETTER,'Q'),	K(KT_LETTER,'W'),	K(KT_LETTER,'E'),	K(KT_LETTER,'R'),
/* 14-17 */	K(KT_LETTER,'T'),	K(KT_LETTER,'Y'),	K(KT_LETTER,'U'),	K(KT_LETTER,'I'),
/* 18-1b */	K(KT_LETTER,'O'),	K(KT_LETTER,'P'),	'{',	'}',
/* 1c-1f */	K_ENTER,	K_CTRL,	K(KT_LETTER,'A'),	K(KT_LETTER,'S'),
/* 20-23 */	K(KT_LETTER,'D'),	K(KT_LETTER,'F'),	K(KT_LETTER,'G'),	K(KT_LETTER,'H'),
/* 24-27 */	K(KT_LETTER,'J'),	K(KT_LETTER,'K'),	K(KT_LETTER,'L'),	':',
/* 28-2b */	'"',	'~',	K_SHIFT,	'|',
/* 2c-2f */	K(KT_LETTER,'Z'),	K(KT_LETTER,'X'),	K(KT_LETTER,'C'),	K(KT_LETTER,'V'),
/* 30-33 */	K(KT_LETTER,'B'),	K(KT_LETTER,'N'),	K(KT_LETTER,'M'),	'<',
/* 34-37 */	'>',	'?',	K_SHIFT,	K_HOLE,
/* 38-3b */	K_ALT,	' ',	K_CAPS,	K_F11,
/* 3c-3f */	K_F12,	K_F13,	K_F14,	K_F15,
/* 40-43 */	K_F16,	K_F17,	K_F18,	K_F19,
/* 44-47 */	K_F20,	K_HOLE,	K_HOLE,	K_SELECT,
/* 48-4b */	K_PGUP,	K_HOLE,	K_PMINUS,	K_LEFT,
/* 4c-4f */	K_HOLE,	K_RIGHT,	K_PPLUS,	K_HOLE,
/* 50-53 */	K_PGDN,	K_HOLE,	K_INSERT,	'?'|0x40,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 5c-5f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 60-63 */	K_HOLE,	K_BREAK,	K_SH_MEM,	'(',
/* 64-67 */	')',	K_PSLASH,	K_PSTAR,	K_P7,
/* 68-6b */	K_P8,	K_P9,	K_P4,	K_P5,
/* 6c-6f */	K_P6,	K_P1,	K_P2,	K_P3,
/* 70-73 */	K_P0,	K_PDOT,	K_PENTER,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* altgr */ /* this can't happen on Atari (no AltGr key!) */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	'@',
/* 04-07 */	K_HOLE,	'$',	K_HOLE,	K_HOLE,
/* 08-0b */	'{',	'[',	']',	'}',
/* 0c-0f */	'\\',	K_HOLE,	'H'&0x1f,	K_HOLE,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	'~',
/* 1c-1f */	K_ENTER,	K_CTRL,	K_HOLE,	K_HOLE,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 38-3b */	K_ALT,	K_HOLE,	K_CAPS,	K(KT_CONS,12),
/* 3c-3f */	K(KT_CONS,13),	K(KT_CONS,14),	K(KT_CONS,15),	K(KT_CONS,16),
/* 40-43 */	K(KT_CONS,17),	K(KT_CONS,18),	K(KT_CONS,19),	K(KT_CONS,20),
/* 44-47 */	K(KT_CONS,21),	K_HOLE,	K_HOLE,	K_FIND,
/* 48-4b */	K_UP,	K_HOLE,	K_PMINUS,	K_LEFT,
/* 4c-4f */	K_HOLE,	K_RIGHT,	K_PPLUS,	K_HOLE,
/* 50-53 */	K_DOWN,	K_HOLE,	K_INSERT,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 5c-5f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 60-63 */	K_HOLE,	K_HOLE,	K_HELP,	K_HOLE,
/* 64-67 */	K_HOLE,	K_PSLASH,	K_PSTAR,	K_P7,
/* 68-6b */	K_P8,	K_P9,	K_P4,	K_P5,
/* 6c-6f */	K_P6,	K_P1,	K_P2,	K_P3,
/* 70-73 */	K_P0,	K_PDOT,	K_PENTER,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* shift altgr */ /* this can't happen on Atari (no AltGr key!) */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	'H'&0x1f,	K_HOLE,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_ENTER,	K_CTRL,	K_HOLE,	K_HOLE,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 38-3b */	K_ALT,	K_HOLE,	K_CAPS,	K_HOLE,
/* 3c-3f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 40-43 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 44-47 */	K_HOLE,	K_HOLE,	K_HOLE,	K_FIND,
/* 48-4b */	K_UP,	K_HOLE,	K_PMINUS,	K_LEFT,
/* 4c-4f */	K_HOLE,	K_RIGHT,	K_PPLUS,	K_HOLE,
/* 50-53 */	K_DOWN,	K_HOLE,	K_INSERT,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 5c-5f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 60-63 */	K_HOLE,	K_HOLE,	K_HELP,	K_HOLE,
/* 64-67 */	K_HOLE,	K_PSLASH,	K_PSTAR,	K_P7,
/* 68-6b */	K_P8,	K_P9,	K_P4,	K_P5,
/* 6c-6f */	K_P6,	K_P1,	K_P2,	K_P3,
/* 70-73 */	K_P0,	K_PDOT,	K_PENTER,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* control */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	'@'&0x1f,
/* 04-07 */	'['&0x1f,	'\\'&0x1f,	']'&0x1f,	'^'&0x1f,
/* 08-0b */	'_'&0x1f,	'?'|0x40,	K_HOLE,	K_HOLE,
/* 0c-0f */	'_'&0x1f,	K_HOLE,	'H'&0x1f,	K_HOLE,
/* 10-13 */	'Q'&0x1f,	'W'&0x1f,	'E'&0x1f,	'R'&0x1f,
/* 14-17 */	'T'&0x1f,	'Y'&0x1f,	'U'&0x1f,	'I'&0x1f,
/* 18-1b */	'O'&0x1f,	'P'&0x1f,	'['&0x1f,	']'&0x1f,
/* 1c-1f */	K_ENTER,	K_CTRL,	'A'&0x1f,	'S'&0x1f,
/* 20-23 */	'D'&0x1f,	'F'&0x1f,	'G'&0x1f,	'H'&0x1f,
/* 24-27 */	'J'&0x1f,	'K'&0x1f,	'L'&0x1f,	K_HOLE,
/* 28-2b */	'G'&0x1f,	'@'&0x1f,	K_SHIFT,	'\\'&0x1f,
/* 2c-2f */	'Z'&0x1f,	'X'&0x1f,	'C'&0x1f,	'V'&0x1f,
/* 30-33 */	'B'&0x1f,	'N'&0x1f,	'M'&0x1f,	K_HOLE,
/* 34-37 */	K_HOLE,	'?'|0x40,	K_SHIFT,	K_HOLE,
/* 38-3b */	K_ALT,	'@'&0x1f,	K_CAPS,	K_F1,
/* 3c-3f */	K_F2,	K_F3,	K_F4,	K_F5,
/* 40-43 */	K_F6,	K_F7,	K_F8,	K_F9,
/* 44-47 */	K_F10,	K_HOLE,	K_HOLE,	K_FIND,
/* 48-4b */	K_UP,	K_HOLE,	K_PMINUS,	K_LEFT,
/* 4c-4f */	K_HOLE,	K_RIGHT,	K_PPLUS,	K_HOLE,
/* 50-53 */	K_DOWN,	K_HOLE,	K_INSERT,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 5c-5f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 60-63 */	K_HOLE,	K_F24,	K_SH_REGS,	'(',
/* 64-67 */	')',	K_PSLASH,	K_PSTAR,	K_P7,
/* 68-6b */	K_P8,	K_P9,	K_P4,	K_P5,
/* 6c-6f */	K_P6,	K_P1,	K_P2,	K_P3,
/* 70-73 */	K_P0,	K_PDOT,	K_PENTER,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* shift control */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	'H'&0x1f,	K_HOLE,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_ENTER,	K_CTRL,	K_HOLE,	K_HOLE,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 38-3b */	K_ALT,	K_HOLE,	K_CAPS,	K_HOLE,
/* 3c-3f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 40-43 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 44-47 */	K_HOLE,	K_HOLE,	K_HOLE,	K_FIND,
/* 48-4b */	K_UP,	K_HOLE,	K_PMINUS,	K_LEFT,
/* 4c-4f */	K_HOLE,	K_RIGHT,	K_PPLUS,	K_HOLE,
/* 50-53 */	K_DOWN,	K_HOLE,	K_INSERT,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 5c-5f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 60-63 */	K_HOLE,	K_HOLE,	K_HOLE,	'(',
/* 64-67 */	')',	K_PSLASH,	K_PSTAR,	K_P7,
/* 68-6b */	K_P8,	K_P9,	K_P4,	K_P5,
/* 6c-6f */	K_P6,	K_P1,	K_P2,	K_P3,
/* 70-73 */	K_P0,	K_PDOT,	K_PENTER,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* control altgr */ /* this can't happen on Atari (no AltGr key!) */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	'H'&0x1f,	K_HOLE,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_ENTER,	K_CTRL,	K_HOLE,	K_HOLE,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 38-3b */	K_ALT,	K_HOLE,	K_CAPS,	K_HOLE,
/* 3c-3f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 40-43 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 44-47 */	K_HOLE,	K_HOLE,	K_HOLE,	K_FIND,
/* 48-4b */	K_UP,	K_HOLE,	K_PMINUS,	K_LEFT,
/* 4c-4f */	K_HOLE,	K_RIGHT,	K_PPLUS,	K_HOLE,
/* 50-53 */	K_DOWN,	K_HOLE,	K_INSERT,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 5c-5f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 60-63 */	K_HOLE,	K_HOLE,	K_HELP,	K_HOLE,
/* 64-67 */	K_HOLE,	K_PSLASH,	K_PSTAR,	K_P7,
/* 68-6b */	K_P8,	K_P9,	K_P4,	K_P5,
/* 6c-6f */	K_P6,	K_P1,	K_P2,	K_P3,
/* 70-73 */	K_P0,	K_BOOT,	K_PENTER,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* shift control altgr */ /* this can't happen on Atari (no AltGr key!) */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	'H'&0x1f,	K_HOLE,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_ENTER,	K_CTRL,	K_HOLE,	K_HOLE,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 38-3b */	K_ALT,	K_HOLE,	K_CAPS,	K_HOLE,
/* 3c-3f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 40-43 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 44-47 */	K_HOLE,	K_HOLE,	K_HOLE,	K_FIND,
/* 48-4b */	K_UP,	K_HOLE,	K_PMINUS,	K_LEFT,
/* 4c-4f */	K_HOLE,	K_RIGHT,	K_PPLUS,	K_HOLE,
/* 50-53 */	K_DOWN,	K_HOLE,	K_INSERT,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 5c-5f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 60-63 */	K_HOLE,	K_HOLE,	K_HELP,	K_HOLE,
/* 64-67 */	K_HOLE,	K_PSLASH,	K_PSTAR,	K_P7,
/* 68-6b */	K_P8,	K_P9,	K_P4,	K_P5,
/* 6c-6f */	K_P6,	K_P1,	K_P2,	K_P3,
/* 70-73 */	K_P0,	K_PDOT,	K_PENTER,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* alt */
/* 00-03 */	K_HOLE,	K(KT_META,'['&0x1f),	K(KT_META,'1'),	K(KT_META,'2'),
/* 04-07 */	K(KT_META,'3'),	K(KT_META,'4'),	K(KT_META,'5'),	K(KT_META,'6'),
/* 08-0b */	K(KT_META,'7'),	K(KT_META,'8'),	K(KT_META,'9'),	K(KT_META,'0'),
/* 0c-0f */	K(KT_META,'-'),	K(KT_META,'='),	K(KT_META,'H'&0x1f),	K(KT_META,'I'&0x1f),
/* 10-13 */	K(KT_META,'q'),	K(KT_META,'w'),	K(KT_META,'e'),	K(KT_META,'r'),
/* 14-17 */	K(KT_META,'t'),	K(KT_META,'y'),	K(KT_META,'u'),	K(KT_META,'i'),
/* 18-1b */	K(KT_META,'o'),	K(KT_META,'p'),	K(KT_META,'['),	K(KT_META,']'),
/* 1c-1f */	K(KT_META,'M'&0x1f),	K_CTRL,	K(KT_META,'a'),	K(KT_META,'s'),
/* 20-23 */	K(KT_META,'d'),	K(KT_META,'f'),	K(KT_META,'g'),	K(KT_META,'h'),
/* 24-27 */	K(KT_META,'j'),	K(KT_META,'k'),	K(KT_META,'l'),	K(KT_META,';'),
/* 28-2b */	K(KT_META,'\''),	K(KT_META,'`'),	K_SHIFT,	K(KT_META,'\\'),
/* 2c-2f */	K(KT_META,'z'),	K(KT_META,'x'),	K(KT_META,'c'),	K(KT_META,'v'),
/* 30-33 */	K(KT_META,'b'),	K(KT_META,'n'),	K(KT_META,'m'),	K(KT_META,','),
/* 34-37 */	K(KT_META,'.'),	K(KT_META,'/'),	K_SHIFT,	K_HOLE,
/* 38-3b */	K_ALT,	K(KT_META,' '),	K_NUM,	K(KT_CONS,0),
/* 3c-3f */	K(KT_CONS,1),	K(KT_CONS,2),	K(KT_CONS,3),	K(KT_CONS,4),
/* 40-43 */	K(KT_CONS,5),	K(KT_CONS,6),	K(KT_CONS,7),	K(KT_CONS,8),
/* 44-47 */	K(KT_CONS,9),	K_HOLE,	K_HOLE,	K_FIND,
/* 48-4b */	K_SCROLLBACK,	K_HOLE,	K_PMINUS,	K_LEFT,
/* 4c-4f */	K_HOLE,	K_RIGHT,	K_PPLUS,	K_HOLE,
/* 50-53 */	K_SCROLLFORW,	K_HOLE,	K_HOLD,	K(KT_META,'?'|0x40),
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 5c-5f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 60-63 */	K_HOLE,	K_CONS,	K_SH_STAT,	'(',
/* 64-67 */	')',	K_PSLASH,	K_PSTAR,	K_ASC7,
/* 68-6b */	K_ASC8,	K_ASC9,	K_ASC4,	K_ASC5,
/* 6c-6f */	K_ASC6,	K_ASC1,	K_ASC2,	K_ASC3,
/* 70-73 */	K_ASC0,	K_PDOT,	K_PENTER,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* shift alt */
/* 00-03 */	K_HOLE,	K(KT_META,'['&0x1f),	K(KT_META,'!'),	K(KT_META,'@'),
/* 04-07 */	K(KT_META,'#'),	K(KT_META,'$'),	K(KT_META,'%'),	K(KT_META,'^'),
/* 08-0b */	K(KT_META,'&'),	K(KT_META,'*'),	K(KT_META,'('),	K(KT_META,')'),
/* 0c-0f */	K(KT_META,'_'),	K(KT_META,'+'),	K(KT_META,'H'&0x1f),	K(KT_META,'I'&0x1f),
/* 10-13 */	K(KT_META,'Q'),	K(KT_META,'W'),	K(KT_META,'E'),	K(KT_META,'R'),
/* 14-17 */	K(KT_META,'T'),	K(KT_META,'Y'),	K(KT_META,'U'),	K(KT_META,'I'),
/* 18-1b */	K(KT_META,'O'),	K(KT_META,'P'),	K(KT_META,'{'),	K(KT_META,'}'),
/* 1c-1f */	K(KT_META,K_ENTER),	K_CTRL,	K(KT_META,'A'),	K(KT_META,'S'),
/* 20-23 */	K(KT_META,'D'),	K(KT_META,'F'),	K(KT_META,'G'),	K(KT_META,'H'),
/* 24-27 */	K(KT_META,'J'),	K(KT_META,'K'),	K(KT_META,'L'),	K(KT_META,':'),
/* 28-2b */	K(KT_META,'"'),	K(KT_META,'~'),	K_SHIFT,	K(KT_META,'|'),
/* 2c-2f */	K(KT_META,'Z'),	K(KT_META,'X'),	K(KT_META,'C'),	K(KT_META,'V'),
/* 30-33 */	K(KT_META,'B'),	K(KT_META,'N'),	K(KT_META,'M'),	K(KT_META,'<'),
/* 34-37 */	K(KT_META,'>'),	K(KT_META,'?'),	K_SHIFT,	K_HOLE,
/* 38-3b */	K_ALT,	K(KT_META,' '),	K_CAPS,	K(KT_CONS,10),
/* 3c-3f */	K(KT_CONS,11),	K(KT_CONS,12),	K(KT_CONS,13),	K(KT_CONS,14),
/* 40-43 */	K(KT_CONS,15),	K(KT_CONS,16),	K(KT_CONS,17),	K(KT_CONS,18),
/* 44-47 */	K(KT_CONS,19),	K_HOLE,	K_HOLE,	K_SELECT,
/* 48-4b */	K_PGUP,	K_HOLE,	K_PMINUS,	K_LEFT,
/* 4c-4f */	K_HOLE,	K_RIGHT,	K_PPLUS,	K_HOLE,
/* 50-53 */	K_PGDN,	K_HOLE,	K_INSERT,	K(KT_META,'?'|0x40),
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 5c-5f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 60-63 */	K_HOLE,	K_HOLE,	K_HOLE,	'(',
/* 64-67 */	')',	K_PSLASH,	K_PSTAR,	K_P7,
/* 68-6b */	K_P8,	K_P9,	K_P4,	K_P5,
/* 6c-6f */	K_P6,	K_P1,	K_P2,	K_P3,
/* 70-73 */	K_P0,	K_PDOT,	K_PENTER,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* alt altgr */ /* this can't happen on Atari (no AltGr key!) */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	'H'&0x1f,	K_HOLE,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_ENTER,	K_CTRL,	K_HOLE,	K_HOLE,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 38-3b */	K_ALT,	K_HOLE,	K_CAPS,	K_HOLE,
/* 3c-3f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 40-43 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 44-47 */	K_HOLE,	K_HOLE,	K_HOLE,	K_FIND,
/* 48-4b */	K_UP,	K_HOLE,	K_PMINUS,	K_LEFT,
/* 4c-4f */	K_HOLE,	K_RIGHT,	K_PPLUS,	K_HOLE,
/* 50-53 */	K_DOWN,	K_HOLE,	K_INSERT,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 5c-5f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 60-63 */	K_HOLE,	K_HOLE,	K_HELP,	K_HOLE,
/* 64-67 */	K_HOLE,	K_PSLASH,	K_PSTAR,	K_P7,
/* 68-6b */	K_P8,	K_P9,	K_P4,	K_P5,
/* 6c-6f */	K_P6,	K_P1,	K_P2,	K_P3,
/* 70-73 */	K_P0,	K_PDOT,	K_PENTER,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* shift alt altgr */ /* this can't happen on Atari (no AltGr key!) */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	'H'&0x1f,	K_HOLE,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_ENTER,	K_CTRL,	K_HOLE,	K_HOLE,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 38-3b */	K_ALT,	K_HOLE,	K_CAPS,	K_HOLE,
/* 3c-3f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 40-43 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 44-47 */	K_HOLE,	K_HOLE,	K_HOLE,	K_FIND,
/* 48-4b */	K_UP,	K_HOLE,	K_PMINUS,	K_LEFT,
/* 4c-4f */	K_HOLE,	K_RIGHT,	K_PPLUS,	K_HOLE,
/* 50-53 */	K_DOWN,	K_HOLE,	K_INSERT,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 5c-5f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 60-63 */	K_HOLE,	K_HOLE,	K_HELP,	K_HOLE,
/* 64-67 */	K_HOLE,	K_PSLASH,	K_PSTAR,	K_P7,
/* 68-6b */	K_P8,	K_P9,	K_P4,	K_P5,
/* 6c-6f */	K_P6,	K_P1,	K_P2,	K_P3,
/* 70-73 */	K_P0,	K_PDOT,	K_PENTER,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* control alt */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K(KT_META,'@'&0x1f),
/* 04-07 */	K(KT_META,'['&0x1f),	K(KT_META,'\\'&0x1f),	K(KT_META,']'&0x1f),	K(KT_META,'^'&0x1f),
/* 08-0b */	K(KT_META,'_'&0x1f),	K(KT_META,'?'|0x40),	K_HOLE,	K_HOLE,
/* 0c-0f */	K(KT_META,'_'&0x1f),	K_HOLE,	K(KT_META,'H'&0x1f),	K_HOLE,
/* 10-13 */	K(KT_META,'Q'&0x1f),	K(KT_META,'W'&0x1f),	K(KT_META,'E'&0x1f),	K(KT_META,'R'&0x1f),
/* 14-17 */	K(KT_META,'T'&0x1f),	K(KT_META,'Y'&0x1f),	K(KT_META,'U'&0x1f),	K(KT_META,'I'&0x1f),
/* 18-1b */	K(KT_META,'O'&0x1f),	K(KT_META,'P'&0x1f),	K(KT_META,'['&0x1f),	K(KT_META,']'&0x1f),
/* 1c-1f */	K(KT_META,K_ENTER),	K_CTRL,	K(KT_META,'A'&0x1f),	K(KT_META,'S'&0x1f),
/* 20-23 */	K(KT_META,'D'&0x1f),	K(KT_META,'F'&0x1f),	K(KT_META,'G'&0x1f),	K(KT_META,'H'&0x1f),
/* 24-27 */	K(KT_META,'J'&0x1f),	K(KT_META,'K'&0x1f),	K(KT_META,'L'&0x1f),	K_HOLE,
/* 28-2b */	K(KT_META,'G'&0x1f),	K(KT_META,'@'&0x1f),	K_SHIFT,	K(KT_META,'\\'&0x1f),
/* 2c-2f */	K(KT_META,'Z'&0x1f),	K(KT_META,'X'&0x1f),	K(KT_META,'C'&0x1f),	K(KT_META,'V'&0x1f),
/* 30-33 */	K(KT_META,'B'&0x1f),	K(KT_META,'N'&0x1f),	K(KT_META,'M'&0x1f),	K_HOLE,
/* 34-37 */	K_HOLE,	K(KT_META,'?'|0x40),	K_SHIFT,	K_HOLE,
/* 38-3b */	K_ALT,	K(KT_META,'@'&0x1f),	K_CAPS,	K_F1,
/* 3c-3f */	K_F2,	K_F3,	K_F4,	K_F5,
/* 40-43 */	K_F6,	K_F7,	K_F8,	K_F9,
/* 44-47 */	K_F10,	K_HOLE,	K_HOLE,	K_FIND,
/* 48-4b */	K_UP,	K_HOLE,	K_PMINUS,	K_LEFT,
/* 4c-4f */	K_HOLE,	K_RIGHT,	K_PPLUS,	K_HOLE,
/* 50-53 */	K_DOWN,	K_HOLE,	K_INSERT,	K_BOOT,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 5c-5f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 60-63 */	K_HOLE,	K_HOLE,	K_HOLE,	'(',
/* 64-67 */	')',	K_PSLASH,	K_PSTAR,	K_P7,
/* 68-6b */	K_P8,	K_P9,	K_P4,	K_P5,
/* 6c-6f */	K_P6,	K_P1,	K_P2,	K_P3,
/* 70-73 */	K_P0,	K_PDOT,	K_PENTER,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* shift control alt */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	K(KT_META,'H'&0x1f),	K_HOLE,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K(KT_META,K_ENTER),	K_CTRL,	K_HOLE,	K_HOLE,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 38-3b */	K_ALT,	K_HOLE,	K_CAPS,	K_HOLE,
/* 3c-3f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 40-43 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 44-47 */	K_HOLE,	K_HOLE,	K_HOLE,	K_FIND,
/* 48-4b */	K_UP,	K_HOLE,	K_PMINUS,	K_LEFT,
/* 4c-4f */	K_HOLE,	K_RIGHT,	K_PPLUS,	K_HOLE,
/* 50-53 */	K_DOWN,	K_HOLE,	K_INSERT,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 5c-5f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 60-63 */	K_HOLE,	K_HOLE,	K_HOLE,	'(',
/* 64-67 */	')',	K_PSLASH,	K_PSTAR,	K_P7,
/* 68-6b */	K_P8,	K_P9,	K_P4,	K_P5,
/* 6c-6f */	K_P6,	K_P1,	K_P2,	K_P3,
/* 70-73 */	K_P0,	K_PDOT,	K_PENTER,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* control alt altgr */ /* this can't happen on Atari (no AltGr key!) */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	'H'&0x1f,	K_HOLE,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_ENTER,	K_CTRL,	K_HOLE,	K_HOLE,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 38-3b */	K_ALT,	K_HOLE,	K_CAPS,	K_HOLE,
/* 3c-3f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 40-43 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 44-47 */	K_HOLE,	K_HOLE,	K_HOLE,	K_FIND,
/* 48-4b */	K_UP,	K_HOLE,	K_PMINUS,	K_LEFT,
/* 4c-4f */	K_HOLE,	K_RIGHT,	K_PPLUS,	K_HOLE,
/* 50-53 */	K_DOWN,	K_HOLE,	K_INSERT,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 5c-5f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 60-63 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 64-67 */	K_HOLE,	K_PSLASH,	K_PSTAR,	K_P7,
/* 68-6b */	K_P8,	K_P9,	K_P4,	K_P5,
/* 6c-6f */	K_P6,	K_P1,	K_P2,	K_P3,
/* 70-73 */	K_P0,	K_PDOT,	K_PENTER,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    }, { /* shift control alt altgr */ /* this can't happen on Atari (no AltGr key!) */
/* 00-03 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 04-07 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 08-0b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 0c-0f */	K_HOLE,	K_HOLE,	'H'&0x1f,	K_HOLE,
/* 10-13 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 14-17 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 18-1b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 1c-1f */	K_ENTER,	K_CTRL,	K_HOLE,	K_HOLE,
/* 20-23 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 24-27 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 28-2b */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 2c-2f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 30-33 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 34-37 */	K_HOLE,	K_HOLE,	K_SHIFT,	K_HOLE,
/* 38-3b */	K_ALT,	K_HOLE,	K_CAPS,	K_HOLE,
/* 3c-3f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 40-43 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 44-47 */	K_HOLE,	K_HOLE,	K_HOLE,	K_FIND,
/* 48-4b */	K_UP,	K_HOLE,	K_PMINUS,	K_LEFT,
/* 4c-4f */	K_HOLE,	K_RIGHT,	K_PPLUS,	K_HOLE,
/* 50-53 */	K_DOWN,	K_HOLE,	K_INSERT,	K_HOLE,
/* 54-57 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 58-5b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 5c-5f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 60-63 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 64-67 */	K_HOLE,	K_PSLASH,	K_PSTAR,	K_P7,
/* 68-6b */	K_P8,	K_P9,	K_P4,	K_P5,
/* 6c-6f */	K_P6,	K_P1,	K_P2,	K_P3,
/* 70-73 */	K_P0,	K_PDOT,	K_PENTER,	K_HOLE,
/* 74-77 */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 78-7b */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
/* 7c-7f */	K_HOLE,	K_HOLE,	K_HOLE,	K_HOLE,
    },
};



typedef enum kb_state_t
{
    KEYBOARD, AMOUSE, RMOUSE, JOYSTICK, CLOCK
} KB_STATE_T;

typedef struct keyboard_state
{
    unsigned char  buf[6];
    int 	   len;
    KB_STATE_T	   state;
} KEYBOARD_STATE;

KEYBOARD_STATE kb_state;

#define	DEFAULT_KEYB_REP_DELAY	(HZ/4)
#define	DEFAULT_KEYB_REP_RATE	(HZ/25)

/* These could be settable by some ioctl() in future... */
static unsigned int key_repeat_delay = DEFAULT_KEYB_REP_DELAY;
static unsigned int key_repeat_rate  = DEFAULT_KEYB_REP_RATE;

static unsigned char rep_scancode;
static struct timer_list atakeyb_rep_timer = { NULL, NULL, 0, 0, atakeyb_rep };


static void atakeyb_rep( unsigned long ignore )

{	unsigned long flags;
	
	/* process_scancode() mucks around with the tty queues without
	 * disabling interrupts! It assumes that it is called from an irq...
	 */
	save_flags(flags);
	cli();

	del_timer( &atakeyb_rep_timer );
	atakeyb_rep_timer.expires = key_repeat_rate;
	add_timer( &atakeyb_rep_timer );

	process_scancode( rep_scancode );

	restore_flags(flags);
}


static void keyboard_interrupt(struct  intframe *fp, void *data)
{
  u_char acia_stat;
  int scancode;
  int break_flag;

 repeat:
  if (acia.mid_ctrl & ACIA_IRQ)
	MIDI_int();
  acia_stat = acia.key_ctrl;
  /* check out if the interrupt came from this ACIA */
  if (!((acia_stat | acia.mid_ctrl) & ACIA_IRQ))
	return;

    if (acia_stat & ACIA_OVRN)
    {
	/* a very fast typist or a slow system, give a warning */
	printk( "Keyboard overrun\n" );
	goto get_key;
    }

    if (acia_stat & ACIA_RDRF)	/* received a character */
    {
    get_key:
	scancode = acia.key_data;	/* get it or reset the ACIA, I'll get it! */
	switch (kb_state.state)
	{
	  case KEYBOARD:
	    switch (scancode)
	    {
	      case 0xF7:
		kb_state.state = AMOUSE;
		kb_state.len = 0;
		break;

	      case 0xF8:
	      case 0xF9:
     	      case 0xFA:
	      case 0xFB:
		kb_state.state = RMOUSE;
	    	kb_state.len = 1;
		kb_state.buf[0] = scancode;
		break;

	      case 0xFC:
		kb_state.state = CLOCK;
		kb_state.len = 0;
		break;

	      case 0xFE:
	      case 0xFF:
		kb_state.state = JOYSTICK;
		kb_state.len = 1;
		kb_state.buf[0] = scancode;
		break;

	      default:
		break_flag = scancode & BREAK_MASK;
		scancode &= ~BREAK_MASK;

		if (break_flag) {
		    del_timer( &atakeyb_rep_timer );
		    rep_scancode = 0;
		}
		else {
		    del_timer( &atakeyb_rep_timer );
		    rep_scancode = scancode;
		    atakeyb_rep_timer.expires = key_repeat_delay;
		    add_timer( &atakeyb_rep_timer );
		}

		process_scancode( break_flag | scancode );
		break;
	    }
	    break;

	  case AMOUSE:
	    kb_state.buf[kb_state.len++] = scancode;
	    if (kb_state.len == 5)
	    {
		kb_state.state = KEYBOARD;
		/* not yet used */
		/* wake up someone waiting for this */
	    }
	    break;	

	  case RMOUSE:
	    kb_state.buf[kb_state.len++] = scancode;
	    if (kb_state.len == 3)
	    {
		kb_state.state = KEYBOARD;
		atari_mouse_interrupt(kb_state.buf);
	    }
	    break;

	  case JOYSTICK:
	    kb_state.buf[1] = scancode;
	    kb_state.state = KEYBOARD;
	    joystick_interrupt(kb_state.buf);
	    break;

	  case CLOCK:
	    kb_state.buf[kb_state.len++] = scancode;
	    if (kb_state.len == 6)
	    {
		kb_state.state = KEYBOARD;
		/* wake up someone waiting for this.
		   But will this ever be used, as Linux keeps its own time.
		   Perhaps for synchronization purposes? */
		/* wake_up_interruptible(&clock_wait); */
	    }
	    break;
	}
    }

#ifdef KEYB_WRITE_INTERRUPT
    if (acia_stat & ACIA_TDRE)	/* transmit of character is finished */
    {
	if (kb_state.buf)
	{
	    acia.key_data = *kb_state.buf++;
	    kb_state.len--;
	    if (kb_state.len == 0)
	    {
		kb_state.buf = NULL;
		if (!kb_state.kernel_mode)
			/* unblock something */;
	    }
	}
    }
#endif

#if 0
    if (acia_stat & ACIA_CTS)
	/* cannot happen */;
#endif

    if (acia_stat & (ACIA_FE | ACIA_PE))
    {
	printk("Error in keyboard communication");
    }

    /* process_scancode() can take a lot of time, so check again if
	 * some character arrived
	 */
    goto repeat;
}

#ifdef KEYB_WRITE_INTERRUPT
void ikbd_write(const char *str, int len)
{
    u_char acia_stat;

    if (kb_stat.buf)
	/* wait */;
    acia_stat = acia.key_ctrl;
    if (acia_stat & ACIA_TDRE)
    {
	if (len != 1)
	{
	    kb_stat.buf = str + 1;
	    kb_stat.len = len - 1;
	}
	acia.key_data = *str;
	/* poll */
    }
}
#else
/*
 * I write to the keyboard without using interrupts, I poll instead.
 * This takes for the maximum length string allowed (7) at 7812.5 baud
 * 8 data 1 start 1 stop bit: 9.0 ms
 * If this takes too long for normal operation, interrupt driven writing
 * is the solution. (I made a feeble attempt in that direction but I
 * kept it simple for now.)
 */
void ikbd_write(const char *str, int len)
{
    u_char acia_stat;

    if ((len < 1) || (len > 7))
	panic("ikbd: maximum string length exceeded");
    while (len)
    {
	acia_stat = acia.key_ctrl;
	if (acia_stat & ACIA_TDRE)
	{
	    acia.key_data = *str++;
	    len--;
	}
    }
}
#endif

/* Reset (without touching the clock) */
void ikbd_reset(void)
{
    static const char cmd[2] = { 0x80, 0x01 };
    
    ikbd_write(cmd, 2);

    /* if allswell code 0xF1 is returned, else the break codes of
       all keys making contact */
}

/* Set mouse button action */
void ikbd_mouse_button_action(int mode)
{
    char cmd[2] = { 0x07, mode };

    ikbd_write(cmd, 2);
}

/* Set relative mouse position reporting */
void ikbd_mouse_rel_pos(void)
{
    static const char cmd[1] = { 0x08 };

    ikbd_write(cmd, 1);
}

/* Set absolute mouse position reporting */
void ikbd_mouse_abs_pos(int xmax, int ymax)
{
    char cmd[5] = { 0x09, xmax>>8, xmax&0xFF, ymax>>8, ymax&0xFF };

    ikbd_write(cmd, 5);
}

/* Set mouse keycode mode */
void ikbd_mouse_kbd_mode(int dx, int dy)
{
    char cmd[3] = { 0x0A, dx, dy };

    ikbd_write(cmd, 3);
}

/* Set mouse threshold */
void ikbd_mouse_thresh(int x, int y)
{
    char cmd[3] = { 0x0B, x, y };

    ikbd_write(cmd, 3);
}

/* Set mouse scale */
void ikbd_mouse_scale(int x, int y)
{
    char cmd[3] = { 0x0C, x, y };

    ikbd_write(cmd, 3);
}

/* Interrogate mouse position */
void ikbd_mouse_pos_get(int *x, int *y)
{
    static const char cmd[1] = { 0x0D };

    ikbd_write(cmd, 1);

    /* wait for returning bytes */
}

/* Load mouse position */
void ikbd_mouse_pos_set(int x, int y)
{
    char cmd[6] = { 0x0E, 0x00, x>>8, x&0xFF, y>>8, y&0xFF };

    ikbd_write(cmd, 6);
}

/* Set Y=0 at bottom */
void ikbd_mouse_y0_bot(void)
{
    static const char cmd[1] = { 0x0F };

    ikbd_write(cmd, 1);
}

/* Set Y=0 at top */
void ikbd_mouse_y0_top(void)
{
    static const char cmd[1] = { 0x10 };

    ikbd_write(cmd, 1);
}

/* Resume */
void ikbd_resume(void)
{
    static const char cmd[1] = { 0x11 };

    ikbd_write(cmd, 1);
}

/* Disable mouse */
void ikbd_mouse_disable(void)
{
    static const char cmd[1] = { 0x12 };

    ikbd_write(cmd, 1);
}

/* Pause output */
void ikbd_pause(void)
{
    static const char cmd[1] = { 0x13 };

    ikbd_write(cmd, 1);
}

/* Set joystick event reporting */
void ikbd_joystick_event_on(void)
{
    static const char cmd[1] = { 0x14 };

    ikbd_write(cmd, 1);
}

/* Set joystick interrogation mode */
void ikbd_joystick_event_off(void)
{
    static const char cmd[1] = { 0x15 };

    ikbd_write(cmd, 1);
}

/* Joystick interrogation */
void ikbd_joystick_get_state(void)
{
    static const char cmd[1] = { 0x16 };

    ikbd_write(cmd, 1);
}

#if 0
/* This disables all other ikbd activities !!!! */
/* Set joystick monitoring */
void ikbd_joystick_monitor(int rate)
{
    static const char cmd[2] = { 0x17, rate };

    ikbd_write(cmd, 2);

    kb_state.state = JOYSTICK_MONITOR;
}
#endif

/* some joystick routines not in yet (0x18-0x19) */

/* Disable joysticks */
void ikbd_joystick_disable(void)
{
    static const char cmd[1] = { 0x1A };

    ikbd_write(cmd, 1);
}

/* Time-of-day clock set */
void ikbd_clock_set(int year, int month, int day, int hour, int minute, int second)
{
    char cmd[7] = { 0x1B, year, month, day, hour, minute, second };

    ikbd_write(cmd, 7);
}

/* Interrogate time-of-day clock */
void ikbd_clock_get(int *year, int *month, int *day, int *hour, int *minute, int second)
{
    static const char cmd[1] = { 0x1C };

    ikbd_write(cmd, 1);
}

/* Memory load */
void ikbd_mem_write(int address, int size, char *data)
{
    panic("Attempt to write data into keyboard memory");
}

/* Memory read */
void ikbd_mem_read(int address, char data[6])
{
    char cmd[3] = { 0x21, address>>8, address&0xFF };

    ikbd_write(cmd, 3);

    /* receive data and put it in data */
}

/* Controller execute */
void ikbd_exec(int address)
{
    char cmd[3] = { 0x22, address>>8, address&0xFF };

    ikbd_write(cmd, 3);
}

/* Status inquiries (0x87-0x9A) not yet implemented */

unsigned long atari_keyb_init(unsigned long kmem_start)
{
  /* setup key map */
  key_map = atakey_map;
  /* say that we don't have an AltGr key */
  mach_keyboard_type = KB_84;

  acia.key_ctrl = ACIA_RESET;		/* reset ACIA */

  /* divide 500kHz by 64 gives 7812.5 baud */
  /* 8 data no parity 1 start 1 stop bit */
  /* receive interrupt enabled */
#ifdef KEYB_WRITE_INTERRUPT
  /* RTS low, transmit interrupt enabled */
  acia.key_ctrl = (ACIA_DIV64|ACIA_D8N1S|ACIA_RLTIE|ACIA_RIE);
#else
  /* RTS low, transmit interrupt disabled */
  acia.key_ctrl = (ACIA_DIV64|ACIA_D8N1S|ACIA_RLTID|ACIA_RIE);
#endif

    add_isr(IRQ_MFP_ACIA, keyboard_interrupt, 0, NULL);

    ikbd_reset();
    ikbd_mouse_disable();
    ikbd_joystick_disable();

    kb_state.state = KEYBOARD;
    kb_state.len = 0;

	/* enable ACIA Interrupts */ 
	mfp.active_edge &= ~0x10;
	mfp.int_en_b |= 0x40;

  return kmem_start;
}


int atari_kbdrate( struct kbd_repeat *k )

{
	if (k->delay > 0) {
		/* convert from msec to jiffies */
		key_repeat_delay = (k->delay * HZ + 500) / 1000;
		if (key_repeat_delay < 1)
			key_repeat_delay = 1;
	}
	if (k->rate > 0) {
		key_repeat_rate = (k->rate * HZ + 500) / 1000;
		if (key_repeat_rate < 1)
			key_repeat_rate = 1;
	}

	k->delay = key_repeat_delay * 1000 / HZ;
	k->rate  = key_repeat_rate  * 1000 / HZ;
	
	return( 0 );
}
