/*
 * acsi.c -- Device driver for Atari ACSI hard disks
 *
 * Copyright 1994 Roman Hodek
 *   EMail: rnhodek@cip.informatik.uni-erlangen.de (Internet)
 *      or: Roman_Hodek@n.maus.de (MausNet, NO mail > 16 KB!)
 *
 * Some parts are based on hd.c by Linus Thorvalds
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 *
 */

/*
 * Still to in this file:
 *  - There is also the "bits get set somewhere while DMA" problem as
 *    with the floppy driver. See there. Maybe this is a hardware problem.
 *  - If a command ends with an error status (!= 0), the following
 *    REQUEST SENSE commands (4 to fill the ST-DMA FIFO) are done by
 *    polling the _IRQ signal (not interrupt-driven). This should be
 *    avoided in future because it takes up a non-neglectible time in
 *    the interrupt service routine while interrupts are disabled.
 *    Maybe a timer interrupt will get lost :-(
 */

/*
 * General notes:
 *
 *  - All ACSI devices (disks, CD-ROMs, ...) use major number 28.
 *    Minors are organized like it is with SCSI: The upper 4 bits
 *    identify the device, the lower 4 bits the partition.
 *    The device numbers (the upper 4 bits) are given in the same
 *    order as the devices are found on the bus.
 *  - Up to 8 LUNs are supported for each target, but only a total of
 *    16 devices (due to minor numbers...). Note that Atari allows
 *    only a maximum of 4 targets (i.e. controllers, not devices) on
 *    the ACSI bus!
 *  - The SLM laser printer is not supported. In many ways, it behaves
 *    different than other ACSI devices and there are other problems,
 *    too (DMA size is not a multiple of 512, intervention necessary
 *    while DMA-ing the page data).
 *  - A optimizing scheme similar to SCSI scatter-gather is implemented.
 *  - Removable media are supported. After a medium change to device
 *    is reinitialized (partition check etc.). Also, if the device
 *    knows the PREVENT/ALLOW MEDIUM REMOVAL command, the door should
 *    be locked and unlocked when mounting the first or unmounting the
 *    last filesystem on the device. The code is untested, because I
 *    don't have a removable harddisk.
 *
 */

 
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/genhd.h>
#include <linux/delay.h>

#include <linux/bootinfo.h>
#include <linux/interrupt.h>
#include <linux/atarihw.h>
#include <linux/atari_stdma.h>
#include <linux/atari_rootsec.h>
#include <linux/config.h>

#include <asm/system.h>
#include <asm/segment.h>


#define MAJOR_NR ACSI_MAJOR
#include "blk.h"

#define DEBUG
#undef DEBUG_DETECT
#undef NO_WRITE
#undef LUN0_ONLY

#define MAX_ERRORS     		8	/* Max read/write errors/sector */
#define MAX_LUN				8	/* Max LUNs per target */
#define MAX_DEV		   		16

#define ACSI_BUFFER_SIZE	(16*1024)
#define ACSI_BUFFER_SECTORS	(ACSI_BUFFER_SIZE/512)

#define ACSI_TIMEOUT		(4*HZ)

/* minimum delay between two commands */

#define COMMAND_DELAY 500

typedef enum {
	NONE, HARDDISK, CDROM
} ACSI_TYPE;

struct acsi_info_struct {
	ACSI_TYPE		type;			    /* type of device             */
	unsigned		target;			    /* target number              */
	unsigned		lun;			    /* LUN in target controller   */
	unsigned		removable : 1;	    /* Flag for removable media   */
	unsigned		read_only : 1;	    /* Flag for read only devices */
	unsigned		old_atari_disk : 1; /* Is an old Atari disk       */
	unsigned		changed : 1;	    /* Medium has been changed    */
	unsigned long 	size;			    /* #blocks                    */
} acsi_info[MAX_DEV];


/* The data returned my MODE SENSE differ between the old Atari
 * harddisks and SCSI disks connected to ACSI. In the following, both
 * formats are defined and some macros to operate on them potably.
 */

typedef struct {
	unsigned long	dummy[2];
	unsigned long	sector_size;
	unsigned char	format_code;
#define ATARI_SENSE_FORMAT_FIX	1	
#define ATARI_SENSE_FORMAT_CHNG	2
	unsigned char	cylinders_h;
	unsigned char	cylinders_l;
	unsigned char	heads;
	unsigned char	reduced_h;
	unsigned char	reduced_l;
	unsigned char	precomp_h;
	unsigned char	precomp_l;
	unsigned char	landing_zone;
	unsigned char	steprate;
	unsigned char	type;
#define ATARI_SENSE_TYPE_FIXCHNG_MASK		4
#define ATARI_SENSE_TYPE_SOFTHARD_MASK		8
#define ATARI_SENSE_TYPE_FIX				4
#define ATARI_SENSE_TYPE_CHNG				0
#define ATARI_SENSE_TYPE_SOFT				0
#define ATARI_SENSE_TYPE_HARD				8
	unsigned char	sectors;
} ATARI_SENSE_DATA;

#define ATARI_CAPACITY(sd) \
	(((int)((sd).cylinders_h<<8)|(sd).cylinders_l) * \
	 (sd).heads * (sd).sectors)


typedef struct {
	unsigned char   dummy1;
	unsigned char   medium_type;
	unsigned char   dummy2;
	unsigned char   descriptor_size;
	unsigned long   block_count;
	unsigned long   sector_size;
	/* Page 0 data */
	unsigned char	page_code;
	unsigned char	page_size;
	unsigned char	page_flags;
	unsigned char	qualifier;
} SCSI_SENSE_DATA;

#define SCSI_CAPACITY(sd) 	((sd).block_count & 0xffffff)


typedef union {
	ATARI_SENSE_DATA	atari;
	SCSI_SENSE_DATA		scsi;
} SENSE_DATA;

#define SENSE_TYPE_UNKNOWN	0
#define SENSE_TYPE_ATARI	1
#define SENSE_TYPE_SCSI		2

#define SENSE_TYPE(sd)										\
	(((sd).atari.dummy[0] == 8 &&							\
	  ((sd).atari.format_code == 1 ||						\
	   (sd).atari.format_code == 2)) ? SENSE_TYPE_ATARI :	\
	 ((sd).scsi.dummy1 >= 11) ? SENSE_TYPE_SCSI :			\
	 SENSE_TYPE_UNKNOWN)
	 
#define CAPACITY(sd)							\
	(SENSE_TYPE(sd) == SENSE_TYPE_ATARI ?		\
	 ATARI_CAPACITY((sd).atari) :				\
	 SCSI_CAPACITY((sd).scsi))

#define SECTOR_SIZE(sd)							\
	(SENSE_TYPE(sd) == SENSE_TYPE_ATARI ?		\
	 (sd).atari.sector_size :					\
	 (sd).scsi.sector_size & 0xffffff)

#define CARTRCH_STAT(dev,buf)					\
	(acsi_info[(dev)].old_atari_disk ?			\
	 (((buf)[0] & 0x7f) == 0x28) :					\
	 ((((buf)[0] & 0x70) == 0x70) ?					\
	  (((buf)[2] & 0x0f) == 0x06) :					\
	  (((buf)[0] & 0x0f) == 0x06)))					\


static char 			*acsi_buffer;
static unsigned long 	phys_acsi_buffer;
static int				NDevices = 0;
static int				acsi_sizes[MAX_DEV<<4] = { 0, };
static int				acsi_blocksizes[MAX_DEV<<4] = { 0, };
static struct hd_struct	acsi_part[MAX_DEV<<4] = { {0,0}, };
static int 				access_count[MAX_DEV] = { 0, };
static char 			busy[MAX_DEV] = { 0, };
static struct wait_queue *busy_wait;

static int				CurrentNReq;
static int				CurrentNSect;
static char				*CurrentBuffer;


#define ENABLE_IRQ()							\
    do {										\
		mfp.int_pn_b = 0x7f;					\
		mfp.int_en_b |= 0x80;					\
	} while(0)

#define DISABLE_IRQ()							\
    do {										\
		mfp.int_en_b &= ~0x80;					\
	} while(0)

#define SET_TIMER()								\
	do {										\
		del_timer( &acsi_timer );				\
		acsi_timer.expires = ACSI_TIMEOUT;		\
		add_timer( &acsi_timer );				\
	} while(0)

#define CLEAR_TIMER()							\
	do {										\
		del_timer( &acsi_timer );				\
	} while(0)

static unsigned long	STramMask;
#define STRAM_ADDR(a)	(((a) & STramMask) == 0)



/* ACSI commands */

static char tur_cmd[6]        = { 0x00, 0, 0, 0, 0, 0 };
static char modesense_cmd[6]  = { 0x1a, 0, 0, 0, 24, 0 };
static char inquiry_cmd[6]    = { 0x12, 0, 0, 0,255, 0 };
static char reqsense_cmd[6]   = { 0x03, 0, 0, 0, 4, 0 };
static char read_cmd[6]       = { 0x08, 0, 0, 0, 0, 0 };
static char write_cmd[6]      = { 0x0a, 0, 0, 0, 0, 0 };
static char pa_med_rem_cmd[6] = { 0x1e, 0, 0, 0, 0, 0 };

#define CMDSET_TARG_LUN(cmd,targ,lun)			\
    do {										\
		cmd[0] = (cmd[0] & ~0xe0) | (targ)<<5;	\
		cmd[1] = (cmd[1] & ~0xe0) | (lun)<<5;	\
	} while(0)

#define CMDSET_BLOCK(cmd,blk)						\
    do {											\
		unsigned long __blk = (blk);				\
		cmd[3] = __blk; __blk >>= 8;				\
		cmd[2] = __blk; __blk >>= 8;				\
		cmd[1] = (cmd[1] & 0xe0) | (__blk & 0x1f);	\
	} while(0)

#define CMDSET_LEN(cmd,len)						\
	do {										\
		cmd[4] = (len);							\
	} while(0)

#define min(a,b)	(((a)<(b))?(a):(b))


/* ACSI errors (from REQUEST SENSE); There are two tables, one for the
 * old Atari disks and one for SCSI on ACSI disks.
 */

struct acsi_error {
	unsigned char	code;
	const char		*text;
} atari_acsi_errors[] = {
	{ 0x00, "No error (??)" },
	{ 0x01, "No index pulses" },
	{ 0x02, "Seek not complete" },
	{ 0x03, "Write fault" },
	{ 0x04, "Drive not ready" },
	{ 0x06, "No Track 00 signal" },
	{ 0x10, "ECC error in ID field" },
	{ 0x11, "Uncorrectable data error" },
	{ 0x12, "ID field address mark not found" },
	{ 0x13, "Data field address mark not found" },
	{ 0x14, "Record not found" },
	{ 0x15, "Seek error" },
	{ 0x18, "Data check in no retry mode" },
	{ 0x19, "ECC error during verify" },
	{ 0x1a, "Access to bad block" },
	{ 0x1c, "Unformatted or bad format" },
	{ 0x20, "Invalid command" },
	{ 0x21, "Invalid block address" },
	{ 0x23, "Volume overflow" },
	{ 0x24, "Invalid argument" },
	{ 0x25, "Invalid drive number" },
	{ 0x26, "Byte zero parity check" },
	{ 0x28, "Cartride changed" },
	{ 0x2c, "Error count overflow" },
	{ 0x30, "Controller selftest failed" }
},

	scsi_acsi_errors[] = {
	{ 0x00, "No error (??)" },
	{ 0x01, "Recovered error" },
	{ 0x02, "Drive not ready" },
	{ 0x03, "Uncorrectable medium error" },
	{ 0x04, "Hardware error" },
	{ 0x05, "Illegal request" },
	{ 0x06, "Unit attention (Reset or cartridge changed)" },
	{ 0x07, "Data protection" },
	{ 0x08, "Blank check" },
	{ 0x0b, "Aborted Command" },
	{ 0x0d, "Volume overflow" }
};



/***************************** Prototypes *****************************/

static int wait_for_IRQ( unsigned timeout );
static int wait_for_noIRQ( unsigned timeout );
static int acsicmd_dma( const char *cmd, char *buffer, int blocks, int
                        rwflag, int enable);
static int acsicmd_nodma( const char *cmd, int enable);
static int acsi_reqsense( char *buffer, int targ, int lun);
static int acsi_getstatus( void );
static void acsi_print_error( const unsigned char *errblk, int dev );
static void acsi_interrupt( struct intframe *fp, void *data );
static void unexpected_acsi_interrupt( void );
static void bad_rw_intr( void );
static void read_intr( void );
static void write_intr( void);
static void acsi_times_out( unsigned long dummy );
static void copy_to_acsibuffer( void );
static void copy_from_acsibuffer( void );
static void do_end_requests( void );
static void do_acsi_request( void );
static void redo_acsi_request( void );
static int acsi_ioctl( struct inode *inode, struct file *file, unsigned int
                       cmd, unsigned long arg );
static int acsi_open( struct inode * inode, struct file * filp );
static void acsi_release( struct inode * inode, struct file * file );
static void acsi_prevent_removal( int target, int flag );
static int acsi_mode_sense( int target, int lun, SENSE_DATA *sd );
static void acsi_geninit( void );
int revalidate_acsidisk( int dev, int maxusage );

/************************* End of Prototypes **************************/


struct timer_list acsi_timer = { NULL, NULL, 0, 0, acsi_times_out };



/***********************************************************************
 *
 *   ACSI primitives
 *
 **********************************************************************/


/*
 * The following two functions wait for _IRQ to become Low or High,
 * resp., with a timeout. The 'timeout' parameter is in jiffies
 * (10ms).
 * If the functions are called with timer interrupts on (int level <
 * 6), the timeout is based on the 'jiffies' variable to provide exact
 * timeouts for device probing etc.
 * If interrupts are disabled, the number of tries is based on the
 * 'loops_per_sec' variable. A rough estimation is sufficient here...
 */

#define INT_LEVEL													\
	({	unsigned __sr;												\
		__asm__ __volatile__ ( "movew	sr,%0" : "=dm" (__sr) );	\
		(__sr >> 8) & 7;											\
	})

static int wait_for_IRQ( unsigned timeout )

{
	if (INT_LEVEL < 6) {
		unsigned long maxjif;
		for( maxjif = jiffies + timeout; jiffies < maxjif; )
			if (!(mfp.par_dt_reg & 0x20)) return( 1 );
	}
	else {
		long tries = loops_per_sec / HZ / 8 * timeout;
		while( --tries >= 0 )
			if (!(mfp.par_dt_reg & 0x20)) return( 1 );
	}		
	return( 0 ); /* timeout! */
}


static int wait_for_noIRQ( unsigned timeout )

{
	if (INT_LEVEL < 6) {
		unsigned long maxjif;
		for( maxjif = jiffies + timeout; jiffies < maxjif; )
			if (mfp.par_dt_reg & 0x20) return( 1 );
	}
	else {
		long tries = loops_per_sec * timeout / HZ / 8;
		while( tries-- >= 0 )
			if (mfp.par_dt_reg & 0x20) return( 1 );
	}		
	return( 0 ); /* timeout! */
}

static struct timeval start_time;
extern struct timeval xtime;

/* maybe a global defined do_gettimeofday in time.c would be better */

static void do_gettimeofday(struct timeval *tv)
{
	int flags;
	save_flags(flags);
	cli();
	*tv = xtime;
	tv->tv_usec += mach_gettimeoffset();
	if (tv->tv_usec >= 1000000) {
		tv->tv_usec -= 1000000;
		tv->tv_sec++;
	}
	restore_flags(flags);
}

static void
delay_start(void)
{
	do_gettimeofday(&start_time);
}

/* wait from delay_start to now usec (<1E6) usec */

static void
delay_end(long usec)
{
	struct timeval end_time;
	long deltau,deltas;
	do_gettimeofday(&end_time);
	deltau=end_time.tv_usec - start_time.tv_usec;
	deltas=end_time.tv_sec - start_time.tv_sec;
	if (deltas > 1 || deltas < 0)
		return;
	if (deltas > 0)
		deltau += 1000*1000;
	if (deltau >= usec)
		return;
	udelay(usec-deltau);
}

/* Send one data byte over the bus and set mode for next operation
 * with one move.l -- Atari recommends this...
 */

#define DMA_LONG_WRITE(data,mode)												\
    do {																		\
		*((unsigned long *)&dma_wd.fdc_acces_seccount) = ((data)<<16) | (mode);	\
	} while(0)


/* acsicmd_dma() sends an ACSI command and sets up the DMA to transfer
 * 'blocks' blocks of 512 bytes from/to 'buffer'.
 * Because the _IRQ signal is used for handshaking the command bytes,
 * the ACSI interrupt has to be disabled in this function. If the end
 * of the operation should be signalled by a real interrupt, it has to be
 * reenabled afterwards.
 */

static int acsicmd_dma( const char *cmd, char *buffer, int blocks, int rwflag, int enable)

{	unsigned long	flags, paddr;
	int				i;

#ifdef NO_WRITE
	if (rwflag || *cmd == 0x0a) {
		printk( "ACSI: Write commands disabled!\n" );
		return( 0 );
	}
#endif
	
	rwflag = rwflag ? 0x100 : 0;
	paddr = VTOP( buffer );

	delay_end(COMMAND_DELAY);
	DISABLE_IRQ();

	save_flags(flags);  
	cli();
	/* Low on A1 */
	dma_wd.dma_mode_status = 0x88 | rwflag;
	nop(); nop();

	/* set DMA address */
	dma_wd.dma_lo = (unsigned char)paddr;
	paddr >>= 8;
	nop();
	dma_wd.dma_md = (unsigned char)paddr;
	paddr >>= 8;
	nop();
	if (boot_info.bi_atari.hw_present & ATARIHW_EXTD_DMA)
		st_dma_ext_dmahi = (unsigned short)paddr;
	else
		dma_wd.dma_hi = (unsigned char)paddr;
	nop();
	restore_flags(flags);

	/* send the command bytes except the last */
	for( i = 0; i < 5; ++i ) {
		DMA_LONG_WRITE( *cmd++, 0x8a | rwflag );
		udelay(20);
		if (!wait_for_IRQ( HZ/2 )) return( 0 ); /* timeout */
	}

	/* Clear FIFO and switch DMA to correct direction */  
	dma_wd.dma_mode_status = 0x92 | (rwflag ^ 0x100);  
	nop();
	dma_wd.dma_mode_status = 0x92 | rwflag;
	nop();

	/* How many sectors for DMA */
	dma_wd.fdc_acces_seccount = blocks;
	nop();
	
	/* send last command byte */
	dma_wd.dma_mode_status = 0x8a | rwflag;
	nop();
	DMA_LONG_WRITE( *cmd++, 0x0a | rwflag );
	if (enable)
		ENABLE_IRQ();
	udelay(80);

	return( 1 );
}


/*
 * acsicmd_nodma() sends an ACSI command that requires no DMA.
 */

static int acsicmd_nodma( const char *cmd, int enable)

{	int	i;

	delay_end(COMMAND_DELAY);
	DISABLE_IRQ();

	/* send first command byte */
	dma_wd.dma_mode_status = 0x88;
	nop(); nop();
	DMA_LONG_WRITE( *cmd++, 0x8a );
	udelay(20);
	if (!wait_for_IRQ( HZ/2 )) return( 0 ); /* timeout */

	/* send the intermediate command bytes */
	for( i = 0; i < 4; ++i ) {
		DMA_LONG_WRITE( *cmd++, 0x8a );
		udelay(20);
		if (!wait_for_IRQ( HZ/2 )) return( 0 ); /* timeout */
	}

	/* send last command byte */
	DMA_LONG_WRITE( *cmd++, 0x0a );
	if (enable)
		ENABLE_IRQ();
	udelay(80);
	
	return( 1 );
	/* Note that the ACSI interrupt is still disabled after this
	 * function. If you want to get the IRQ delivered, enable it manually!
	 */
}


static int acsi_reqsense( char *buffer, int targ, int lun)

{
	CMDSET_TARG_LUN( reqsense_cmd, targ, lun);
	if (!acsicmd_dma( reqsense_cmd, buffer, 1, 0, 0 )) return( 0 );
	if (!wait_for_IRQ( 10 )) return( 0 );
	acsi_getstatus();
	if (!acsicmd_nodma( reqsense_cmd, 0 )) return( 0 );
	if (!wait_for_IRQ( 10 )) return( 0 );
	acsi_getstatus();
	if (!acsicmd_nodma( reqsense_cmd, 0 )) return( 0 );
	if (!wait_for_IRQ( 10 )) return( 0 );
	acsi_getstatus();
	if (!acsicmd_nodma( reqsense_cmd, 0 )) return( 0 );
	if (!wait_for_IRQ( 10 )) return( 0 );
	acsi_getstatus();
	cache_clear( VTOP(buffer), 16 );
	
	return( 1 );
}	


/*
 * ACSI status phase: get the status byte from the bus
 *
 * I've seen several times that a 0xff status is read, propably due to
 * a timing error. In this case, the procedure is repeated after the
 * next _IRQ edge.
 */

static int acsi_getstatus( void )

{	int	status;

	for(;;) {
		if (!wait_for_IRQ( 100 )) {
			delay_start();
			return( -1 );
		}
		dma_wd.dma_mode_status = 0x8a;
		nop(); nop();
		status = dma_wd.fdc_acces_seccount;
		if (status != 0xff) break;
#ifdef DEBUG
		printk("ACSI: skipping 0xff status byte\n" );
#endif
		udelay(40);
		wait_for_noIRQ( 20 );
	}
	dma_wd.dma_mode_status = 0x80;
	udelay(40);
	wait_for_noIRQ( 20 );

	delay_start();
	return( status & 0x1f ); /* mask of the device# */
}


#if 0

/* Receive data in an extended status phase. Never needed... */

static int acsi_extstatus( char *buffer, int cnt )

{	int	status, i = 0;

	udelay(80);
	while( cnt-- >= 0 ) {
		if (!wait_for_IRQ( 40 )) return( 0 );
		dma_wd.dma_mode_status = 0x8a;
		nop(); nop();
		status = dma_wd.fdc_acces_seccount;
		*buffer++ = status & 0xff;
		udelay(80);
	}
	return( 1 );
}

#endif


static void acsi_print_error( const unsigned char *errblk, int dev  )

{	int atari_err, i, errcode;
	struct acsi_error *arr;
	
	atari_err = acsi_info[dev].old_atari_disk;
	if (atari_err)
		errcode = errblk[0] & 0x7f;
	else
		if ((errblk[0] & 0x70) == 0x70)
			errcode = errblk[2] & 0x0f;
		else
			errcode = errblk[0] & 0x0f;
	
	printk( "ACSI error 0x%02x", errcode );

	if (errblk[0] & 0x80)
		printk( " for sector %d",
				((errblk[1] & 0x1f) << 16) |
				(errblk[2] << 8) | errblk[0] );

	arr = atari_err ? atari_acsi_errors : scsi_acsi_errors;
	i = atari_err ? sizeof(atari_acsi_errors)/sizeof(*atari_acsi_errors) :
		            sizeof(scsi_acsi_errors)/sizeof(*scsi_acsi_errors);
	
	for( --i; i >= 0; --i )
		if (arr[i].code == errcode) break;
	if (i >= 0)
		printk( ": %s\n", arr[i].text );
}




/*******************************************************************
 *
 * ACSI interrupt routine
 *   Test, if this is a ACSI interrupt and call the irq handler
 *   Otherwise ignore this interrupt.
 *
 *******************************************************************/

static void acsi_interrupt( struct intframe *fp, void *data )

{	void (*acsi_irq_handler)(void) = DEVICE_INTR;

	DEVICE_INTR = NULL;
	CLEAR_TIMER();

	if (!acsi_irq_handler)
		acsi_irq_handler = unexpected_acsi_interrupt;
	acsi_irq_handler();
}


/******************************************************************
 *
 * The Interrupt handlers
 *
 *******************************************************************/


static void unexpected_acsi_interrupt( void )

{
	printk("Unexpected ACSI interrupt\n");
}


/* This function is called in case of errors. Because we cannot reset
 * the ACSI bus or a single device, there is no other choice than
 * retrying several times :-(
 */

static void bad_rw_intr( void )

{
	if (!CURRENT)
		return;

	if (++CURRENT->errors >= MAX_ERRORS)
		end_request(0);
	/* Otherwise just retry */
}


static void read_intr( void )

{	int		status;
	
	status = acsi_getstatus();
	if (status != 0) {
		int dev = DEVICE_NR(MINOR(CURRENT->dev));
		printk( "ad%c: ", dev+'a' );
		if (!acsi_reqsense( acsi_buffer, acsi_info[dev].target, 
					acsi_info[dev].lun))
			printk( "ACSI error and REQUEST SENSE failed (status=0x%02x)\n", status );
		else {
			acsi_print_error( acsi_buffer, dev );
			if (CARTRCH_STAT( dev, acsi_buffer ))
				acsi_info[dev].changed = 1;
		}
		ENABLE_IRQ();
		bad_rw_intr();
		redo_acsi_request();
		return;
	}

	cache_clear( VTOP(CurrentBuffer), CurrentNSect*512 );
	if (CurrentBuffer == acsi_buffer)
		copy_from_acsibuffer();

	do_end_requests();
	redo_acsi_request();
}


static void write_intr(void)

{	int	status;

	status = acsi_getstatus();
	if (status != 0) {
		int	dev = DEVICE_NR(MINOR(CURRENT->dev));
		printk( "ad%c: ", dev+'a' );
		if (!acsi_reqsense( acsi_buffer, acsi_info[dev].target,
					acsi_info[dev].lun))
			printk( "ACSI error and REQUEST SENSE failed (status=0x%02x)\n", status );
		else {
			acsi_print_error( acsi_buffer, dev );
			if (CARTRCH_STAT( dev, acsi_buffer ))
				acsi_info[dev].changed = 1;
		}
		bad_rw_intr();
		redo_acsi_request();
		return;
	}

	do_end_requests();
	redo_acsi_request();
}


static void acsi_times_out( unsigned long dummy )

{
	DEVICE_INTR = NULL;
	printk("ACSI timeout\n");
	if (!CURRENT) return;
	if (++CURRENT->errors >= MAX_ERRORS) {
#ifdef DEBUG
		printk("ACSI: too many errors.\n");
#endif
		end_request(0);
	}

	redo_acsi_request();
}



/***********************************************************************
 *
 *  Scatter-gather utility functions
 *
 ***********************************************************************/


static void copy_to_acsibuffer( void )

{	int					i;
	char				*src, *dst;
	struct buffer_head	*bh;
	
	src = CURRENT->buffer;
	dst = acsi_buffer;
	bh = CURRENT->bh;

	if (!bh)
		memcpy( dst, src, CurrentNSect*512 );
	else
		for( i = 0; i < CurrentNReq; ++i ) {
			memcpy( dst, src, bh->b_size );
			dst += bh->b_size;
			if ((bh = bh->b_reqnext))
				src = bh->b_data;
		}
}


static void copy_from_acsibuffer( void )

{	int					i;
	char				*src, *dst;
	struct buffer_head	*bh;
	
	dst = CURRENT->buffer;
	src = acsi_buffer;
	bh = CURRENT->bh;

	if (!bh)
		memcpy( dst, src, CurrentNSect*512 );
	else
		for( i = 0; i < CurrentNReq; ++i ) {
			memcpy( dst, src, bh->b_size );
			src += bh->b_size;
			if ((bh = bh->b_reqnext))
				dst = bh->b_data;
		}
}


static void do_end_requests( void )

{	int		i, n;

	if (!CURRENT->bh) {
		CURRENT->nr_sectors -= CurrentNSect;
		CURRENT->current_nr_sectors -= CurrentNSect;
		CURRENT->sector += CurrentNSect;
		if (CURRENT->nr_sectors == 0)
			end_request(1);
	}
	else {
		for( i = 0; i < CurrentNReq; ++i ) {
			n = CURRENT->bh->b_size >> 9;
			CURRENT->nr_sectors -= n;
			CURRENT->current_nr_sectors -= n;
			CURRENT->sector += n;
			end_request(1);
		}
	}
}




/***********************************************************************
 *
 *  do_acsi_request and friends
 *
 ***********************************************************************/

static void do_acsi_request( void )

{
	stdma_lock( acsi_interrupt, NULL );
	redo_acsi_request();
}


static void redo_acsi_request( void )

{	unsigned			block, dev, target, lun, nsect;
	char 				*buffer;
	unsigned long		pbuffer;
	struct buffer_head	*bh;
	
	if (CURRENT && CURRENT->dev < 0) {
		if (!DEVICE_INTR) stdma_release();
		return;
	}

	if (DEVICE_INTR)
		return;

  repeat:
	CLEAR_TIMER();
	/* Another check here: An interrupt or timer event could have
	 * happened since the last check!
	 */
	if (CURRENT && CURRENT->dev < 0) {
		if (!DEVICE_INTR) stdma_release();
		return;
	}
	if (DEVICE_INTR)
		return;

	if (!CURRENT) {
		CLEAR_INTR;
		stdma_release();
		return;
	}
	
	if (MAJOR(CURRENT->dev) != MAJOR_NR)
		panic(DEVICE_NAME ": request list destroyed");
	if (CURRENT->bh) {
		if (!CURRENT->bh->b_lock)
			panic(DEVICE_NAME ": block not locked");
	}

	dev = MINOR(CURRENT->dev);
	block = CURRENT->sector;
	if (DEVICE_NR(dev) >= NDevices ||
		block+CURRENT->nr_sectors >= acsi_part[dev].nr_sects) {
#ifdef DEBUG
		printk( "ad%c: attempted access for block %d past end of device at block %ld.\n",
				DEVICE_NR(dev)+'a', block, acsi_part[dev].nr_sects);
#endif
		end_request(0);
		goto repeat;
	}
	if (acsi_info[DEVICE_NR(dev)].changed) {
		printk( "ad%c: request denied because cartridge has been changed.\n",
				DEVICE_NR(dev)+'a' );
		end_request(0);
		goto repeat;
	}
	
	block += acsi_part[dev].start_sect;
	target = acsi_info[DEVICE_NR(dev)].target;
	lun    = acsi_info[DEVICE_NR(dev)].lun;

	/* Find out how many sectors should be transferred from/to
	 * consecutive buffers and thus can be done with a single command.
	 */
	buffer      = CURRENT->buffer;
	pbuffer     = VTOP(buffer);
	nsect       = CURRENT->current_nr_sectors;
	CurrentNReq = 1;

	if ((bh = CURRENT->bh) && bh != CURRENT->bhtail) {
		if (!STRAM_ADDR(pbuffer)) {
			/* If transfer is done via the ACSI buffer anyway, we can
			 * assemble as much bh's as fit in the buffer.
			 */
			while( (bh = bh->b_reqnext) ) {
				if (nsect + (bh->b_size>>9) > ACSI_BUFFER_SECTORS) break;
				nsect += bh->b_size >> 9;
				++CurrentNReq;
				if (bh == CURRENT->bhtail) break;
			}
			buffer = acsi_buffer;
		}
		else {
			unsigned long pendadr, pnewadr;
			pendadr = pbuffer + nsect*512;
			while( (bh = bh->b_reqnext) ) {
				pnewadr = VTOP(bh->b_data);
				if (!STRAM_ADDR(pnewadr) || pendadr != pnewadr) break;
				nsect += bh->b_size >> 9;
				pendadr += pnewadr + bh->b_size;
				++CurrentNReq;
				if (bh == CURRENT->bhtail) break;
			}
		}
	}
	else {
		if (!STRAM_ADDR(pbuffer)) {
			buffer = acsi_buffer;
			if (nsect > ACSI_BUFFER_SECTORS)
				nsect = ACSI_BUFFER_SECTORS;
		}
	}
	CurrentBuffer = buffer;
	CurrentNSect  = nsect;
	
	if (CURRENT->cmd == WRITE) {
		CMDSET_TARG_LUN( write_cmd, target, lun );
		CMDSET_BLOCK( write_cmd, block );
		CMDSET_LEN( write_cmd, nsect );
		if (buffer == acsi_buffer)
			copy_to_acsibuffer();
		cache_push( pbuffer, nsect*512 );
		SET_INTR(write_intr);
		if (!acsicmd_dma( write_cmd, buffer, nsect, 1, 1)) {
			printk( "ACSI: Timeout in command block\n" );
			bad_rw_intr();
			goto repeat;
		}
		SET_TIMER();
		return;
	}
	if (CURRENT->cmd == READ) {
		CMDSET_TARG_LUN( read_cmd, target, lun );
		CMDSET_BLOCK( read_cmd, block );
		CMDSET_LEN( read_cmd, nsect );
		SET_INTR(read_intr);
		if (!acsicmd_dma( read_cmd, buffer, nsect, 0, 1)) {
			printk( "ACSI: Timeout in command block\n" );
			bad_rw_intr();
			goto repeat;
		}
		SET_TIMER();
		return;
	}
	panic("unknown ACSI command");
}



/***********************************************************************
 *
 *  Misc functions: ioctl, open, release, check_change, ...
 *
 ***********************************************************************/


static int acsi_ioctl( struct inode *inode, struct file *file,
					   unsigned int cmd, unsigned long arg )
{	int dev, err;

	if (!inode)
		return -EINVAL;
	dev = DEVICE_NR(MINOR(inode->i_rdev));
	if (dev >= NDevices)
		return -EINVAL;
	switch (cmd) {
		/* I left out the GETGEO cmd; This doesn't make much sense for
		 * ACSI disks...
		 */
	  case BLKGETSIZE:   /* Return device size */
		if (!arg)  return -EINVAL;
		err = verify_area(VERIFY_WRITE, (long *) arg, sizeof(long));
		if (err)
			return err;
		put_fs_long(acsi_part[MINOR(inode->i_rdev)].nr_sects,
					(long *) arg);
		return 0;
	  case BLKFLSBUF:
		if(!suser())  return -EACCES;
		if(!inode->i_rdev) return -EINVAL;
		fsync_dev(inode->i_rdev);
		invalidate_buffers(inode->i_rdev);
		return 0;

	  case BLKRRPART: /* Re-read partition tables */
		return revalidate_acsidisk(inode->i_rdev, 1);
	  RO_IOCTLS(inode->i_rdev,arg);
	  default:
		return -EINVAL;
	}
}


/*
 * Open a device, check for read-only and lock the medium if it is
 * removable.
 */

static int acsi_open( struct inode * inode, struct file * filp )

{	int target;
	target = DEVICE_NR(MINOR(inode->i_rdev));

	while (busy[target])
		sleep_on(&busy_wait);

	if (access_count[target] == 0 && acsi_info[target].removable) {
		check_disk_change( inode->i_rdev );
		acsi_prevent_removal( target, 1 );
	}
	access_count[target]++;

	if (filp && filp->f_mode) {
		check_disk_change( inode->i_rdev );
		if (filp->f_mode & 2) {
			if (acsi_info[target].read_only) {
				acsi_release( inode, filp );
				return( -EROFS );
			}
		}
	}

	return 0;
}

/*
 * Releasing a block device means we sync() it, so that it can safely
 * be forgotten about...
 */

static void acsi_release( struct inode * inode, struct file * file )

{	int target;
	
	sync_dev(inode->i_rdev);

	target = DEVICE_NR(MINOR(inode->i_rdev));
	if (--access_count[target] == 0 && acsi_info[target].removable)
		acsi_prevent_removal( target, 0 );
}


/*
 * Prevent or allow a media change for removable devices.
 */

static void acsi_prevent_removal( int target, int flag )

{
	stdma_lock( NULL, NULL );
	
	CMDSET_TARG_LUN( pa_med_rem_cmd,
					 acsi_info[target].target,
					 acsi_info[target].lun );
	CMDSET_LEN( pa_med_rem_cmd, flag );
	
	if (!acsicmd_nodma( pa_med_rem_cmd, 0)) goto the_end;
	if (!wait_for_IRQ( 3*HZ )) goto the_end;
	acsi_getstatus();
	/* Do not report errors -- some devices may not know this command. */
  the_end:
	ENABLE_IRQ();
	stdma_release();
}


int check_acsi_media_change( int full_dev, int flag )

{	int target = DEVICE_NR(MINOR(full_dev));
	
	if (!acsi_info[target].removable) return( 0 );

	if (acsi_info[target].changed)
		/* We can be sure that the medium has been changed -- REQUEST
		 * SENSE has reported this earlier.
		 */
		return( 1 );

	/* If the flag isn't set, make a test by reading block 0.
	 * If errors happen, it seems to be better to say "changed"...
	 */
	stdma_lock( NULL, NULL );
	CMDSET_TARG_LUN( read_cmd, acsi_info[target].target, acsi_info[target].lun );
	CMDSET_BLOCK( read_cmd, 0 );
	CMDSET_LEN( read_cmd, 1 );
	if (acsicmd_dma( read_cmd, acsi_buffer, 1, 0, 0) &&
		wait_for_IRQ( HZ )) {
		if (acsi_getstatus()) {
			if (acsi_reqsense( acsi_buffer, 
			acsi_info[target].target, acsi_info[target].lun)) {
				if (CARTRCH_STAT( target, acsi_buffer ))
					acsi_info[target].changed = 1;
			}
			else {
				printk( "ad%c: REQUEST SENSE failed in test for medium change; assuming a change\n",
						target+'a' );
				acsi_info[target].changed = 1;
			}
		}
	}
	else {
		printk( "ad%c: Test for medium changed timed out; assuming a change\n",
				target+'a' );
		acsi_info[target].changed = 1;
	}
	ENABLE_IRQ();
	stdma_release();

	/* Now, after reading a block, the changed status is surely valid. */
	return( acsi_info[target].changed );
}


static int acsi_mode_sense( int target, int lun, SENSE_DATA *sd )

{
	CMDSET_TARG_LUN( modesense_cmd, target, lun );
	if (!acsicmd_dma( modesense_cmd, acsi_buffer, 1, 0, 0 ) ||
		!wait_for_IRQ( 3*HZ ) ||
		acsi_getstatus() != 0 ||
		/* read twice to jump over the second 16-byte border! */
		(udelay(300), !wait_for_noIRQ( 20 )) ||
		!acsicmd_nodma( modesense_cmd, 0 ) ||
		!wait_for_IRQ( 3*HZ ) ||
		acsi_getstatus() != 0) {
		return( 0 );
	}
	cache_clear( phys_acsi_buffer, sizeof(SENSE_DATA) );
	*sd = *(SENSE_DATA *)acsi_buffer;

	/* Validity check, depending on type of data */
	
	switch( SENSE_TYPE(*sd) ) {

	  case SENSE_TYPE_ATARI:
		if (CAPACITY(*sd) == 0)
			goto invalid_sense;
		break;

	  case SENSE_TYPE_SCSI:
		if (sd->scsi.descriptor_size != 8 || CAPACITY(*sd) == 0)
			goto invalid_sense;
		break;

	  case SENSE_TYPE_UNKNOWN:

		printk( "ACSI target %d, lun %d: Cannot interpret sense data\n",
				target, lun ); 
		
	  invalid_sense:

#ifdef DEBUG
		{	int i;
			printk( "Mode sense data for ACSI target %d, lun %d seem not valid:",
					target, lun );
			for( i = 0; i < sizeof(SENSE_DATA); ++i )
				printk( "%02x ", (unsigned char)acsi_buffer[i] );
			printk( "\n" );
		}
#endif
		return( 0 );
	}
		
	return( 1 );
}



/*******************************************************************
 *
 *  Initialization
 *
 ********************************************************************/


static struct gendisk acsi_gendisk = {
	MAJOR_NR,			/* Major number */	
	"ad",				/* Major name */
	4,					/* Bits to shift to get real from partition */
	1 << 4,				/* Number of partitions per real */
	MAX_DEV,			/* maximum number of real */
	acsi_geninit,		/* init function */
	acsi_part,			/* hd struct */
	acsi_sizes,			/* block sizes */
	0,					/* number */
	(void *)acsi_info,	/* internal */
	NULL				/* next */
};
	
#define MAX_SCSI_DEVICE_CODE 10

static const char *const scsi_device_types[MAX_SCSI_DEVICE_CODE] =
{
 "Direct-Access    ",
 "Sequential-Access",
 "Printer          ",
 "Processor        ",
 "WORM             ",
 "CD-ROM           ",
 "Scanner          ",
 "Optical Device   ",
 "Medium Changer   ",
 "Communications   "
};

static void print_inquiry(unsigned char *data)
{
	int i;

	printk("  Vendor: ");
	for (i = 8; i < 16; i++)
		{
	        if (data[i] >= 0x20 && i < data[4] + 5)
			printk("%c", data[i]);
		else
			printk(" ");
		}

	printk("  Model: ");
	for (i = 16; i < 32; i++)
		{
	        if (data[i] >= 0x20 && i < data[4] + 5)
			printk("%c", data[i]);
		else
			printk(" ");
		}

	printk("  Rev: ");
	for (i = 32; i < 36; i++)
		{
	        if (data[i] >= 0x20 && i < data[4] + 5)
			printk("%c", data[i]);
		else
			printk(" ");
		}

	printk("\n");

	i = data[0] & 0x1f;

	printk("  Type:   %s ",
	       i < MAX_SCSI_DEVICE_CODE ? scsi_device_types[i] : "Unknown          " );
	printk("                 ANSI SCSI revision: %02x", data[2] & 0x07);
	if ((data[2] & 0x07) == 1 && (data[3] & 0x0f) == 1)
	  printk(" CCS\n");
	else
	  printk("\n");
}

static void acsi_geninit( void )

{	int 			i;
	int 			target, lun, status;
	ACSI_TYPE		type;
	int				rmf, rdonly, old_atari_disk, got_inquiry;
	SENSE_DATA		sense;
	unsigned char 		reqsense,extsense;
	unsigned long	size=0;
	
	printk ("Probing ACSI devices:\n");
	NDevices = 0;
	stdma_lock( NULL, NULL );

	for( target = 0; target < 8 && NDevices < MAX_DEV; ++target ) {

		lun = 0;
#ifndef LUN0_ONLY
	  repeat_for_lun:
#endif

		/* Do a TEST UNIT READY command to test the presence of a
		 * device
		 */
		CMDSET_TARG_LUN( tur_cmd, target, lun );
		if (!acsicmd_nodma( tur_cmd, 0)) {
			/* timed out -> no device here */
#ifdef DEBUG_DETECT
			printk("target %d lun %d: timeout\n",target,lun);
#endif
			continue;
		}
		
		status = acsi_getstatus();
		if (status) {
			if (status == 0x12)
				/* The SLM printer should be the only device that
				 * responds with the error code in the status byte. In
				 * correct status bytes, bit 4 is never set.
				 */
				printk( "Detected SLM printer (not supported) at id %d lun %d\n",
						target, lun );
			/* ignore CHECK CONDITION, since some devices send a
			   UNIT ATTENTION */
			if ((status & 0x1e) != 0x2) {
#ifdef DEBUG_DETECT
				printk("target %d lun %d: status %d\n",target,lun,status);
#endif
				continue;
			}
		}

		if (!acsi_reqsense(acsi_buffer, target, lun)) {
			printk("acsi_reqsense failed\n");
			acsi_buffer[0]=0;
			acsi_buffer[2]=0;
		}
		reqsense=acsi_buffer[0];
		extsense=acsi_buffer[2];

		if (status) {
			if ((reqsense & 0x70) == 0x70) {/* extended sense */
				if (extsense != 0x6) { /* UNIT ATTENTION */
#ifdef DEBUG_DETECT
					printk("target %d lun %d: extended sense %d\n",target,lun,extsense);
#endif
					continue;
				}
				reqsense = acsi_buffer[2];
			}
			else {
				if (reqsense & 0x7f) {
#ifdef DEBUG_DETECT
					printk("target %d lun %d: sense %d\n",target,lun,reqsense);
#endif
					continue;
				}
			}
		}
		else 
			if (reqsense == 0x4) {	/* SH204 Bug workaround */
#ifdef DEBUG_DETECT
				printk("target %d lun %d status=0 sense=4\n",target,lun);
#endif
				continue;
			}

		/* Assume default values */
		type = HARDDISK;
		rmf = 1;
		rdonly = 0;
		old_atari_disk = 0;
		got_inquiry = 0;
		
		/* Do an INQUIRY command to get more infos on this device;
		 * some older Atari disks don't know about INQUIRY, so use the
		 * defaults above in case of error...
		 */
		/* Fake inquiry result for old atari disks */
		memcpy(acsi_buffer,"\000\000\001\000    Adaptec 40xx                    ",40);
		CMDSET_TARG_LUN( inquiry_cmd, target, lun );
		if (acsicmd_dma( inquiry_cmd, acsi_buffer, 1, 0, 0) &&
			acsi_getstatus() == 0) {
			acsicmd_nodma( inquiry_cmd, 0);
			acsi_getstatus();
			cache_clear( phys_acsi_buffer, 256 );
			got_inquiry = 1;
			rmf = !!(acsi_buffer[1] & 0x80); /* removable? */
		}
		
		print_inquiry(acsi_buffer);

		switch( acsi_buffer[0] ) {
		case 0:
			type = HARDDISK;
			break;
		case 5:
			type = CDROM;
			rdonly = 1;
			break;
		default:
			type = NONE;
			goto device_ok;
		}

		if (!acsi_mode_sense( target, lun, &sense )) {
			printk( "No mode sense data.\n" );
			continue;
		}
		if (SECTOR_SIZE(sense) != 512) {
			printk( "Sector size != 512 not supported.\n" );
			continue;
		}
		size = CAPACITY(sense);
		if (!got_inquiry && SENSE_TYPE(sense) == SENSE_TYPE_ATARI) {
			/* If INQUIRY failed and the sense data suggest an old
			 * Atari disk (SH20x, Megafile), the disk is not removable
			 */
			rmf = 0;
			old_atari_disk = 1;
		}

	  device_ok:
		printk("Detected ");
		switch (type)
		{
		case NONE:
			printk("unsupported device");
			break;
		case HARDDISK:
			printk("disk");
			break;
		case CDROM:
			printk("cdrom");
			break;
		}	
		if (type != NONE)
			printk(" ad%c",'a'+NDevices);
		printk(" at id %d lun %d ",target,lun);
		if (type == NONE) {
			printk("\n");
			continue;
		}
		/* Fill in the acsi_info for this device... */
		acsi_info[NDevices].type           = type;
		acsi_info[NDevices].target         = target;
		acsi_info[NDevices].lun            = lun;
		acsi_info[NDevices].removable      = rmf;
		acsi_info[NDevices].read_only      = rdonly;
		acsi_info[NDevices].old_atari_disk = old_atari_disk;
		acsi_info[NDevices].changed        = 0;
		acsi_info[NDevices].size           = size;

		/* ...and inform the user about it. */

		if (rmf) printk( "(removable) " );
		if (rdonly) printk( "(read-only) " );
		printk( "%ld MByte\n", (size*512+1024*1024/2)/(1024*1024) );
		
		NDevices++;

		/* If a test for this target was successful, try the next LUN, too */
#ifndef LUN0_ONLY
		if (++lun < MAX_LUN)
			goto repeat_for_lun;
#endif
	}

	/* reenable interrupt */
	ENABLE_IRQ();
	stdma_release();

	printk( "Found %d ACSI device(s) total.\n", NDevices );
					 
	for( i = 0; i < NDevices; ++i ) {
		acsi_part[i<<4].start_sect = 0;
		acsi_part[i<<4].nr_sects = acsi_info[i].size;
	}
	acsi_gendisk.nr_real = NDevices;
	for( i = 0; i < (MAX_DEV << 4); i++ )
		acsi_blocksizes[i] = 1024;
	blksize_size[MAJOR_NR] = acsi_blocksizes;
}


static struct file_operations acsi_fops = {
	NULL,			/* lseek - default */
	block_read,		/* read - general block-dev read */
	block_write,	/* write - general block-dev write */
	NULL,			/* readdir - bad */
	NULL,			/* select */
	acsi_ioctl,		/* ioctl */
	NULL,			/* mmap */
	acsi_open,		/* open */
	acsi_release,	/* release */
	block_fsync		/* fsync */
};


unsigned long acsi_init( unsigned long mem_start, unsigned long mem_end )

{
	if (boot_info.machtype != MACH_ATARI ||
		!(boot_info.bi_atari.hw_present & ATARIHW_STND_ACSI))
		return( mem_start );

	if (register_blkdev( MAJOR_NR, "ad", &acsi_fops )) {
		printk( "Unable to get major %d for ACSI\n", MAJOR_NR );
		return( mem_start );
	}

	if (!(acsi_buffer = atari_stram_alloc(
						  (boot_info.bi_atari.hw_present&ATARIHW_EXTD_DMA) ?
						    512 : ACSI_BUFFER_SIZE,
						  &mem_start ))) {
		printk( "Unable to get ACSI ST-Ram buffer.\n" );
		return( mem_start );
	}
	phys_acsi_buffer = VTOP( acsi_buffer );
	STramMask = (boot_info.bi_atari.hw_present & ATARIHW_EXTD_DMA) ?
				  0x00000000 : 0xff000000;
	
	blk_dev[MAJOR_NR].request_fn = DEVICE_REQUEST;
	read_ahead[MAJOR_NR] = 8;		/* 8 sector (4kB) read-ahead */
	acsi_gendisk.next = gendisk_head;
	gendisk_head = &acsi_gendisk;

	return( mem_start );
}


#define DEVICE_BUSY busy[target]
#define USAGE access_count[target]
#define GENDISK_STRUCT acsi_gendisk

/*
 * This routine is called to flush all partitions and partition tables
 * for a changed scsi disk, and then re-read the new partition table.
 * If we are revalidating a disk because of a media change, then we
 * enter with usage == 0.  If we are using an ioctl, we automatically have
 * usage == 1 (we need an open channel to use an ioctl :-), so this
 * is our limit.
 */

int revalidate_acsidisk( int dev, int maxusage )

{	int target, major;
	struct gendisk * gdev;
	int max_p;
	int start;
	int i;
	SENSE_DATA sense;
	
	target =  DEVICE_NR(MINOR(dev));
	gdev = &GENDISK_STRUCT;

	cli();
	if (DEVICE_BUSY || USAGE > maxusage) {
		sti();
		return -EBUSY;
	};
	DEVICE_BUSY = 1;
	sti();

	max_p = gdev->max_p;
	start = target << gdev->minor_shift;
	major = MAJOR_NR << 8;

	for( i = max_p - 1; i >= 0 ; i-- ) {
		sync_dev( major | start | i );
		invalidate_inodes( major | start | i );
		invalidate_buffers( major | start | i );
		gdev->part[start+i].start_sect = 0;
		gdev->part[start+i].nr_sects = 0;
	};

	stdma_lock( NULL, NULL );
	if (!acsi_mode_sense( acsi_info[target].target, acsi_info[target].lun, &sense )) {
		printk( "ACSI device %04x: mode sense failed.\n", dev );
		acsi_info[target].size = 0x1fffff;
	}
	else {
		acsi_info[target].size = CAPACITY(sense);
	}
	ENABLE_IRQ();
	stdma_release();
	
	acsi_info[target].changed = 0;

	gdev->part[start].nr_sects = acsi_info[target].size;
	resetup_one_dev( gdev, target );

	DEVICE_BUSY = 0;
	wake_up(&busy_wait);
	return 0;
}
