/*
 *  linux/amiga/amiflop.c
 *
 *  Copyright (C) 1993  Greg Harp
 *  Portions of this driver are based on code contributed by Brad Pepers
 */

#include <linux/config.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/kernel.h>
#include <linux/timer.h>
#include <linux/amifdreg.h>
#include <linux/fd.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/interrupt.h>

#include <linux/amigahw.h>
#include <linux/amigaints.h>

#define MAJOR_NR FLOPPY_MAJOR
#include "blk.h"

/*
 *  Defines
 */
#define MAX_SECTORS	22

/*
 *  Error codes
 */
#define FD_OK		0	/* operation succeeded */
#define FD_ERROR	-1	/* general error (seek, read, write, etc) */
#define FD_NOUNIT	1	/* unit does not exist */
#define FD_UNITBUSY	2	/* unit already active */
#define FD_NOTACTIVE	3	/* unit is not active */
#define FD_NOTREADY	4	/* unit is not ready (motor not on/no disk) */

/*
 *  Floppy ID values
 */
#define FD_NODRIVE	0x00000000  /* response when no unit is present */
#define FD_DD_3 	0xffffffff  /* double-density 3.5" (880K) drive */
#define FD_HD_3 	0x55555555  /* high-density 3.5" (1760K) drive */
#define FD_DD_5 	0xaaaaaaaa  /* double-density 5.25" (440K) drive */


/*
 *  Macros
 */
#define MOTOR_ON	(ciab.prb &= ~DSKMOTOR)
#define MOTOR_OFF	(ciab.prb |= DSKMOTOR)
#define SELECT(mask)    (ciab.prb &= ~mask)
#define DESELECT(mask)  (ciab.prb |= mask)
#define SELMASK(drive)  (1 << (3 + (drive & 3)))

#define DRIVE(x) ((x) & 3)
#define PROBE(x) ((x) >> 2) & 1)
#define TYPE(x)  ((x) >> 3) & 2)
#define DATA(x)  ((x) >> 5) & 3)

static struct fd_drive_type drive_types[] = {
/*  code	name	   tr he   rdsz   wrsz sm pc1 pc2    sd     st	  st*/
 { FD_DD_3,	"DD 3.5", 160, 2, 14716, 13630, 1, 80,161, 3000, 18000, 1000},
 { FD_HD_3,	"HD 3.5", 160, 2, 25000, 25000, 2, 80,161, 3000, 18000, 1000},
 { FD_NODRIVE, "No Drive", 0, 0,     0,     0, 0,  0,  0,    0,     0,    0}
};
static int num_dr_types = sizeof(drive_types) / sizeof(drive_types[0]);

static struct fd_data_type data_types[] = {
  { "Amiga", 11 }
};
static int num_da_types = sizeof(data_types) / sizeof(data_types[0]);

#if 0
  {1440,  9,   2, 80, "IBM Double Density 3.5"},    /* 720K */
  {2880, 18,   2, 80, "IBM High Density 3.5"},      /* 1.44M */
  { 880, 11,   2, 40, "Amiga Double Density 5.25"}, /* 440K */
  { 720,  9,   2, 40, "IBM Double Density 5.25"},   /* 360K */
#endif

/* current info on each unit */
static struct floppy_struct unit[FD_MAX_UNITS];

/* track buffer */
static int savedtrack = -1;
static int writepending = 0;
static int writefromint = 0;
static unsigned char trackdata[MAX_SECTORS * 512];
static char *raw_buf;

/*
 * These are global variables, as that's the easiest way to give
 * information to interrupts. They are the data used for the current
 * request.
 */
static char block_flag = 0;
static int selected = 0;
static struct wait_queue *wait_fd_block = NULL;

/* Synchronization of FDC access. */
static volatile int fdc_busy = 0;
static struct wait_queue *fdc_wait = NULL;
static struct wait_queue *motor_wait = NULL;

/*
 * Functions
 */
/*======================================================================
  Turn off the motor of the given drive.  Unit must already be active.
  Returns standard floppy error code.
======================================================================*/
static void fd_motor_off(unsigned long drive)
{
	unsigned long flags;
	unsigned char prb = ~0;

	save_flags(flags);
	cli();

	if (unit[drive].track % 2 != 0)
		prb &= ~DSKSIDE;
	ciab.prb |= (SELMASK(0)|SELMASK(1)|SELMASK(2)|SELMASK(3));
	ciab.prb = prb;
	prb &= ~SELMASK(drive);
	ciab.prb = prb;
	udelay (1);
	prb |= (SELMASK(0)|SELMASK(1)|SELMASK(2)|SELMASK(3));
	ciab.prb = prb;
	selected = -1;
	unit[drive].motor = 0;

	restore_flags(flags);
}

static struct timer_list flush_track_timer;

static struct timer_list motor_on_timer;
static struct timer_list motor_off_timer[FD_MAX_UNITS];
static int on_attempts;

static void motor_on_callback(unsigned long nr)
{
	if (!(ciaa.pra & DSKRDY) || --on_attempts == 0) {
		unit[nr].motor = 1;
		wake_up (&motor_wait);
	} else {
		motor_on_timer.expires = HZ/10;
		add_timer(&motor_on_timer);
	}
}

static int motor_on(int nr)
{
	unsigned long flags;
	unsigned char prb = ~0;

	save_flags (flags);
	cli();
	del_timer(motor_off_timer + nr);

	if (!unit[nr].motor) {
		del_timer(&motor_on_timer);
		motor_on_timer.data = nr;
		motor_on_timer.expires = HZ/2;
		add_timer(&motor_on_timer);
		on_attempts = 10;


		prb &= ~DSKMOTOR;
		if (unit[nr].track % 2 != 0)
			prb &= ~DSKSIDE;
		ciab.prb |= (SELMASK(0)|SELMASK(1)|SELMASK(2)|SELMASK(3));
		ciab.prb = prb;
		prb &= ~SELMASK(nr);
		ciab.prb = prb;
		selected = nr;

		while (!unit[nr].motor)
			sleep_on (&motor_wait);
	}
	restore_flags(flags);

	if (on_attempts == 0) {
		printk ("motor_on failed, turning motor off\n");
		fd_motor_off (nr);
		return 0;
	}

	return 1;
}

static void floppy_on (unsigned int nr)
{
	static int useless_var = 0;

	if (useless_var)
		floppy_on (nr);
}

static void floppy_off (unsigned int nr)
{
	del_timer(motor_off_timer+nr);
	motor_off_timer[nr].expires = 3*HZ;
	add_timer(motor_off_timer+nr);
}

static void fd_select (int drive)
{
	unsigned char prb = ~0;

	if (drive == selected)
		return;
	selected = drive;

	if (unit[drive].track % 2 != 0)
		prb &= ~DSKSIDE;
	if (unit[drive].motor == 1)
		prb &= ~DSKMOTOR;
	ciab.prb |= (SELMASK(0)|SELMASK(1)|SELMASK(2)|SELMASK(3));
	ciab.prb = prb;
	prb &= ~SELMASK(drive);
	ciab.prb = prb;
}

static void fd_deselect (int drive)
{
	unsigned char prb;
	unsigned long flags;

	if (drive != selected)
		return;

	save_flags (flags);
	sti();

	selected = -1;

	prb = ciab.prb;
	prb |= (SELMASK(0)|SELMASK(1)|SELMASK(2)|SELMASK(3));
	ciab.prb = prb;

	restore_flags (flags);

}

/*======================================================================
  Seek the drive to track 0.
  The drive must be active and the motor must be running.
  Returns standard floppy error code.
======================================================================*/
static int fd_calibrate(int drive)
{
	unsigned char prb;
	int n;

	if (!motor_on (drive))
		return 0;
	fd_select (drive);
	prb = ciab.prb;
	prb |= DSKSIDE;
	prb &= ~DSKDIREC;
	ciab.prb = prb;
	for (n = unit[drive].type->tracks/4; n != 0; --n) {
		if (ciaa.pra & DSKTRACK0)
			break;
		prb &= ~DSKSTEP;
		ciab.prb = prb;
		prb |= DSKSTEP;
		udelay (2);
		ciab.prb = prb;
		udelay(unit[drive].type->step_delay);
	}
	udelay (unit[drive].type->settle_time);
	prb |= DSKDIREC;
	n = unit[drive].type->tracks/2 + 20;
	for (;;) {
		prb &= ~DSKSTEP;
		ciab.prb = prb;
		prb |= DSKSTEP;
		udelay (2);
		ciab.prb = prb;
		udelay(unit[drive].type->step_delay + 1000);
		if ((ciaa.pra & DSKTRACK0) == 0)
			break;
		if (--n == 0) {
			printk ("calibrate failed, turning motor off\n");
			fd_motor_off (drive);
			unit[drive].track = -1;
			return 0;
		}
	}
	unit[drive].track = 0;
	udelay(unit[drive].type->settle_time);

	return 1;
}

/*======================================================================
  Seek the drive to the requested cylinder.
  The drive must have been calibrated at some point before this.
  The drive must also be active and the motor must be running.
======================================================================*/
static int fd_seek(int drive, int track)
{
	unsigned char prb;
	int cnt;

	if (unit[drive].track == track)
		return 1;
	if (!motor_on(drive))
		return 0;
	fd_select (drive);
	if (unit[drive].track < 0 && !fd_calibrate(drive))
		return 0;

	cnt = unit[drive].track/2 - track/2;
	prb = ciab.prb;
	prb |= DSKSIDE | DSKDIREC;
	if (track % 2 != 0)
		prb &= ~DSKSIDE;
	if (cnt < 0) {
		cnt = - cnt;
		prb &= ~DSKDIREC;
	}
	ciab.prb = prb;
	if (track % 2 != unit[drive].track % 2)
		udelay (unit[drive].type->side_time);
	unit[drive].track = track;
	if (cnt == 0)
		return 1;
	do {
		prb &= ~DSKSTEP;
		ciab.prb = prb;
		prb |= DSKSTEP;
		udelay (1);
		ciab.prb = prb;
		udelay (unit[drive].type->step_delay);
	} while (--cnt != 0);
	udelay (unit[drive].type->settle_time);

	return 1;
}

static void encode(unsigned long data, unsigned long *dest)
{
  unsigned long data2;

  data &= 0x55555555;
  data2 = data ^ 0x55555555;
  data |= ((data2 >> 1) | 0x80000000) & (data2 << 1);

  if (*(dest - 1) & 0x00000001)
    data &= 0x7FFFFFFF;

  *dest = data;
}

static void encode_block(unsigned long *dest, unsigned long *src, int len)
{
  int cnt, to_cnt = 0;
  unsigned long data;

  /* odd bits */
  for (cnt = 0; cnt < len / 4; cnt++) {
    data = src[cnt] >> 1;
    encode(data, dest + to_cnt++);
  }

  /* even bits */
  for (cnt = 0; cnt < len / 4; cnt++) {
    data = src[cnt];
    encode(data, dest + to_cnt++);
  }
}

unsigned long checksum(unsigned long *addr, int len)
{
	unsigned long csum = 0;

	len /= sizeof(*addr);
	while (len-- > 0)
		csum ^= *addr++;
	csum = ((csum>>1) & 0x55555555)  ^  (csum & 0x55555555);

	return csum;
}

struct header {
	unsigned char magic;
	unsigned char track;
	unsigned char sect;
	unsigned char ord;
	unsigned char labels[16];
	unsigned long hdrchk;
	unsigned long datachk;
};

static unsigned long *putsec(unsigned long *raw, int track, int cnt,
			     unsigned char *data)
{
	struct header hdr;
	int i;

	*raw = (raw[-1]&1) ? 0x2AAAAAAA : 0xAAAAAAAA;
	raw++;
	*raw++ = 0x44894489;

	hdr.magic = 0xFF;
	hdr.track = track;
	hdr.sect = cnt;
	hdr.ord = 11-cnt;
	for (i = 0; i < 16; i++)
		hdr.labels[i] = 0;
	hdr.hdrchk = checksum((ulong *)&hdr,
			      (char *)&hdr.hdrchk-(char *)&hdr);
	hdr.datachk = checksum((ulong *)data, 512);

	encode_block(raw, (ulong *)&hdr.magic, 4);
	raw += 2;
	encode_block(raw, (ulong *)&hdr.labels, 16);
	raw += 8;
	encode_block(raw, (ulong *)&hdr.hdrchk, 4);
	raw += 2;
	encode_block(raw, (ulong *)&hdr.datachk, 4);
	raw += 2;
	encode_block(raw, (ulong *)data, 512);
	raw += 256;

	return raw;
}


/*==========================================================================
  amiga_write converts track/labels data to raw track data
==========================================================================*/
static void amiga_write(int disk, unsigned long raw, unsigned char *data,
			int track)
{
	int cnt;
	unsigned long *ptr = (unsigned long *)raw;

	/* gap space */
	for (cnt = 0; cnt < 415; cnt++)
		*ptr++ = 0xaaaaaaaa;

	/* sectors */
	for (cnt = 0; cnt < 11; cnt++)
		ptr = putsec (ptr, track, cnt, data + cnt*512);
	*(ushort *)ptr = (ptr[-1]&1) ? 0x2AA8 : 0xAAA8;
	raw = (unsigned long)ptr + 2;
}

static unsigned long decode (unsigned long *data, unsigned long *raw,
				   int len)
{
	ulong *odd, *even;

	/* convert length from bytes to longwords */
	len >>= 2;
	odd = raw;
	even = odd + len;

	/* prepare return pointer */
	raw += len * 2;

	do {
		*data++ = ((*odd++ & 0x55555555) << 1) | (*even++ & 0x55555555);
	} while (--len != 0);

	return (ulong)raw;
}

#define MFM_NOSYNC	1
#define MFM_HEADER	2
#define MFM_DATA	3
#define MFM_TRACK	4

/*==========================================================================
 scan_sync - looks for the next start of sector marked by a sync. d3 is the
		sector number (10..0). When d3 = 10, can't be certain of a
		starting sync.
==========================================================================*/
static unsigned long scan_sync(unsigned long raw, unsigned long end)
{
	ushort *ptr = (ushort *)raw, *endp = (ushort *)end;

	while (ptr < endp && *ptr++ != 0x4489)
		;
	if (ptr < endp) {
		while (*ptr == 0x4489 && ptr < endp)
			ptr++;
		return (ulong)ptr;
	}
	return 0;
}

/*==========================================================================
  amiga_read reads a raw track of data into a track buffer
==========================================================================*/
static int amiga_read(int drive, unsigned char *track_data,
		      unsigned long raw, int track)
{
	unsigned long end;
	int scnt;
	unsigned long csum;
	struct header hdr;

	end = raw + unit[drive].type->read_size;

	for (scnt = unit[drive].sects-1; scnt >= 0; scnt--) {
		if (!(raw = scan_sync(raw, end))) {
			printk ("can't find sync for sector %d\n", scnt);
			return MFM_NOSYNC;
		}

		raw = decode ((ulong *)&hdr.magic, (ulong *)raw, 4);
		raw = decode ((ulong *)&hdr.labels, (ulong *)raw, 16);
		raw = decode ((ulong *)&hdr.hdrchk, (ulong *)raw, 4);
		raw = decode ((ulong *)&hdr.datachk, (ulong *)raw, 4);
		csum = checksum((ulong *)&hdr,
				(char *)&hdr.hdrchk-(char *)&hdr);

#ifdef DEBUG
		printk ("(%x,%d,%d,%d) (%lx,%lx,%lx,%lx) %lx %lx\n",
			hdr.magic, hdr.track, hdr.sect, hdr.ord,
			*(ulong *)&hdr.labels[0], *(ulong *)&hdr.labels[4],
			*(ulong *)&hdr.labels[8], *(ulong *)&hdr.labels[12],
			hdr.hdrchk, hdr.datachk);
#endif

		if (hdr.hdrchk != csum) {
			printk("MFM_HEADER: %08lx,%08lx\n", hdr.hdrchk, csum);
			return MFM_HEADER;
		}

		/* verify track */
		if (hdr.track != track) {
			printk("MFM_TRACK: %d, %d\n", hdr.track, track);
			return MFM_TRACK;
		}

		raw = decode ((ulong *)(track_data + hdr.sect*512),
			      (ulong *)raw, 512);
		csum = checksum((ulong *)(track_data + hdr.sect*512), 512);

		if (hdr.datachk != csum) {
			printk("MFM_DATA: (%x:%d:%d:%d) sc=%d %lx, %lx\n",
			       hdr.magic, hdr.track, hdr.sect, hdr.ord, scnt,
			       hdr.datachk, csum);
			printk ("data=(%lx,%lx,%lx,%lx)\n",
				((ulong *)(track_data+hdr.sect*512))[0],
				((ulong *)(track_data+hdr.sect*512))[1],
				((ulong *)(track_data+hdr.sect*512))[2],
				((ulong *)(track_data+hdr.sect*512))[3]);
			return MFM_DATA;
		}
	}

	return 0;
}

/*
 * Note that MAX_ERRORS=X doesn't imply that we retry every bad read
 * max X times - some types of errors increase the errorcount by 2 or
 * even 3, so we might actually retry only X/2 times before giving up.
 */
#define MAX_ERRORS 12

/*
 * The driver is trying to determine the correct media format
 * while probing is set. rw_interrupt() clears it after a
 * successful access.
 */
static int probing = 0;

/* Prevent "aliased" accesses. */
static fd_ref[4] = { 0,0,0,0 };
static fd_device[4] = { 0,0,0,0 };

/*
 * Current device number. Taken either from the block header or from the
 * format request descriptor.
 */
#define CURRENT_DEVICE (CURRENT->dev)

/* Current error count. */
#define CURRENT_ERRORS (CURRENT->errors)

static void request_done(int uptodate)
{
  timer_active &= ~(1 << FLOPPY_TIMER);
  end_request(uptodate);
}

/*
 * floppy-change is never called from an interrupt, so we can relax a bit
 * here, sleep etc. Note that floppy-on tries to set current_DOR to point
 * to the desired drive, but it will probably not survive the sleep if
 * several floppies are used at the same time: thus the loop.
 */
int amiga_floppy_change(struct buffer_head * bh)
{
	int drive = bh->b_dev & 3;
	int changed;

	if (!bh)
		return 0;

	if (MAJOR(bh->b_dev) != MAJOR_NR) {
		printk("floppy_change: not a floppy\n");
		return 0;
	}

	fd_select (drive);
	changed = !(ciaa.pra & DSKCHANGE);
	fd_deselect (drive);

	if (changed) {
		unit[drive].track = -1;
		selected = -1;
		savedtrack = -1;
		writepending = 0; /* if this was true before, too bad! */
		writefromint = 0;
		return 1;
	}
	return 0;
}

static __inline__ void copy_buffer(void *from, void *to)
{
  ulong *p1,*p2;
  int cnt;

  p1 = (ulong *)from;
  p2 = (ulong *)to;

  for (cnt = 0; cnt < 512/4; cnt++)
    *p2++ = *p1++;
}

static void raw_read(int drive, int track, char *ptrack, int len)
{
	/* setup adkcon bits correctly */
	custom.adkcon = ADK_MSBSYNC;
	custom.adkcon = ADK_SETCLR|ADK_WORDSYNC|ADK_FAST;

	custom.dsksync = MFM_SYNC;

	custom.dsklen = 0;
	udelay (unit[drive].type->side_time);
	custom.dskptr = (u_char *)ptrack;
	custom.dsklen = len/sizeof(short) | DSKLEN_DMAEN;
	custom.dsklen = len/sizeof(short) | DSKLEN_DMAEN;

	block_flag = 1;

	while (block_flag == 1)
		sleep_on (&wait_fd_block);

	custom.dsklen = 0;
}

static int raw_write(int drive, int track, char *ptrack, int len)
{
	ushort adk;

	if ((ciaa.pra & DSKPROT) == 0)
		return 0;

	/* clear adkcon bits */
	custom.adkcon = ADK_PRECOMP1|ADK_PRECOMP0|ADK_WORDSYNC|ADK_MSBSYNC;
	/* set appropriate adkcon bits */
	adk = ADK_SETCLR|ADK_FAST;
	if ((ulong)track >= unit[drive].type->precomp2)
		adk |= ADK_PRECOMP1;
	else if ((ulong)track >= unit[drive].type->precomp1)
		adk |= ADK_PRECOMP0;
	custom.adkcon = adk;

	custom.dsklen = DSKLEN_WRITE;
#if 0
	udelay (unit[drive].type->side_time);
#endif
	custom.dskptr = (u_char *)ptrack;
	custom.dsklen = len/sizeof(short) | DSKLEN_DMAEN|DSKLEN_WRITE;
	custom.dsklen = len/sizeof(short) | DSKLEN_DMAEN|DSKLEN_WRITE;

	block_flag = 2;
	return 1;
}

static void post_write (void)
{
  udelay (2000);    /* 2ms post-write delay */
  custom.dsklen = 0;
  writepending = 0;
  writefromint = 0;
}

static int get_track(int drive, int track)
{
	int error;

	if (savedtrack == track)
		return 0;

	raw_read(drive, track, raw_buf, unit[drive].type->read_size);
	savedtrack = -1;
	error = amiga_read(drive, trackdata, (unsigned long)raw_buf, track);
	switch (error) {
	    case 0:
		savedtrack = track;
		return 0;
	    case MFM_TRACK:
		unit[drive].track = -1;
		/* fall through */
	    default:
		return -1;
	}
}

static void flush_track_callback(unsigned long nr)
{
  writefromint = 1;
  amiga_write(nr, (unsigned long)raw_buf, trackdata, savedtrack);
  if (!raw_write(nr, savedtrack, raw_buf, unit[nr].type->write_size)) {
    printk ("floppy disk write protected\n");
    writefromint = 0;
    writepending = 0;
  }
}

static int non_int_flush_track (unsigned long nr)
{
  writefromint = 0;
  amiga_write(nr, (unsigned long)raw_buf, trackdata, savedtrack);
  if (!raw_write(nr, savedtrack, raw_buf, unit[nr].type->write_size)) {
    printk ("floppy disk write protected in write!\n");
    writepending = 0;
    return 0;
  }
  while (block_flag == 2)
    sleep_on (&wait_fd_block);
  post_write();
  return 1;
}

static void redo_fd_request(void)
{
	unsigned int block, track, sector;
	int device, drive, cnt;
	struct floppy_struct *floppy;
	char *data;
	unsigned long flags;

	if (CURRENT && CURRENT->dev < 0) return;

    repeat:
	if (!CURRENT) {
		if (!fdc_busy)
			printk("FDC access conflict!");
		fdc_busy = 0;
		wake_up(&fdc_wait);
		CLEAR_INTR;
		return;
	}

	if (MAJOR(CURRENT->dev) != MAJOR_NR)
		panic(DEVICE_NAME ": request list destroyed");

	if (CURRENT->bh) {
		if (!CURRENT->bh->b_lock)
			panic(DEVICE_NAME ": block not locked");
	}

	probing = 0;
	device = MINOR(CURRENT_DEVICE);
	if (device > 3) {
		/* manual selection */
		drive = device & 3;
		floppy = unit + drive;
	} else {
		/* Auto-detection */
		/* printk("redo_fd_request: can't handle auto detect\n");*/
		/* printk("redo_fd_request: default to normal\n");*/
		drive = device & 3;
		floppy = unit + drive;
	}

	save_flags (flags);
	cli();
	if (drive != selected && writepending) {
	  del_timer (&flush_track_timer);
	  restore_flags (flags);
	  if (!non_int_flush_track (selected)) {
	    end_request(0);
	    goto repeat;
	  }
	} else
	  restore_flags (flags);

	for (cnt = 0; cnt < 2; cnt++) {
		block = CURRENT->sector + cnt;
		if ((int)block > floppy->blocks) {
			request_done(0);
			goto repeat;
		}

		track = block / floppy->dtype->sects;
		sector = block - track * floppy->dtype->sects;
		data = CURRENT->buffer + 512 * cnt;

		save_flags (flags);
		cli();
		if (track != savedtrack && writepending) {
		  del_timer (&flush_track_timer);
		  restore_flags (flags);
		  if (!non_int_flush_track (selected)) {
		    end_request(0);
		    goto repeat;
		  }
		} else
		  restore_flags (flags);

		switch (CURRENT->cmd) {
		    case READ:
			if (!motor_on (drive)) {
				end_request(0);
				goto repeat;
			}
			fd_select (drive);
			if (!fd_seek(drive, track)) {
				end_request(0);
				goto repeat;
			}
			if (get_track(drive, track) == -1) {
				end_request(0);
				goto repeat;
			}
			copy_buffer(trackdata + sector * 512, data);
			break;

		    case WRITE:
			if (!motor_on (drive)) {
				end_request(0);
				goto repeat;
			}
			fd_select (drive);
			if (!fd_seek(drive, track)) {
				end_request(0);
				goto repeat;
			}
			if (get_track(drive, track) == -1) {
				end_request(0);
				goto repeat;
			}
			copy_buffer(data, trackdata + sector * 512);
			/*
			 * setup a callback to write the track buffer
			 * after a short (1 tick) delay.
			 */
			save_flags (flags);
			cli();

			if (writepending)
			    /* reset the timer */
			    del_timer (&flush_track_timer);
			    
			writepending = 1;
			flush_track_timer.data = drive;
			flush_track_timer.expires = 1;
			add_timer (&flush_track_timer);
			restore_flags (flags);
			break;

		    default:
			printk("do_fd_request: unknown command\n");
			request_done(0);
			goto repeat;
		}
	}

	request_done(1);
	goto repeat;
}

static void do_fd_request(void)
{
	cli();
	while (fdc_busy) sleep_on(&fdc_wait);
	fdc_busy = 1;
	sti();
	redo_fd_request();
}

static int fd_ioctl(struct inode *inode, struct file *filp,
		    unsigned int cmd, unsigned long param)
{
	printk("fd_ioctl: %d\n", cmd);
	return 0;
}

/*======================================================================
  Return unit ID number of given disk
======================================================================*/
static unsigned long get_drive_id(int drive)
{
	int i;
	ulong id = 0;

	/* set up for ID */
	MOTOR_ON;
	udelay(2);
	SELECT(SELMASK(drive));
	udelay(2);
	DESELECT(SELMASK(drive));
	udelay(2);
	MOTOR_OFF;
	udelay(2);
	SELECT(SELMASK(drive));
	udelay(2);
	DESELECT(SELMASK(drive));
	udelay(2);

	/* loop and read disk ID */
	for (i=0; i<32; i++) {
		SELECT(SELMASK(drive));
		udelay(2);

		/* read and store value of DSKRDY */
		id <<= 1;
		id |= (ciaa.pra & DSKRDY) ? 0 : 1;	/* cia regs are low-active! */

		DESELECT(SELMASK(drive));
	}

	selected = -1;

        /*
         * RB: At least A500/A2000's df0: don't identify themselves.
         * As every (real) Amiga has at least a 3.5" DD drive as df0:
         * we default to that if df0: doesn't identify as a certain
         * type.
         */
        if(drive == 0 && id == FD_NODRIVE)
                id = FD_DD_3;

	/* return the ID value */
	return (id);
}

static void fd_probe(int drive)
{
	unsigned long code;
	int type, data;

	code = get_drive_id(drive);

	/* get drive type */
	unit[drive].type = NULL;
	for (type = 0; type < num_dr_types; type++)
		if (drive_types[type].code == code)
			break;

	if (type >= num_dr_types) {
		printk("fd_probe: unsupported drive type %08lx found\n",
		       code);
		return;
	}

	/* get data type */
	unit[drive].dtype = NULL;
	for (data = 0; data < num_da_types; data++)
		break;

	if (data >= num_da_types) {
		printk("fd_probe: can't get data type\n");
		return;
	}

	unit[drive].type = &drive_types[type];
	unit[drive].dtype = &data_types[data];
	unit[drive].track = -1;

	unit[drive].sects = data_types[data].sects * drive_types[type].sect_mult;
	unit[drive].blocks = drive_types[type].heads * drive_types[drive].tracks *
		unit[drive].sects;

	unit[drive].disk = -1;
	unit[drive].motor = 0;
	unit[drive].busy = 0;
	unit[drive].status = -1;
}

static void config_types(void)
{
	int drive;

	printk("Probing floppy drive(s):\n");
	for (drive = 0; drive < FD_MAX_UNITS; drive++) {
		fd_probe(drive);
		if (unit[drive].type->code != FD_NODRIVE)
			printk("Unit %d: %s\n", drive, unit[drive].type->name);
	}
}

/*
 * floppy_open check for aliasing (/dev/fd0 can be the same as
 * /dev/PS0 etc), and disallows simultaneous access to the same
 * drive with different device numbers.
 */
static int floppy_open(struct inode *inode, struct file *filp)
{
  int drive;
  int old_dev;

  drive = inode->i_rdev & 3;
  old_dev = fd_device[drive];

  if (fd_ref[drive])
    if (old_dev != inode->i_rdev)
      return -EBUSY;

  fd_ref[drive]++;
  fd_device[drive] = inode->i_rdev;

  if (old_dev && old_dev != inode->i_rdev)
    invalidate_buffers(old_dev);

  if (filp && filp->f_mode)
    check_disk_change(inode->i_rdev);

  if (filp && (filp->f_flags & (O_WRONLY|O_RDWR))) {
	  int wrprot;

	  fd_select (drive);
	  wrprot = !(ciaa.pra & DSKPROT);
	  fd_deselect (drive);

	  if (wrprot)
		  return -EROFS;
  }

  return 0;
}

static void floppy_release(struct inode * inode, struct file * filp)
{
  unsigned long flags;

  sync_dev(inode->i_rdev);
  save_flags (flags);
  cli();
  if ((inode->i_rdev & 3) == selected && writepending) {
    del_timer (&flush_track_timer);
    restore_flags (flags);
    non_int_flush_track (selected);
  } else
    restore_flags (flags);
  
  if (!fd_ref[inode->i_rdev & 3]--) {
    printk("floppy_release with fd_ref == 0");
    fd_ref[inode->i_rdev & 3] = 0;
  }
}

static struct file_operations floppy_fops = {
	NULL,			/* lseek - default */
	block_read,		/* read - general block-dev read */
	block_write,		/* write - general block-dev write */
	NULL,			/* readdir - bad */
	NULL,			/* select */
	fd_ioctl,		/* ioctl */
	NULL,			/* mmap */
	floppy_open,		/* open */
	floppy_release, 	/* release */
	block_fsync		/* fsync */
};

static void fd_block_done(struct intframe *fp, void *data)
{
  if (block_flag)
    custom.dsklen = 0x4000;

  block_flag = 0;
  wake_up (&wait_fd_block);

  if (writefromint) {
    /* 
     * if it was a write from an interrupt,
     * we will call post_write from here
     */
    post_write ();
  }

}

void amiga_floppy_init(void)
{
  int i;

  if (register_blkdev(MAJOR_NR,"fd",&floppy_fops)) {
    printk("Unable to get major %d for floppy\n",MAJOR_NR);
    return;
  }

  /* initialize variables */
  motor_on_timer.next = NULL;
  motor_on_timer.prev = NULL;
  motor_on_timer.expires = 0;
  motor_on_timer.data = 0;
  motor_on_timer.function = motor_on_callback;
  for (i = 0; i < FD_MAX_UNITS; i++) {
	  motor_off_timer[i].next = NULL;
	  motor_off_timer[i].prev = NULL;
	  motor_off_timer[i].expires = 0;
	  motor_off_timer[i].data = i;
	  motor_off_timer[i].function = fd_motor_off;

	  unit[i].track = -1;
  }

  flush_track_timer.next = NULL;
  flush_track_timer.prev = NULL;
  flush_track_timer.expires = 0;
  flush_track_timer.data = 0;
  flush_track_timer.function = flush_track_callback;
  
  /* blk_size[MAJOR_NR] = floppy_sizes; */
  blk_dev[MAJOR_NR].request_fn = DEVICE_REQUEST;

  timer_table[FLOPPY_TIMER].fn = NULL;
  timer_active &= ~(1 << FLOPPY_TIMER);

  config_types();

  raw_buf = (char *)amiga_chip_alloc (30000);

  /* make sure that disk DMA is enabled */
  custom.dmacon = DMAF_SETCLR | DMAF_DISK;

  add_isr(IRQ_FLOPPY, fd_block_done, 0, NULL);
}
