/*
 *  linux/kernel/blk_drv/floppy.c
 *
 *  Copyright (C) 1993  Greg Harp
 *  Atari Support by Bjoern Brauel, Roman Hodek
 *
 *  Big cleanup Sep 11..14 1994 Roman Hodek:
 *   - Driver now works interrupt driven
 *   - Support for two drives; should work, but I cannot test that :-(
 *   - Reading is done in whole tracks and buffered to speed up things
 *   - Disk change detection and drive deselecting after motor-off
 *     similar to TOS
 *   - Autodetection of disk format (DD/HD); untested yet, because I
 *     don't have an HD drive :-(
 *
 *  Fixes Nov 13 1994 Martin Schaller:
 *   - Autodetection works now
 *   - Support for 5 1/4" disks
 *   - Removed drive type (unknown on atari)
 *   - Do seeks with 8 Mhz
 *
 *  Changes by Andreas Schwab:
 *   - After errors in multiple read mode try again reading single sectors
 *  (Feb 1995):
 *   - Clean up error handling
 *   - Set blk_size for proper size checking
 *   - Initialize track register when testing presence of floppy
 *   - Implement some ioctl's
 *
 *  Things left to do:
 *   - Formatting
 *   - Maybe a better strategy for disk change detection (does anyone
 *     know one?)
 *   - There are some strange problems left: The strangest one is
 *     that, at least on my TT (4+4MB), the first 2 Bytes of the last
 *     page of the TT-Ram (!) change their contents (some bits get
 *     set) while a floppy DMA is going on. But there are no accesses
 *     to these memory locations from the kernel... (I tested that by
 *     making the page read-only). I cannot explain what's going on...
 *   - Sometimes the drive-change-detection stops to work. The
 *     function is still called, but the WP bit always reads as 0...
 *     Maybe a problem with the status reg mode or a timing problem.
 *     Note 10/12/94: The change detection now seems to work reliably.
 *     There is no proof, but I've seen no hang for a long time...
 */

#include <linux/config.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/kernel.h>
#include <linux/timer.h>
#include <linux/atafdreg.h>
#include <linux/fd.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/mm.h>
#include <linux/bootinfo.h>
#include <asm/system.h>
#include <asm/bitops.h>

#include <linux/atarihw.h>
#include <linux/atariints.h>
#include <linux/atari_stdma.h>

#define MAJOR_NR FLOPPY_MAJOR
#include "blk.h"


/* Note: FD_MAX_UNITS could be redefined to 2 for the Atari (with
 * little additional rework in this file). But I'm not yet sure if
 * some other code depends on the number of floppies... (It is defined
 * in a public header!)
 */
#if 0
#undef FD_MAX_UNITS
#define	FD_MAX_UNITS	2
#endif

#define TRACKBUFFER 1
#undef DEBUG

/* Disk types: DD or HD */
static struct atari_disk_type {
	const char	*name;
	unsigned	spt;		/* sectors per track */
	unsigned	blocks;		/* total number of blocks */
	unsigned	fdc_speed;	/* fdc_speed setting */
	unsigned 	stretch;	/* track doubling ? */
} disk_type[] = {
	{ "d360",  9, 720, 0, 0},	/* 360kB diskette */
	{ "h1200",15,2400, 3, 0},	/* 1.2MB diskette */
	{ "D360",  9, 720, 0, 1},	/* 360kb in 720kb drive */
	{ "D720",  9,1440, 0, 0},	/* 720kb diskette (DD) */	
/* the following 2 are only aliases on the atari, since we can't read/write 
   with 300kbps */
	{ "h360",  9, 720, 0, 1},	/* 360kb in 1.2MB drive */
	{ "h720",  9,1440, 0, 0},	/* 720kb in 1.2MB drive */
	{ "H1440",18,2880, 3, 0},	/* 1.4 MB diskette (HD) */
	{ "D820", 10,1640, 0, 0},	/* DD disk with 82 tracks/10 sectors */
	{ "H1640",20,3280, 3, 0},	/* HD disk with 82 tracks/20 sectors */
};

#define	NUM_DISK_TYPES (sizeof(disk_type)/sizeof(*disk_type))

/*
 * Maximum disk size (in kilobytes). This default is used whenever the
 * current disk size is unknown.
 */
#define MAX_DISK_SIZE 1640

static int floppy_sizes[] ={
	MAX_DISK_SIZE, MAX_DISK_SIZE, MAX_DISK_SIZE, MAX_DISK_SIZE,
	 360, 360 ,360, 360,
	1200,1200,1200,1200,
	 360, 360, 360, 360,
	 720, 720, 720, 720,
	 360, 360, 360, 360,
	 720, 720, 720, 720,
	1440,1440,1440,1440,
	 820, 820, 820, 820,
	1640,1640,1640,1640
};

/* current info on each unit */
static struct atari_floppy_struct {
	int connected;				/* !=0 : drive is connected */
	int autoprobe;				/* !=0 : do autoprobe	    */

	struct atari_disk_type	*disktype;	/* current type of disk */

	         int			track;		/* current head position or -1
										 * if unknown */
	unsigned int			steprate;	/* steprate setting */
	unsigned int			wpstat;		/* current state of WP signal
										 * (for disk change detection) */
} unit[FD_MAX_UNITS];


#define FDC_READ(reg) ({						\
	/* unsigned long __flags; */				\
	unsigned short __val;						\
    /* save_flags(__flags); cli(); */			\
    dma_wd.dma_mode_status = 0x80 | (reg);		\
	udelay(40);									\
	__val = dma_wd.fdc_acces_seccount;			\
    /* restore_flags(__flags); */				\
	__val & 0xff;								\
})

#define FDC_WRITE(reg,val)						\
    do {										\
        /* unsigned long __flags; */			\
        /* save_flags(__flags); cli(); */		\
		dma_wd.dma_mode_status = 0x80 | (reg);	\
		udelay(40);								\
		dma_wd.fdc_acces_seccount = (val);		\
        /* restore_flags(__flags); */			\
	} while(0)


/* Buffering variables:
 * First, there is a DMA buffer in ST-RAM that is used for floppy DMA
 * operations. Second, a track buffer is used to cache a whole track
 * of the disk to save read operations. These are two seperate buffers
 * because that allows write operations without clearing the track buffer.
 */

#define	MAX_SECTORS	22

unsigned char *DMABuffer;			  /* buffer for writes */
static unsigned long PhysDMABuffer;   /* physical address */

#ifdef TRACKBUFFER
unsigned char *TrackBuffer;			  /* buffer for reads */
static unsigned long PhysTrackBuffer; /* physical address */
static int BufferDrive, BufferSide, BufferTrack;
static int read_track;		/* non-zero if we are reading whole tracks */

#define	SECTOR_BUFFER(sec)	(TrackBuffer + ((sec)-1)*512)
#define	IS_BUFFERED(drive,side,track) \
    (BufferDrive == (drive) && BufferSide == (side) && BufferTrack == (track))
#endif

/*
 * These are global variables, as that's the easiest way to give
 * information to interrupts. They are the data used for the current
 * request.
 */
static int SelectedDrive = 0;
static int ReqCmd, ReqBlock;
static int ReqSide, ReqTrack, ReqSector, ReqCnt;
static int HeadSettleFlag = 0;
static unsigned char *ReqData, *ReqBuffer;
static int MotorOn = 0, MotorOffTrys;

/* Synchronization of FDC access. */
static volatile int fdc_busy = 0;
static struct wait_queue *fdc_wait = NULL;

static unsigned int changed_floppies = 0xff, fake_change = 0;
#define	CHECK_CHANGE_DELAY	HZ/2

#define	FD_MOTOR_OFF_DELAY	(3*HZ)
#define	FD_MOTOR_OFF_MAXTRY	(10*20)

#define FLOPPY_TIMEOUT		(6*HZ)
#define RECALIBRATE_ERRORS	3	/* Atfer this many errors the drive
								 * will be recalibrated. */
#define MAX_ERRORS			6	/* After this many errors the driver
								 * will give up. */


#define	START_MOTOR_OFF_TIMER(delay)			\
    do {										\
        motor_off_timer.expires = (delay);		\
        add_timer( &motor_off_timer );			\
        MotorOffTrys = 0;						\
	} while(0)

#define	START_CHECK_CHANGE_TIMER(delay)							\
    do {														\
        timer_table[FLOPPY_TIMER].expires = jiffies + (delay);	\
        timer_active |= (1 << FLOPPY_TIMER);					\
	} while(0)

#define	START_TIMEOUT()							\
    do {										\
        del_timer( &timeout_timer );			\
        timeout_timer.expires = FLOPPY_TIMEOUT;	\
        add_timer( &timeout_timer );			\
	} while(0)

#define	STOP_TIMEOUT()							\
    do {										\
        del_timer( &timeout_timer );			\
	} while(0)


/*
 * The driver is trying to determine the correct media format
 * while Probing is set. fd_rwsec_done() clears it after a
 * successful access.
 */
static int Probing = 0;

/* This flag is set when a dummy seek is necesary to make the WP
 * status bit accessible.
 */
static int NeedSeek = 0;


#ifdef DEBUG
#define DPRINT(a)	printk a
#else
#define DPRINT(a)
#endif

/***************************** Prototypes *****************************/

static void fd_select_side( int side );
static void fd_select_drive( int drive );
static void fd_deselect( void );
static void fd_motor_off_timer( unsigned long dummy );
static void check_change( void );
static __inline__ void set_head_settle_flag( void );
static __inline__ int get_head_settle_flag( void );
static void floppy_irq( struct intframe *fp, void *data );
static void fd_error( void );
static void do_fd_action( int drive );
static void fd_calibrate( void );
static void fd_calibrate_done( int status );
static void fd_seek( void );
static void fd_seek_done( int status );
static void fd_rwsec( void );
#ifdef TRACKBUFFER
static void fd_readtrack_check( unsigned long dummy );
#endif
static void fd_rwsec_done( int status );
static void fd_times_out( unsigned long dummy );
static void finish_fdc( void );
static void fd_finish_done (int status);
static void floppy_off( unsigned int nr);
static __inline__ void copy_buffer( void *from, void *to);
static void setup_req_params( int drive );
static void redo_fd_request( void);
static int fd_ioctl( struct inode *inode, struct file *filp, unsigned int
                     cmd, unsigned long param);
static void fd_probe( int drive );
static int fd_test_drive_present( int drive );
static void config_types( void );
static int floppy_open( struct inode *inode, struct file *filp );
static void floppy_release( struct inode * inode, struct file * filp );

/************************* End of Prototypes **************************/

static struct timer_list motor_off_timer =
						 { NULL, NULL, 0, 0, fd_motor_off_timer };
#ifdef TRACKBUFFER
static struct timer_list readtrack_timer =
						 { NULL, NULL, 0, 0, fd_readtrack_check };
#endif

static struct timer_list timeout_timer =
						 { NULL, NULL, 0, 0, fd_times_out };



/* Select the side to use. */

static void fd_select_side( int side )

{	unsigned long flags;

	save_flags(flags);
	cli();
  
	sound_ym.rd_data_reg_sel = 14; /* Select PSG Port A */
	sound_ym.wd_data = (side == 0) ? sound_ym.rd_data_reg_sel | 0|01 :
	                                 sound_ym.rd_data_reg_sel & 0xfe;

	restore_flags(flags);
}


/* Select a drive, update the FDC's track register and set the correct
 * clock speed for this disk's type.
 */

static void fd_select_drive( int drive )

{
	unsigned long flags;
	unsigned char tmp;
  
	if (drive == SelectedDrive)
	  return;

	save_flags(flags);
	cli();
	sound_ym.rd_data_reg_sel = 14; /* Select PSG Port A */
	tmp = sound_ym.rd_data_reg_sel;
	sound_ym.wd_data = (tmp | DSKDRVNONE) & ~(drive == 0 ? DSKDRV0 : DSKDRV1);
	restore_flags(flags);

	/* restore track register to saved value */
	FDC_WRITE( FDCREG_TRACK, unit[drive].track );
	udelay(40);

	/* select 8/16 MHz */
	if (unit[drive].disktype)
		dma_wd.fdc_speed = unit[drive].disktype->fdc_speed;
	
	SelectedDrive = drive;
}


/* Deselect both drives. */

static void fd_deselect( void )

{	unsigned long flags;

	save_flags(flags);
	cli();
    sound_ym.rd_data_reg_sel=14;	/* Select PSG Port A */
	sound_ym.wd_data = sound_ym.rd_data_reg_sel | 7; /* no drives selected */
	restore_flags(flags);

	SelectedDrive = -1;
}


/* This timer function deselects the drives when the FDC switched the
 * motor off. The deselection cannot happen earlier because the FDC
 * counts the index signals, which arrive only if one drive is selected.
 */

static void fd_motor_off_timer( unsigned long dummy )

{	unsigned long flags;
	unsigned char status;
	int		      delay;

	del_timer( &motor_off_timer );
	
	if (SelectedDrive < 0)
		/* no drive selected, needn't deselect anyone */
		return;

	save_flags(flags);
	cli();
	
	if (stdma_islocked())
		goto retry;

	status = FDC_READ( FDCREG_STATUS );

	if (!(status & 0x80)) {
		/* motor already turned off by FDC -> deselect drives */
		fd_deselect();
		restore_flags(flags);
		MotorOn = 0;
		return;
   }
	/* not yet off, try again */

  retry:
  restore_flags(flags);
	/* Test again later; if tested too often, it seems there is no disk
	 * in the drive and the FDC will leave the motor on forever (or,
	 * at least until a disk is inserted). So we'll test only twice
	 * per second from then on...
	 */
	delay = (MotorOffTrys < FD_MOTOR_OFF_MAXTRY) ?
		        (++MotorOffTrys, HZ/20) : HZ/2;
	START_MOTOR_OFF_TIMER( delay );
}


/* This function is repeatedly called to detect disk changes (as good
 * as possible) and keep track of the current state of the write protection.
 */

static void check_change( void )

{	static int    drive = 0;

	unsigned long flags;
	unsigned char old_porta;
	int			  stat;

	if (++drive > 1 || !unit[drive].connected)
		drive = 0;

	if (!stdma_islocked()) {

		save_flags(flags);
		cli();
		sound_ym.rd_data_reg_sel = 14;
		old_porta = sound_ym.rd_data_reg_sel;
		sound_ym.wd_data = (old_porta | DSKDRVNONE) &
			               ~(drive == 0 ? DSKDRV0 : DSKDRV1);
		stat = !!(FDC_READ( FDCREG_STATUS ) & FDCSTAT_WPROT);
		sound_ym.wd_data = old_porta;
		restore_flags(flags);

		if (stat != unit[drive].wpstat) {
			DPRINT(( "wpstat[%d] = %d\n", drive, stat ));
			unit[drive].wpstat = stat;
			set_bit (drive, &changed_floppies);
		}
	}
	
	START_CHECK_CHANGE_TIMER( CHECK_CHANGE_DELAY );
}

 
/* Handling of the Head Settling Flag: This flag should be set after each
 * seek operation, because we dont't use seeks with verify.
 */

static __inline__ void set_head_settle_flag( void )

{
	HeadSettleFlag = FDCCMDADD_E;
}

static __inline__ int get_head_settle_flag( void )

{
	int	tmp = HeadSettleFlag;
	HeadSettleFlag = 0;
	return( tmp );
}

  
  

/* General Interrupt Handling */

static void (*FloppyIRQHandler)( int status ) = NULL;

static void floppy_irq( struct intframe *fp, void *data )

{
	unsigned char status;
	void (*handler)( int );

	handler = FloppyIRQHandler;
	FloppyIRQHandler = NULL;

	if (handler) {
		nop();
		status = FDC_READ( FDCREG_STATUS );
		DPRINT(("FDC irq, status = %02x handler = %08lx\n",status,(unsigned long)handler));
		udelay(40);
		handler( status );
	}
	else {
		DPRINT(("FDC irq, no handler\n"));
	}
}


/* Error handling: If some error happened, retry some times, then
 * recalibrate, then try again, and fail after MAX_ERRORS.
 */

static void fd_error( void )

{
	if (!CURRENT) return;
	CURRENT->errors++;
	if (CURRENT->errors >= MAX_ERRORS) {
		printk( "fd%d: too many errors.\n", SelectedDrive );
		end_request( 0 );
	}
	else if (CURRENT->errors == RECALIBRATE_ERRORS) {
		printk( "fd%d: recalibrating\n", SelectedDrive );
		if (SelectedDrive != -1)
			unit[SelectedDrive].track = -1;
	}
	redo_fd_request();
}



#define	SET_IRQ_HANDLER(proc) do { FloppyIRQHandler = (proc); } while(0)


/* do_fd_action() is the general procedure for a fd request: All
 * required parameter settings (drive select, side select, track
 * position) are checked and set if needed. For each of these
 * parameters and the actual reading or writing exist two functions:
 * one that starts the setting (or skips it if possible) and one
 * callback for the "done" interrupt. Each done func calls the next
 * set function to propagate the request down to fd_rwsec_done().
 */

static void do_fd_action( int drive )

{
	DPRINT(("do_fd_action\n"));
	
#ifdef TRACKBUFFER
  repeat:

	if (IS_BUFFERED( drive, ReqSide, ReqTrack )) {
		if (ReqCmd == READ) {
			copy_buffer( SECTOR_BUFFER(ReqSector), ReqData );
			if (++ReqCnt < 2) {
				/* read next sector */
				setup_req_params( drive );
				goto repeat;
			}
			else {
				/* all sectors finished */
				end_request( 1 );
				redo_fd_request();
				return;
			}
		}
		else {
			/* cmd == WRITE, pay attention to track buffer
			 * consistency! */
			copy_buffer( ReqData, SECTOR_BUFFER(ReqSector) );
		}
	}
#endif

	if (SelectedDrive != drive)
		fd_select_drive( drive );
    
	if (unit[drive].track == -1)
		fd_calibrate();
	else if (unit[drive].track != ReqTrack << unit[drive].disktype->stretch)
		fd_seek();
	else
		fd_rwsec();
}


/* Seek to track 0 if the current track is unknown */

static void fd_calibrate( void )

{
	if (unit[SelectedDrive].track >= 0) {
		fd_calibrate_done( 0 );
		return;
	}

	dma_wd.fdc_speed = 0; 	/* always seek with 8 Mhz */;
	DPRINT(("fd_calibrate\n"));
	SET_IRQ_HANDLER( fd_calibrate_done );
	/* we can't verify, since the speed may be incorrect */
	FDC_WRITE( FDCREG_CMD, FDCCMD_RESTORE | unit[SelectedDrive].steprate );

	NeedSeek = 1;
	MotorOn = 1;
	START_TIMEOUT();
	/* wait for IRQ */
}


static void fd_calibrate_done( int status )

{
	DPRINT(("fd_calibrate_done()\n"));
	STOP_TIMEOUT();

	/* set the correct speed now */
	dma_wd.fdc_speed = unit[SelectedDrive].disktype->fdc_speed;
	if (status & FDCSTAT_RECNF) {
		printk( "fd%d: restore failed\n", SelectedDrive );
		fd_error();
	}
	else {
		unit[SelectedDrive].track = 0;
		fd_seek();
	}
}
  
  
/* Seek the drive to the requested track. The drive must have been
 * calibrated at some point before this.
 */
  
static void fd_seek( void )
  
{
	if (unit[SelectedDrive].track == ReqTrack << 
		unit[SelectedDrive].disktype->stretch) {
		fd_seek_done( 0 );
		return;
	}

 	dma_wd.fdc_speed = 0;	/* always seek witch 8 Mhz */

	DPRINT(("fd_seek() to track %d\n",ReqTrack));
	FDC_WRITE( FDCREG_DATA, ReqTrack << 
		unit[SelectedDrive].disktype->stretch);
	udelay(40);
	SET_IRQ_HANDLER( fd_seek_done );
	FDC_WRITE( FDCREG_CMD, FDCCMD_SEEK | unit[SelectedDrive].steprate );

	MotorOn = 1;
	set_head_settle_flag();
	START_TIMEOUT();
	/* wait for IRQ */
}


static void fd_seek_done( int status )
  
{
	DPRINT(("fd_seek_done()\n"));
	STOP_TIMEOUT();

	/* set the correct speed */
 	dma_wd.fdc_speed = unit[SelectedDrive].disktype->fdc_speed;	
	if (status & FDCSTAT_RECNF) {
		printk( "fd%d: seek error (to track %d)\n",
				SelectedDrive, ReqTrack );
		/* we don't know exactly which track we are on now! */
		unit[SelectedDrive].track = -1;
		fd_error();
	}
	else {
		unit[SelectedDrive].track = ReqTrack << 
			unit[SelectedDrive].disktype->stretch;
		NeedSeek = 0;
		fd_rwsec();
	}
}


/* This does the actual reading/writing after positioning the head
 * over the correct track.
 */

#ifdef TRACKBUFFER
static int MultReadInProgress = 0;
#endif


static void fd_rwsec( void )
	
{
	unsigned long paddr, flags;
	unsigned int  rwflag, old_motoron;
	unsigned int track;
	
	DPRINT(("fd_rwsec(), Sec=%d, Access=%c\n",ReqSector, ReqCmd == WRITE ? 'w' : 'r' ));
	if (ReqCmd == WRITE) {
		if (boot_info.bi_atari.hw_present & ATARIHW_EXTD_DMA) {
			cache_push( (unsigned long)ReqData, 512 );
			paddr = (unsigned long)ReqData;
		}
		else {
			copy_buffer( ReqData, DMABuffer );
			cache_push( PhysDMABuffer, 512 );
			paddr = PhysDMABuffer;
		}
		rwflag = 0x100;
	}
	else {
#ifdef TRACKBUFFER
		if (read_track)
		  paddr = PhysTrackBuffer;
		else
		  paddr = PhysDMABuffer;
#else
		paddr = PhysDMABuffer;
#endif
		rwflag = 0;
	}

	fd_select_side( ReqSide );
  
	/* Start sector of this operation */
#ifdef TRACKBUFFER
	FDC_WRITE( FDCREG_SECTOR, !read_track ? ReqSector : 1 );
#else
	FDC_WRITE( FDCREG_SECTOR, ReqSector );
#endif
	/* Cheat for track if stretch != 0 */
	if (unit[SelectedDrive].disktype->stretch) {
		track = FDC_READ( FDCREG_TRACK);
		FDC_WRITE( FDCREG_TRACK, track >> 
			unit[SelectedDrive].disktype->stretch);
	}
	udelay(40);
  
	/* Setup DMA */
	save_flags(flags);  
	cli();
	dma_wd.dma_lo = (unsigned char)paddr;
	paddr >>= 8;
	nop();
	dma_wd.dma_md = (unsigned char)paddr;
	paddr >>= 8;
	nop();
	if (boot_info.bi_atari.hw_present & ATARIHW_EXTD_DMA)
		st_dma_ext_dmahi = (unsigned short)paddr;
	else
		dma_wd.dma_hi = (unsigned char)paddr;
	nop();
	restore_flags(flags);
  
	/* Clear FIFO and switch DMA to correct mode */  
	dma_wd.dma_mode_status = 0x90 | rwflag;  
	nop();
	dma_wd.dma_mode_status = 0x90 | (rwflag ^ 0x100);  
	nop();
	dma_wd.dma_mode_status = 0x90 | rwflag;
	nop();
  
	/* How many sectors for DMA */
#ifdef TRACKBUFFER
	dma_wd.fdc_acces_seccount =
		!read_track ? 1 : unit[SelectedDrive].disktype->spt;
#else
	dma_wd.fdc_acces_seccount = 1;
#endif
  
	udelay(40);  
  
	/* Start operation */
	dma_wd.dma_mode_status = FDCSELREG_STP | rwflag;
	udelay(40);
	SET_IRQ_HANDLER( fd_rwsec_done );
	dma_wd.fdc_acces_seccount =
		(rwflag ? (FDCCMD_WRSEC | FDCCMDADD_P) :
#ifdef TRACKBUFFER
		          (FDCCMD_RDSEC | (read_track ? FDCCMDADD_M : 0))
#else
		          FDCCMD_RDSEC
#endif					  
         ) | get_head_settle_flag();

	old_motoron = MotorOn;
	MotorOn = 1;
	NeedSeek = 1;
	/* wait for interrupt */

#ifdef TRACKBUFFER
	if (read_track) {
		/* If reading a whole track, wait about one disk rotation and
		 * then check if all sectors are read. The FDC will even
		 * search for the first non-existant sector and need 1 sec to
		 * recognise that it isn't present :-(
		 */
		readtrack_timer.expires = HZ/5 + (old_motoron ? 0 : HZ);
		                          /* 1 rot. + 5 rot.s if motor was off  */
		add_timer( &readtrack_timer );
		MultReadInProgress = 1;
 }
#endif
	START_TIMEOUT();
}

    
#ifdef TRACKBUFFER

static void fd_readtrack_check( unsigned long dummy )
  
{	unsigned long flags, addr;
  
	save_flags(flags);  
	cli();
  
	del_timer( &readtrack_timer );
  
	if (!MultReadInProgress) {
		/* This prevents a race condition that could arise if the
		 * interrupt is triggered while the calling of this timer
		 * callback function takes place. The IRQ function then has
		 * already cleared 'MultReadInProgress'  when control flow
		 * gets here.
		 */
        restore_flags(flags);
		return;
	}
  
	/* get the current DMA address */
	addr = dma_wd.dma_lo & 0xff;
	nop();
    addr |= (dma_wd.dma_md & 0xff) << 8;
	nop();
	addr |= (dma_wd.dma_hi & 0xff) << 16;
	nop();
  
	if (addr >= PhysTrackBuffer + unit[SelectedDrive].disktype->spt*512) {
		/* already read enough data, force an FDC interrupt to stop
		 * the read operation
		 */
		SET_IRQ_HANDLER( NULL );
		restore_flags(flags);
		DPRINT(("fd_readtrack_check(): done\n"));
		FDC_WRITE( FDCREG_CMD, FDCCMD_FORCI );
		udelay(40);

		/* No error until now -- the FDC would have interrupted
		 * otherwise!
		 */
		fd_rwsec_done( 0 );
	}
	else {
		/* not yet finished, wait another tenth rotation */
		restore_flags(flags);
		DPRINT(("fd_readtrack_check(): not yet finished\n"));
		readtrack_timer.expires = HZ/5/10;
		add_timer( &readtrack_timer );
	}
}
  
#endif

static void fd_rwsec_done( int status )
  
{
	unsigned int track;

	DPRINT(("fd_rwsec_done()\n"));

#ifdef TRACKBUFFER
	if (read_track && !MultReadInProgress) return;
	MultReadInProgress = 0;

	STOP_TIMEOUT();
	
	if (read_track)
		del_timer( &readtrack_timer );
#endif

	/* Correct the track if stretch != 0 */
	if (unit[SelectedDrive].disktype->stretch) {
		track = FDC_READ( FDCREG_TRACK);
		FDC_WRITE( FDCREG_TRACK, track << 
			unit[SelectedDrive].disktype->stretch);
	}

#ifndef TRACKBUFFER
	dma_wd.dma_mode_status = 0x90;
	nop();
	if (!(dma_wd.dma_mode_status & 0x01)) {
		printk( "fd%d: DMA error\n", SelectedDrive );
		goto err_end;
	}
#endif
    
	if (ReqCmd == WRITE && (status & FDCSTAT_WPROT)) {
		printk( "fd%d: is write protected\n", SelectedDrive );
		goto err_end;
	}	
	if ((status & FDCSTAT_RECNF)
#ifdef TRACKBUFFER
		/* RECNF is no error after a multiple read when the FDC
		 * searched for a non-existant sector!
		 */
		&& !(read_track &&
			 FDC_READ(FDCREG_SECTOR) > unit[SelectedDrive].disktype->spt)
#endif
	   ) {
		if (Probing) {
			if (unit[SelectedDrive].disktype > disk_type) {
				/* try another disk type */
				unit[SelectedDrive].disktype--;
				floppy_sizes[SelectedDrive]
				  = unit[SelectedDrive].disktype->blocks >> 1;
			}
			else
				Probing=0;
		} else {	
/* record not found, but not probing. Maybe stretch wrong ? Restart probing */
			if (unit[SelectedDrive].autoprobe) {
				unit[SelectedDrive].disktype = disk_type + NUM_DISK_TYPES-1;
				floppy_sizes[SelectedDrive]
				  = unit[SelectedDrive].disktype->blocks >> 1;
				Probing = 1;
			}
		}
		if (Probing) {
			dma_wd.fdc_speed = unit[SelectedDrive].disktype->fdc_speed;
			setup_req_params( SelectedDrive );
#ifdef TRACKBUFFER
			BufferDrive = -1;
#endif
			do_fd_action( SelectedDrive );
			return;
		}

		printk( "fd%d: sector %d not found (side %d, track %d)\n",
				SelectedDrive, FDC_READ (FDCREG_SECTOR), ReqSide, ReqTrack );
		goto err_end;
	}
	if (status & FDCSTAT_CRC) {
		printk( "fd%d: CRC error (side %d, track %d, sector %d)\n",
				SelectedDrive, ReqSide, ReqTrack, FDC_READ (FDCREG_SECTOR) );
		goto err_end;
	}
	if (status & FDCSTAT_LOST) {
		printk( "fd%d: lost data (side %d, track %d, sector %d)\n",
				SelectedDrive, ReqSide, ReqTrack, FDC_READ (FDCREG_SECTOR) );
		goto err_end;
	}

	Probing = 0;
	
	if (ReqCmd == READ) {
#ifdef TRACKBUFFER
		if (!read_track)
		  {
		    cache_clear (PhysDMABuffer, 512);
		    copy_buffer (DMABuffer, ReqData);
		  }
		else
		  {
		    cache_clear (PhysTrackBuffer, MAX_SECTORS * 512);
		    BufferDrive = SelectedDrive;
		    BufferSide  = ReqSide;
		    BufferTrack = ReqTrack;
		    copy_buffer (SECTOR_BUFFER (ReqSector), ReqData);
		  }
#else
		cache_clear( PhysDMABuffer, 512 );
		copy_buffer( DMABuffer, ReqData );
#endif
	}
  
	if (++ReqCnt < 2) {
		/* read next sector */
		setup_req_params( SelectedDrive );
		do_fd_action( SelectedDrive );
	}
	else {
		/* all sectors finished */
		end_request( 1 );
		redo_fd_request();
	}
	return;
  
  err_end:
#ifdef TRACKBUFFER
	BufferDrive = -1;
#endif
	fd_error();
}


static void fd_times_out( unsigned long dummy )

{
	SET_IRQ_HANDLER( NULL );
	/* If the timeout occured while the readtrack_check timer was
	 * active, we need to cancel it, else bad things will happen */
	del_timer( &readtrack_timer );
	FDC_WRITE( FDCREG_CMD, FDCCMD_FORCI );
	udelay( 40 );
	
	printk( "floppy timeout\n" );
	fd_error();
}


/* The (noop) seek operation here is needed to make the WP bit in the
 * FDC status register accessible for check_change. If the last disk
 * operation would have been a RDSEC, this bit would always read as 0
 * no matter what :-( To save time, the seek goes to the track we're
 * already on.
 */

static void finish_fdc (void)
{
  dma_wd.fdc_speed = 0;

  if (!NeedSeek)
    {
      fd_finish_done (0);
      return;
    }

  DPRINT(("finish_fdc: dummy seek started\n"));
  FDC_WRITE (FDCREG_DATA, unit[SelectedDrive].track);
  SET_IRQ_HANDLER (fd_finish_done);
  FDC_WRITE (FDCREG_CMD, FDCCMD_SEEK);
  MotorOn = 1;
  START_TIMEOUT();
  /* wait for IRQ */
}	

static void fd_finish_done (int status)
{
	unsigned long flags;

	DPRINT(("finish_fdc_done entered\n"));
	STOP_TIMEOUT();
	NeedSeek = 0;
	
	if ((timer_active & (1 << FLOPPY_TIMER)) &&
	    timer_table[FLOPPY_TIMER].expires < jiffies + 5)
		/* If the check for a disk change is done too early after this
		 * last seek command, the WP bit still reads wrong :-((
		 */
		timer_table[FLOPPY_TIMER].expires = jiffies + 5;
	else
		START_CHECK_CHANGE_TIMER( CHECK_CHANGE_DELAY );
	del_timer( &motor_off_timer );
	START_MOTOR_OFF_TIMER( FD_MOTOR_OFF_DELAY );

	save_flags(flags);
	cli();
	stdma_release();
	fdc_busy = 0;
	wake_up( &fdc_wait );
	restore_flags(flags);

	DPRINT(("finish_fdc() finished\n"));
}


/* Prevent "aliased" accesses. */
static fd_ref[4] = { 0,0,0,0 };
static fd_device[4] = { 0,0,0,0 };

/*
 * Current device number. Taken either from the block header or from the
 * format request descriptor.
 */
#define CURRENT_DEVICE (CURRENT->dev)

/* Current error count. */
#define CURRENT_ERRORS (CURRENT->errors)


/* dummy for blk.h */
static void floppy_off( unsigned int nr) {}


/* The detection of disk changes is a dark chapter in Atari history :-(
 * Because the "Drive ready" signal isn't present in the Atari
 * hardware, one has to rely on the "Write Protect". This works fine,
 * as long as no write protected disks are used. TOS solves this
 * problem by introducing tri-state logic ("maybe changed") and
 * looking at the serial number in block 0. This isn't possible for
 * Linux, since the floppy driver can't make assumptions about the
 * filesystem used on the disk and thus the contents of block 0. I've
 * choosen the method to always say "The disk was changed" if it is
 * unsure whether it was. This implies that every open or mount
 * invalidates the disk buffers if you work with write protected
 * disks. But at least this is better than working with incorrect data
 * due to unrecognised disk changes.
 */

int atari_floppy_change( struct buffer_head * bh )

{	unsigned int drive = (bh->b_dev & 0x03);

	if (MAJOR(bh->b_dev) != MAJOR_NR) {
		printk("floppy_changed: not a floppy\n");
		return 0;
	}
	
	if (test_bit (drive, &fake_change)) {
		/* simulated change (e.g. after formatting) */
		clear_bit (drive, &fake_change);
		goto is_changed;
	}
	if (test_bit (drive, &changed_floppies)) {
		/* surely changed (the WP signal changed at least once) */
	  is_changed:
		clear_bit (drive, &changed_floppies);
		goto ret_changed;
	}
	if (unit[drive].wpstat) {
		/* WP is on -> could be changed: to be sure, buffers should be
         * invalidated...
		 */
	  ret_changed:
#ifdef TRACKBUFFER
		BufferDrive = -1;
#endif
		
		unit[drive].disktype = NULL;
		floppy_sizes[drive] = MAX_DISK_SIZE;
		return 1;
	}

	/* It doesn't make sense to read or write block 0 (described by
	 * bh) here, because it doesn't answer our question :-(
	 */

	return 0;
}


static __inline__ void copy_buffer(void *from, void *to)

{	ulong	*p1 = (ulong *)from, *p2 = (ulong *)to;
	int		cnt;

	for( cnt = 512/4; cnt; cnt-- )
	  *p2++ = *p1++;
}


/* This sets up the global variables describing the current request. */

static void setup_req_params( int drive )

{	int block = ReqBlock + ReqCnt;

	ReqTrack = block / unit[drive].disktype->spt;
	ReqSector = block - ReqTrack * unit[drive].disktype->spt + 1;
	ReqSide = ReqTrack & 1;
	ReqTrack >>= 1;
	ReqData = ReqBuffer + 512 * ReqCnt;

#ifdef TRACKBUFFER
	read_track = (ReqCmd == READ && CURRENT_ERRORS == 0);
#endif

	DPRINT(("Request params: Si=%d Tr=%d Se=%d Data=%08lx\n",ReqSide,
		ReqTrack, ReqSector, (unsigned long)ReqData ));
}


static void redo_fd_request(void)

{
	int device, drive, type;
	struct atari_floppy_struct *floppy;
  
	DPRINT(("redo_fd_request: CURRENT=%08lx CURRENT->dev=%04x CURRENT->sector=%ld\n",
		(unsigned long)CURRENT, CURRENT ? CURRENT->dev : 0,
		CURRENT ? CURRENT->sector : 0 ));

	if (CURRENT && CURRENT->dev < 0)
		goto the_end;

repeat:
    
	if (!CURRENT)
		goto the_end;

	if (MAJOR(CURRENT->dev) != MAJOR_NR)
		panic(DEVICE_NAME ": request list destroyed");

	if (CURRENT->bh) {
		if (!CURRENT->bh->b_lock)
			panic(DEVICE_NAME ": block not locked");
	}

	device = MINOR(CURRENT_DEVICE);
	drive = device & 3;
	type = device >> 2;
	floppy = &unit[drive];
	
	if (!floppy->connected) {
		/* drive not connected */
		printk( "Unknown Device: fd%d\n", drive );
		end_request(0);
		goto repeat;
	}
		
	if (type == 0) {
		if (!floppy->disktype) {
			Probing = 1;
			floppy->disktype  = disk_type + NUM_DISK_TYPES-1;
			floppy_sizes[drive] = floppy->disktype->blocks >> 1;
			floppy->autoprobe = 1;
		}
	} 
	else {
		/* user supplied disk type */
		--type;
		if (type >= NUM_DISK_TYPES) {
			printk( "fd%d: invalid disk format", drive );
			end_request( 0 );
			goto repeat;
		}
		floppy->disktype = &disk_type[type];
		floppy_sizes[drive] = disk_type[type].blocks >> 1;
		floppy->autoprobe = 0;
	}
	
	if (CURRENT->sector + 1 > floppy->disktype->blocks) {
			end_request(0);
			goto repeat;
		}

	/* stop deselect timer */
	del_timer( &motor_off_timer );
		
	ReqCnt = 0;
	ReqCmd = CURRENT->cmd;
	ReqBlock = CURRENT->sector;
	ReqBuffer = CURRENT->buffer;
	setup_req_params( drive );
	do_fd_action( drive );

	return;

  the_end:
	finish_fdc();
}


void do_fd_request(void)

{ 	unsigned long flags;

	DPRINT(("do_fd_request for pid %d\n",current->pid));
	save_flags(flags);
	cli();
	while( fdc_busy ) sleep_on( &fdc_wait );
	fdc_busy = 1;
	stdma_lock( floppy_irq, 0 );
	restore_flags(flags);

	redo_fd_request();
}


static int
invalidate_drive (int rdev)
{
  /* invalidate the buffer track to force a reread */
#ifdef TRACKBUFFER
  BufferDrive = -1;
#endif
  set_bit (rdev & 3, &fake_change);
  check_disk_change (rdev);
  return 0;
}

static int fd_ioctl(struct inode *inode, struct file *filp,
					unsigned int cmd, unsigned long param)
{
#define IOCTL_MODE_BIT 8
#define OPEN_WRITE_BIT 16
#define IOCTL_ALLOWED (filp && (filp->f_mode & IOCTL_MODE_BIT))

  int drive, device;

  device = inode->i_rdev;
  switch (cmd)
    {
      RO_IOCTLS (inode->i_rdev, param);
    }
  drive = MINOR (device);
  switch (cmd)
    {
    case FDFMTBEG:
      return 0;
    case FDCLRPRM:
      unit[drive].disktype = NULL;
      floppy_sizes[drive] = MAX_DISK_SIZE;
      return invalidate_drive (device);
    case FDFMTEND:
    case FDFLUSH:
      return invalidate_drive (drive);
    }
  if (!suser ())
    return -EPERM;
  if (drive < 0 || drive > 3)
    return -EINVAL;
  switch (cmd)
    {
    default:
      return -EINVAL;
    }
  return 0;
}


/* Initialize the 'unit' variable for drive 'drive' */

static void fd_probe( int drive )
{
	unit[drive].connected = 0;
	unit[drive].disktype  = NULL;

	if (!fd_test_drive_present( drive ))
		return;

	unit[drive].connected = 1;
	unit[drive].track     = 0;
	unit[drive].steprate  = FDCSTEP_3;
	MotorOn = 1;	/* from probe restore operation! */
}


/* This function tests the physical presence of a floppy drive (not
 * whether a disk is inserted). This is done by issuing a restore
 * command, waiting max. 2 seconds (that should be enough to move the
 * head across the whole disk) and looking at the state of the "TR00"
 * signal. This should now be raised if there is a drive connected
 * (and there is no hardware failure :-) Otherwise, the drive is
 * declared absent.
 */

static int fd_test_drive_present( int drive )

{	unsigned long timeout;
	unsigned char status;
	int			  ok;
	
	if (drive > 1) return( 0 );
	fd_select_drive( drive );

	/* disable interrupt temporarily */
	mfp.int_en_b &= 0x7f;
	FDC_WRITE (FDCREG_TRACK, 0xff00);
	FDC_WRITE( FDCREG_CMD, FDCCMD_RESTORE | FDCSTEP_6 );

	for( ok = 0, timeout = jiffies + 2*HZ+HZ/2; jiffies < timeout; ) {
		if (!(mfp.par_dt_reg & 0x20))
			break;
	}

	status = FDC_READ( FDCREG_STATUS );
	ok = (status & FDCSTAT_TR00) != 0;

	/* force interrupt to abort restore operation (FDC would try
	 * about 50 seconds!) */
	FDC_WRITE( FDCREG_CMD, FDCCMD_FORCI );
	udelay(500);
	status = FDC_READ( FDCREG_STATUS );
	udelay(20);

	if (ok) {
		/* dummy seek command to make WP bit accessible */
		FDC_WRITE( FDCREG_DATA, 0 );
		FDC_WRITE( FDCREG_CMD, FDCCMD_SEEK );
		while( mfp.par_dt_reg & 0x20 )
			;
		status = FDC_READ( FDCREG_STATUS );
	}

	mfp.int_pn_b = 0x7f;
	mfp.int_en_b |= 0x80;

	return( ok );
}


/* Look how many and which kind of drives are connected. If there are
 * floppies, additionally start the disk-change and motor-off timers.
 */

static void config_types( void )

{
	int drive, cnt = 0;
	unsigned char status;

	/* for probing drives, set the FDC speed to 8 MHz */
	dma_wd.fdc_speed = 0;

	printk("Probing floppy drive(s):\n");
	for( drive = 0; drive < FD_MAX_UNITS; drive++ ) {
		fd_probe( drive );
		if (unit[drive].connected) {
			printk("fd%d\n", drive);
			++cnt;
		}
	}

	if (FDC_READ( FDCREG_STATUS ) & FDCSTAT_BUSY) {
		/* If FDC is still busy from probing, give it another FORCI
		 * command to abort the operation. If this isn't done, the FDC
		 * will interrupt later and its IRQ line stays low, because
		 * the status register isn't read. And this will block any
		 * interrupts on this IRQ line :-(
		 */
		FDC_WRITE( FDCREG_CMD, FDCCMD_FORCI );
		udelay(500);
		status = FDC_READ( FDCREG_STATUS );
		udelay(20);
	}
	
	if (cnt > 0) {
		START_MOTOR_OFF_TIMER( FD_MOTOR_OFF_DELAY );
		if (cnt == 1) fd_select_drive( 0 );
		START_CHECK_CHANGE_TIMER( CHECK_CHANGE_DELAY );
	}
}

/*
 * floppy_open check for aliasing (/dev/fd0 can be the same as
 * /dev/PS0 etc), and disallows simultaneous access to the same
 * drive with different device numbers.
 */

static int floppy_open( struct inode *inode, struct file *filp )

{
  int drive;
  int old_dev;

  if (!filp)
    {
      DPRINT ("Weird, open called with filp=0\n");
      return -EIO;
    }

  drive = inode->i_rdev & 3;
  old_dev = fd_device[drive];

  if (fd_ref[drive])
    if (old_dev != inode->i_rdev)
      return -EBUSY;

  if (fd_ref[drive] == -1 || (fd_ref[drive] && filp->f_flags & O_EXCL))
    return -EBUSY;

  if (filp->f_flags & O_EXCL)
    fd_ref[drive] = -1;
  else
    fd_ref[drive]++;

  fd_device[drive] = inode->i_rdev;

  if (old_dev && old_dev != inode->i_rdev)
    invalidate_buffers(old_dev);

  /* Allow ioctls if we have write-permissions even if read-only open */
  if (filp->f_mode & 2 || permission (inode, 2) == 0)
    filp->f_mode |= IOCTL_MODE_BIT;
  if (filp->f_mode & 2)
    filp->f_mode |= OPEN_WRITE_BIT;

  if (filp->f_flags & O_NDELAY)
    return 0;

	if (filp->f_mode) {
		check_disk_change( inode->i_rdev );
		if (filp->f_mode & 2) {
			if (unit[drive].wpstat) {
				floppy_release(inode, filp);
				return -EROFS;
			}
		}
	}

  return 0;
}


static void floppy_release( struct inode * inode, struct file * filp )

{
  int drive;

  drive = inode->i_rdev & 3;

  if (!filp || (filp->f_mode & (2 | OPEN_WRITE_BIT)))
    /* if the file is mounted OR (writable now AND writable at open
       time) Linus: Does this cover all cases? */
    block_fsync (inode, filp);

  if (fd_ref[drive] < 0)
    fd_ref[drive] = 0;
  else if (!fd_ref[drive]--)
    {
      printk("floppy_release with fd_ref == 0");
      fd_ref[drive] = 0;
    }
}

static struct file_operations floppy_fops = {
	NULL,			/* lseek - default */
	block_read,		/* read - general block-dev read */
	block_write,		/* write - general block-dev write */
	NULL,			/* readdir - bad */
	NULL,			/* select */
	fd_ioctl,		/* ioctl */
	NULL,			/* mmap */
	floppy_open,		/* open */
	floppy_release, 	/* release */
	block_fsync		/* fsync */
};


void atari_floppy_init(void)

{	int i;

	if (register_blkdev(MAJOR_NR,"fd",&floppy_fops)) {
		printk("Unable to get major %d for floppy\n",MAJOR_NR);
		return;
	}

	/* initialize variables */
	SelectedDrive = -1;
#ifdef TRACKBUFFER
	BufferDrive = -1;
#endif

	/* initialize motor-off timer */
	timer_table[FLOPPY_TIMER].fn = check_change;
	timer_active &= ~(1 << FLOPPY_TIMER);

	/* The ST-Ram buffers are now allocated in
	 * atari_floppy_alloc_buffers(), which is called by
	 * blk_dev_init(). This is because floppy_init() is called after
	 * mem_init() and stram_alloc() isn't possible any more here.
	 */
	PhysDMABuffer = (unsigned long) VTOP(DMABuffer);
#ifdef TRACKBUFFER
	PhysTrackBuffer = (unsigned long) VTOP(TrackBuffer);
	BufferDrive = BufferSide = BufferTrack = -1;
#endif

	for (i = 0; i < FD_MAX_UNITS; i++) {
		unit[i].track = -1;
  }

  blk_size[MAJOR_NR] = floppy_sizes;
	blk_dev[MAJOR_NR].request_fn = DEVICE_REQUEST;

	config_types();
}


/* This function is called by blk_dev_init() to give the floppy driver
 * a chance to allocate some ST-Ram
 */

unsigned long atari_floppy_alloc_buffers( unsigned long mem_start )

{
#ifdef TRACKBUFFER
	DMABuffer = atari_stram_alloc( (MAX_SECTORS+1)*512, &mem_start );
	TrackBuffer = DMABuffer + 512;
#else
	DMABuffer = atari_stram_alloc( 512, &mem_start );
#endif
	return( mem_start );
}


/* Local Variables: */
/* tab-width: 4     */
/* End:             */
