

/*
 * drivers/char/amiga_ser.c: Amiga built-in serial port driver.
 *
 * Copyright 1994 Roman Hodek, 1994 Hamish Macdonald
 *       Based on the Atari MFP driver by Roman Hodek
 * 
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file README.legal in the main directory of this archive
 * for more details.
 *
 */


/*
 * This file implements the driver for the Amiga built-in serial port.
 */

#include <linux/types.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/errno.h>
#include <linux/tty.h>
#include <linux/termios.h>
#include <linux/bootinfo.h>
#include <linux/amigahw.h>
#include <linux/amigaints.h>

#include "serial.h"

/* some serial hardware definitions */
#define SDR_OVRUN   (1<<15)
#define SDR_TBE     (1<<13)
#define SDR_TSRE    (1<<12)

#define AC_SETCLR   (1<<15)
#define AC_UARTBRK  (1<<11)

#define SER_DTR     (1<<7)
#define SER_RTS     (1<<6)
#define SER_DCD     (1<<5)
#define SER_CTS     (1<<4)
#define SER_DSR     (1<<3)

/***************************** Prototypes *****************************/

static void ser_rx_int( struct intframe *fp, struct async_struct *info );
static void ser_tx_int( struct intframe *fp, struct async_struct *info );
static void ser_vbl_int( struct intframe *fp, struct async_struct *info );
static void ser_init( struct async_struct *info );
static void ser_deinit( struct async_struct *info, int leave_dtr );
static void ser_enab_tx_int( struct async_struct *info, int enab_flag );
static int  ser_check_custom_divisor( struct async_struct *info, int divisor );
static void ser_change_speed( struct async_struct *info );
static void ser_restart( struct async_struct *info );
static void ser_throttle( struct async_struct *info, int status );
static void ser_set_break( struct async_struct *info, int break_flag );
static void ser_get_serial_info( struct async_struct *info,
				struct serial_struct *retinfo );
static unsigned int ser_get_modem_info( struct async_struct *info );
static int ser_set_modem_info( struct async_struct *info, int new_dtr,
			      int new_rts );

/************************* End of Prototypes **************************/



/* SERIALSWITCH structure for the Amiga serial port
 */

static SERIALSWITCH amiga_ser_switch = {
    ser_init, ser_deinit, ser_enab_tx_int,
    ser_check_custom_divisor, ser_change_speed,
    ser_restart, ser_throttle, ser_set_break,
    ser_get_serial_info, ser_get_modem_info,
    ser_set_modem_info, NULL
};

/* Divisors for standard speeds (NTSC and PAL) */
static int baud_table[2][15] = {
    /* NTSC */
    {
	/* B0     */ 0,
	/* B50    */ 71589,
	/* B75    */ 47726,
	/* B110   */ 32540,
	/* B134	  */ 26712,
	/* B150	  */ 23862,
	/* B200	  */ 17896,
	/* B300	  */ 11930,
	/* B600	  */ 5964,
	/* B1200  */ 2892,
	/* B1800  */ 1987,
	/* B2400  */ 1490,
	/* B4800  */ 744,
	/* B9600  */ 371,
	/* B19200 */ 185
    },
    /* PAL */
    {
	/* B0     */ 0,
	/* B50    */ 70936,
	/* B75    */ 47290,
	/* B110   */ 32243,
	/* B134	  */ 26468,
	/* B150	  */ 23644,
	/* B200	  */ 17733,
	/* B300	  */ 11821,
	/* B600	  */ 5910,
	/* B1200  */ 2954,
	/* B1800  */ 1969,
	/* B2400  */ 1476,
	/* B4800  */ 737,
	/* B9600  */ 368,
	/* B19200 */ 183
    }
};
	
static __inline__ void ser_DTRoff(void)
{
    ciab.pra |= SER_DTR; /* active low */
}

static __inline__ void ser_DTRon(void)
{
    ciab.pra &= ~SER_DTR; /* active low */
}

static __inline__ void ser_RTSoff(void)
{
    ciab.pra |= SER_RTS; /* active low */
}

static __inline__ void ser_RTSon(void)
{
    ciab.pra &= ~SER_RTS; /* active low */
}

void amiga_serinit( struct async_struct *info )
{	
    unsigned long flags;

    save_flags (flags);
    cli();

    /* set ISRs, and then disable the rx interrupts */
    add_isr( IRQ_AMIGA_TBE, (isrfunc)ser_tx_int, 0, info);
    add_isr( IRQ_AMIGA_RBF, (isrfunc)ser_rx_int, 0, info);
    add_isr( IRQ_AMIGA_VERTB, (isrfunc)ser_vbl_int, 0, info );

    /* turn off both Tx and Rx interrupts */
    custom.intena = IF_TBE | IF_RBF;
    /* clear any pending interrupts */
    custom.intreq = IF_RBF | IF_TBE;
    restore_flags (flags);

    /*
     * set the appropriate directions for the modem control flags,
     * and clear RTS and DTR
     */
    ciab.ddra |= (SER_DTR | SER_RTS);   /* outputs */
    ciab.ddra &= ~(SER_DCD | SER_CTS | SER_DSR);  /* inputs */
    ciab.pra |= (SER_DTR | SER_RTS); /* active low */
    
    info->type            = SER_AMIGA;
    info->base            = &custom.serdatr; /* dummy value */
    info->RI.addr         = 0; /* NO RI */
    info->RI.bitno        = 0;
    info->RI.active_state = 0;
    info->sw              = &amiga_ser_switch;
}


static void ser_rx_int( struct intframe *fp, struct async_struct *info )
{
    int		ch, err = 0;
	
    ch = custom.serdatr;
    if (ch & SDR_OVRUN)
	err = TTY_OVERRUN;
    
    /* 
     * clear the interrupt right away, so that another character can
     * be read
     */
    custom.intreq = IF_RBF;

    ch &= 0xff;
	
    rs_put_char_to_queue( info, ch, err );
    rs_rx_throttle_if_needed( info );
}


static void ser_tx_int( struct intframe *fp, struct async_struct *info )
{
	int	ch;

	if ((ch = rs_get_char_from_queue( info )) < 0) {
		/* disable tx interrupts */
		custom.intena = IF_TBE;
	} else {
		/* clear previous interrupt */
		custom.intreq = IF_TBE;
		custom.serdat = ch | 0x100;
	}
	rs_tx_wakeup_if_needed( info );
}


/* This variable holds the current state of the DCD/CTS bits */
static unsigned char current_ctl_bits;

static void ser_vbl_int( struct intframe *fp, struct async_struct *info )
{	
    unsigned char bits;

    if (!(info->flags & ASYNC_INITIALIZED))
	return;

    bits = ciab.pra & (SER_DCD | SER_CTS);

    if (bits ^ current_ctl_bits) {

	if ((bits ^ current_ctl_bits) & SER_DCD)
	    rs_dcd_changed( info, !(bits & SER_DCD)); /* active low */

	if ((bits ^ current_ctl_bits) & SER_CTS)
	    rs_cts_changed( info, !(bits & SER_CTS)); /* active low */
    }

    current_ctl_bits = bits;
}

static void ser_init( struct async_struct *info )
{	
    /* enable Rx interrupts */
    custom.intena = IF_SETCLR | IF_RBF;

    /* setup a pending transmit interrupt */
    custom.intreq = IF_SETCLR | IF_TBE;

    /* turn on DTR and RTS */
    ser_DTRon();
    ser_RTSon();

    /* remember current state of the DCD and CTS bits */
    current_ctl_bits = ciab.pra & (SER_DCD | SER_CTS);
}


static void ser_deinit( struct async_struct *info, int leave_dtr )
{
    /* disable Rx and Rx interrupts */
    custom.intena = IF_RBF | IF_TBE;

    /* wait for last byte to be completely shifted out */
    while( !(custom.serdatr & SDR_TSRE) )
	;

    /* drop RTS and DTR if required */
    ser_RTSoff();
    if (!leave_dtr)
	ser_DTRoff();
}

static void ser_enab_tx_int (struct async_struct *info, int enab_flag)
{
    if (enab_flag) {
	/* enable tx interrupts (should result in an immediate interrupt) */
	custom.intena = IF_SETCLR | IF_TBE;
    } else
	/* disable tx interrupts */
	custom.intena = IF_TBE;
}

static int ser_check_custom_divisor( struct async_struct *info, int divisor )
{
    /* allow any divisor */
    return 1;
}


static void ser_change_speed( struct async_struct *info )

{
    unsigned	cflag, baud, chsize, stopb, parity, aflags;
    unsigned	div = 0;
	
    if (!info->tty || !info->tty->termios) return;

    cflag  = info->tty->termios->c_cflag;
    baud   = cflag & CBAUD;
    chsize = cflag & CSIZE;
    stopb  = cflag & CSTOPB;
    parity = cflag & (PARENB | PARODD);
    aflags = info->flags & ASYNC_SPD_MASK;

    if (baud == 15 && aflags == ASYNC_SPD_CUST)
	div = info->custom_divisor;
    else {
	/* Maximum  speed is 19200 :-( */
	if (baud > 14) baud = 14;
	if (boot_info.bi_amiga.eclock == NTSC_ECLOCK)
	    div = baud_table[0][baud];
	else
	    div = baud_table[1][baud];
    }

    if (!div) {
	/* speed == 0 -> drop DTR */
	ser_DTRoff();
	return;
    }

    /* setup the serial port period register */
    custom.serper = div;
}
	
	
static void ser_restart( struct async_struct *info )
{
	/* 
	 * nothing needed; the transmit interrupt has just been
	 * enabled, and an interrupt should be pending.  The
	 * Tx ISR will be called, which will fire out a character
	 * if necessary.
	 */
}


static void ser_throttle( struct async_struct *info, int status )
{
    if (status == TTY_THROTTLE_RQ_FULL)
	ser_RTSoff();
    else if (status == TTY_THROTTLE_RQ_AVAIL)
	ser_RTSon();
}


static void ser_set_break( struct async_struct *info, int break_flag )
{
    if (break_flag)
	custom.adkcon = AC_SETCLR | AC_UARTBRK;
    else
	custom.adkcon = AC_UARTBRK;
}


static void ser_get_serial_info( struct async_struct *info,
				struct serial_struct *retinfo )
{
    retinfo->baud_base = 0;
    retinfo->custom_divisor = info->custom_divisor;
}


static unsigned int ser_get_modem_info( struct async_struct *info )
{
    unsigned int minfo = ciab.pra;

    return(
	   ((minfo & SER_DTR) ? 0 : TIOCM_DTR) |
	   ((minfo & SER_RTS) ? 0 : TIOCM_RTS) |
	   ((minfo & SER_DCD) ? 0 : TIOCM_CAR) |
	   ((minfo & SER_CTS) ? 0 : TIOCM_CTS) |
	   ((minfo & SER_DSR) ? 0 : TIOCM_DSR) |
	   /* TICM_RNG */ 0
	   );
}


static int ser_set_modem_info( struct async_struct *info,
			      int new_dtr, int new_rts )
{
    if (new_dtr == 0)
	ser_DTRoff();
    else if (new_dtr == 1)
	ser_DTRon();
    
    if (new_rts == 0)
	ser_RTSoff();
    else if (new_rts == 1)
	ser_RTSon();

    return( 0 );
}
