#ifndef _LINUX_LP_H
#define _LINUX_LP_H

/*
 * usr/include/linux/lp.h  modified for Amiga by Michael Rausch
 * modified for Atari by Andreas Schwab
 * bug fixed by Jes Sorensen 18/8-94:
 *     It was not possible to compile the kernel only for Atari or Amiga.
 *
 * linux i386 version  c.1991-1992 James Wiegand
 * many modifications copyright (C) 1992 Michael K. Johnson
 * Interrupt support added 1993 Nigel Gamble
 */

#include <linux/autoconf.h>
#include <linux/bootinfo.h>

/*
 * Per POSIX guidelines, this module reserves the LP and lp prefixes
 * These are the lp_table[minor].flags flags...
 */
#define LP_EXIST 0x0001
#define LP_BUSY	 0x0004
#define LP_ABORT 0x0040

/* timeout for each character.  This is relative to bus cycles -- it
 * is the count in a busy loop.  THIS IS THE VALUE TO CHANGE if you
 * have extremely slow printing, or if the machine seems to slow down
 * a lot when you print.  If you have slow printing, increase this
 * number and recompile, and if your system gets bogged down, decrease
 * this number.  This can be changed with the tunelp(8) command as well.
 */

#define LP_INIT_CHAR 400000

/* This is the amount of time that the driver waits for the printer to
 * catch up when the printer's buffer appears to be filled.  If you
 * want to tune this and have a fast printer (i.e. HPIIIP), decrease
 * this number, and if you have a slow printer, increase this number.
 * This is in hundredths of a second, the default 2 being .05 second.
 * Or use the tunelp(8) command, which is especially nice if you want
 * change back and forth between character and graphics printing, which
 * are wildly different...
 */

#define LP_INIT_TIME 2

/* IOCTL numbers */
#define LPCHAR   0x0001  /* corresponds to LP_INIT_CHAR */
#define LPTIME   0x0002  /* corresponds to LP_INIT_TIME */
#define LPABORT  0x0004  /* call with TRUE arg to abort on error,
			    FALSE to retry.  Default is retry.  */
#define LPSETIRQ 0x0005  /* call with new IRQ number,
			    or 0 for polling (no IRQ) */
#define LPGETIRQ 0x0006  /* get the current IRQ number */
#define LPWAIT   0x0008  /* corresponds to LP_INIT_WAIT */

/* timeout for printk'ing a timeout, in jiffies (100ths of a second).
   This is also used for re-checking error conditions if LP_ABORT is
   not set.  This is the default behavior. */

#define LP_TIMEOUT_INTERRUPT	(60 * HZ)
#define LP_TIMEOUT_POLLED	(10 * HZ)

#define LP_F	lp_table.flags		/* flags for busy, etc. */
#define LP_CHAR	lp_table.chars		/* busy timeout */
#define LP_TIME	lp_table.time		/* wait time */
#define LP_WAIT	lp_table.wait		/* strobe wait */
#define LP_IRQ	lp_table.irq		/* interrupt # */
					/* 0 means polled */

#define LP_BUFFER_SIZE 256

struct lp_struct {
	unsigned int irq;
	int flags;
	unsigned int chars;
	unsigned int time;
	struct wait_queue *lp_wait_q;
	char *lp_buffer;
};

struct lp_struct lp_table = {
	0, 0, LP_INIT_CHAR, LP_INIT_TIME, NULL, NULL,
}; 
#define LP_NO 1


#define AMIGA_LP_OUT(c)	ciaa.prb = c

/*  busy is low-active, and so are the cia bits */
#define AMIGA_LP_IS_BUSY()	(ciab.pra & 1)
#define AMIGA_LP_HAS_POUT()	(ciab.pra & 2)
#define AMIGA_LP_IS_ONLINE()	(ciab.pra & 4)

/* LP_OUT must be executed with interrupts blocked */
#define ATARI_LP_OUT(c) (sound_ym.rd_data_reg_sel = 15, \
			 sound_ym.wd_data = c, \
			 sound_ym.rd_data_reg_sel = 14, \
			 sound_ym.wd_data = sound_ym.rd_data_reg_sel & ~(1 << 5), \
			 udelay (1), \
			 sound_ym.wd_data = sound_ym.rd_data_reg_sel | (1 << 5))

#define ATARI_LP_IS_BUSY()	(mfp.par_dt_reg & 1)
#define ATARI_LP_HAS_POUT()	0
#define ATARI_LP_IS_ONLINE()	!ATARI_LP_IS_BUSY() /* or 1 XXX */

#if defined (CONFIG_AMIGA) + defined (CONFIG_ATARI) == 1

#ifdef CONFIG_AMIGA
#define LP_OUT(c) AMIGA_LP_OUT(c)
#define LP_IS_BUSY() AMIGA_LP_IS_BUSY()
#define LP_HAS_POUT() AMIGA_LP_HAS_POUT()
#define LP_IS_ONLINE() AMIGA_LP_IS_ONLINE()
#endif
#ifdef CONFIG_ATARI
#define LP_OUT(c) ATARI_LP_OUT(c)
#define LP_IS_BUSY() ATARI_LP_IS_BUSY()
#define LP_HAS_POUT() ATARI_LP_HAS_POUT()
#define LP_IS_ONLINE() ATARI_LP_IS_ONLINE()
#endif

#else /* multi configuration */

static __inline__ void
LP_OUT (int c)
{
  switch (boot_info.machtype)
    {
#ifdef CONFIG_AMIGA
    case MACH_AMIGA:
      AMIGA_LP_OUT (c);
      break;
#endif
#ifdef CONFIG_ATARI
    case MACH_ATARI:
      ATARI_LP_OUT (c);
      break;
#endif
    }
}

static __inline__ int
LP_IS_BUSY (void)
{
  switch (boot_info.machtype)
    {
#ifdef CONFIG_AMIGA
    case MACH_AMIGA:
      return AMIGA_LP_IS_BUSY ();
#endif
#ifdef CONFIG_ATARI
    case MACH_ATARI:
      return ATARI_LP_IS_BUSY ();
#endif
    default:
      return 0;
    }
}

static __inline__ int
LP_HAS_POUT (void)
{
  switch (boot_info.machtype)
    {
#ifdef CONFIG_AMIGA
    case MACH_AMIGA:
      return AMIGA_LP_HAS_POUT ();
#endif
#ifdef CONFIG_ATARI
    case MACH_ATARI:
      return ATARI_LP_HAS_POUT ();
#endif
    default:
      return 0;
    }
}

static __inline__ int
LP_IS_ONLINE (void)
{
  switch (boot_info.machtype)
    {
#ifdef CONFIG_AMIGA
    case MACH_AMIGA:
      return AMIGA_LP_IS_ONLINE ();
#endif
#ifdef CONFIG_ATARI
    case MACH_ATARI:
      return ATARI_LP_IS_ONLINE ();
#endif
    default:
      return 0;
    }
}

#endif

/*
 * function prototypes
 */

extern long lp_init(long);

#endif
